#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import os
import sys
import errno
import codecs
import argparse
import locale
from lxml import etree, html


VERSION = "3.6.1.0"

class ArgumentParserWriteUTF8(argparse.ArgumentParser):
    """Subclass of ArgumentParser that outputs UTF-8 properly

    Standard ArgumentParser in argparse outputs help messages without
    converting to UTF-8. This fails if the message is a unicode string
    containing non-ASCII characters and output is redirected (to a
    pipe or a file). Unredirected output seems OK. See
    http://bugs.python.org/issue9779
    This class fixes it. It will probably not be needed in Python 3.
    """

    def _print_message(self, message, file=None):
        """Output message to file, encoded as UTF-8 """
        if message:
            if file is None:
                file = sys.stderr
            file.write(message.encode(locale.getpreferredencoding()))


class BfaHocr():
    def __init__(self, hocrfile, parse_word_pos=False):
        self.hocrfile = hocrfile
        # if not os.path.isfile(self.hocrfile):
        #     raise EnvironmentError (
        #         errno.ENOENT, os.strerror(errno.ENOENT), self.hocrfile)
        with open(self.hocrfile, "rb") as f:
            self.hocrdoc = etree.parse(f)
        self.words = {}
        self.page_orientation = {}
        if parse_word_pos is True:
            self.parse_word_pos()

    def parse_word_pos(self):
        self.words = {}
        pages = self.hocrdoc.xpath("//*[@class='ocr_page']")
        
        for page in pages:
            pageno = int(page.get("id").replace("page_", ""))
            orientation = page.get("orientation")
            if orientation is not None:
                orientation = int(orientation)               
            else:
                orientation = 0
            self.page_orientation[pageno] = orientation
            #height, width = 0
            width, height = [int(y) for y in [x.strip().split(" ")[1:] for x in page.get("title").split(";") if x.strip().lower().startswith("bbox")][0]][2:4]

            self.words[pageno] = {}
            wordnodes = page.xpath(".//*[@class='ocrx_word']")
            for node in wordnodes:
                if node.text.strip() not in self.words[pageno]:
                    self.words[pageno][node.text.strip()] = []
                title = node.get("title")
                bbox = [int(x) for x in title.split(";")[0].split(" ")[1:5]]
                # bbox startx, starty, endx, endy
                #calcbbox = self.calc_position(bbox, orientation, height, width)
                self.words[pageno][node.text.strip()].append([(bbox[0], bbox[1]), (bbox[2], bbox[3])])


    def calc_position(self, bbox, rotation, height, width):
        startx, starty, endx, endy = bbox
        if rotation == 0:
            calcbbox = bbox
        elif rotation == 180:
            calcbbox = [width-endx, height-endy, width-startx, height-starty]
        elif rotation == 90:
            calcbbox = [height - endy, startx, height-starty, endx]
        elif rotation == 270:
            calcbbox = [starty, width - endx, endy, width - startx]
        try:
            calcbbox = [(calcbbox[0], calcbbox[1]), (calcbbox[2], calcbbox[3])]
        except:
            pass
        return calcbbox

    def get_wordpos(self, word, pageno=None, exact=False, ignorecase=True, returndict=False):
        """
        returndict True: {1: [{'tl': (645, 376), 'br': (1517, 473)}, {'tl': (754, 2931), 'br': (994, 2964)}], 2: [{'tl': (645, 376), 'br': (1517, 473)}, {'tl': (754, 2931), 'br': (994, 2964)}], 3: [{'tl': (2070, 2108), 'br': (2093, 2447)}], 4: [{'tl': (645, 376), 'br': (1517, 473)}, {'tl': (754, 2931), 'br': (994, 2964)}]}
        returndict False: [[1, [[(645, 376), (1517, 473)], [(754, 2931), (994, 2964)]]], [2, [[(645, 376), (1517, 473)], [(754, 2931), (994, 2964)]]], [3, [[(2070, 2108), (2093, 2447)]]], [4, [[(645, 376), (1517, 473)], [(754, 2931), (994, 2964)]]]]
        """
        if pageno is None:
            pages = self.words.keys()
        else:
            pages = [pageno]
        resultlist = []
        resultdict = {}
        lowerword = word.lower()
        for page in pages:
            for realword in self.words[page]:
                if ignorecase is True:
                    iword = word.lower()
                    irealword = realword.lower()
                else:
                    iword = word
                    irealword = realword
                if exact is False:
                    if iword in irealword:
                        if page not in resultdict:
                            resultdict[page] = []
                        for position in self.words[page][realword]:
                            resultdict[page].append(
                                {"tl": position[0], "br": position[1]})
                else:
                    #print iword, irealword
                    if iword == "testseite":
                        pass
                    if iword == irealword:
                        if page not in resultdict:
                            resultdict[page] = []
                        for position in self.words[page][realword]:
                            resultdict[page].append(
                                {"tl": position[0], "br": position[1]})
        return {word: resultdict}

    def set_orientation(self, pagedict):
        """
        input args:
        pagedict -> dictionary, key is pagenumber (int), value is orientation 
                    in degrees (0,90,180,270) to set (int)
        pagedict[1]=270 -> page 1 orientation will be set to 270
        """
        pages = self.hocrdoc.xpath("//*[@class='ocr_page']")
        for page in pages:
            pageno = page.get("id").replace("page_", "")
            if pageno is None:
                raise Exception("page id not found")
            if int(pageno) in pagedict or "all" in pagedict:
                if "all" in pagedict:
                    new_orientation = pagedict["all"]
                else:
                    new_orientation = pagedict[int(pageno)]
                if new_orientation not in (0, 90, 180, 270):
                    raise Exception(
                        "orientation should be one of the following (int): 0,90,180,270")
                page.set("orientation", str(new_orientation))
        hocrstring = etree.tostring(self.hocrdoc).decode('UTF-8')
        with codecs.open(self.hocrfile, "wb", encoding="utf8") as fobj:
            fobj.write(hocrstring)

    def rotate(self, pagedict):
        """
        input args:
        pagedict -> dictionary, key is pagenumber (int), value is rotation 
                    in degrees (0,90,180,270) to set (int)
        pagedict[1]=270 -> page 1 ocrx_word nodes bbox values for 270 degree rotation (clockwise)
        """
        pages = self.hocrdoc.xpath("//*[@class='ocr_page']")
        for page in pages:
            pageno = page.get("id").replace("page_", "")
            if pageno is None:
                raise Exception("page id not found")
            if int(pageno) in pagedict or "all" in pagedict:
                if "all" in pagedict:
                    rotation = pagedict["all"]
                else:
                    rotation = pagedict[int(pageno)]
                if rotation not in (0, 90, 180, 270):
                    raise Exception(
                        "orientation should be one of the following (int): 0,90,180,270")
                # change page resolution in case we rotate 90/270
                # <div class="ocr_page" id="page_1" title="image &quot;c:\users\bitfarm\appdata\local\temp\tessocr_1ksits\tesstmp1-1.tif&quot;; bbox 0 0 3508 2479; ppageno 0" orientation="270">
                old_orientation = page.get("orientation")
                if old_orientation is None:
                    old_orientation = 0
                width, height = [int(y) for y in [x.strip().split(" ")[1:] for x in page.get("title").split(";") if x.strip().lower().startswith("bbox")][0]][2:4]
                if rotation in (90,270):
                    page_title = page.get("title")
                    titleparts = page_title.split(";")
                    bbox_ori = titleparts[1]
                    bbox = bbox_ori.split(" ")[-4:]
                    new_bbox = bbox[0:2]+[bbox[3]]+[bbox[2]]
                    new_bbox = " bbox "+ " ".join(new_bbox)
                    titleparts[1] = " "+new_bbox
                    new_title = ";".join(titleparts)
                    #print new_title
                    page.set("title", new_title)
                    #width, height = height, width
                wordnodes = page.xpath(".//*[@class='ocrx_word']")
                for node in wordnodes:
                    title = node.get("title")
                    title_parts = title.split(";")
                    bbox = [int(x) for x in title.split(";")[0].split(" ")[1:5]]
                    # bbox startx, starty, endx, endy

                    calcbbox = self.calc_position(bbox, rotation, height, width)
                    
                    new_title = "bbox "+ "%d %d %d %d" %(calcbbox[0][0], calcbbox[0][1], calcbbox[1][0], calcbbox[1][1]) + ";"+title_parts[1]
                    # print title
                    # print new_title
                    node.set("title", new_title)
                new_rotation = (int(old_orientation) + rotation) % 360
                page.set("orientation", str(new_rotation))
        hocrstring = etree.tostring(self.hocrdoc).decode('UTF-8')
        with codecs.open(self.hocrfile, "wb", encoding="utf8") as fobj:
            fobj.write(hocrstring)
        #self.getWordPos()



    def get_orientation(self):
        for page in sorted(self.page_orientation):
            yield self.page_orientation[page]

def test():
    from bitfarm_xml import OcrXmlParser
    oxp = OcrXmlParser()
    word = "Testseite"
    oxp.readfile(
        r"C:\bitfarm-daten\Archiv_RS\2019\05\22\19522162618VMTH233405653.ocr_xml", resolution=[(2479, 3508), (2479, 3508), (3508, 2479), (2479, 3508)])
    oxpresult = oxp.getWordPos(
        word, exact=False, ignorecase=True, returndict=True)
    print(oxpresult)
    hocr = BfaHocr(
        r"C:\Users\bitfarm\Desktop\tesseract-test\tif\4page-mit-quer.hocr", parse_word_pos=True)
    print(hocr.get_wordpos(word, exact=False, ignorecase=True, returndict=True))
    #hocr.set_orientation({"all": 90})
    # TODO: commandline-interface for set_orientation and get_wordpos

def test():
    pth = r"C:\bitfarm-daten\Archiv_RS\2019\06\03\196315528VMTH224170271.hocr"
    hocr = BfaHocr(pth, parse_word_pos=True)
    word=u"Informationen"
    res = hocr.get_wordpos(word, exact=False, ignorecase=True, returndict=True)
    print (res["Informationen"][1])
    sys.exit()

def test_rotate():
    pth = r"C:\bitfarm-daten\Archiv_RS\2019\06\05\1965162021VMTH237831770.hocr"
    hocr = BfaHocr(pth, parse_word_pos=True)
    hocr.rotate({1:90})

if __name__ == "__main__":
    # test_rotate()
    # sys.exit()
    parser = ArgumentParserWriteUTF8()
    parser.add_argument("hocrfile")
    parser.add_argument("-search", help=u'Suchbegriff', default=None)
    parser.add_argument(
        '-exact', help=u'exakten Suchbegriff finden', default=False)
    parser.add_argument(
        '-ignorecase', help=u'Groß- Kleinschreibung ignorieren', default=True)
    parser.add_argument(
        '-orientation', help=u'orientation in hocr-File setzen (Format: Seite1:Ausrichtung1,Seite2:Ausrichtung2 oder "all":Ausrichtung - Ausrichtung = (0,90,180,270))', default=None)
    parser.add_argument(
        '-rotate', help=u'orientation in hocr-File ändern -> berechnet auch die bbox Infos der einzelnen Worte neu! (Format: Seite1:Ausrichtung1,Seite2:Ausrichtung2 oder "all":Ausrichtung - Ausrichtung = (0,90,180,270))', default=None)
    parser.add_argument("-getorientation", help=u"show orientation per page", default=False, type=bool)
    args = parser.parse_args()
    #print args
    hocr = BfaHocr(args.hocrfile, parse_word_pos=True)
    #print hocr
    if args.search is not None:
        print(hocr.get_wordpos(args.search, exact=args.exact, ignorecase=args.ignorecase, returndict=True))
    elif args.orientation is not None:
        pagedict = {}
        pages = args.orientation.split(",")
        if len(pages) == 1 and pages[0].split(":")[0] == "all":
            pagedict = {"all":int(pages[0].split(":")[1])}
        else:
            for page in pages:
                pno, orientation = page.split(":")
                pagedict[int(pno)] = int(orientation)
        #print pagedict
        hocr.set_orientation(pagedict)
    elif args.getorientation is True:
        for orientation in hocr.get_orientation():
            print(orientation)
    elif args.rotate is not None:
        pagedict = {}
        pages = args.rotate.split(",")
        if len(pages) == 1 and pages[0].split(":")[0] == "all":
            pagedict = {"all":int(pages[0].split(":")[1])}
        else:
            for page in pages:
                pno, orientation = page.split(":")
                pagedict[int(pno)] = int(orientation)
        #print pagedict
        hocr.rotate(pagedict)
