#!/usr/bin/env python
# -*- coding: utf-8*-

import argparse
import base64
import ConfigParser
import codecs
from collections import defaultdict
import locale
import logging
import os
import re
import sys
import time
import atexit
import smtplib
import traceback

from email import generator
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from datetime import datetime
from dpapi import decryptData, cryptData
from bfaServerClient import BfaServerClient, _set_logger
from pprint import pprint, pformat

version = "3.6.1.4"


codecs.register(lambda name: codecs.lookup(
    'utf8') if name == 'cp65001' else None)
ADDTYPES = {
    0: u"Universal",
    1: u"Datum",
    2: u"Auswahl",
    3: u"Merfach-Auswahl",
    5: u"Intervall",
    6: u"Ganzzahl",
    7: u"Fließkomma",
    8: u"Währung",
    11: u"DLM-Auswahl",
    12: u"DLM-Datum"
}

RIGHTS_BITMASKS = {
    "LIST":             1,  # + E | Suchen: Suchen und Trefferliste anzeigen.
    "READ":             2,  # *   | Anzeigen: Dokument aus dem Archiv anzeigen
    "MODIFY":           4,  # + E | Ändern/Bearbeiten: ändern von Metadaten
    # + E | Schreiben: wenn was auf Platte geschrieben wird (Oridatei laden)
    "WRITE":            8,
    "MAIL":            16,  # + E | Email-Versand: Dok als Mail versenden -> share
    "PRINT":           32,  # + - | Drucken: Dok drucken            -> share
    "LINK":            64,  # -   | Verknüpfung: Verknüpfung anlegen
    "IMPORT":         128,  # -   | Importieren: in Arc REIN schieben - importieren
    "HISTORY":        256,  # *   | History anzeigen: kla ne
    "MOVE":           512,  # -   | Verschieben: aus Arc RAUS schieben
    "AWRITE":        1024,  # -   | Anmerkungen schreiben: NEUE annos anbringen  -> public
    "AREAD":         2048,  # + E | Anmerkungen lesen: kla ne                    -> public
    "AMODIFY":       4096,  # -   | Anmerkungen ändern: annos ändern             -> public
    "APWRITE":       8192,  # -   | NEUE annos anbringen                         -> private
    "APREAD":       16384,  # + E | annos lesen                                  -> private
    "WORKFLOWBREAK": 32768,  # -   | workflow abbrechen: kla ne
    # *   | Lesen protokollieren: Log Eintrag machen beim lesen/anschauen (Funktion)
    "LOGREAD":      65536,
    "ADMIN":       131072,  # -   | Admin-Funktionen: z.B. OriDok wiederherstellen
    "DELETE":      262144,  # -   | Löschen: NRS Dok können mit diesem Recht gelöscht werden
    "EXPORT":      524288,  # -   | Exportieren: bezieht sich auf alle Export Funktionen
    "REVSAVE":    1048576,  # -   | Revisionssicher ablegen: darf Revisionssicher ablegen
    "APMODIFY":   2097152,  # -   | annos ändern                                 -> private
    "field": {
        # + E | Feldwert kann nur gelesen werden bzw. für die Suche verwendet werden
        "READ":         1,
        # + E | Feldwert kann beschrieben werden, aber nur wenn es noch keinen Wert hat.
        "WRITE":        2,
        # + E | Feldwert kann immer beschrieben werden.
        "MODIFY":       4
    }
}


def noconvert(x): return x


def datetostr(x): return datetime.strftime(x, '%Y-%m-%d %H:%M:%S')


def floattostr(x): return str(x).replace(".", ",")


def truncto2digit(x): return float("%.2f" % x)


ADDCONVERTER = defaultdict(lambda: None, {  # key: input addtype, subkey: output addtype
    0: defaultdict(lambda: None, {0: noconvert, 2: noconvert, 3: noconvert, 11: noconvert}),
    1: defaultdict(lambda: None, {1: noconvert, 0: datetostr, 2: datetostr, 3: datetostr, 11: datetostr, 12: noconvert}),
    2: defaultdict(lambda: None, {2: noconvert, 0: noconvert, 3: noconvert, 11: noconvert}),
    3: defaultdict(lambda: None, {3: noconvert, 0: noconvert, 2: noconvert, 11: noconvert}),
    5: defaultdict(lambda: None, {5: noconvert, 0: str, 2: str, 3: str, 11: str}),
    6: defaultdict(lambda: None, {6: noconvert, 0: str, 2: str, 3: str, 11: str, 7: float, 8: float}),
    7: defaultdict(lambda: None, {7: noconvert, 0: floattostr, 2: floattostr, 3: floattostr, 8: truncto2digit, 11: floattostr}),
    8: defaultdict(lambda: None, {8: noconvert, 0: floattostr, 2: floattostr, 3: floattostr, 7: noconvert, 11: floattostr}),
    11: defaultdict(lambda: None, {11: noconvert, 0: noconvert, 2: noconvert, 3: noconvert}),
    12: defaultdict(lambda: None, {12: noconvert, 0: datetostr, 1: noconvert, 2: datetostr, 3: datetostr, 11: datetostr})
})

class LogReport(object):

    def __init__(self, log, config):
        self.log = log
        self.reportdata = [[]]
        self.config = config
        self.current_gdocid = None
        self.gdocids = [None]
        
    def error(self, msg, report=False):
        msg = self.render_msg(msg)
        self.log.error(msg)
        if report is True:
            self.reportdata[-1].append(msg)
        

    def info(self, msg, report=False):
        msg = self.render_msg(msg)
        self.log.info(msg)
        if report is True:
            self.reportdata[-1].append(msg)
    
    def debug(self, msg, report=False):
        msg = self.render_msg(msg)
        self.log.debug(msg)
        if report is True:
            self.reportdata[-1].append(msg)

    def critical(self, msg, report=False):
        msg = self.render_msg(msg)
        self.log.critical(msg)
        if report is True:
            self.reportdata[-1].append(msg)

    def warning(self, msg, report=False):
        msg = self.render_msg(msg)
        self.log.warning(msg)
        if report is True:
            self.reportdata[-1].append(msg)

    def render_msg(self, msg):
        #datestr = datetime.strftime(datetime.now(), "%Y-%m-%d %H:%M:%S")
        #_msg = datestr+ " "
        if self.current_gdocid is not None:
            msg = u"-gdocid %d- " % (self.current_gdocid) + msg
        return msg

    def set_gdocid(self, gdocid):
        self.gdocids.append(gdocid)
        self.current_gdocid = gdocid
        self.reportdata.append([])

    def send_report(self):
        for gdocid, data in zip(self.gdocids, self.reportdata):
            for row in data:
                print "    " + row.replace("\n", "\n    ")
            print "--------------------"
        pass

def convert_add(src_add, dst_add):
    if type(src_add) == dict:
        src_type = src_add["type"]
        src_val = src_add["value"]
    else:
        src_type = src_add.type
        src_val = src_add.value

    if type(dst_add) == dict:
        dst_type = dst_add["type"]
    else:
        dst_type = dst_add.type
    func = ADDCONVERTER[src_type][dst_type]
    if func is None:
        raise Exception("convert_add: unconvertable types")
    else:
        return func(src_val)



class SyncResponseHandler():
    def handle_respAuth(self, webclient, response):
        webclient.userid = response.userid
        webclient.loggedin = True
        return (True, response.userid)

    def handle_default(self, client, response):
        return response.data

    def handle_respGetBookmarks(self, client, response):
        return response.data

    def handle_respAddfields(self, client, response):
        fields = response.data
        addfields = {}
        for field in fields:
            addfields[field["id"]] = {"name": field["name"], "add": field["add"],
                                      "indextableid": field["indextableid"], "type": field["type"]}
            arcs = []
            for arc in field['rights']["_rights"]:
                try:
                    arcs.append({"arcid": arc["_BitfarmFieldRight__arcid"],
                                 "fieldpos": arc["_BitfarmFieldRight__fieldpos"]})
                except:
                    pass
            addfields[field["id"]]["arcs"] = arcs
        return addfields

    def handle_respAuthError(self, webclient, response):
        return (False, response.data)

    def handle_respDenied(self, webclient, response):
        return False

    def handle_respOK(self, client, response):
        return True

    def handle_respSendFile(self, client, response):
        return True

    def handle_respArchives(self, webclient, response):
        arcs = {}
        statefields = {}
        addfields = {}
        sf_keys = ["addfieldid", "name",
                   "_BitfarmStatefield__bitmask", "indexcolumnid", "pos", "id"]
        sf_mapping = ["addfieldid", "name",
                      "bitmask", "indexcolumnid", "pos", "id"]
        af_keys = ['indextableid', 'name', 'pos',
                   'add', 'mode', 'type', 'options', "rights"]
        af_mapping = ['indextableid', 'name',
                      'pos', 'add', 'mode', 'type', 'options', "rights"]
        for arc in response.data:
            #print arc["arcname"]
            # pprint.pprint(arc)
            if arc["enabled"] == u"enabled":
                arcs[arc["id"]] = {"name": arc["arcname"], "lgrid": arc["lgrid"],
                                   "type": arc["type"], "statefields": [], "addfields": [], "rights": arc["rights"]['_BitfarmRights__rights']}
                if "statefields" in arc:
                    for sf in arc["statefields"]:
                        mysf = {sf_mapping[sf_keys.index(
                            k)]: v for k, v in sf.items() if k in sf_keys}
                        arcs[arc["id"]]["statefields"].append(mysf)
                        if sf["id"] not in statefields:
                            mysf["arcrights"] = {arc["id"]:0}
                            statefields[sf["id"]] = mysf
                        for right in sf["rights"]["_rights"]:
                            if arc["id"] not in statefields[sf["id"]]["arcrights"]:
                                statefields[sf["id"]]["arcrights"][arc["id"]] = 0
                            statefields[sf["id"]]["arcrights"][arc["id"]] += right["_BitfarmFieldRight__rights"]
                if "addfields" in arc:
                    #print arc["addfields"]
                    for af in arc["addfields"]:
                        myaf = {af_mapping[af_keys.index(
                            k)]: v for k, v in af.items() if k in af_keys}
                        arcs[arc["id"]]["addfields"].append(myaf)
                        if af["add"] not in addfields:
                            myaf["arcrights"] = {arc["id"]:0}
                            addfields[af["add"]] = myaf
                            addfields[af["add"]]["arcids"] = [arc["id"]]
                        else:
                            addfields[af["add"]]["arcids"].append(arc["id"])
                        for right in af["rights"]["_rights"]:
                            if arc["id"] not in addfields[af["add"]]["arcrights"]:
                                addfields[af["add"]]["arcrights"][arc["id"]] = 0
                            addfields[af["add"]]["arcrights"][arc["id"]] += right["_BitfarmFieldRight__rights"]
        for stsid in statefields:
            statefields[stsid]["arcids"] = []
            for arcid, arc in arcs.items():
                for sf in arc["statefields"]:
                    if sf["id"] == stsid:
                        statefields[stsid]["arcids"].append(arcid)
        return (arcs, statefields, addfields)

    def handle_respEditDenied(self, webclient, response):
        userid = response.userid
        return (False, userid)
    
    def handle_respAccess(self, webclient, response):
        #print response
        doc = response.data
        return (True, doc)

    def handle_respSearch(self, webclient, response):
        if response.data is None:
            return []
        else:
            return response.data
        
    def handle_respError(self, client, response):
        return (False, response.data)

def _(value):
    return value


class ConsoleStreamHandler(logging.StreamHandler):
    # def __init__(self, *args, **kwargs):
    #     logging.StreamHandler.__init__(self, *args, **kwargs)

    def format(self, record):
        record.msg = record.msg.encode("cp1252")
        logging.StreamHandler.format(self, record)

    # def emit(self, record):
    #     msg = self.format(record)
    #     logging.StreamHandler.emit(self, msg.encode("cp1252"))


class MyConfigParser(ConfigParser.RawConfigParser):
    """Modified ConfigParser that allow ':' in keys and only '=' as separator.
    """
    OPTCRE = re.compile(
        r'(?P<option>[^=\s][^=]*)'          # allow only =
        r'\s*(?P<vi>[=])\s*'                # for option separator
        r'(?P<value>.*)$'
    )


class BfaSync():
    def __init__(self, configfile=None, log=None):
        self.configfile = configfile
        self.bfapath = os.getenv("bitfarm-archiv")
        self.bindir = os.path.join(self.bfapath, "bin")
        if self.configfile is None:
            self.configfile = os.path.join(self.bindir, "bfa_sync.ini")
        if not os.path.isfile(self.configfile):
            raise Exception(u"configfile not found!")
        else:
            self.config = MyConfigParser()
            # self.config.optionxform = unicode
            with codecs.open(self.configfile, "rb", encoding="utf8") as f:
                self.config.readfp(f)

        if log is not None:
            self._log = log
        else:
            self._log = logging.getLogger("bfa_sync")

            
            ch = logging.StreamHandler(sys.stdout)
            #ch.setLevel(logging.DEBUG)
            formatter = logging.Formatter(
                u'%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            self._log.addHandler(ch)

            trfh = logging.handlers.RotatingFileHandler(self.config.get("main", "logfile"), backupCount=self.config.getint("main", "logbackups"))
            trfh.setFormatter(formatter)
            self._log.addHandler(trfh)
            self._log.propagate = False
            self._log.handlers[1].doRollover()
            LOGLEVEL = {"DEBUG": logging.DEBUG, "INFO": logging.INFO, "ERROR": logging.ERROR}
            self._log.setLevel(LOGLEVEL[self.config.get("main", "loglevel")])
        self.log = LogReport(self._log, self.config)
        if "logserver" not in self.config.options("main"):
            self.config.set("main", "logserver", False)
        if self.config.getboolean("main", "logserver") is True:
            _set_logger(self._log)
        atexit.register(self.quit)
        self.log.info("bfa_sync started")
        self.client = None
        self.dst_clients = {}
        self.reportdata = {}
        self.configs_checked = {}

    def quit(self):
        pass
        #self.log.send_report()

    def connect_client(self, profile, username, userpass, loadmeta=True):
        self.log.info(u"connecting user %s (profile %s)" %(username, profile))
        client = BfaServerClient(profile)
        client.responsehandler = SyncResponseHandler()
        try:
            conresult = client.connect(username=username, userpass=userpass)
        except Exception as e:
            self.log.error("error connecting bfaserver\r\n"+str(e))
            raise Exception("error connecting bfaserver\r\n"+e)

        if conresult[0] is True:
            self.log.info(u"user %s (profile %s) connected" %(username, profile))
            client.userid = conresult[1]
            client.username = username
            client.sock.settimeout(5)
        else:
            self.log.error(u"user %s (profile %s) failed to connect" %(username, profile))
            raise Exception("login error")
        if loadmeta is True:
            client.meta = self.load_server_meta(client)
            # arcs, statefields, addfields = client.getReadArchives(asjson=True)
            # client.arcs = arcs
            # client.statefields = statefields
            # client.addfields = addfields
        return client

    def load_server_meta(self, client):
        # warte-Cursor anzeigen
        if client is not None:
            try:
                arcs, statefields, addfields = client.getReadArchives(
                    asjson=True)
                bookmarks = client.getBookmarks(
                    globale=False, locale=True, history=False, asjson=False)
            except Exception as e:
                self.usermessage(_(u"keine Verbindung zum Server"))
                return None
        else:
            # self.usermessage(_(u"keine Verbindung zum Server"))
            return None
        # normalen cursor anzeigen
        meta = {}
        meta["archives"] = arcs
        meta["addfields"] = addfields
        meta["statefields"] = statefields
        meta["bookmarks"] = bookmarks
        if meta["bookmarks"] is None:
            meta["bookmarks"] = []

        return meta

    def check_config_new(self, client):
        # srcclient -> angemeldete bfaServerClient-Instanz (section main)
        # dstclients -> dict mit key section-Name (dst_\d+) -> value angemeldete bfaServerClient-Instanz des Zielsystems
        self.log.debug("starting config check")
        configerrors = {}
        if client.section == "main":
            check_sections = [
                x for x in self.config.sections() if re.match("dst_\d+_mapping", x)]
            for csect in check_sections:
                # Quellserver Check
                configerrors[csect] = {}
                
                src_arcs = [((csect, x), self.config.getint(csect, x))
                            for x in self.config.options(csect) if re.match("src_arc_\d+", x)]
                meta = client.meta
                bookmarks = [(csect, x) for x in self.config.options(csect) if re.match("bookmark_\d+", x)]
                sectionbmids = [self.config.get(x[0], x[1]) for x in bookmarks]
                bmidlist = [x.id for x in meta["bookmarks"]]
                bookmarks_missg = [x[1] for x in bookmarks]

                # arcid muss vorhanden sein, Lesen- und Suchen-Recht muss gegeben sein
                arcs_missing = [x
                                for x in src_arcs if x[1] not in meta["archives"]]
                arcs_missing_right = [x for x in src_arcs if x[1] in meta["archives"] and x[1] not in arcs_missing and meta["archives"][x[1]]["rights"] & (
                    RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]) != RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]]
                arcids_missing = [x[1] for x in arcs_missing]
                arcids_missing_right = [x[1] for x in arcs_missing_right]
                configerrors[csect] = {"arcs_missing":arcs_missing, "arcs_missing_right":arcs_missing_right}
                configerrors[csect]["meta"] = {}
                for ((subsection, option), arcid) in src_arcs:
                    
                    if arcid in arcids_missing+arcids_missing_right:
                        continue
                    configerrors[csect]["meta"][arcid] = {}
                    src_index = option[-4:]
                    # Statusfelder mit lesendem Zugriff ermitteln
                    sts_need_read = [((csect, x), re.findall("%sts_\d+%", self.config.get(csect, x))[0][1:-1]) for x in self.config.options(
                        csect) if re.match("meta_"+src_index+"_dst_sts", x) and len(re.findall("%sts_\d+%", self.config.get(csect, x))) > 0]
                    # Statusfelder mit schreibendem Zugriff ermitteln
                    sts_need_write = [((csect, x), re.findall("meta_"+src_index+"_src_(sts_\d+)", x)[0])
                        for x in self.config.options(csect) if re.match("meta_"+src_index+"_src_sts_\d+", x)]
                    # Zusatzfelder mit lesendem Zugriff ermitteln
                    add_need_read = [((csect, x), re.findall("%add_\d+%", self.config.get(csect, x))[0][1:-1]) for x in self.config.options(
                        csect) if re.match("meta_"+src_index+"_dst_add", x) and len(re.findall("%add_\d+%", self.config.get(csect, x))) > 0]
                    # Zusatzfelder mit schreibendem Zugriff ermitteln
                    add_need_write = [((csect, x), re.findall("meta_"+src_index+"_src_(add_\d+)", x)[0])
                        for x in self.config.options(csect) if re.match("meta_"+src_index+"_src_add_\d+", x)]

                    # fehlende Statusfelder (nur aktuelles Archiv!):
                    sts_missing = [x for x in sts_need_read + sts_need_write if int(
                        x[1][4:]) not in meta["statefields"] or arcid not in meta["statefields"][int(x[1][4:])]["arcids"]]
                    # Statusfelder mit fehlendem Leserecht
                    sts_missing_read = []
                    for ((sts_sect, sts_opt), sts) in sts_need_read:
                        if sts not in [x[1] for x in sts_missing]:
                            stsid = int(sts[4:])
                            if arcid not in meta["statefields"][stsid]["arcrights"]:
                                sts_missing_read.append(
                                    ((sts_sect, sts_opt), sts))
                            else:
                                sts_rights = meta["statefields"][stsid]["arcrights"][arcid]
                                if RIGHTS_BITMASKS["field"]["READ"] != RIGHTS_BITMASKS["field"]["READ"] & sts_rights and \
                                        RIGHTS_BITMASKS["field"]["WRITE"] != RIGHTS_BITMASKS["field"]["WRITE"] & sts_rights and \
                                        RIGHTS_BITMASKS["field"]["MODIFY"] != RIGHTS_BITMASKS["field"]["MODIFY"] & sts_rights: \
                                    sts_missing_read.append(((sts_sect, sts_opt), sts))

                    sts_missing_write = []
                    for ((sts_sect, sts_opt), sts) in sts_need_write:
                        if sts not in [x[1] for x in sts_missing]:
                            stsid = int(sts[4:])
                            if arcid not in meta["statefields"][stsid]["arcrights"]:
                                sts_missing_write.append(((sts_sect, sts_opt), sts))
                            else:
                                sts_rights = meta["statefields"][stsid]["arcrights"][arcid]
                                if RIGHTS_BITMASKS["field"]["WRITE"] != RIGHTS_BITMASKS["field"]["WRITE"] & sts_rights and \
                                        RIGHTS_BITMASKS["field"]["MODIFY"] != RIGHTS_BITMASKS["field"]["MODIFY"] & sts_rights: \
                                    sts_missing_write.append(((sts_sect, sts_opt), sts))

                    add_missing = [x for x in add_need_read + add_need_write if x[1] not in meta["addfields"] or arcid not in meta["addfields"][x[1]]["arcids"]]

                    add_missing_read = []
                    for ((add_sect, add_opt), addid) in add_need_read:
                        if addid not in [x[1] for x in add_missing]:
                            if arcid not in meta["addfields"][addid]["arcids"]:
                                add_missing_read.append(((add_sect, add_opt), addid))
                            else:
                                add_rights = meta["addfields"][addid]["arcrights"][arcid]
                                if RIGHTS_BITMASKS["field"]["READ"] != RIGHTS_BITMASKS["field"]["READ"] & add_rights and \
                                        RIGHTS_BITMASKS["field"]["WRITE"] != RIGHTS_BITMASKS["field"]["WRITE"] & add_rights and \
                                        RIGHTS_BITMASKS["field"]["MODIFY"] != RIGHTS_BITMASKS["field"]["MODIFY"] & add_rights: \
                                    add_missing_read.append(((add_sect, add_opt), addid))

                    add_missing_write = []
                    for ((add_sect, add_opt), addid) in add_need_write:
                        if addid not in [x[1] for x in add_missing]:
                            if arcid not in meta["addfields"][addid]["arcids"]:
                                add_missing_write.append(((add_sect, add_opt), addid))
                            else:
                                add_rights = meta["addfields"][addid]["arcrights"][arcid]
                                if RIGHTS_BITMASKS["field"]["WRITE"] != RIGHTS_BITMASKS["field"]["WRITE"] & add_rights and \
                                        RIGHTS_BITMASKS["field"]["MODIFY"] != RIGHTS_BITMASKS["field"]["MODIFY"] & add_rights: \
                                    add_missing_write.append(((add_sect, add_opt), addid))

                    configerrors[csect]["meta"][arcid]["sts_missing"] = sts_missing
                    configerrors[csect]["meta"][arcid]["sts_missing_read"] = sts_missing_read
                    configerrors[csect]["meta"][arcid]["sts_missing_write"] = sts_missing_write
                    configerrors[csect]["meta"][arcid]["add_missing"] = add_missing
                    configerrors[csect]["meta"][arcid]["add_missing_read"] = add_missing_read
                    configerrors[csect]["meta"][arcid]["add_missing_write"] = add_missing_write

        else:
            # Zielserver
            configerrors = {}
            
            mapping_section = client.section + "_mapping"
            configerrors[mapping_section] = {}
            
            dst_arcs = [((mapping_section, x), self.config.getint(mapping_section, x))
                for x in self.config.options(mapping_section) if re.match("dst_arc_\d+", x)]
            meta = client.meta
            
        

            # arcid muss vorhanden sein, Lesen- und Suchen-Recht muss gegeben sein
            arcs_missing = [x for x in dst_arcs if x[1] not in meta["archives"]]

            arcs_missing_right = [x for x in dst_arcs if x[1] in meta["archives"] and x[1] not in arcs_missing and meta["archives"][x[1]]["rights"] & (
                    RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]) != RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]]
            arcids_missing = [x[1] for x in arcs_missing]
            arcids_missing_right = [x[1] for x in arcs_missing_right]

            
            configerrors[mapping_section] = {"arcs_missing":arcs_missing, "arcs_missing_right":arcs_missing_right, "meta":{}}

            for ((subsection, option), arcid) in dst_arcs:
                # bookmarks = [(csect, x) for x in self.config.options(csect) if re.match("bookmark_\d+", x)]
                # sectionbmids = [self.config.get(x[0], x[1]) for x in bookmarks]
                # bmidlist = [x.id for x in meta["bookmarks"]]
                # bookmarks_missg = [x[1] for x in bookmarks]
                sts_missing = []
                sts_missing_read = []
                sts_missing_write = []
                add_missing = []
                add_missing_read = []
                add_missing_write = []
                if arcid in arcids_missing+arcids_missing_right:
                    continue
                configerrors[mapping_section]["meta"][arcid] = {}
                dst_index = option[-4:]
                # Zusatzfelder ermitteln
                add_need_write = [((mapping_section, x), re.findall("meta_"+dst_index+"_dst_(add_\d+)", x)[0])
                    for x in self.config.options(mapping_section) if re.match("meta_" + dst_index + "_dst_add_\d+", x)]
                # fehlende Zusatzfelder
                add_missing = [x for x in add_need_write if x[1] not in meta["addfields"] or arcid not in meta["addfields"][x[1]]["arcids"]]
                # Zusatzfelder mit unzureichenden Berechtigungen
                add_missing_write = []
                for ((add_sect, add_opt), addid) in add_need_write:
                    if addid not in [x[1] for x in add_missing]:
                        if arcid not in meta["addfields"][addid]["arcids"]:
                            add_missing_write.append(((add_sect, add_opt), addid))
                        else:
                            add_rights = meta["addfields"][addid]["arcrights"][arcid]
                            if RIGHTS_BITMASKS["WRITE"] != RIGHTS_BITMASKS["WRITE"] & add_rights and \
                                    RIGHTS_BITMASKS["MODIFY"] != RIGHTS_BITMASKS["MODIFY"] & add_rights: \
                                add_missing_write.append(((add_sect, add_opt), addid))
                # Statusfelder ermitteln
                sts_need_write = [((mapping_section, x), re.findall("meta_"+dst_index+"_dst_(sts_\d+)", x)[0])
                    for x in self.config.options(mapping_section) if re.match("meta_" + dst_index + "_dst_sts_\d+", x)]
                # fehlende Statusfelder
                sts_missing = [x for x in sts_need_write if int(
                        x[1][4:]) not in meta["statefields"] or arcid not in meta["statefields"][int(x[1][4:])]["arcids"]]

                sts_missing_write = []
                for ((sts_sect, sts_opt), sts) in sts_need_write:
                    if sts not in [x[1] for x in sts_missing]:
                        stsid = int(sts[4:])
                        if arcid not in meta["statefields"][stsid]["arcrights"]:
                            sts_missing_write.append(((sts_sect, sts_opt), sts))
                        else:
                            sts_rights = meta["statefields"][stsid]["arcrights"][arcid]
                            if RIGHTS_BITMASKS["WRITE"] != RIGHTS_BITMASKS["WRITE"] & sts_rights and \
                                    RIGHTS_BITMASKS["MODIFY"] != RIGHTS_BITMASKS["MODIFY"] & sts_rights: \
                                sts_missing_write.append(((sts_sect, sts_opt), sts))

                
                
                configerrors[subsection]["meta"][arcid]["sts_missing"] = sts_missing
                configerrors[subsection]["meta"][arcid]["sts_missing_read"] = sts_missing_read
                configerrors[subsection]["meta"][arcid]["sts_missing_write"] = sts_missing_write
                configerrors[subsection]["meta"][arcid]["add_missing"] = add_missing
                configerrors[subsection]["meta"][arcid]["add_missing_read"] = add_missing_read
                configerrors[subsection]["meta"][arcid]["add_missing_write"] = add_missing_write
            

        return configerrors

    def check_config(self, meta, client):
        section = client.section
        if section == "main":  # Quell-Client -> src-archive, addfields und statefields jeder mapping section prüfen
            check_sections = [
                x for x in self.config.sections() if re.match("dst_\d+_mapping", x)]
            for csect in check_sections:
                arcs = [((csect, x), self.config.getint(csect, x))
                        for x in self.config.options(csect) if re.match("src_arc_\d+", x)]
                arcid = arcs[0][1]
                # Zusatzfelder, die auf der Quelle gesetzt werden
                adds = [((csect, x), re.findall("meta_\d+_src_(add_\d+)", x)[0])
                        for x in self.config.options(csect) if re.match("meta_\d+_src_add_\d+", x)]
                # Quell-Zusatzfelder, deren Wert für ein Ziel-Zusatzfeld benötigt wird
                adds.extend([((csect, x), re.findall("%add_\d+%", self.config.get(csect, x))[0][1:-1]) for x in self.config.options(
                    csect) if re.match("meta_\d+_dst_add", x) and len(re.findall("%add_\d+%", self.config.get(csect, x))) > 0])
                # Statusfelder, die auf der Quelle gesetzt werden
                sts = [((csect, x), re.findall("meta_\d+_src_(sts_\d+)", x)[0])
                       for x in self.config.options(csect) if re.match("meta_\d+_src_sts_\d+", x)]
                # Quell-Statusfelder, deren Wert für ein Zielstatus benötigt wird
                sts.extend([((csect, x), re.findall("%sts_\d+%", self.config.get(csect, x))[0][1:-1]) for x in self.config.options(
                    csect) if re.match("meta_\d+_dst_sts", x) and len(re.findall("%sts_\d+%", self.config.get(csect, x))) > 0])
                missing_arcs = [x[1] for x in arcs if x[1] not in meta["archives"] or meta["archives"][x[1]]["rights"] & (
                    RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]) != RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]]
                arcs_configpaths = [x[0] for x in arcs if x[1]
                                    not in meta["archives"] or x[1] in missing_arcs]
                missing_adds = [x[1] for x in adds if x[1] not in meta["addfields"]
                                or arcid not in meta["addfields"][x[1]]["arcids"]]
                adds_configpaths = [x[0] for x in adds if x[1] not in meta["addfields"]
                                    or arcid not in meta["addfields"][x[1]]["arcids"]]
                missing_sts = [x[1] for x in sts if int(
                    x[1][4:]) not in meta["statefields"] or arcid not in meta["statefields"][int(x[1][4:])]["arcids"]]
                sts_configpaths = [x[0] for x in sts if int(
                    x[1][4:]) not in meta["statefields"] or arcid not in meta["statefields"][int(x[1][4:])]["arcids"]]
                # print missing_arcs
                # print missing_adds
                # print missing_sts
                # prints
        else:
            csect = section+"_mapping"
            arcs = [((csect, x), self.config.getint(csect, x))
                    for x in self.config.options(csect) if re.match("dst_arc_\d+", x)]
            arcid = arcs[0][1]
            adds = [((csect, x), re.findall("meta_\d+_dst_(add_\d+)", x)[0])
                    for x in self.config.options(csect) if re.match("meta_\d+_dst_add_\d+", x)]

            sts = [((csect, x), re.findall("meta_\d+_dst_(sts_\d+)", x)[0])
                   for x in self.config.options(csect) if re.match("meta_\d+_dst_sts_\d+", x)]

            missing_arcs = [x[1] for x in arcs if x[1] not in meta["archives"] or meta["archives"][x[1]]["rights"] & (
                RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["IMPORT"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]) != RIGHTS_BITMASKS["READ"]+RIGHTS_BITMASKS["IMPORT"]+RIGHTS_BITMASKS["LIST"]+RIGHTS_BITMASKS["MODIFY"]]
            arcs_configpaths = [x[0] for x in arcs if x[1]
                                not in meta["archives"] or x[1] in missing_arcs]
            missing_adds = [x[1] for x in adds if x[1] not in meta["addfields"]
                            or arcid not in meta["addfields"][x[1]]["arcids"]]
            adds_configpaths = [x[0] for x in adds if x[1] not in meta["addfields"]
                                or arcid not in meta["addfields"][x[1]]["arcids"]]
            missing_sts = [x[1] for x in sts if int(
                x[1][4:]) not in meta["statefields"] or arcid not in meta["statefields"][int(x[1][4:])]["arcids"]]
            sts_configpaths = [x[0] for x in sts if int(
                x[1][4:]) not in meta["statefields"] or arcid not in meta["statefields"][int(x[1][4:])]["arcids"]]
            # print missing_arcs
            # print missing_adds
            # print missing_sts
            # print
        msg = u""
        result = {}
        if len(missing_arcs) > -1:
            result["missing_arcs"] = missing_arcs
            result["missing_arcs_options"] = arcs_configpaths
            # msg += _(u"Archiv-IDs: ") + ", ".join(missing_arcs) +"\n\n"
        if len(missing_adds) > -1:
            result["missing_adds"] = missing_adds
            result["missing_adds_options"] = adds_configpaths
            # msg += _(u"Zusatzfeld-IDs: ") + ", ".join(missing_adds) +"\n\n"
        if len(missing_sts) > -1:
            result["missing_sts"] = missing_sts
            result["missing_sts_options"] = sts_configpaths
            # msg += _(u"Statusfeld-IDs: ") + ", ".join(missing_sts) +"\n\n"
        #
        return result

    def get_sync_meta(self, doc, section, sub_section):
        # ermittelt Metadaten für die Synchronisation des Dokuments mit der angegebenen gdocid und section/sub_section
        # Ergebnis wird als dict zurückgeliefert
        self.log.debug(u"get_sync_meta called")
        sync_tstamp = datetime.now()
        sync_tstamp_str = sync_tstamp.strftime('%Y-%m-%d %H:%M:%S')

        profile = self.config.get(section, "profile")
        # # client Verbindungen herstellen, falls noch nicht vorhanden
        # if profile not in self.dst_clients:
        #     try:
        #         bfauser = decryptData(base64.b64decode(
        #             self.config.get(section, "bfauser")))
        #         bfapass = decryptData(base64.b64decode(
        #             self.config.get(section, "bfapass")))
        #     except:
        #         raise Exception("error decrypting user/password")
        #     self.dst_clients[profile] = self.connect_client(
        #         profile, bfauser, bfapass)
        src_addfields = {}
        src_stsfields = {}
        dst_addfields = {}
        dst_stsfields = {}

        mapping_section = section + "_mapping"
        # self.client.section = "main"
        # src_errors = self.check_config(self.client.meta, self.client)
        # errorcount = sum(
        #     [len(value) for key, value in src_errors.items() if not key.endswith("_options")])

        # self.dst_clients[profile].section = section
        # dst_errors = self.check_config(
        #     self.dst_clients[profile].meta, self.dst_clients[profile])
        # errorcount += sum([len(value) for key,
        #                    value in dst_errors.items() if not key.endswith("_options")])

        # if errorcount > 0:
        #     # TODO: handle config errors
        #     self.log.error("Konfigurationsfehler!\n%s\n%s" % (pformat(dst_errors), pformat(src_errors)))
        #     raise Exception("Konfigurationsfehler!\n%s\n%s" % (pformat(dst_errors), pformat(src_errors)))

        pattern = "(meta_" + sub_section + "_src_(add_\d+))"
        src_adds = {re.match(pattern, x).group(2): self.config.get(mapping_section, re.match(pattern, x).group(1))
                    for x in self.config.options(mapping_section) if re.match(pattern, x) is not None}

        pattern = "(meta_" + sub_section + "_dst_(add_\d+))"
        dst_adds = {re.match(pattern, x).group(2): self.config.get(mapping_section, re.match(pattern, x).group(1))
                    for x in self.config.options(mapping_section) if re.match(pattern, x) is not None}

        pattern = "(meta_" + sub_section + "_src_(sts_\d+))"
        src_sts = {re.match(pattern, x).group(2): self.config.get(mapping_section, re.match(pattern, x).group(1))
                   for x in self.config.options(mapping_section) if re.match(pattern, x) is not None}

        pattern = "(meta_" + sub_section + "_dst_(sts_\d+))"
        dst_sts = {re.match(pattern, x).group(2): self.config.get(mapping_section, re.match(pattern, x).group(1))
                   for x in self.config.options(mapping_section) if re.match(pattern, x) is not None}

        dst_arcid = self.config.getint(
            mapping_section, "dst_arc_"+sub_section)

        # Zusatzfeldwerte für Zieldokument ermitteln
        self.log.info(u"dst-document addfields:")
        for dst_addid, rawval in dst_adds.items():

            # Zielzusatzfeld ermitteln
            if dst_addid not in self.dst_clients[profile].meta["addfields"]:
                self.log.error(
                    u"addfield %s unknown in profile %s!" % (dst_addid, profile))
                continue

            dstadd = self.dst_clients[profile].meta["addfields"][dst_addid]

            realval = rawval
            self.log.debug(_(u"addfield: '%s' (type '%s'), config-value: '%s'" %
                             (dstadd["name"], ADDTYPES[dstadd["type"]], rawval)))
            if rawval == "%timestamp%":
                if dstadd["type"] in (1, 12):  # Datumsfeld/DLM-Datumsfeld
                    realval = sync_tstamp
                # universal/Auswahl/Mehrfachauswahl
                elif dstadd["type"] in (0, 2, 3, 11):
                    realval = sync_tstamp_str
                else:
                    self.log.error("type error for addfield '%s'. cannot map timestamp to field type '%s'." % (
                        dst_addid, ADDTYPES[dstadd["type"]]))
                    continue
            elif rawval == "%srcgdocid%":
                if dstadd["type"] == 6:  # Ganzzahl
                    realval = doc.gdocid
                # universal/Auswahl/Mehrfachauswahl
                elif dstadd["type"] in (0, 2, 3, 11):
                    realval = str(doc.gdocid)
                else:
                    self.log.error("type error for field %s. cannot map gdocid to field type %s." % (
                        dst_addid, ADDTYPES[dstadd["type"]]))
                    continue

            elif "%srcgdocid%" in rawval:
                if dstadd["type"] in (0, 2, 3, 11):
                    realval = realval.replace("%srcgdocid%", str(doc.gdocid))
                else:
                    self.log.error("type error for field %s. cannot map gdocid with string to field type %s." % (
                        dst_addid, ADDTYPES[dstadd["type"]]))
                    continue
            if "%srcuser%" in rawval or "%dstuser%" in rawval:
                if "%srcuser%" in rawval:
                    username = self.client.username
                    realval = realval.replace("%srcuser%", username)
                if "%dstuser%" in rawval:
                    username = self.dst_clients[profile].username
                    realval = realval.replace("%dstuser%", username)
                # universal/Auswahl/Mehrfachauswahl
                if dstadd["type"] not in (0, 2, 3, 11):
                    self.log.error("type error for field %s. cannot map username to field type %s." % (
                        dst_addid, ADDTYPES[dstadd["type"]]))
                    continue
            elif re.match("^%add_\d+%$", rawval):  # Zusatzfeldwert exklusiv

                add_id = re.findall("%(add_\d+)%", realval)[0]
                if add_id not in self.client.meta["addfields"]:
                    self.log.error(
                        "field %s unknown for profile %s" % (add_id, profile))
                    continue
                elif doc.arcid not in self.client.meta["addfields"][add_id]["arcids"]:
                    self.log.error(
                        "field %s ('%s') unknown in archive %d ('%s')" % (add_id, self.client.meta["addfields"][add_id]["name"], doc.arcid, self.client.meta["archives"][doc.arcid]["name"]))
                    continue
                doc_add = doc.fields[add_id]
                self.log.debug("mapping field '%s' (type '%s') to field '%s' (type '%s')" % (
                    doc_add.name, ADDTYPES[doc_add.type], dstadd["name"], ADDTYPES[dstadd["type"]]))
                try:
                    try:
                        self.log.debug("source field-value '%s'" % str(doc_add.value))
                    except:
                        try:
                            self.log.debug("source field-value '%s'" % unicode(doc_add.value))
                        except:
                            pass
                    realval = convert_add(doc_add, dstadd)
                except Exception as e:
                    self.log.error(u"source addfield '%s' (type '%s') and destination addfield '%s' (type '%s') incompatible types!" % (
                        doc_add.name, ADDTYPES[doc_add.type], dstadd["name"], ADDTYPES[dstadd["type"]]))
                    continue
                self.log.debug(
                    u"value conversion success. New value: %s" % realval)
            else:   # zusammengesetzter Wert, kann dann nur in universal/Auswahl/Mehrfachauswahl/DLM-Auswahl geschrieben werden
                add_ids = re.findall("%(add_\d+)%", realval)
                add_error = False
                for add_id in add_ids:
                    if add_id not in self.client.meta["addfields"]:
                        self.log.error(
                            "field %s unknown for profile %s" % (add_id, profile))
                        add_error = True
                        break
                    doc_add = doc.fields[add_id]
                    try:
                        addval = convert_add(doc_add, {"type": 0})
                    except Exception as e:
                        self.log.error(u"destination addfield '%s' (type '%s') error converting" % (
                            doc_add.name, ADDTYPES[doc_add.type]))
                        add_error = True
                        break
                    if addval is None:
                        addval = ""
                    realval = realval.replace("%"+add_id+"%", addval)
                if add_error is True:
                    continue

            dst_addfields[dst_addid] = realval
        # print "Zusatzfelder Ziel:"
        # print dst_addfields

        # Statusfeldwerte für Zieldokument ermitteln
        self.log.debug(u"dst-document statefields:")
        for dst_stsid, rawval in dst_sts.items():
            # Ziel-Statusfeld ermitteln
            stsid = int(dst_stsid.split("_")[1])
            if stsid not in self.dst_clients[profile].meta["statefields"]:
                self.log.error(
                    u"statefield %s unknown in profile %s!" % (dst_stsid, profile))
                continue

            dststs = self.dst_clients[profile].meta["statefields"][stsid]
            self.log.debug(u"statefield: '%s', config-value: '%s'" %(dststs["name"], rawval))
            if rawval in ("True", "False"):
                dst_stsfields[stsid] = bool(rawval)
                self.log.debug(u"new value: %s" %rawval)
            elif re.match("^%sts_\d+%$", rawval):
                src_stsid = int(re.findall("^%sts_(\d+)%$", rawval)[0])
                src_val = None
                for stsname, sts in doc.statefields.items():
                    if sts.id == src_stsid:
                        src_val = doc.statefields[stsname].value
                        break
                if src_val is not None:
                    dst_stsfields[stsid] = src_val
                    self.log.debug(u"source statefield: '%s', value: %s" % (stsname, str(src_val)))
                    self.log.debug(u"new value: %s" %str(src_val))
        # print "Statusfelder Ziel:"
        # print dst_stsfields

        # Zusatzfeldwerte für Quelldokument ermitteln
        self.log.debug(u"source-document addfields")
        for src_addid, rawval in src_adds.items():
            #print src_addid, rawval
            realval = None
            
            self.log.debug(_(u"addfield: '%s' (type '%s'), config-value: '%s'" %
                             (doc.fields[src_addid].name, ADDTYPES[doc.fields[src_addid].type], rawval)))
            if rawval == "%timestamp%":
                # Datumsfeld/DLM-Datumsfeld
                if doc.fields[src_addid].type in (1, 12):
                    realval = sync_tstamp
                # universal/Auswahl/Mehrfachauswahl
                elif doc.fields[src_addid].type in (0, 2, 3, 11):
                    realval = sync_tstamp_str
                else:
                    self.log.error("type error for addfield '%s'. cannot map timestamp to field type '%s'." % (
                        src_addid, ADDTYPES[doc.fields[src_addid].type]))
                    continue
            elif rawval == "%srcuser%" or rawval == "%dstuser%":
                if rawval == "%srcuser%":
                    realval = self.client.username
                if rawval == "%dstuser%":
                    realval = self.dst_clients[profile].username
                # universal/Auswahl/Mehrfachauswahl/DLM-Auswahl
                if doc.fields[src_addid].type not in (0, 2, 3, 11):
                    self.log.error("type error for field %s. cannot map username to field type %s." % (
                        src_addid, ADDTYPES[dstadd["type"]]))
                    continue
            elif rawval == r"%dstgdocid%":
                if doc.fields[src_addid].type in (0, 2, 3, 11):
                    realval = rawval + "_as_str"
                    # Wert muss nach erfolgter synchronisation ermittelt und gesetzt werden!
                elif doc.fields[src_addid].type == 6:
                    # Wert muss nach erfolgter synchronisation ermittelt und gesetzt werden!
                    realval = rawval+"_as_int"
                else:
                    self.log.error("type error for field %s. cannot map gdocid to field type %s." % (
                        src_addid, ADDTYPES[dstadd["type"]]))
                    continue
            elif rawval == r"%srcgdocid%":
                if doc.fields[src_addid].type in (0, 2, 3, 11):
                    realval = str(doc.gdocid)
                elif doc.fields[src_addid].type == 6:
                    realval = doc.gdocid
                else:
                    self.log.error("type error for field %s. cannot map gdocid to field type %s." % (
                        src_addid, ADDTYPES[dstadd["type"]]))
                    continue
            else:  # fester Wert -> kann nur in Feldtypen gemappt werden, die strings/unicode entgegennehmen
                if doc.fields[src_addid].type in (0, 2, 3, 11):
                    realval = rawval
                else:
                    self.log.error("type error for field %s. cannot map username to field type %s." % (
                        src_addid, ADDTYPES[dstadd["type"]]))
            self.log.debug(u"new value: %s" %str(realval))
            src_addfields[src_addid] = realval
        # print "Zusatzfelder Quelle:"
        # print src_addfields

        # Statusfeldwerte für Zieldokument ermitteln
        self.log.debug(u"source-document statefields")
        for src_stsid, rawval in src_sts.items():
            sf = None
            for stsname, sts in doc.statefields.items():
                    if sts.id == src_stsid:
                        sf = doc.statefields[stsname]
                        break
            if sf is not None:
                self.log.debug(u"statefield: '%s', config-value: '%s'" %(sf.name, rawval))
                
            stsid = int(src_stsid.split("_")[1])
            self.log.debug(u"new value: %s" %rawval)
            src_stsfields[stsid] = bool(rawval)

        # print "Statusfelder Quelle:"
        # print src_stsfields

        docloc = None
        oriloc = None
        extloc = None
        srcloc = None

        docname = doc.title
        docdate = doc.date
        sha1 = doc.sha1
        revsave = doc.is_revisionsave
        pagecount = doc.pages
        doctype = None
        filetype = doc.filetype
        filename = doc.filename
        ftx = doc.fulltxt
        tags = doc.tags
        fieldtxt = None
        annotxt = None
        links = doc.links
        annotations = [x for x in doc.annotations if x.layer == 0]
        for anno in annotations:
            anno.modified = True

        return {"dst_arcid": dst_arcid,
                "dstclient": self.dst_clients[profile],
                "srcclient": self.client,
                "is_revisionsave": doc.is_revisionsave,
                "src_sts": src_stsfields,
                "dst_sts": dst_stsfields,
                "src_add": src_addfields,
                "dst_add": dst_addfields,
                "other_meta": {
                    "title": docname,
                    "docdate": docdate,
                    "revsave": revsave,
                    "annotations": annotations,
                    "pagecount": pagecount,
                    "doctype": doctype,
                    "filetype": filetype,
                    "sha1": sha1,
                    "filename": filename,
                    "ftx": ftx,
                    "tags": tags,
                    "fieldtxt": fieldtxt,
                    "annotxt": annotxt,
                    "links": links}
                }

    def send_arc_bybookmark(self, arcid):
        self.log.info(u"syncing arcid %d" %arcid)
        sections = sync.get_matching_sections(int(arcid))
        
        successdocs = []
        errordocs = []
        errors = []
        globalerrors = []
        globalmessages = []
        
        for section, subsections in sections.items():
            mapping_section = section + "_mapping"
            if mapping_section in self.config.sections():
                for subsection in subsections:
                    bmid = None
                    globalmessages.append(u"syncing arcid %d with section '%s', subsection '%s'" %(arcid, mapping_section, subsection))
                    if "bookmark_" + subsection in self.config.options(mapping_section):
                        bmid = self.config.get(mapping_section, "bookmark_" + subsection)
                        if len(bmid) > 0:
                            self.log.info(u"bookmark found in section %s: %s" %(mapping_section, bmid))
                            pass
                        else:
                            bmid = None
                            globalerrors.append(u"no bookmark found in section %s" % (mapping_section))
                            self.log.info(u"no bookmark found in section %s" % (mapping_section))
                    else:
                        globalerrors.append(u"no bookmark found in section %s" % (mapping_section))
                        self.log.info(u"no bookmark found in section %s" % (mapping_section))
                    if bmid is not None:
                        try:
                            self.client = self.get_client("main")
                        except Exception as e:
                            reporttxt = "critical: could not connect bitfarm-server, see logfile for details\n\n"
                            txt = self.create_txt_report([], [], [], reporttxt)
                            self.send_report_email(txt)
                            self.log.error(u"client error! Exception:\n%s" % str(e), report=True)
                            raise e
                        self.log.info("executing bookmark %s" %bmid)
                        result = self.client.getExecuteBookmarks(*bmid.split(","), asjson=True)
                        self.log.info("bookmark result %d documents" % len(result))
                        globalmessages.append("bookmark result %d documents" % len(result))
                        self.log.info("bookmark result: %s" %result)
                        for row in result:
                            dstdoc = None
                            try:
                                docs = self.send_doc(row[0], {section: [subsection]})
                                if type(docs) == tuple:
                                    srcdoc, dstdoc = docs
                                if dstdoc:
                                    syncresult = True
                                else:
                                    syncresult = False
                                if syncresult is True:
                                    successdocs.append(((srcdoc.profile, srcdoc.gdocid, srcdoc.arcid, srcdoc.archive), (dstdoc.profile, dstdoc.gdocid, dstdoc.arcid, dstdoc.archive)))
                                else:
                                    errordocs.append(row[0])
                                    errors.append("unknown -> check logfile")
                            except Exception as e:
                                errordocs.append(row[0])
                                errors.append("unknown -> check logfile")
                                self.log.error(u"syncing gdocid %d failed with error: %s" %(row[0], str(e)))
                    else:
                        globalerrors.append(u"no bookmark found in section %s" % (mapping_section))        
        if self.config.get("main", "report") == "always" or (len(errors) > 0 and self.config.get("main", "report") == "erroronly"):
            txtreport = self.create_txt_report(successdocs, errordocs, errors, "\n".join(globalerrors + globalmessages)+"\n\n")
            try:
                self.send_report_email(txtreport)
            except Exception as e:
                self.log.error(u"could not send report, error:\n%s" %str(e))

    def sync_arc_bybookmark(self, arcid):
        pass

    def sync_docs(self, gdocids):
        for gdocid in gdocids:
            doc = self.sync_doc(gdocid)

    def sync_doc(self, gdocid):


        self.log.set_gdocid(gdocid)

        self.log.info(u"sync_doc document %d started" %(gdocid), report=True)
        try:
            self.client = self.get_client("main")
        except Exception as e:
            self.log.error(u"client error! Exception:\n%s" % str(e), report=True)
            raise e

        self.log.debug(u"loading document")
        try:
            doc = self.client.getDoc(
                gdocid, includetext=True, include_annostream=True)
        except Exception as e:
            self.log.error("error loading document: %s" % str(e), report=True)
            raise e
        if not doc:
            self.log.error("error loading document",report=True)
            raise Exception("error loading document")
        self.log.debug(u"load -> success")
        
        # sektionen ermitteln, für die die Quellarcid des Dokuments hinterlegt ist
        self.log.debug(u"determining sections via arcid")
        try:
            relevant_sections = self.get_matching_sections(doc.arcid)
        except Exception as e:
            self.log.error("error determining relevant sections", report=True)
            raise e
        self.log.debug(u"section/subsection -> %s" %str(relevant_sections))
        if len(relevant_sections) < 1:
            self.log.error(u"no configuration found for source-archive %s" % doc.archive)
            #raise Exception(u"no configuration found for source-archive %s" % doc.archive)
        
        for section in relevant_sections:
            for sub_section in relevant_sections[section]:
                try:
                    dstclient = self.get_client(section)
                except Exception as e:
                    self.log.error(u"error connection clients. Exception:\n%s" % (str(e)), report=True)
                    raise e
                # config-check
                try:
                    errormessage_src = ""
                    errormessage_dst = ""
                    # src_errors = self.check_config(self.client.meta, self.client)
                    # dst_errors = self.check_config(dstclient.meta, dstclient)
                    if self.client.profile not in self.configs_checked:
                        src_errors = self.check_config_new(self.client)
                        errormessage_src, configlist_src = self.create_config_errormessage(src_errors, self.client)
                        self.configs_checked[self.client.profile] = True
                    if dstclient.profile not in self.configs_checked:
                        dst_errors = self.check_config_new(dstclient)
                        errormessage_dst, configlist_dst = self.create_config_errormessage(dst_errors, dstclient)
                        self.configs_checked[dstclient.profile] = True
                except Exception as e:
                    self.log.error(u"config check failed, document will not be synced!", report=True)
                    self.log.error(u"exception: %s" % str(e))
                    self.log.error(u"traceback: %s" % str(traceback.format_exc()))
                    raise e
                if len(errormessage_src) + len(errormessage_dst) > 0:
                    # TODO: handle config errors
                    self.log.error("config errors!\n%s\n%s" % ("\n".join(errormessage_src), "\n".join(errormessage_dst)), report=True)
                    raise Exception("config errors!, see logfile for for details")
                
                # dstgdocid Konfiguration in aktueller section finden
                # meta_0002_src_add_44 = %dstgdocid%
                # falls nicht vorhanden srcgdocid in aktueller section finden
                # meta_0002_dst_add_14 = %srcgdocid%
                src = [(opt, val) for (opt, val) in self.config.items(section+"_mapping") if re.match("meta_"+sub_section+"_(src|dst)_add_(\d+)", opt) and "%srcgdocid%" in val]
                dst = [(opt, val) for (opt, val) in self.config.items(section+"_mapping") if re.match("meta_"+sub_section+"_(src|dst)_add_(\d+)", opt) and "%dstgdocid%" in val]

                # remote Dokument mit dstclient laden
                

                cfg_srcarcid = self.config.getint(section + "_mapping", "src_arc_" + sub_section)
                if cfg_srcarcid != doc.arcid:
                    self.log.warning(u"no config for archive '%s' in section %s/%s" % (doc.archive, section + "_mapping", "src_arc_" + sub_section))
                    raise Exception(u"no config for archive %s" % doc.archive)
                self.log.debug("config check success")

                self.log.info(u"syncing document with section %s/subsection %s" %(section, sub_section))
                self.log.info(u"determining sync-metadata")
                try:
                    meta = self.get_sync_meta(doc, section, sub_section)
                except Exception as e:
                    self.log.error("get_sync_meta error: \n%s" % (str(e)), report=True)
                    raise e
                print meta

    def send_doc(self, gdocid, relevant_sections=None):
        """
        sendet ein Quelldokument an Server gemäß der übergebenen relevant_sections
        relevant_sections ist dabei ein dict mit key -> Haupt-Sektionsname (z.B. dst_0001)
        und value eine Liste der indexe der zuverwendenen Untersektionen in der Mapping Sektion.
        Beispiel: Sektion "dst_0001", sub-Sektion "0001" (src_arc_"0001")
        Werden keine relevant_sections übergeben, werden alle zum Quellarchiv passenden Sektionen
        durchlaufen.
        Lesezeichen werden beim Aufruf dieser Funktion nicht ausgwewertet, die übergebene gdocid wird
        auch dann synchronisiert, wenn das Dokument nicht vom hinterlegten Lesezeichen gefunden wird.
        """


        """
        pro Dokument folgende Schritte ausführen und Loggen
        - Dokument laden
        - config prüfen
        - Metadaten ermitteln
        - Ziel-Dokument senden/erzeugen
        - Ziel-Dokumenten-Metadaten speichern 
        - Quell-Dokumentendaten speichern
        - Report mit gdocid, Syncerfolg/-misserfolg
        """




        self.log.set_gdocid(gdocid)

        self.log.info(u"send_doc document %d started" %(gdocid), report=True)
        try:
            self.client = self.get_client("main")
        except Exception as e:
            self.log.error(u"client error! Exception:\n%s" % str(e), report=True)
            raise e

        self.log.debug(u"loading document")
        try:
            doc = self.client.getDoc(
                gdocid, includetext=True, include_annostream=True)
        except Exception as e:
            self.log.error("error loading document: %s" % str(e), report=True)
            raise e
        if not doc:
            self.log.error("error loading document",report=True)
            raise Exception("error loading document")
        self.log.debug(u"load -> success")
        # wenn relevant_sections nicht mitgegeben wurde -> 
        # sektionen ermitteln, für die die Quellarcid des Dokuments hinterlegt ist
        if relevant_sections is None:
            self.log.debug(u"no section/subsection given, determining sections via arcid")
            try:
                relevant_sections = self.get_matching_sections(doc.arcid)
            except Exception as e:
                self.log.error("error determining relevant sections", report=True)
                raise e
        self.log.debug(u"section/subsection -> %s" %str(relevant_sections))
        if len(relevant_sections) < 1:
            self.log.error(u"no configuration found for source-archive %s" % doc.archive)
            raise Exception(u"no configuration found for source-archive %s" % doc.archive)
        for section in relevant_sections:
            for sub_section in relevant_sections[section]:
                try:
                    dstclient = self.get_client(section)
                except Exception as e:
                    self.log.error(u"error connection clients. Exception:\n%s" % (str(e)), report=True)
                    raise e
                
                try:
                    errormessage_src = ""
                    errormessage_dst = ""
                    # src_errors = self.check_config(self.client.meta, self.client)
                    # dst_errors = self.check_config(dstclient.meta, dstclient)
                    if self.client.profile not in self.configs_checked:
                        src_errors = self.check_config_new(self.client)
                        errormessage_src, configlist_src = self.create_config_errormessage(src_errors, self.client)
                        self.configs_checked[self.client.profile] = True
                    if dstclient.profile not in self.configs_checked:
                        dst_errors = self.check_config_new(dstclient)
                        errormessage_dst, configlist_dst = self.create_config_errormessage(dst_errors, dstclient)
                        self.configs_checked[dstclient.profile] = True
                except Exception as e:
                    self.log.error(u"config check failed, document will not be synced!", report=True)
                    self.log.error(u"exception: %s" % str(e))
                    self.log.error(u"traceback: %s" % str(traceback.format_exc()))
                    raise e
                
                # errorcount = sum(
                #     [len(value) for key, value in src_errors.items() if not key.endswith("_options")])
                # errorcount += sum([len(value) for key,
                #     value in dst_errors.items() if not key.endswith("_options")])

                if len(errormessage_src) + len(errormessage_dst) > 0:
                    # TODO: handle config errors
                    self.log.error("config errors!\n%s\n%s" % ("\n".join(errormessage_src), "\n".join(errormessage_dst)), report=True)
                    raise Exception("config errors!, see logfile for for details")
                
                cfg_srcarcid = self.config.getint(section + "_mapping", "src_arc_" + sub_section)
                if cfg_srcarcid != doc.arcid:
                    self.log.warning(u"no config for archive '%s' in section %s/%s" % (doc.archive, section + "_mapping", "src_arc_" + sub_section))
                    raise Exception(u"no config for archive %s" % doc.archive)
                self.log.debug("config check success")

                self.log.info(u"sending document with section %s/subsection %s" %(section, sub_section))
                self.log.info(u"determining sync-metadata")
                try:
                    meta = self.get_sync_meta(doc, section, sub_section)
                except Exception as e:
                    self.log.error("get_sync_meta error: \n%s" % (str(e)), report=True)
                    raise e
                documentfiles = []
                documentstreams = []
                for loc in ["docloc", "srcloc", "extloc", "oriloc"]:
                    try:
                        docfile = doc.__getattribute__(loc)
                    except:
                        self.log.critical(
                            u"gdocid %d, attrib %s not available." % (doc.gdocid, loc), report=True)
                        continue
                    if docfile not in ("", None):
                        fname = os.path.expandvars(docfile)
                        if not os.path.isfile(fname):
                            self.log.critical(
                                u"gdocid %d, %s '%s' not found!" % (doc.gdocid, loc, fname), report=True)
                            continue
                        else:
                            with open(fname, "rb") as fobj:
                                documentstream = fobj.read()
                                documentstreams.append(documentstream)
                                rs_nrs = docfile.split(os.path.sep)[1]
                                if rs_nrs.lower() == self.client.con.archive_nrs.lower():
                                    revsave = "nrs"
                                else:
                                    revsave = "rs"
                                documentfiles.append(
                                    (loc, len(documentstream), os.path.splitext(docfile)[1], revsave))
                self.log.info("creating new document")
                success, result = meta["dstclient"].getCreateDocument(
                    [meta["dst_arcid"], self.client.profile, doc.gdocid] + documentfiles, documentstreams)

                if success is True:
                    self.log.info("document created (gdocid %d)" %result.data.gdocid)
                    dstdoc = result.data
                else:
                    #TODO: Fehlerbehandlung
                    self.log.error(u"error creating document on remote server: %s" %result)
                    raise Exception(u"error creating document on remote server")
                self.log.debug(u"mapping netadata to src-doc/dst-doc")
                # update doc-objects
                try:
                    dstdoc, srcdoc = self.map_meta(meta, dstdoc, doc)
                except Exception as e:
                    self.log.error(u"error mapping metadata to documents.\n%s" % str(e), report=True)
                    raise e

                # Dokumente speichern
                success, result = meta["dstclient"].getEdit(dstdoc.gdocid)
                if success is False:
                    self.log.critical(u"failed to set editmode for new document (gdocid: %d)" % dstdoc.gdocid, report=True)
                    raise Exception(u"failed to set editmode for new document (gdocid: %d)" % dstdoc.gdocid)
                saveresult = meta["dstclient"].saveDoc(dstdoc)
                if saveresult is not True:
                    self.log.critical(u"failed to save document (gdocid: %d)" % dstdoc.gdocid, report=True)
                    raise Exception(u"failed to save document (gdocid: %d)" % dstdoc.gdocid)
                cancel = meta["dstclient"].getCancel(dstdoc.gdocid)
                if cancel is not True:
                    self.log.critical(u"failed to disable editmode for document (gdocid: %d)" % dstdoc.gdocid, report=True)
                    raise Exception(u"failed to disable editmode for document (gdocid: %d)" % dstdoc.gdocid)

                success, result = meta["srcclient"].getEdit(srcdoc.gdocid)
                if success is False:
                    self.log.critical(u"failed to set editmode for source document (gdocid: %d)" % srcdoc.gdocid, report=True)
                    raise Exception(u"failed to set editmode for source document (gdocid: %d)" % srcdoc.gdocid)
                saveresult = meta["srcclient"].saveDoc(srcdoc)
                if saveresult is not True:
                    self.log.critical(u"failed to save source document (gdocid: %d)" % srcdoc.gdocid, report=True)
                    raise Exception(u"failed to save source document (gdocid: %d)" % srcdoc.gdocid)
                cancel = meta["srcclient"].getCancel(srcdoc.gdocid)
                if cancel is not True:
                    self.log.critical(u"failed to disable editmode for source document (gdocid: %d)" % srcdoc.gdocid, report=True)
                    raise Exception(u"failed to disable editmode for source document (gdocid: %d)" % srcdoc.gdocid, report=True)
                self.log.info("document synced")
                return (srcdoc, dstdoc)

    def create_config_errormessage(self, errors, client):

        arcs_missing = []
        arcs_missing_right = []
        sts_missing_read = []
        sts_missing_write = []
        sts_missing = []
        add_missing = []
        add_missing_read = []
        add_missing_write = []

        for section, vals in errors.items():
            print section
            print vals
            arcs_missing.extend(vals["arcs_missing"])
            arcs_missing_right.extend(vals["arcs_missing_right"])
            
            for arcid in vals["meta"]:
                sts_missing.extend(vals["meta"][arcid]["sts_missing"])
                sts_missing_read.extend(vals["meta"][arcid]["sts_missing_read"])
                sts_missing_write.extend(vals["meta"][arcid]["sts_missing_write"])
                add_missing.extend(vals["meta"][arcid]["add_missing"])
                add_missing_read.extend(vals["meta"][arcid]["add_missing_read"])
                add_missing_write.extend(vals["meta"][arcid]["add_missing_write"])

        msg_rows = []
        configlist = []
        if len(arcs_missing + arcs_missing_right + sts_missing_read + sts_missing_write + sts_missing + add_missing + add_missing_read + add_missing_write) > 0:
            msg_rows.append(_(u"Profil: %s, Benutzer %s:\n" %(client.profile, client.username)))
            if len(arcs_missing) > 0:
                missing_ids = []
                msg_rows.append(_(u"folgende Archive sind nicht verfügbar:"))
                for ((section, option), arcid) in arcs_missing:
                    configlist.append(((section, option), arcid))
                    if arcid not in missing_ids:
                        missing_ids.append(arcid)
                        if arcid not in client.meta["archives"]:
                            msg_rows.append("    arcid %d" % arcid)
                        else:
                            msg_rows.append("    Archiv %s" % client.meta["archives"][arcid]["name"])
            if len(arcs_missing_right) > 0:
                missing_ids = []
                msg_rows.append(_(u"auf folgende Archive sind unzureichende Rechte gegeben:"))
                for ((section, option), arcid) in arcs_missing_right:
                    configlist.append(((section, option), arcid))
                    if arcid not in missing_ids:
                        missing_ids.append(arcid)
                        if arcid not in client.meta["archives"]:
                            msg_rows.append("    -arcid %d" % arcid)
                        else:
                            msg_rows.append("    -Archiv %s" % client.meta["archives"][arcid]["name"]) 
            if len(sts_missing) > 0:
                missing_ids = []
                msg_rows.append(_(u"folgende Statusfelder sind nicht verfügbar:"))
                for ((section, option), sts) in sts_missing:
                    configlist.append(((section, option), sts))
                    stsid = int(sts[4:])
                    if stsid not in missing_ids:
                        missing_ids.append(stsid)
                        if stsid not in client.meta["statefields"]:
                            msg_rows.append("    -stsid %s" % sts)
                        else:
                            msg_rows.append("    -Statusfeld %s" % client.meta["statefields"][stsid]["name"])
            if len(sts_missing_read) > 0:
                missing_ids = []
                msg_rows.append(_(u"Statusfelder ohne Leserecht:"))
                for ((section, option), sts) in sts_missing_read:
                    configlist.append(((section, option), sts))
                    stsid = int(sts[4:])
                    if stsid not in missing_ids:
                        missing_ids.append(stsid)
                        if stsid not in client.meta["statefields"]:
                            msg_rows.append("    -stsid %s" % sts)
                        else:
                            msg_rows.append("    -Statusfeld %s" % client.meta["statefields"][stsid]["name"])
            if len(sts_missing_write) > 0:
                missing_ids = []
                msg_rows.append(_(u"Statusfelder ohne Ändernrecht:"))
                for ((section, option), sts) in sts_missing_write:
                    configlist.append(((section, option), sts))
                    stsid = int(sts[4:])
                    if stsid not in missing_ids:
                        missing_ids.append(stsid)
                        if stsid not in client.meta["statefields"]:
                            msg_rows.append("    -stsid %s" % sts)
                        else:
                            msg_rows.append("    -Statusfeld %s" % client.meta["statefields"][stsid]["name"])
            if len(add_missing) > 0:
                missing_ids = []
                msg_rows.append(_(u"folgende Zusatzfelder sind nicht verfügbar:"))
                for ((section, option), add) in add_missing:
                    configlist.append(((section, option), add))
                    if add not in missing_ids:
                        missing_ids.append(add)
                        if add not in client.meta["addfields"]:
                            msg_rows.append("    -addid %s" % add)
                        else:
                            msg_rows.append("    -Zusatzfeld %s" % client.meta["addfields"][add]["name"])

            if len(add_missing_read) > 0:
                missing_ids = []
                msg_rows.append(_(u"Zusatzfelder ohne Leserecht:"))
                for ((section, option), add) in add_missing_read:
                    configlist.append(((section, option), add))
                    if add not in missing_ids:
                        missing_ids.append(add)
                        if add not in client.meta["addfields"]:
                            msg_rows.append("    -addid %s" % add)
                        else:
                            msg_rows.append("    -Zusatzfeld %s" % client.meta["addfields"][add]["name"])
            if len(add_missing_write) > 0:
                missing_ids = []
                msg_rows.append(_(u"Zusatzfelder ohne Ändernrecht:"))
                for ((section, option), add) in add_missing_write:
                    configlist.append(((section, option), add))
                    if add not in missing_ids:
                        missing_ids.append(add)
                        if add not in client.meta["addfields"]:
                            msg_rows.append("    -addid %s" % add)
                        else:
                            msg_rows.append("    -Zusatzfeld %s" % client.meta["addfields"][add]["name"])
                    

        return msg_rows, configlist

    def map_meta(self, meta, dstdoc, srcdoc):
        """
        mappt alle per get_sync_meta ermittelten Metadaten in Quell- und Zieldokument.
        Speichern der Dokumente erfolgt ausserhalb dieser Funktion!
        """
        # Zieldokumentendaten
        dstdoc.userid = meta["dstclient"].userid
        for prop in meta["other_meta"]:
            value = meta["other_meta"][prop]
            dstdoc.__setattr__(prop, value)
        for add, value in meta["dst_add"].items():
            dstdoc.fields[add].value = value
        for sts, value in meta["dst_sts"].items():
            for name, sfield in dstdoc.statefields.items():
                if sfield.id == sts:
                    dstdoc.statefields[name].value = value
                    break
        dstdoc.modified = True

        for addid, value in meta["src_add"].items():
            if type(value) in (unicode, str):
                if "%dstgdocid%_as_str" in value:
                    value = value.replace("%dstgdocid%_as_str", str(dstdoc.gdocid))
                elif value == "%dstgdocid%_as_int":
                    value = dstdoc.gdocid
                elif "%dstgdocid%" in value:
                    value = value.replace("%dstgdocid%", str(dstdoc.gdocid))
            
            srcdoc.fields[addid].value = value
            
        for sts, value in meta["src_sts"].items():
            for name, sfield in srcdoc.statefields.items():
                if sfield.id == sts:
                    srcdoc.statefields[name].value = value
                    break
        srcdoc.modified = True
        
        return dstdoc, srcdoc

    def send_docs(self, gdocids, report=True):
        """
        sendet alle übergebenen Dokumente (gdocids) 
        gemäß Konfiguration an ein oder mehrere Zielsysteme
        gdocids: Liste von ints
        report: True/oder False -> sendet einen Report gemäß Konfiguration
        """
        successdocs = []
        errordocs = []
        errors = []
        for gdocid in gdocids:
            try:
                docs = self.send_doc(gdocid)
                if type(docs) == tuple:
                    srcdoc, dstdoc = docs
                if dstdoc:
                    result=True
                if result is True:
                    successdocs.append(((srcdoc.profile, srcdoc.gdocid, srcdoc.arcid, srcdoc.archive), (dstdoc.profile, dstdoc.gdocid, dstdoc.arcid, dstdoc.archive)))
                else:
                    errordocs.append(gdocid)
                    errors.append("unknown -> check logfile")

            except Exception as e:
                errordocs.append(gdocid)
                errors.append("Exception: %s" % e)
        if report is True:
            reportlevel = self.config.get("main", "report")
            if reportlevel == "always" or (reportlevel == "erroronly" and len(errors)>0):
                txtreport = self.create_txt_report(successdocs, errordocs, errors)
                try:
                    self.send_report_email(txtreport)
                except Exception as e:
                    self.log.error(u"could not send report, error:\n%s" %str(e))

    def send_report_email(self, txtreport, attachlog=True):
        recipient = self.config.get("main", "recipient")
        sender = self.config.get("main", "sender")      
        usetls = self.config.get("main", "usetls")
        usessl = self.config.get("main", "usessl")
        smtppass = self.config.get("main", "smtppass")
        smtpuser = self.config.get("main", "smtpuser")
        try:
            smtpport = self.config.getint("main", "smtpport")
        except:
            self.log.warning(u"cannot send report email, configuration incomplete")
        smtp = self.config.get("main", "smtp")

        if smtp == u"" or recipient == u"" or sender == u"":
            self.log.warning(u"cannot send report email, configuration incomplete")
            return

        if usessl is True:
            try:
                s = smtplib.SMTP_SSL(smtp, smtpport)
            except Exception as e:
                self.log.error("error connecting smtpserver.\nException:\n" + str(e))
                raise e
        else:
            try:
                s = smtplib.SMTP(smtp, smtpport)
            except Exception as e:
                self.log.error("error connecting smtpserver.\nException:\n" + str(e))
                raise e
        #s.set_debuglevel(1)

        try:
            ehlo_helo_result = s.ehlo_or_helo_if_needed()
        except Exception as e:
            self.log.error("error ehlo/helo.\nException:\n" + str(e))
            raise e
                
        if usetls is True:
            try:
                s.ehlo_or_helo_if_needed()
                s.starttls()
                s.ehlo_or_helo_if_needed()
            except Exception as e:
                self.log.error("error starttls.\nException:\n" + str(e))
                raise e

        if smtpuser not in (None, "") and smtppass not in (None, ""):
            try:
                s.login(smtpuser, smtppass)
            except Exception as e:
                self.log.error("error smtp-login.\nException:\n" + str(e))
                raise e

        msg = MIMEMultipart()
        part1 = MIMEText(txtreport, 'plain')

        
        msg['Subject'] = "bfa_sync report"
        msg['From'] = sender
        msg['To'] = recipient
        msg.attach(part1)
        try:
            with open(self.config.get("main", "logfile"), "rb") as fil:
                part = MIMEApplication(
                    fil.read(),
                    Name=os.path.basename(self.config.get("main", "logfile"))
                )
        except:
            pass
        # After the file is closed
        part['Content-Disposition'] = 'attachment; filename="%s"' % os.path.basename(self.config.get("main", "logfile"))
        msg.attach(part)

        # with open('filename.eml', 'w') as out:
        #     gen = generator.Generator(out)
        #     gen.flatten(msg)

        try:
            s.sendmail(msg['From'], [msg['To']], msg.as_string())
        except Exception as e:
            self.log.error("error sendmail.\nException:\n" + str(e))
            raise e
        s.quit()

    def create_txt_report(self, successdocs, errordocs, errors, globalmessage=""):
        txt = "bfa_sync Report\n"
        txt += "----------------------\n"
        if globalmessage != "":
            txt += globalmessage
        txt += "total docs: %d\n" % (len(successdocs) + len(errordocs))
        txt += "success: %d\n" % len(successdocs)
        txt += "failed: %d\n" % len(errordocs)
        if len(errordocs) > 0:
            txt += "\nerrors:\n"
            for gdocid, error in zip(errordocs, errors):
                txt += "  -gdocid %d: %s\n" % (gdocid, error)
        
        if len(successdocs) > 0:
            txt += "\nsuccess:\n"
            for ((src_profile, src_gdocid, src_arcid, src_archive), (dst_profile, dst_gdocid, dst_arcid, dst_archive)) in successdocs:
                txt += "  -gdocid %d (profile '%s', archive '%s') synced to gdocid %d (profile '%s', archive '%s')\n" % (src_gdocid, src_profile, src_archive, dst_gdocid, dst_profile, dst_archive)
        return txt
        


    def get_client(self, section):
        self.log.debug("get_client for section %s, profile %s" %(section, self.config.get(section, "profile")))
        if section == "main":
            client = self.client
        else:    
            profile = self.config.get(section, "profile")
            if profile in self.dst_clients:
                client = self.dst_clients[profile]
            else:
                client = None
        if client is None or client.loggedin is not True:
            self.log.debug(u"client not connected!")
            client = self.connect_client(**self.get_client_condata(section))
            
        client.section = section
        if section != "main":
            self.dst_clients[profile] = client
        return client

    def get_client_condata(self, section):
        result = {}
        result["profile"] = self.config.get(section, "profile")
        result["username"] = decryptData(base64.b64decode(
                self.config.get(section, "bfauser")))
        result["userpass"] = decryptData(base64.b64decode(
                self.config.get(section, "bfapass")))
        return result

    def get_matching_sections(self, arcid):
        self.log.info(u"find relevant config sections for arcid %d" %arcid)
        result = {}
        sections = [section for section in self.config.sections()
                    if re.match("dst_\d{4}$", section)]
        for section in sections:
            # src_arc_0001
            mapping_section = section+"_mapping"
            options = [option[-4:] for option in self.config.options(mapping_section) if option.startswith(
                "src_arc_") and self.config.getint(mapping_section, option) == arcid]
            if len(options) > 0:
                result[section] = options
        if len(result) > 0:
            self.log.info(u"relevant config sections for arcid %d: %s" % (arcid, str(result)))
        else:
            self.log.info(u"no relevant config sections for arcid %d found." %arcid)
        return result


if __name__ == "__main__":
    """
    commandline-Tool zum syncen von Dokumenten
    synct entweder einzelne per commandline-arg übergebene gdocids
    oder übergebene Archive (arcids).
    Werden gdocids übergeben, wird die passende Konfiguration 
    über das zugehörige Archiv ermittelt, Lesezeichen werden dabei 
    ignoriert.
    Werden arcids übergeben, muss zu den einzelnen Archiven ein 
    Lesezeichen konfiguriert sein, mit Hilfe dessen die zu 
    synchronisierenden Dokumente ermittelt werden.
    """
    sync = BfaSync()
    parser = argparse.ArgumentParser()
    parser.add_argument("-g", "--gdocids", help=u"semikolon-getrennte Liste von gdocids", type=str)
    parser.add_argument("-a", "--arcids", help=u"semikolon-getrennte Liste von arcids oder 'all' für alle in der Konfiguration hinterlegten Quellarchive", type=str)
    parser.add_argument("--sync", dest='sync', required=False, default=False, action='store_true',
                        help=u"synchronisiere remote-Dokument, wenn bereits im Zielsystem vorhanden")

    args = parser.parse_args()
    sync.log.info("started with args: %s" %args)
    if args.gdocids:
        gdocids = [int(x) for x in args.gdocids.split(";")]
        if args.sync is True:
            sync.sync_docs(gdocids)
        else:
            sync.send_docs(gdocids)
    elif args.arcids:
        #print args
        for arcid in args.arcids.split(";"):
            if args.sync is True:
                sync.sync_arc_bybookmark(int(arcid))
            else:
                sync.send_arc_bybookmark(int(arcid))

            
    
   