import configparser
import sys
import os
import argparse
import logging
import locale
import threading
import tempfile
import codecs
import glob
import shutil
import subprocess
from lxml import etree, html
import time
import contextlib

from multiprocessing.pool import ThreadPool

import thtiflib

from logging.handlers import TimedRotatingFileHandler


# constants
LOGLEVEL = {"debug": logging.DEBUG, "info": logging.INFO,
            "error": logging.ERROR, "critical": logging.CRITICAL, "warning": logging.WARNING}
VERSION = "3.6.1.2"

configtemplate = """[main]
tesscmd=%bitfarm-archiv%\\Tesseract-OCR\\tesseract.exe
logfile=%bitfarm-archiv%\\bin\\logs\\tessocr.log
logbackup=7
loglevel=DEBUG"""

stime = time.time()


class ArgumentParserWriteUTF8(argparse.ArgumentParser):
    """Subclass of ArgumentParser that outputs UTF-8 properly

    Standard ArgumentParser in argparse outputs help messages without
    converting to UTF-8. This fails if the message is a unicode string
    containing non-ASCII characters and output is redirected (to a
    pipe or a file). Unredirected output seems OK. See
    http://bugs.python.org/issue9779
    This class fixes it. It will probably not be needed in Python 3.
    """

    def _print_message(self, message, file=None):
        """Output message to file, encoded as UTF-8 """
        if message:
            if file is None:
                file = sys.stderr
            file.write(message.encode(locale.getpreferredencoding()))


class LogWriter:
    def __init__(self, logger, log_level=logging.CRITICAL):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())


class BfaTessOCR():
    def __init__(self, configfile=None, logid="", logger=None):
        if configfile is None:
            configfile = os.path.join(os.path.dirname(sys.argv[0]), "bfa_tessocr.ini")
            if not os.path.isfile(configfile):
                configfile = os.path.join(os.path.expandvars("%bitfarm-archiv%"), "bfa_tessocr.ini")
        # loading the config
        if not os.path.isfile(configfile):
            if os.path.isdir(os.path.expandvars("%bitfarm-archiv%")):
                with codecs.open(os.path.join(os.path.expandvars("%bitfarm-archiv%"), "bfa_tessocr.ini"), "w", encoding="utf8") as fobj:
                    fobj.write(configtemplate)
                configfile = os.path.join(os.path.expandvars("%bitfarm-archiv%"), "bfa_tessocr.ini")

        self.config = configparser.ConfigParser(allow_no_value=True, interpolation=None)
        # self.config.optionxform = unicode

        # with codecs.open(configfile, "r", encoding="utf8") as fobj:
        # with open(configfile, "r", encoding="utf8") as fobj:
        #     self.config.read_file(fobj)
        self.config.read_file(open(configfile, "r", encoding="utf8"))
        # set up logging
        self.logid = logid

        if logger is not None:
            self.log = logger
        else:
            self.log = logging.getLogger("bfa_tessocr")
            #self.log.propagate=False
            logfilepath = os.path.expandvars(self.config.get("main", "logfile"))
            logdir = os.path.dirname(logfilepath)
            if logdir != "" and not os.path.isdir(logdir):
                os.makedirs(logdir)

            logHandler = TimedRotatingFileHandler(logfilepath, when="midnight", backupCount=self.config.get("main", "logbackup"))
            logFormatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            logHandler.setFormatter(logFormatter)
            self.log.addHandler(logHandler)
            logStreamHandler = logging.StreamHandler()
            logStreamHandler.setFormatter(logFormatter)
            self.log.addHandler(logStreamHandler)

            # redirect out/err to log
            if __name__ == "__main__":
                sys.stderr = LogWriter(self.log, logging.CRITICAL)
                sys.stdout = LogWriter(self.log, logging.DEBUG)

            self.log.setLevel(
                LOGLEVEL[self.config.get("main", "loglevel").lower()])
        self.log.info("bfa_tessocr init success, version %s" % VERSION)

    def ocr(self, inputfile, outputbase="", outformat=["txt", "hocr"], numproc=1,
            pbstring=None, languages=["deu"], tessdatadir=None, ompthreadlimit=1,
            tessargs="", overwrite=False, encoding="utf8", addorientation=False):
        """
        \r\n

        Runs tesseract ocr on a given image and save or return the result.
        the method splits the inputimage into single pages, and use 
        multiple Tesseract-processes to speed up ocr.


        returns:
        --------
        The return-value depends on outputbase: 

        If outputbase was given:
            returns True, if we have ocr-result and it could be saved.
            returns False, if ocr-result or saving the result did not work
        If outputbase was NOT given (outputbase=""):
            returns a dictionary containing the result. key for text is "txt", 
            key for hocr-result is "hocr"
            returns False, if ocr or saving the ocr-result did not work



        required arguments:
        -------------------
        - inputfile (string):    path to tif-file


        optional arguments:
        -------------------
        - outputbase (string):  base-filename for the 
                                text/hocr-result. resulting txt/hocr 
                                files will be saved using utf8 encoding, 
                                except argument encoding is set. 
                                Omit or use an empty string to make the 
                                function return the result as a dictionary 
                                instead of saving it to a file.
        - outformat (list):     outputformats to be created by tesseract. 
                                possible values include "txt" and "hocr". 
                                Defaults to ["txt", "hocr"].
        - numproc (int):        maximum number of parallel Tesseract-instances, 
                                you should not use more than you have cores in 
                                your machine. Default 1.
        - pbstring (string):    string that will be inserted into text after 
                                each page. If you insert "%d" in this string it 
                                will be replaced by the actual page-number. 
                                Default None (no pagebreak-strings).
        - languages (list):     language-models to be used by Tesseract. Use 
                                "tesseract.exe --list-langs" to get a list of 
                                languages availabe in your tessdata-dir. 
                                Defaults to ["deu"].
        - tessdatadir (string): representing the path to the tessdata-dir that 
                                should be used (different versions avaliable at 
                                https://github.com/tesseract-ocr). Default None.
        - ompthreadlimit (int): controlls how many threads Tesseract uses 
                                internally for ocr. Defaults to 1, at the moment 
                                leave it as that for best performance.
        - tessargs (str):       additional args to be added, when calling 
                                Tesseract, default "".
        - overwrite (bool):     If set to True, existing files will be overwritten. 
                                If set to False and files do already exist then we 
                                do not perform ocr. Default is False.
        - encoding (string):    Target encoding for text. Default is 'utf8'.
        - addorientation (bool):include orientation per page in hocr-result 
                                (works only with 'hocr' in outformat, results in an
                                additional call to tessearact.exe (possibly 
                                time-consuming)

        """

        self.log.info("running ocr on '%s'" % inputfile)
        # input/outputfile check
        if not os.path.isfile(inputfile):
            self.log.error("image '%s' not found." % inputfile)
            raise Exception("inputfile '%s' not found!" % inputfile)

        if outputbase != "":
            self.log.debug("%s - ouput-filename conflict check" %self.logid)
            outputfile_conflicts = glob.glob(outputbase+".*")
            extensions = [os.path.splitext(x)[1].lower()
                          for x in outputfile_conflicts]
            if "txt" in outformat:
                if ".txt" in extensions:
                    if overwrite is False:
                        self.log.error("outputbase matches existing files")
                        raise Exception("outputbase matches existing files")
                    else:
                        self.log.warning(
                            "outputbase matches existing files, '%s.txt' will be overwritten" % (outputbase))
            if "hocr" in outformat:
                if ".hocr" in extensions:
                    if overwrite is False:
                        self.log.error("outputbase matches existing files")
                        raise Exception("outputbase matches existing files")
                    else:
                        self.log.warning(
                            "outputbase matches existing files, '%s.hocr' will be overwritten" % (outputbase))

        with tempdir(prefix="tessocr_") as td:
            pagetifs = self.split(inputfile, td)
            tessenv = self.set_tessenv(tessdatadir, ompthreadlimit)
            resultdict = {}
            tp = ThreadPool(numproc)
            tessexe = os.path.expandvars(self.config.get("main", "tesscmd"))
            for temptif in pagetifs:
                tempoutputbase = os.path.splitext(temptif)[0]
                cmd = [tessexe, temptif, tempoutputbase,
                       "-l", "+".join(languages)]
                if len(tessargs) > 0:
                    cmd = cmd + tessargs.split(" ")
                cmd = cmd + outformat + ["quiet"]
                self.log.debug("%s adding command to threadpool: " %(self.logid))
                self.log.debug(self.logid + " '"+"' '".join(cmd)+"'")
                tp.apply_async(run_tesseract, (cmd, tessenv,
                                               tempoutputbase, resultdict,))
            tp.close()
            tp.join()
            for pagekey in resultdict:
                for typekey in resultdict[pagekey]:
                    if type(resultdict[pagekey][typekey]) not in (type(None), str):
                        self.log.error("%s ocr error page %d: %s" %
                                       (self.logid, pagekey, resultdict[pagekey][typekey]))
                        raise(resultdict[pagekey][typekey])
            orientation = None
            if "hocr" in outformat and addorientation is True:
                try:
                    orientation = self.get_orientation(inputfile)
                except:
                    orientation = None
            result = self.merge_result(resultdict, pbstring, orientation)

            if encoding != "utf8":
                try:
                    result["txt"] = result["txt"].encode(
                        encoding, errors="replace")
                except:
                    self.log.critical(
                        "%s could not convert utf8-text to '%s'" %(self.logid, encoding))
            if outputbase != "":
                return self.save_result(result, outputbase, encoding=encoding)
            else:
                return result

    def get_orientation(self, tif):
        tessexe = os.path.expandvars(self.config.get("main", "tesscmd"))
        cmd = [tessexe, tif, "stdout", "--psm", "0"]
        try:
            output = subprocess.check_output(cmd)
        except Exception as e:
            self.log.error("%s get_orientation failed with error: %s" % (self.logid, str(e)))
            raise Exception("get_orientation failed with error: %s" % (str(e)))
        result = {}
        page = None
        for row in output.splitlines():
            if row.lower().startswith("page number"):
                page = int(row.split(":")[-1].strip())+1
            if row.lower().startswith("orientation in degrees"):
                orientation = row.split(":")[-1].strip()
                result[page] = orientation
        return result

    def save_result(self, result, outputbase, encoding):
        try:
            if "txt" in result:
                with open(outputbase+".txt", "wb") as fobj:
                    fobj.write(result["txt"])
            if "hocr" in result:
                with codecs.open(outputbase+".hocr", "wb", encoding="utf8") as fobj:
                    fobj.write(result["hocr"])
        except:
            return False
        return True

    def merge_result(self, resultdict, pbstring=None, orientation=None):
        hocrs = []
        txts = []
        xmls = []
        pages = sorted(resultdict)
        text = u""
        hocrdoc = None
        self.log.info("%s merge results"% (self.logid))
        self.log.debug("%s pbstring/type: %s/%s" % (self.logid, pbstring, type(pbstring)))
        for pageno in sorted(resultdict):
            if hocrdoc is None and resultdict[pageno]["hocr"] is not None:
                hocrdoc = etree.parse(resultdict[pageno]["hocr"])
                pages = hocrdoc.xpath("//*[@class='ocr_page']")
                if orientation is not None:
                    element = pages[-1].xpath("//*[@class='ocr_page']")[0]
                    element.set("orientation", orientation[pageno])
                container = pages[-1].getparent()
            elif hocrdoc is not None:
                if resultdict[pageno]["hocr"] is not None:
                    doc2 = etree.parse(resultdict[pageno]["hocr"])
                    pages = doc2.xpath("//*[@class='ocr_page']")
                    for page in pages:
                        if orientation is not None:
                            if pageno in orientation:
                                page.set("orientation", orientation[pageno])
                            else:
                                self.log.warning(
                                "%s no orientation for page %d, assuming no rotation" % (self.logid, pageno))
                                page.set("orientation", "0")
                        page.set('id', "page_"+str(pageno))
                        container.append(page)
            if resultdict[pageno]["txt"] != None:
                with codecs.open(resultdict[pageno]["txt"], "rb", encoding="utf8") as fobj:
                    txt = fobj.read().replace('\x0C', '')

                    if pbstring is not None:

                        temp_pbstring = pbstring.replace("%d", str(pageno))
                        temp_pbstring = temp_pbstring.replace("\\r", "\r")
                        temp_pbstring = temp_pbstring.replace("\\n", "\n")
                        pberror = False
                        try:
                            txt = txt + temp_pbstring
                        except Exception as e:
                            self.log.warning(
                                "%s pagebreak-string could not be added, error '%s'" %(self.logid, e))
                            pberror = True
                        if pberror is True:
                            try:
                                self.log.debug(
                                    "%s try adding pagebreak latin1 decoded" %(self.logid))
                                txt = txt + temp_pbstring.decode("latin1")
                            except Exception as e:
                                self.log.warning(
                                    "%s pagebreak-string could not be added" %(self.logid))
                                pberror = True

                    text += txt
        result = {}
        result["txt"] = text.strip()

        if hocrdoc is not None:
            result["hocr"] = etree.tostring(hocrdoc).decode('UTF-8')
        return result

    def set_tessenv(self, tessdatadir=None, ompthreadlimit=1):
        # TESSDATA_PREFIX & OMP_THREAD_LIMIT are set via environment
        self.log.debug("%s setting up tesseract-environment" %(self.logid))
        env = os.environ.copy()
        if tessdatadir is not None:
            if not os.path.isdir(tessdatadir):
                self.log.error(u"%s tessdata-dir '%s' not found"  %(self.logid, tessdatadir))
                raise Exception("tessdata-dir '%s' not found" % tessdatadir)
            self.log.debug(u"%s using TESSDATA_PREFIX '%s'" %(self.logid, tessdatadir))
            env["TESSDATA_PREFIX"] = tessdatadir
        if ompthreadlimit is not None:
            self.log.debug(u"%s using OMP_THREAD_LIMIT '%s'" %(self.logid, ompthreadlimit))
            env["OMP_THREAD_LIMIT"] = str(ompthreadlimit)
        self.log.debug("%s using this tesseract-environment:" %(self.logid))
        self.log.debug(self.logid + " " + str(env))
        return env

    def split(self, tiffile, tempdir):
        # try splitting with thtiflib (fast)
        self.log.info("%s splitting inputfile (thtiflib)"  %(self.logid))
        try:
            tif = thtiflib.Tiff(tiffile)
            tif.split(1, basename=os.path.join(tempdir, "tesstmp"))
            # raw_input("now")
            return sorted(glob.glob(os.path.join(tempdir, "tesstmp*.*")), key=lambda x: int(x.split("-")[-1].split(".")[0]))
        except Exception as e:
            self.log.warning(
                "%s splitting tif with thtiflib failed. Error %s"  %(self.logid, str(e)))

        # in case splitting via thtiflib did not work -> try SplitterV4
        # locate SplitterV4.exe
        self.log.warning(
            "%s splitting with thtiflib failed, trying SplitterV4 now." %(self.logid))
        bfapth = os.environ.get("bitfarm-archiv")
        splittexe = os.path.join(bfapth, "SplitterV4.exe")
        if not os.path.isfile(splittexe):
            self.log.critical("%s SplitterV4.exe not found, unable to split..." %(self.logid))
            raise Exception(
                "unable to split tif, thtiflib failed, SplitterV4 not found!")

        cmd = [splittexe, "-s", tiffile]
        if not os.path.isfile(splittexe):
            self.log.critical("%s SplitterV4.exe not found." %(self.logid))
            raise Exception("tif splitting failed")
        self.log.debug("%s SplitterV4 found: '%s'" %(self.logid, splittexe))
        # split via subprocess call
        newtif = os.path.join(tempdir, "tesstmp.tif")
        shutil.copyfile(tiffile, newtif)
        cmd = [splittexe, "-s", newtif]
        self.log.debug("%s running SplitterV4:" %(self.logid))
        self.log.debug(self.logid +" " + "'"+" ".join(cmd)+"'")
        result = None
        try:
            result = subprocess.check_call(cmd)
        except Exception as e:
            self.log.critical("%s SplitterV4.exe failed. Error: %s" %(self.logid, str(e)))
            raise Exception(
                "tif splitting failed with thtiflib and SplitterV4")
        if result != 0:
            self.log.critical("%s tif splitting failed" %(self.logid))
            raise Exception("tif splitting failed")

        # splitnames: tesstmp__page000.tif ...
        # rename splitted tiffiles ->
        # tesstmp__page000.tif -> tesstmp-0.tif
        # tesstmp__page001.tif -> tesstmp-1.tif
        try:
            for fn in glob.glob(os.path.join(tempdir, "tesstmp_page*.tif")):
                newname = fn.replace("tesstmp_page", "tesstmp-")
                pageno = os.path.basename(newname).split("-")[-1].split(".")[0]
                newname = newname.replace(pageno, str(int(pageno)))
                os.rename(fn, newname)
            sorted_tifs = sorted(glob.glob(os.path.join(
                tempdir, "tesstmp-*.*")), key=lambda x: int(x.split("-")[-1].split(".")[0]))
            self.log.debug("%s tif split success" %(self.logid))
            self.log.debug(self.logid + " " + "'"+" ".join(sorted_tifs)+"'")
        except:
            self.log.critical("%s error sorting splitted tif pages" %(self.logid))
            raise Exception("error sorting splitted tif pages")
        return sorted_tifs



@contextlib.contextmanager
def tempdir(*args, **kwargs):
    d = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield d
    finally:
        shutil.rmtree(d)


def run_tesseract(cmd, tessenv, outputbase, resultdict):
    error = None
    result = None
    page = int(os.path.basename(outputbase).split("-")[-1])
    #cmd.extend(["--psm", "1"])
    if tessenv is not None:
        try:
            result = subprocess.check_output(
                cmd, env=tessenv, stderr=subprocess.STDOUT)
        except Exception as e:
            result = e
            error = e
    else:
        try:
            result = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except Exception as e:
            result = e
            error = e
    if result is not None:
        txtfiles = sorted([x for x in glob.glob(outputbase+".*") if x.lower().endswith(
            ".txt") is True], key=lambda x: int(x.split("-")[-1].split(".")[0]))
        hocrfiles = sorted([x for x in glob.glob(outputbase+".*") if x.lower().endswith(
            ".hocr") is True], key=lambda x: int(x.split("-")[-1].split(".")[0]))
        for txtfile in txtfiles:
            page = int(txtfile.split("-")[-1].split(".")[0])
            resultdict[page] = {"txt": txtfile, "hocr": None}
        for hocrfile in hocrfiles:
            page = int(hocrfile.split("-")[-1].split(".")[0])
            if page not in resultdict:
                resultdict[page] = {}
            resultdict[page]["hocr"] = hocrfile
            if "txt" not in resultdict[page]:
                resultdict[page]["txt"] = None
    else:
        if error is not None:
            resultdict[page] = {}
            resultdict[page]["txt"] = error
            resultdict[page]["hocr"] = error


if __name__ == "__main__":
    # argument parsing
    stime = time.time()
    parser = ArgumentParserWriteUTF8()
    parser.add_argument("inputfile")
    parser.add_argument(
        "outputbase", help=u'basename for the resulting files, omit extension')
    parser.add_argument('--numproc', dest='numproc', action='store', required=False,
                        type=int,
                        default=1,
                        help=u'max number of parallel tesseract processes')
    parser.add_argument("--pbstring", dest='pbstring', action='store',
                        required=False, default="\r\n--- page %d ---\r\n")
    parser.add_argument("--lang", dest='lang', action='store',
                        required=False, default="deu")
    parser.add_argument("--outformat", dest='outformat',
                        action='store', required=False, default="txt+hocr")
    parser.add_argument('--tessargs', dest='tessargs', action='store', required=False, default='',
                        help=u'additional arguments for tesseract commandline')
    parser.add_argument("--overwrite", dest='overwrite', required=False, default=False, action='store_true',
                        help=u"overwrite existing files, defaults to false")
    parser.add_argument("--tessdata", dest='tessdata',
                        required=False, default=None, help=u"path to tessdata dir")
    parser.add_argument("--ompthreadlimit", dest='ompthreadlimit',
                        required=False, default="1", help=u"path to tessdata dir")
    parser.add_argument("--txtencoding", dest='txtencoding',
                        required=False, default="latin1", help=u"text-encoding")
    parser.add_argument("--addorientation", dest='addorientation',
                        required=False, default=True, type=bool, help=u"add orientation information in hocr-result")                        
    parser.add_argument("--config", dest='config',
                        required=False, default=os.path.join(os.path.dirname(sys.argv[0]), "bfa_tessocr.ini"), help=u"configfile") 
    args = parser.parse_args()

    # loading the config
    configfile = args.config

    myocr = BfaTessOCR(configfile)

    result = myocr.ocr(inputfile=args.inputfile,
                       outputbase=args.outputbase,
                       outformat=args.outformat.split("+"),
                       numproc=args.numproc,
                       pbstring=args.pbstring,
                       languages=args.lang.split("+"),
                       tessdatadir=args.tessdata,
                       ompthreadlimit=args.ompthreadlimit,
                       tessargs=args.tessargs,
                       overwrite=args.overwrite,
                       encoding=args.txtencoding,
                       addorientation=args.addorientation)

    etime = time.time()
    myocr.log.debug("runtime %.04f" % (etime-stime))
