﻿#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Dieses Modul enthält die Basisklasse für Dokumente und Workflows"""

import os
import sys
import locale
import codecs
import configparser
import logging
import re
import uuid
import datetime
import collections
import win32file
import base64
import dpapi
try:
    import bitfarm_enterprise as bitfarm_base
except:
    import bitfarm_core as bitfarm_base

from bitfarm_history import *
from bitfarm_errors import *

__version__ = 'v3.6.2'

# OrderedDict zum Einlesen der Scripts.ini. Da in der Scripts.ini viele nicht eindeutige Optionen vorhanden (zB dropkeybarcode)
# werden hier die Mehrfacheinträge zu einer Liste an Werten. Dann muss man sich den passenden anhand des Index heraussuchen
class bfaOrderedDict(collections.OrderedDict):
    def __setitem__(self, key, value):
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            super().__setitem__(key, value)

class BitfarmBaseObject(object):
    """Basis-Klasse für BitfarmDocument und BitfarmWorkflow. Die Basisklasse enthält alle Attribute die benötigt werden,
    um ein Objekt mittels Task-Manager zu verarbeiten. Sollen weitere Objekte über den Taskmanager verarbeitet werden, so müssen diese
    von BitfarmBaseObject erben"""

    def __init__(self):
        self.__arcid = None
        """@ivar: Die Arcid zu der das Objekt gehört"""
        self.__gdocid = None
        """@ivar:gdocid des Dokuments bzw. des referenzierenden Workflows"""
        self.profile = None
        """@ivar:das aktuelle Profil, dem das Objekt zugehörig ist
        @type: str"""
        self.version = 36
        """die Version des Objekts, derzeit Version 3.6
        @type: int"""
        self.result = False
        """@ivar: das Ergebnis eines Tasks, kann für folgende Tasks in einer Condition abgefragt werden.
        @type: bool"""
        self.q_id = None
        """@ivar: die aktuelle q_id in der sys_queue
        @type: int"""
        self.wf_id = None
        """@ivar: wf_id eines WF aus der sys_wfd Tabelle, bei Dokumenten normalerweise 0
        @type: int"""
        self.tim_id = None
        """@ivar: tim_id eines WF aus der sys_timer Tabelle
        @type: int"""
        self.tim_ref = None
        """@ivar: tim_ref eines WF aus der sys_timer Tabelle
        @type: int"""
        self.prio = 2
        """@ivar: Attribute zur Queueverarbeitung, die Prio des Dokuments in der Verarbeitung, default = 2
        0 = adhoc (wird nur bei Benutzerinteraktion gesetzt) 1 = hoch, 2 = normal, 3 = niedrig, es können aber beliebig weitere Abstufungen hinzugefügt werden
        @type: int"""
        self.debug = False
        """@ivar: Debug-Modus für dieses Dokument, default = False
        @type: bool"""
        self.userid = None
        """@ivar:UserID des Objects
        @type: int"""
        self.groupid = None
        """@ivar:GruppenID des Objects
        @type: int"""
        self.options = None
        """@ivar:Optionen des Objects
        @type: int"""
        self.required = None
        """@ivar:Required-Zähler für Gruppenwiedervorlagen, wieviele einer Gruppe müssen fortsetzen, 0 = Alle müssen fertigsetzen
        @type: int"""
        self.date = None
        """@ivar:aktuelles Datum des Objects
        @type: datetime.datetime"""
        self.alarm = None
        """@ivar:Tage bis eine Wiedervorlage als fällig markiert wird
        @type: int"""
        self.alarm_date = None
        """@ivar:Das Datum der Fälligkeit
        @type: dateteime.datetime"""
        self.created = None
        """@ivar:Erstellungsdatum
        @type: datetime.datetime"""
        self.modified = False
        """@ivar: Wurde das Objekt verändert?
        @type: bool"""
        self.modified_date = None
        """@ivar:Änderungsdatum
        @type: datetime.datetime"""
        self.modified_userid = None
        """@ivar:User der die Änderung vorgenommen hat
        @type: int"""
        self.error = False
        """@ivar: Error-Flag für dieses Dokument, default = False, wenn es gesetzt wird, sollte result auch auf false gesetzt werden.
        @type: bool"""
#        self.duplicate = False
#        """@ivar: Flag für ein Objekt, wenn es als Duplette erkannt wird. Das Objekt kann dann vom Taskmanager entfernt werden.
#        So kann noch ein Notify-Task ausgeführt werden, falls in Taskwarning definiert. Über conditions kann man die Duplettenprüfung an Bedingungen knüpfen.
#        @type: bool"""
        self.tasks = []
        """@ivar: Liste mit BitfarmTaskObjects"""
        self.attachments = []
        """@ivar: ein Liste  von BitfarmAttachment Instanzen
        @type: list"""
        self.tempid = None
        """@ivar: GUID eines neuen, noch nicht gespeicherten Objekts"""

        self.__encoding = None
        self.__errortext = u""

    def _get_gdocid(self):
        return self.__gdocid

    def _set_gdocid(self, _gdocid):
        if self.__gdocid is None:
            if type(_gdocid) is int:
                self.__gdocid = _gdocid
            else:
                self.__gdocid = None

    def _del_gdocid(self):
        self.__gdocid = None

    def _get_arcid(self):
        return self.__arcid

    def _set_arcid(self, _arcid):
        if type(_arcid) is int:
            self.__arcid = _arcid
        else:
            self.__arcid = None

    def _get_errortext(self):
        return self.__errortext

    def _set_errortext(self, _txt):
        if _txt is not None:
            if self.__errortext is None:
                self.__errortext = ''
            self.__errortext += _txt +'\r\n'

    def _get_encoding(self):
        return self.__encoding

    def _set_encoding(self, _enc):
        if self.__encoding is None:
            self.__encoding = _enc

    gdocid = property(_get_gdocid, _set_gdocid, _del_gdocid)
    """die gdocid des Dokuments (readonly). Kann nicht gesetzt werden, sondern wird beim Laden (load) angegeben oder beim Erzeugen (create) erzeugt
    @type: int"""
    arcid = property(_get_arcid, _set_arcid)
    """ die Archiv-ID des aktuellen Dokuments
    @type: int"""
    errortext = property(_get_errortext,  _set_errortext)
    """Fehlermeldungen die während der Taskverarbeitung erzeugt werden, Fehlermeldungen werden nicht überschrieben sondern zeilenweise ergänzt
    @type: unicode"""
    encoding = property(_get_encoding, _set_encoding)
    """Das Encoding des Objekts. Wird bei Laden aus der Datenbank automatisch gesetzt,
    und kann dann nicht mehr während der Laufzeit des Objekts geändert werden,
    entspricht dem Datenbank-Encoding (BitfarmCon.db_encoding)"""


class BitfarmBaseCon(object):
    """
    Bitfarm Configurations-Subklasse.
    Diese Klasse stellt ein einzelnes Profil zur Verfügung, beim Initialisieren kann der Name oder der komplette Pfad auf
    ein Profil übergeben werden.
    con = BitfarmCon('bitfarm')
    Eine Datenbank-Engine wird über die Methode create_db_engine hergestellt, diese wird in con.engine hinterlegt.
    Hier sind auch die Funktionen hinterlegt um User und Gruppen zu lesen, da diese profilabhängig und nicht
    workflow- oder dokumentabhängig sind. Dafür muss aber die DB-Connection übergeben werden, da diese ja erst zur Laufzeit
    erzeugt wird.
    Per Default wird das Con-Objekt als Version 35 erstellt, um abwärtskompatibiltät zur Datenbank 35 zu gewährleisten.
    Jede höhere Version (36, 40) muss in der Con-Datei in der Sektion [Server] angegeben werden.
    @version:
    """
    count = 0
    log = logging.getLogger("Main.BitfarmBaseCon")
    log.setLevel(logging.WARNING)

    if sys.platform == 'win32':
        mypath = os.path.split(os.path.abspath(sys.argv[0]))[0]
    else:
        mypath = os.getcwd()
    mypath = win32file.GetLongPathName(mypath)
    if os.path.split(mypath)[1] == 'bfaServer36':   # laufen wir als OneDir Version?
        mypath = os.path.split(mypath)[0]

    try:
        bfapath = os.environ['bitfarm-archiv']
    except:
        bfapath = mypath

    locale.setlocale(locale.LC_ALL, '')

    def __init__(self, *args, **kwargs):
        """
        Erstellt eine Profil-Instanz. Es kann entweder ein Profilname oder der Pfad auf eine .Con-Datei angegeben werden.
        Pfad auf die Con-Datei kann relativ oder mit Systemvariablen sein. In self.profile wird aber immer der reine Profilname
        (kleingeschrieben) hinterlegt.
        Wird nur ein Profilname angegeben, so wird die Profildatei erst in %bitfarm-archiv% (self.bfapath) gesucht, wenn nicht vorhanden,
        dann im aktuellen Arbeitsverzeichnis (self.mypath).
        """
        self.log.debug("Initialisiere ein bitfarm_con Object")
        # self.__tempname = tempname

        self.gpl = False
        self.config = None
        self.profile = None
        """@ivar: Name des Profils"""
        self.confile = None
        """@ivar: Pfad auf die .Con-Datei"""
        self.engine = None
        """@ivar: hinterlegte DB-Engine"""
        self.sys_queue = None
        """@ivar: Binding der sys_queue Tabelle, falls vorhanden, ansonsten None"""
        self.sys_arc = None
        """@ivar: Binding der sys_arc-Tabelle, sollte vorhanden sein :-) """
        self.sys_arcxs = None
        """@ivar: Binding der sys_arcxs-Tabelle"""
        self.sys_gdocid = None
        """@ivar: Binding der sys_gdocid-Tabelle, nur v35"""
        self.sys_doc = None
        """@ivar: Binding der sys_doc-Tabelle, ab v36"""
        self.sys_docxs = None
        """@ivar: Binding der sys_docxs-Tabelle, ab v36"""
        self.sys_doctxt = None
        """@ivar: Binding der sys_doctxt-Tabelle, ab v36"""
        self.sys_docidx = {}
        """@ivar: Binding der sys_docidx-Tabelle, ab v36
            @type: dict"""
        self.sys_doclifecycle = None
        """@ivar: Binding der sys_doclifecycle-Tabelle, ab v36"""
        self.arc_tbl = None
        """@ivar: Binding an die aktuell ausgewählte Archiv-Tabelle, wird ggf. innerhalb BitfarmDocProcessing oder BitfarmWorkflowProcessing erzeugt, nur v35"""
        self.sys_doclock = None
        """@ivar: Binding der sys_doclock-Tabelle, nur v35"""
        self.sys_doclink = None
        """@ivar: Binding der sys_doclink-Tabelle"""
        self.sys_bookmarks = None
        """@ivar: Binding der sys_bookmarks-Tabelle"""
        self.sys_statefieldxs = None
        """@ivar: Binding der sys_statefieldxs-Tabelle"""
        self.sys_statefields = None
        """@ivar: Binding der sys_statefields-Tabelle"""
        self.sys_addfieldxs = None
        """@ivar: Binding der sys_addfieldxs-Tabelle"""
        self.sys_addfields = None
        """@ivar: Binding der sys_addfields-Tabelle"""
        self.sys_user = None
        """@ivar: Binding der sys_user-Tabelle"""
        self.sys_group = None
        """@ivar: Binding der sys_group-Tabelle"""
        self.sys_groupuser = None
        """@ivar: Binding der sys_groupuser-Tabelle"""
        self.sys_group = None
        """@ivar: Binding der sys_group-Tabelle"""
        self.history = None
        """@ivar: Binding an die aktuelle History-Tabelle, nur v35"""
        self.sys_history = None
        """@ivar: Binding an die History-Tabelle, ab v36 """
        self.sys_timer = None
        """@ivar: Binding an die sys_timer-Tabelle """
        self.sys_timerhist = None
        """@ivar: Binding an die sys_timerhist-Tabelle """
        self.sys_wfd = None
        """@ivar: Binding der sys_wfd-Tabelle, ansonsten None"""
        self.sys_workflow = None
        """@ivar: Binding der sys_workflow-Tabelle, ansonsten None"""
        self.sys_garbage = None
        """@ivar:  Binding der sys_garbage-Tabelle, ansonsten None"""
        self.sys_duplicates = None
        """@ivar: Binding der sys_duplicates-Tabelle, ansonsten None"""
        self.sys_sap = None
        """@ivar: Binding der sys_sap-Tabelle, ansonsten None"""
        self.sys_holiday = None
        """@ivar: Binding der sys_holiday-Tabelle, ansonsten None"""
        self.sys_superior = None
        """@ivar: Binding der sys_superior-Tabelle, ansonsten None"""
        self.sys_svn = None
        """@ivar: Binding der sys_svn-Tabelle, ansonsten None"""
        self.sys_abo = None
        """@ivar: Binding der sys_abo-Tabelle"""
        self.sys_anno = None
        """@ivar: Binding der sys_annotations-Tabelle"""
        self.sys_com = None
        """@ivar: Binding der sys_com-Tabelle"""
        self.db = None
        """@ivar: Datenbank-Connection, über connect_db() erzeugt, sollte nicht innerhalb von Threads verwendet werden """
        self.ini = {}
        """@ivar: Dictionary mit dem Inhalt der scripts.ini """
        self.locale = locale.getdefaultlocale()
        """@ivar: lokale Sprache und Codepage"""
        self.localeconv = locale.localeconv()
        """@ivar: lokale Einstellungen des Gebietsschemas"""
        self.meta = None
        self.history_map = None
        """@ivar: ans Profil gebundene Instanz der HISTORY_MAP Tabelle (dictionary)"""
        self.server_config = {}
        """@ivar: Inhalt der sys_config als Dictionary."""

        try:
            self.history_map = HISTORY_MAP
        except:
            self.history_map = None
            self.log.warning('keine History_map geladen, Historyeinträge nicht möglich')

        self.emptydocs = {}
        """@ivar: die vorhandenen Vorlagen als Bezeichner:Pfad Paare"""

        # bfaServer [SERVER]
        self.bfaserver = None
        """@ivar: Hostname des bfaServers"""
        self.bfaserverport = None
        """@ivar: Port des bfaServers"""
        self.bfaserverports = None
        """@ivar: mehrere semikolongetrennte Ports des bfaServers36"""
        self.certificate = None
        """@ivar: Zertifikatsfile ds bfaServers"""
        self.concurrent = 0
        """@ivar: Anzahl concurrent Nutzer, 1 Nutzer sind default"""
        self.named = 5
        """@ivar: Anzahl named Nutzer, 5 Nutzer sind default"""
        self.ssl = None
        """@ivar: depreceated! nutzt SSL-Verschlüsselung(Achtung Viewer unterstützt derzeit keine SSL Verschlüsselung ab v3.5.2"""
        self.nossl = False
        """@ivar: Schalter um in der Version 3.6.2 die SSL verschlüsselte Verbindung zw Server und Client zu unterbinden"""
        self.auto_reconnect = False
        """@ivar: Der Server hält den angemeldeten Benutzer bei einem Verbindungsabruch noch in der Berechtigungsliste. 
        Der Client kann dann innerhalb dieser Zeit (default 90sec) die Verbindung wiederaufnehmen. Erfordert, dass der Client periodisch
        einen reqHeartBeat an den Server sendet (default alle 30sec.)"""
        self.user_timeout = 90
        """@ivar: Die Zeit, die ein User in der Berechtigungstabelle bleibt, in der vom Client weitere Requests empfangen werden können.
        default = 90 (sec)"""
        # db connection [CONNECTION]
        self.server = None
        """@ivar: Host des Datenbank-Servers"""
        self.port = None
        """@ivar: Port ders Datenbank-Servers"""
        self.db_name = None
        """@ivar: Datenbankname"""
        self.user = None
        """@ivar: Datenbank-User"""
        self.password = None
        """@ivar: Passwort des Datenbankusers"""
        self.domain = None
        """@ivar: Domäne des bitfarm Servers"""
        self.domaincontroller = None
        """@ivar: Domiancontroler, falls dieser explizit angegeben werden muss für die LDAP Verbindung"""
        self.ldapfilter = None
        """@ivar: ldapfilter, falls bei der Authentifizierung nötig, default wird * verwendet"""
        self.ldap_ssl = False
        """@ivar: ldap_ssl, falls der AD Dienst SSL Verschlüsselung benutzt"""
        self.ldap_tls = False
        """@ivar: ldap_tls, falls der AD Dienst TLS Verbindung benötigt. Wird ggf in Zukunft default = True werden, wegen LDAP Signing"""
        self.ldap_ssl_port = 636
        """@ivar: ldap_ssl_port, optional (default 636)"""
        self.ldap_ssl_cert = None
        """@ivar: ldap_ssl_cert, optional, falls eins gesondertes Zertifikat für die SSL Verschlüsselung benötigt wird"""
        self.ldap_debug = False
        self.domains = []
        self.domaincontrollers = {}
        self.use_ldap_ssl = {}
        self.use_ldap_tls = {}
        ldap_ssl = ''
        ldap_tls = ''
        self.develop = None
        """@ivar: Schalter um im Entwicklermodus zu schalten"""

        self.db_engine = None
        """@ivar: Name der Datenbank-Engine, derzeit nur 'mysql' verfügbar"""
        self.db_encoding = None
        """@ivar: das Datenbank-Encoding der Client-Connection, bei 'mysql' wird 'latin1' (v3.5) bzw 'utf8' (v3.6) als default eingetragen"""
        self.mysql_encoding = None
        """@ivar: das Server-Encoding der MySQL entweder 'latin1' (v3.5) bzw 'utf8mb4' (v3.6)"""
        self.db_collation = None
        """@ivar: die Datenbank-Collation, bei 'mysql' wird 'latin1_german_ci' (v3.5) bzw 'utf8mb4_unicode_ci' (v3.6) als default eingetragen"""
        self.tables = None
        """@ivar: Liste der Datenbank-Tabellen der verbundenen Datenbank, wird bei create_db_engine eingelesen"""
        self.version = 35
        """@ivar: Datenbank, bzw bitfarm Serverversion, aktuell Version 3"""
        self.undistributed_arcid = None
        """@ivar: arcid des unverteilt-Archives"""
        self.max_row_size = None
        """@ivar: die maximale Spaltengröße einer Datenbank-Tabelle, der geladenen Datenbank, ist von der Datenbank-Engine abhängig (MySQL = 65535)"""
        self.suuser = None
        """@ivar: lesender Datenbank-User"""
        self.supassword = None
        """@ivar: Passwort des lesenden Datenbankusers"""
        self.db_ssl_ca = 'ca.pem'
        """@ivar: der Pfad auf die ca.pem, für eine SSL verschlüsselte Verbindung zur MySQL8"""
        self.db_ssl_clientcert = 'client-cert.pem'
        """@ivar: der Pfad auf die client-cert.pem, für eine SSL verschlüsselte Verbindung zur MySQL8"""
        self.db_ssl_clientkey = 'client-key.pem'
        """@ivar: der Pfad auf die client-key.pem, für eine SSL verschlüsselte Verbindung zur MySQL8"""

        # archivierung [ARCHIVING]
        self.archiveserver = None
        """@ivar: Host des Archivierungsserver"""
        self.archiveport = None
        """@ivar: Port des Archivierungsserver"""

        # [SCRIPTS]
        self.serverprogram = None
        """@ivar: Lokaler Pfad zum bitfarm-Archiv Installationsverzeichnis (nicht UNC)"""
        self.clientprogram = None
        """@ivar: UNC Pfad zum bitfarm-Archiv Installationsverzeichnis"""
        self.clientspooler = None
        """@ivar: Lokaler Pfad zum Übergabe-Verzeichnis (nicht UNC)"""
        self.uncclientspooler = None
        """@ivar: UNC Pfad zum Übergabe-Verzeichnis """
        self.archive_rs = None
        """@ivar: UNC Pfad zur revisionssicheren Archivablage"""
        self.archive_nrs = None
        """@ivar: UNC Pfad zur nicht revisionssicheren Archivablage"""
        self.storages = []
        """@ivar: Systemvariablen mit UNC Pfad zum Storage als Liste angelegt, ein Dokument verweist über die Variable doc.storage auf sein Storage, v36"""
        self.queue = None
        """@ivar: Pfad des Queue-Verzeichnisses, wenn nicht angegeben dann %bitfarm-archiv%\queue"""
        self.omnipage = None
        """@ivar: Lokaler Pfad und Dateiname zur Omnipageinstallation"""
        self.ghostscript = None
        """@ivar: Lokaler Pfad und Dateiname zur Ghostscriptinstallation (wird auch von gs_detect.exe gesetzt)"""
        self.gsview = None
        """@ivar: Lokaler Pfad und Dateiname zur GSView-Installation"""
        self.openofficepath = None
        """@ivar: Pfad zur OpenOfficeInstallation (oder Portable bei GPL) """
        self.smtpserver = None
        """@ivar: Hostname oder IP-Adresse eines SMTP Servers, der die Erinnerungsmails aus der Wiedervorlage oder Mails aus dem Workflow versendet oder Verknüpfungsmails aus dem Viewer heraus."""
        self.sendername = None
        """@ivar: Absenderadresse für Server-Mails."""
        self.smtpport = None
        """@ivar: Port des smtp-Servers (normalerweise 25)"""
        self.authuser = None
        """@ivar: Benutzername für SMTP Server, falls eine Authentifizierung nötig ist."""
        self.authpass = None
        """@ivar: Passwort für SMTP Server, falls eine Authentifizierung nötig ist."""
        self.auth_cred  = None
        """@ivar: Passwort für SMTP Server mit der dbapi verschlüsselt."""
        self.usetls = False
        """@ivar: STARTTLS verwenden für die SMTP Verbindung."""
        self.smtp_oauth = None
        """@ivar: verwende OAuth2 für SMTP Versand"""
        self.duplicate_testing = False
        """@ivar: aktiviert die Duplettenprüfung"""
        self.adminmail = None
        """@ivar: Empfänger für warnings, falls nichts anderes angegeben"""
        self.templates = None
        """@ivar: UNC Pfad zum bitfarm-Archiv Templates-Verzeichnis"""
        self.workflowfile = None
        """@ivar:  UNC-Pfad und Dateiname der zentralen WFD-Datei (Nur nötig bei Multi-Server Umgebungen)"""
        self.undistributed = None
        """@ivar: Name des unverteilt Archivs"""
        self.workflows_dir = None
        """@ivar: UNC Pfad zum zentralen bitfarm-Archiv Workflow-Verzeichnis (Nur nötig bei Multi-Server Umgebungen)"""
        self.tempdir = None
        """@ivar: Pfad zum Temp-Verzeichnis"""
        self.garbage_id = None
        """@ivar: Arcid des Garbagearchives"""
        self.duplicates_id = None
        """@ivar: Arcid des Dublettenarchives"""

        # [BACKUP]
        self.backuppath = None
        """@ivar: Zielpfad in dem die RAR-Dateien, die vom Backupscript  erstellt werden, abgelegt werden sollen."""
        self.sqldatapath = None
        """@ivar: Lokaler Pfad zu den MySQL Datenbank Dateien. """
        self.backuparchive = False
        """@ivar: """
        self.backupnrsarchive = False
        """@ivar: Soll die revisionssichere Archivablage mit im Backup eingeschlossen werden? In den meisten Fällen false, da dies über einen normales Fileserverbackup läuft."""
        self.backupversion = False
        """@ivar: Soll das Bitfarm-Programmverzeichnis mit im Backup eingeschlossen werden?"""
        self.backupworkflows = False
        """@ivar: Sicherung des workflow-Verzeichnisses"""
        self.backupsvnrepository = False
        """@ivar: Sicherung des SVN-Repositorys (sinnvoll bei Einsatz der Versionierung). In den meisten Fällen false, da dies über einen normales Fileserverbackup läuft"""
        self.backupmysqllog = False
        """@ivar: Backup der MySQL Logdateien"""
        self.filesize = None
        """@ivar: Maximale Größe in KB einer RAR Datei, bevor eine neue RAR Datei erstellt wird."""
        self.backupinterval = None
        """@ivar: Anzahl der Tage, die die RAR Dateien im backuppath aufbewahrt werden, bevor sie automatisch per Script gelöscht werden."""
        self.backupuse7zip = False
        """@ivar: das Backupscript soll 7zip zum Packen verwenden. der Zippath muss in der scripts.ini angegeben sein."""
        self.backupmysqlservice = None
        """@ivar: BDer Dienstname der MySQL, damit das Backupscript diese beenden kann."""

        # [OPTIONS]
        self.logdir = None
        """@ivar: UNC-Pfad zum zentralen Verzeichnis für Client-Logs (Schreibbrechtigung für User setzen; nur zum Debuggen verwenden (Performance))"""
        self.savepass = None
        """@ivar: Hier schaltet man im Viewer die Möglichkeit das Passwort zu speichern aus (0)"""
        self.manuscanplugin = None
        """@ivar: Dateiname des Manuscanplugins, das eingebunden werden soll."""
        self.importerplugin = None
        """@ivar: Dateiname des Importerplugins, das eingebunden werden soll."""
        self.svninfo = None
        """@ivar: Schaltet die SVNInfo-Funktion im Viewer ein (1)"""
        self.linkchars = None
        """@ivar: Ab der hier angegebenen Anzahl von Zeichen beginnt der Viewer bei Verknüpfungen mit der Autovervollständigung."""
        self.regkeyarchived = None
        """@ivar: Wenn dieser Wert höher ist, als der lokal in der Registry gespeicherte Wert, überträgt der Viewer seine Registrierungsdaten an den Server"""
        self.ignorerightsatimport = None
        """@ivar: Hier schaltet man für Importer und Manuscan die Überprüfung der Zusatzfeldrechte aus (1)"""
        self.usesearch35 = False
        """@ivar: Schaltet für Zusatzfelder (Text) auf das Suchverhalten der Version 35"""
        self.usestatus35 = False
        """@ivar: Schaltet für Statusfelder auf die Positionsabhängige Bitmaske der Version 35 (zB für Plugins)"""
        self.resultlist_showinedit = False
        """@ivar: Zeigt in der Trefferliste die Dokumente in Bearbeitung grün an, anstatt Dokumente der Wiedervorlgae"""

        #[DLM]
        self.expired_dir = None
        """@ivar: Das Verzeichnis, in das die abgelaufenen Dokumente gesichert werden."""
        self.check_time = None
        """@ivar: der Zeitpunkt, wann täglich gecheckt werden soll. 
        @type: datetime.time"""
        self.dont_check_expired = False
        """@ivar: Schaltet die Überprüfung der Dokumentenlebensdauer vollständig ab."""

        #[Updater]
        self.update_files = []
        """@ivar: Datei(en, semikolongetrennt) die dem Client Updater zur Verfügung gestellt werden"""
        self.update_dirs = []
        """@ivar: Verzeichniss(e, semikolongetrennt) die dem Client Updater zur Verfügung gestellt werden"""

        #[client]
        self.bfauser = None
        """@ivar: der User für Client Serververbindungen, per dpapi verschlüsselt"""
        self.bfapass = None
        """@ivar: das Passwort der Client Users"""

        # BitfarmBaseCon.count += 1
        if kwargs:
            if kwargs['gpl'] is True:
                self.gpl = True

        if len(args) > 0:
            confile = args[0]
            if not os.path.isfile(os.path.abspath(os.path.expandvars(confile))):    # wenn wir keinen pfad bekommen haben, bauen wir uns einen
                fname = confile
                if fname[-4:].lower() == '.con':  # beliebter Fehler Profilname mit Endung .con angeben! Soll man nicht, ist dann aber jetzt auch egal
                    fname = fname[:-4]
                confile = os.path.join(self.bfapath, fname + '.con')
                if not os.path.exists(confile):                                     # in bfapath nicht gefunden, dann im eigenen Pfad suchen
                    confile = os.path.join(self.mypath, fname + '.con')
                    if not os.path.exists(os.path.abspath(confile)):    # finden wir unter mypath nichts, suchen wir ein Verzeichnis höher
                        searchpath = os.path.split(self.mypath)[0]
                        confile = os.path.join(searchpath, fname + '.con')    # da wir ggf. in einer PyINstaller Onedir App laufen

            self.config = configparser.ConfigParser(allow_no_value=True, interpolation=None)

            confile = os.path.abspath(os.path.expandvars(confile))  # Pfad auflösen
            if os.path.isfile(confile) and os.path.exists(confile):
                try:
                    # self.config.readfp(open(confile))
                    self.config.read_file(codecs.open(confile, "r", "cp1252"))
                except:
                    self.log.error("Die Datei %s konnte nicht gelesen werden. Bitte Aufbau und Encoding (cp1252) prüfen" % confile)
                    raise ParseConError(confile)
            else:
                self.log.debug("Die Datei %s konnte nicht gefunden werden" % confile)
                raise FindConError(confile)

            self.profile = os.path.splitext(os.path.basename(confile))[0].lower()
            self.confile = confile

            if self.config.has_section('Server'):
                try:
                    if self.config.has_option('Server', 'Version'):
                        self.version = self.config.getint('Server', 'Version')
                except:
                    self.version = 35
                if self.config.has_option('Server', 'edition'):
                    if 'gpl' in self.config.get('Server', 'edition').lower():
                        self.gpl = True
                if self.config.has_option('Server', 'Server'):
                    self.bfaserver = self.config.get('Server', 'Server')
                if self.config.has_option('Server', 'Ports'):
                    ports = self.config.get('Server', 'Ports')
                    self.bfaserverports = [int(x) for x in ports.split(";")]
                    self.bfaserverport = self.bfaserverports[0]
                elif self.config.has_option('Server', 'Port'):
                    self.bfaserverport = self.config.getint('Server', 'Port')
                if self.config.has_option('Server', 'Certificate'):
                    self.certificate = self.config.get('Server', 'Certificate')
                if self.version == 35 and self.config.has_option('Server', 'Concurrent'): # in v36 wird die profil.rc datei gelesen
                    try:
                        self.concurrent = self.config.getint('Server', 'Concurrent')
                    except:
                        self.concurrent = 5     # 5 war default bei v35
                # if self.config.has_option('Server', 'SSL'):
                #     try:
                #         self.ssl = self.config.getboolean('Server', 'SSL')
                #     except:
                #         self.ssl = False
                if self.config.has_option('Server', 'nossl'):
                    try:
                        self.nossl = self.config.getboolean('Server', 'nossl')
                    except:
                        self.nossl = False
                if self.config.has_option('Server', 'auto_reconnect'):
                    try:
                        self.auto_reconnect = self.config.getboolean('Server', 'auto_reconnect')
                    except:
                        self.auto_reconnect = False
                if self.config.has_option('Server', 'user_timeout'):
                    try:
                        self.user_timeout = self.config.getint('Server', 'user_timeout')
                        if self.user_timeout < 60:
                            self.user_timeout = 60
                    except:
                        self.user_timeout = 90
                if self.config.has_option('Server', 'develop'):
                    try:
                        self.develop = self.config.getboolean('Server', 'develop')
                    except:
                        self.develop = False

            if self.config.has_section('Connection'):
                if self.config.has_option('Connection', 'server'):
                    self.server = self.config.get('Connection', 'server')
                if self.config.has_option('Connection', 'Port'):
                    self.port = self.config.getint('Connection', 'Port')
                if self.config.has_option('Connection', 'DB'):
                    self.db_name = self.config.get('Connection', 'DB')
                if self.config.has_option('Connection', 'User'):
                    self.user = self.bitdecrypt(self.config.get('Connection', 'User'))
                if self.config.has_option('Connection', 'Pass'):
                    self.password = self.bitdecrypt(self.config.get('Connection', 'Pass'))
                if self.config.has_option('Connection', 'suuser'):
                    self.suuser = self.bitdecrypt(self.config.get('Connection', 'suuser'))
                if self.config.has_option('Connection', 'supass'):
                    self.supassword = self.bitdecrypt(self.config.get('Connection', 'supass'))
                if self.config.has_option('Connection', 'Domain'):
                    self.domain = self.config.get('Connection', 'Domain').lower()
                if self.config.has_option('Connection', 'domaincontroler'):     # alter Rechtschreibfehler, zur backwardcompatibility
                    self.domaincontroller = self.config.get('Connection', 'domaincontroler').lower()
                if self.config.has_option('Connection', 'domaincontroller'):
                    self.domaincontroller = self.config.get('Connection', 'domaincontroller').lower()
                if self.config.has_option('Connection', 'ldapfilter'):
                    self.ldapfilter = self.config.get('Connection', 'ldapfilter')
                if self.config.has_option('Connection', 'ldap_ssl'):
                    ldap_ssl = self.config.get('Connection', 'ldap_ssl')
                if self.config.has_option('Connection', 'ldap_tls'):
                    ldap_tls = self.config.get('Connection', 'ldap_tls')
                if self.config.has_option('Connection', 'ldap_ssl_port'):
                    self.ldap_ssl_port = self.config.getint('Connection', 'ldap_ssl_port')
                if self.config.has_option('Connection', 'ldap_ssl_cert'):
                    self.ldap_ssl_cert = self.config.get('Connection', 'ldap_ssl_cert')
                if self.config.has_option('Connection', 'ldap_debug'):
                    try:
                        self.ldap_debug = self.config.getboolean('Connection', 'ldap_debug')
                    except:
                        self.ldap_debug = False
                if self.config.has_option('Connection', 'db_engine'):
                    self.db_engine = self.config.get('Connection', 'db_engine')
                else:
                    self.db_engine = 'mysql'
                if self.config.has_option('Connection', 'max_row_size'):
                    self.max_row_size = self.config.get('Connection', 'max_row_size')
                else:
                    self.max_row_size = 61440   # eigentlich 65535 in der MySQL, wir machen aber unsere Tabellenspalten
                                                # nicht so voll, damit beim Feldtypändern nicht direkt eine neue Tabelle angelegt werden muss
                if self.config.has_option('Connection', 'db_encoding'):
                    self.db_encoding = self.config.get('Connection', 'db_encoding')
                    if self.db_encoding == 'utf8':      # utf8 ist nur ein Alias für utf8mb4
                        self.mysql_encoding = 'utf8mb4'
                    else:
                        self.mysql_encoding = self.db_encoding
                else:
                    if self.version == 35:
                        self.db_encoding = 'latin1'
                        self.mysql_encoding = 'latin1'
                    else:
                        self.db_encoding = 'utf8'
                        self.mysql_encoding = 'utf8mb4'

                if self.config.has_option('Connection', 'db_collation'):
                    self.db_collation = self.config.get('Connection', 'db_encoding')
                else:
                    if self.version == 35:
                        self.db_collation = 'latin1_german1_ci'
                    else:
                        self.db_collation = 'utf8mb4_unicode_ci'

                # hier werden die AD Connections hinterlegt

                if type(self.domain) is str:
                    self.domains = self.domain.split('|')
                else:
                    self.domains = [self.domain]

                if type(self.domaincontroller) is str:
                    controlers = self.domaincontroller.split('|')
                else:
                    controlers = [self.domaincontroller]

                # self.domaincontrollers = {}
                # Zuordnen der domaincontroller zu den einzelnen Domains
                for i in range(len(self.domains)):
                    if i >= len(controlers):    # Ende der Controlers Liste erreicht? Dann den letzten Eintrag wieder nehmen
                        self.domaincontrollers[self.domains[i]] = controlers[-1]
                    else:
                        self.domaincontrollers[self.domains[i]] = controlers[i]

                # self.use_ldap_ssl = {}
                if type(ldap_ssl) in [str]:
                    ldap_ssl = ldap_ssl.split('|')
                else:
                    ldap_ssl = [str(self.ldap_ssl)]
                if type(ldap_tls) in [str]:
                    ldap_tls = ldap_tls.split('|')
                else:
                    ldap_tls = [str(self.ldap_tls)]

                for i in range(len(self.domains)):
                    if i >= len(ldap_ssl):    # Ende der ldap_ssl Liste erreicht? Dann den letzten Eintrag wieder nehmen
                        self.use_ldap_ssl[self.domains[i]] = ldap_ssl[-1].lower() == 'true'
                    else:
                        self.use_ldap_ssl[self.domains[i]] = ldap_ssl[i].lower() == 'true'
                    if i >= len(ldap_tls):    # Ende der ldap_tls Liste erreicht? Dann den letzten Eintrag wieder nehmen
                        self.use_ldap_tls[self.domains[i]] = ldap_tls[-1].lower() == 'true'
                    else:
                        self.use_ldap_tls[self.domains[i]] = ldap_tls[i].lower() == 'true'

                if self.config.has_option('Connection', 'ssl_ca'):
                    self.db_ssl_ca = self.config.get('Connection', 'ssl_ca')
                if self.config.has_option('Connection', 'ssl_clientcert'):
                    self.db_ssl_clientcert = self.config.get('Connection', 'ssl_clientcert')
                if self.config.has_option('Connection', 'ssl_clientkey'):
                    self.db_ssl_clientkey = self.config.get('Connection', 'ssl_clientkey')

            if self.config.has_section('Archiving'):
                if self.config.has_option('Archiving', 'Server'):
                    self.archiveserver = self.config.get('Archiving', 'Server')
                if self.config.has_option('Archiving', 'Port'):
                    self.archiveport = self.config.getint('Archiving', 'Port')

            if self.config.has_section('Scripts'):
                if self.config.has_option('Scripts', 'serverprogramm'):
                    self.serverprogram = self.config.get('Scripts', 'serverprogramm')
                if self.config.has_option('Scripts', 'serverprogram'):
                    self.serverprogram = self.config.get('Scripts', 'serverprogram')
                if self.config.has_option('Scripts', 'clientprogramm'):
                    self.clientprogram = self.config.get('Scripts', 'clientprogramm')
                if self.config.has_option('Scripts', 'clientprogram'):
                    self.clientprogram = self.config.get('Scripts', 'clientprogram')
                if self.config.has_option('Scripts', 'clientspooler'):
                    self.clientspooler = self.config.get('Scripts', 'clientspooler')
                if self.config.has_option('Scripts', 'uncclientspooler'):
                    self.uncclientspooler = self.config.get('Scripts', 'uncclientspooler')
                if self.config.has_option('Scripts', 'archive_rs'):
                    self.archive_rs = self.config.get('Scripts', 'archive_rs')
                elif self.config.has_option('Scripts', 'archivverzeichnis'):
                    self.archive_rs = self.config.get('Scripts', 'archivverzeichnis')
                if self.config.has_option('Scripts', 'archive_nrs'):
                    self.archive_nrs = self.config.get('Scripts', 'archive_nrs')
                elif self.config.has_option('Scripts', 'archivkopie'):
                    self.archive_nrs = self.config.get('Scripts', 'archivkopie')
                if self.config.has_option('Scripts', 'storages'):
                    self.storages = self.config.get('Scripts', 'storages').split(';')
                else:
                    self.storages = ['bfastore0']
                if self.config.has_option('Scripts', 'queue'):
                    self.queue = self.config.get('Scripts', 'queue')
                else:
                    if self.serverprogram is not None and os.path.isdir(self.serverprogram):
                        self.queue = os.path.join(self.serverprogram, 'queue')
                    else:
                        self.queue = os.path.join(self.mypath, 'queue')
                if self.config.has_option('Scripts', 'omnipage'):
                    self.omnipage = self.config.get('Scripts', 'omnipage')
                if self.config.has_option('Scripts', 'ghostscript'):
                    self.ghostscript = self.config.get('Scripts', 'ghostscript')
                if self.config.has_option('Scripts', 'gsview'):
                    self.gsview = self.config.get('Scripts', 'gsview')
                if self.config.has_option('Scripts', 'smtpserver'):
                    self.smtpserver = self.config.get('Scripts', 'smtpserver')
                if self.config.has_option('Scripts', 'usetls'):
                    try:
                        self.usetls = self.config.getboolean('Scripts', 'usetls')
                    except:
                        self.usetls = False
                if self.config.has_option('Scripts', 'smtpport'):
                    try:
                        self.smtpport = self.config.getint('Scripts', 'smtpport')
                    except:
                        self.smtpport = 25 if not self.usetls else 587
                else:
                    self.smtpport = 25 if not self.usetls else 587
                if self.config.has_option('Scripts', 'sendername'):
                    self.sendername = self.config.get('Scripts', 'sendername')
                if self.config.has_option('Scripts', 'authuser'):
                    self.authuser = self.config.get('Scripts', 'authuser')
                if self.config.has_option('Scripts', 'authpass'):
                    self.authpass = self.config.get('Scripts', 'authpass')
                if self.config.has_option('Scripts', 'auth_cred'):
                    self.authpass = self.config.get('Scripts', 'auth_cred')
                    self.authpass = dpapi.decryptData(base64.b64decode(self.authpass))
                    self.authpass = self.authpass.decode()
                if self.config.has_option('Scripts', 'smtp_oauth'):
                    self.smtp_oauth = self.config.get('Scripts', 'smtp_oauth')
                #                if self.config.has_option('Scripts', 'schattenarchiv'):
                #                    try:
                #                        self.schattenarchiv = self.config.getboolean('Scripts', 'schattenarchiv')
                #                    except:
                #                        self.schattenarchiv = False
                if self.config.has_option('Scripts', 'bfa_server_smtp'):
                    try:
                        self.bfa_server_smtp = self.config.getboolean('Scripts', 'bfa_server_smtp')
                    except:
                        self.bfa_server_smtp = False
                if self.config.has_option('Scripts', 'duplicate_testing'):
                    try:
                        self.duplicate_testing = self.config.getboolean('Scripts', 'duplicate_testing')
                    except:
                        self.duplicate_testing = False
                if self.config.has_option('Scripts', 'adminmail'):
                    self.adminmail = self.config.get('Scripts', 'adminmail')
                if self.config.has_option('Scripts', 'templates'):
                    self.templates = self.config.get('Scripts', 'templates')
                if self.config.has_option('Scripts', 'workflowfile'):
                    self.workflowfile = self.config.get('Scripts', 'workflowfile')
                if self.config.has_option('Scripts', 'openofficepath'):
                    self.openofficepath = self.config.get('Scripts', 'openofficepath')
                if self.config.has_option('Scripts', 'undistributed'):
                    self.undistributed = self.config.get('Scripts', 'undistributed')
                elif self.config.has_option('Scripts', 'unverteilt'):
                    self.undistributed = self.config.get('Scripts', 'unverteilt')
                if self.config.has_option('Scripts', 'undistributed_arcid'):
                    try:
                        self.undistributed_arcid = self.config.getint('Scripts', 'undistributed_arcid')
                    except:
                        self.undistributed_arcid = None
                if self.config.has_option('Scripts', 'workflows'):
                    self.workflows_dir = self.config.get('Scripts', 'workflows')
                if self.config.has_option('Scripts', 'tempdir'):
                    self.tempdir = self.config.get('Scripts', 'tempdir')
                else:
                    self.tempdir = os.path.join(self.mypath, 'temp')
                if self.config.has_option('Scripts', 'garbage_id'):
                    try:
                        self.garbage_id = self.config.getint('Scripts', 'garbage_id')
                    except:
                        self.garbage_id = None
                if self.config.has_option('Scripts', 'duplicates_id'):
                    try:
                        self.duplicates_id = self.config.getint('Scripts', 'duplicates_id')
                    except:
                        self.duplicates_id = None

            if self.config.has_section('Backup'):
                if self.config.has_option('Backup', 'backuppath'):
                    self.backuppath = self.config.get('Backup', 'backuppath')
                if self.config.has_option('Backup', 'sqldatapath'):
                    self.sqldatapath = self.config.get('Backup', 'sqldatapath')
                if self.config.has_option('Backup', 'backuparchiv'):
                    try:
                        self.backuparchive = self.config.getboolean('Backup', 'backuparchiv')
                    except:
                        self.backuparchive = False
                if self.config.has_option('Backup', 'backupnrsarchiv'):
                    try:
                        self.backupnrsarchive = self.config.getboolean('Backup', 'backupnrsarchiv')
                    except:
                        self.backupnrsarchive = False
                if self.config.has_option('Backup', 'backupversion'):
                    try:
                        self.backupversion = self.config.getboolean('Backup', 'backupversion')
                    except:
                        self.backupversion = False
                if self.config.has_option('Backup', 'backupworkflows'):
                    try:
                        self.backupworkflows = self.config.getboolean('Backup', 'backupworkflows')
                    except:
                        self.backupworkflows = False
                if self.config.has_option('Backup', 'backupsvnrepository'):
                    try:
                        self.backupsvnrepository = self.config.getboolean('Backup', 'backupsvnrepository')
                    except:
                        self.backupsvnrepository = False
                if self.config.has_option('Backup', 'backupmysqllog'):
                    try:
                        self.backupmysqllog = self.config.getboolean('Backup', 'backupmysqllog')
                    except:
                        self.backupmysqllog = False
                if self.config.has_option('Backup', 'filesize'):
                    try:
                        self.filesize = self.config.getint('Backup', 'filesize')
                    except:
                        self.filesize = 1024*1024
                if self.config.has_option('Backup', 'backupinterval'):
                    try:
                        self.backupinterval = self.config.getint('Backup', 'backupinterval')
                    except:
                        self.backupinterval = 14
                elif self.config.has_option('Backup', 'backupzeitraum'):
                    try:
                        self.backupinterval = self.config.getint('Backup', 'backupinterval')
                    except:
                        self.backupinterval = 14
                if self.config.has_option('Backup', 'Use7zip'):
                    try:
                        self.backupuse7zip = self.config.getboolean('Backup', 'Use7zip')
                    except:
                        self.backupuse7zip = False

            if self.config.has_section('Options'):
                if self.config.has_option('Options', 'logdir'):
                    self.logdir = self.config.get('Options', 'logdir')
                if self.config.has_option('Options', 'savepass'):
                    try:
                        self.savepass = self.config.getboolean('Options', 'savepass')
                    except:
                        self.savepass = False
                if self.config.has_option('Options', 'manuscanplugin'):
                    self.manuscanplugin = self.config.get('Options', 'manuscanplugin')
                if self.config.has_option('Options', 'importerplugin'):
                    self.importerplugin = self.config.get('Options', 'importerplugin')
                if self.config.has_option('Options', 'SvnInfo'):
                    try:
                        self.svninfo = self.config.getboolean('Options', 'SvnInfo')
                    except:
                        self.svninfo = False
                if self.config.has_option('Options', 'LinkChars'):
                    self.linkchars = self.config.get('Options', 'LinkChars')
                if self.config.has_option('Options', 'regkeyarchived'):
                    try:
                        self.regkeyarchived = self.config.getint('Options', 'regkeyarchived')
                    except:
                        self.regkeyarchived = 0
                if self.config.has_option('Options', 'ignorerightsatimport'):
                    try:
                        self.ignorerightsatimport = self.config.getboolean('Options', 'ignorerightsatimport')
                    except:
                        self.ignorerightsatimport = False
                if self.config.has_option('Options', 'usesearch35'):
                    try:
                        self.usesearch35 = self.config.getboolean('Options', 'usesearch35')
                    except:
                        self.usesearch35 = False
                if self.config.has_option('Options', 'usestatus35'):
                    try:
                        self.usestatus35 = self.config.getboolean('Options', 'usestatus35')
                    except:
                        self.usestatus35 = False
                if self.config.has_option('Options', 'resultlist_showinedit'):
                    try:
                        self.resultlist_showinedit = self.config.getboolean('Options', 'resultlist_showinedit')
                    except:
                        self.resultlist_showinedit = False

            if self.config.has_section('History_ext'):
                for option in self.config.options('History_ext'):
                    if int(option) < 900:
                        raise WrongHistoryIndexError(confile)
                    if option in self.history_map:
                        raise DuplicateHistoryIndexError(confile)
                    else:
                        self.history_map[int(option)] = self.config.get('History_ext', option)

            if self.config.has_section('EmptyDocs'):
                for option in self.config.options('EmptyDocs'):
                    self.emptydocs[option] = self.config.get('EmptyDocs', option)
                    self.emptydocs[option] = os.path.expandvars(self.emptydocs[option])

            if self.config.has_section('DLM'):
                if self.config.has_option('DLM', 'expired_dir'):
                    self.expired_dir = self.config.get('DLM', 'expired_dir')
                if not self.expired_dir:
                    self.expired_dir = os.path.join(self.mypath, 'expired')
                if self.config.has_option('DLM', 'check_time'):
                    try:
                        _check_time = self.config.get('DLM', 'check_time')
                        _check_time = _check_time.split(':')
                        self.check_time = datetime.time(int(_check_time[0]), int(_check_time[1]))
                    except:
                        self.check_time = datetime.time(6, 0)
                else:
                    self.check_time = datetime.time(6, 0)
                if self.config.has_option('DLM', 'dont_check_expired'):
                    try:
                        self.dont_check_expired = self.config.getboolean('DLM', 'dont_check_expired')
                    except:
                        self.dont_check_expired = False
            else:
                self.expired_dir = os.path.join(self.mypath, 'expired')
                self.check_time = datetime.time(6,0)

            if self.config.has_section('Updater'):
                if self.config.has_option('Updater', 'update_files'):
                    self.update_files = self.config.get('Updater', 'update_files')
                    if self.update_files:
                        self.update_files = self.update_files.split(';')
                        self.update_files = [x for x in self.update_files if x]
                if self.config.has_option('Updater', 'update_dirs'):
                    self.update_dirs = self.config.get('Updater', 'update_dirs')
                    if self.update_dirs:
                        self.update_dirs = self.update_dirs.split(';')
                        self.update_dirs = [x for x in self.update_dirs if x]

            if self.config.has_section('client'):
                if self.config.has_option('client', 'bfauser'):
                    self.bfauser = self.config.get('client', 'bfauser')
                    if self.bfauser is not None:
                        try:
                            self.bfauser = dpapi.decryptData(base64.b64decode(self.bfauser))
                        except:
                            self.bfauser = None

            if self.config.has_section('client'):
                if self.config.has_option('client', 'bfapass'):
                    self.bfapass = self.config.get('client', 'bfapass')
                    if self.bfapass is not None:
                        try:
                            self.bfapass = dpapi.decryptData(base64.b64decode(self.bfapass))
                        except:
                            self.bfapass = None

        if self.config is not None:
            self.config = None

    @staticmethod
    def _bitdecrypt(self, key):
        """
        Methode zum Entschlüsseln des User und Passworts in einer .con-Datei
        """
        gkey = bitfarm_base.GLOBALKEY
        _ret = bytearray()

        pos = 0
        for i in range(0, int(len(key) / 2)):
            _ret.append(int(gkey[pos % (len(gkey)): pos % (len(gkey)) + 2], 16) ^ int(key[i * 2:i * 2 + 2], 16))
            pos += 2
        try:
            if _ret[:2] == b'\xff\xfe':
                _ret = _ret.decode('utf-16-le')
            else:
                _ret = _ret.decode('cp1252')    # alles was nicht multibyte ist (unicode) ist ursprünglich seitens Delphi als ANIS encryptet worden
        except:
            self.log.error('_bitdecrypt: Could not decrypt key')
            _ret = None
        return _ret

    def bitdecrypt(self, key):
        """
        Methode zum Entschlüsseln des User und Passworts in einer .con-Datei
        """
        if self.gpl:
            gkey = 'b17fa300900b421333de672c41a00d'
        else:
            gkey = bitfarm_base.GLOBALKEY
        _ret = bytearray()

        pos = 0
        for i in range(0, int(len(key) / 2)):
            _ret.append(int(gkey[pos % (len(gkey)): pos % (len(gkey)) + 2], 16) ^ int(key[i * 2:i * 2 + 2], 16))
            pos += 2
        try:
            if _ret[:2] == b'\xff\xfe':
                _ret = _ret.decode('utf-16-le')
            else:
                _ret = _ret.decode('cp1252')    # alles was nicht multibyte ist (unicode) ist ursprünglich seitens Delphi als ANIS encryptet worden
        except:
            self.log.error('bitdecrypt: Could not decrypt key')
            _ret = None
        return _ret

    @staticmethod
    def _bitencrypt(self, key):
        """
        Methode zum Verschlüsseln des User und Passworts
        """
        gkey = bitfarm_base.GLOBALKEY
        _ret = ''
        pos = 0

        dkey = [int(x) for x in key.encode('utf-16-le')]
        if list(filter(lambda x: x > 0, dkey[1::2])):   # check if any second byte is greater 0, so it's multibyte
            dkey = [255, 254] + dkey                    # add bom
        else:
            dkey = dkey[0::2]

        for i in range(0, len(dkey)):
            _ret += '{:02x}'.format(dkey[i] ^ int(gkey[pos % (len(gkey)):pos % (len(gkey)) + 2], 16))
            pos += 2
        return _ret.upper()

    def bitencrypt(self, key):
        """
        Methode zum Verschlüsseln des User und Passworts
        """
        if self.gpl:
            gkey = 'b17fa300900b421333de672c41a00d'
        else:
            gkey = bitfarm_base.GLOBALKEY
        _ret = ''
        pos = 0

        dkey = [int(x) for x in key.encode('utf-16-le')]
        if list(filter(lambda x: x > 0, dkey[1::2])):   # check if any second byte is greater 0, so it's multibyte
            dkey = [255, 254] + dkey                    # add bom
        else:
            dkey = dkey[0::2]

        for i in range(0, len(dkey)):
            _ret += '{:02x}'.format(dkey[i] ^ int(gkey[pos % (len(gkey)):pos % (len(gkey)) + 2], 16))
            pos += 2
        return _ret.upper()

    def read_ini(self, inifile=None):
        """Hilfsfunktion die die scripts.ini aus dem Programmverzeichnis (default)
        oder eine beliebige INI-Datei liest und in der Con-Instanz unter con.ini als BitfarmINIObject hinterlegt."""
        if inifile is None:
            inifile = os.path.join(self.bfapath, 'scripts.ini')

        if not os.path.exists(inifile):
            return False

        # self.config = configparser.ConfigParser(allow_no_value=True)
        self.config = configparser.RawConfigParser(dict_type=bfaOrderedDict, allow_no_value=True, strict=False)
        # self.config = configparser.RawConfigParser(allow_no_value=True, strict=False)
        if os.path.isfile(inifile) and os.path.exists(inifile):
            try:
                # self.config.readfp(open(inifile))
                self.config.read_file(open(inifile))
            except:
                return False

        for section in self.config.sections():
            self.ini[section.lower()] = BitfarmINIObject()
            for option in self.config.options(section):
                if option not in self.ini[section.lower()].__dict__:
                    if self.config.get(section, option).lower() == 'true':
                        self.ini[section.lower()].__dict__[option] = True
                    elif self.config.get(section, option).lower() == 'false':
                        self.ini[section.lower()].__dict__[option] = False
                    elif self.config.get(section, option).lower() == '':
                        self.ini[section.lower()].__dict__[option] = None
                    elif self.config.get(section, option).isdigit():
                        self.ini[section.lower()].__dict__[option] = int(self.config.get(section, option))
                    else:
                        self.ini[section.lower()].__dict__[option] = self.config.get(section, option)
        if self.config is not None:
            self.config = None

        return True

    def is_guid(self, guid):
        if type(guid) is uuid.UUID:
            return True
        if type(guid) is not str:
            return False

        guid = guid.lower()

        pattern = re.compile('[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')

        return pattern.search(guid) is not None


class BitfarmBaseConnections(object):
    def __init__(self, *name, **kwargs):
        """Erzeugt eine BitfarmConnections Instanz, es kann der Name / oder Pfad eine Profils übergeben werden, bzw. eine Liste von Profilen"""

        self.connections = {}
        """@ivar: Dictonary, welches alle Connections trägt
            @type: dictonary"""
        self.main = None
        """@ivar: hier ist das erste geladene Profil hinterlegt, die Main.Connection. Über diese Connection wird die sys_queue und die config_tabellen erreicht
            @type: BitfarmCon Instanz"""
        self.profiles = []
        """@ivar: die Namen der geladenen Profile
        @type: list"""

        if len(name) > 0:
            if type(name[0]) is list:
                name = list(name[0])
            for profile in name:
                if os.path.isfile(profile) and os.path.exists(profile):
                    profilename = os.path.basename(profile)
                    profilename = os.path.splitext(profilename)[0]
                else:
                    profilename = profile
                profilename = profilename.lower()
                self.connections[profilename] = BitfarmBaseCon(profile, **kwargs)
                self.profiles.append(profilename)

                if self.main is None:
                    self.main = self.connections[profilename]

    def __getitem__(self, name):
        if name not in self.connections:
            return self.main
        else:
            return self.connections[name]

    def __iter__(self):
        for profile in self.profiles:
            try:
                yield self.connections[profile]
            except Exception:
                pass

    # def create_db_engine(self, bind_tables=True, autocommit=True, echo=False, user=None, password=None):
    #     """create_db_engine Methode: erzeugt für alle geladenen Profile eine Datenbank-Engine.
    #         Optional kann bind_tables (default=True) auf False gesetzt werden. Dann werden keine Tabellen-Bindings ans con Objekt
    #         vorgenommen, man muss selbst für die benötigten Bindings sorgen (bsp: self.sys_timer = sqla.Table('sys_timer', self.meta,  autoload = True))
    #         oder über die Funktion bind_tables()
    #     """
    #     for profile in self.connections:
    #         try:
    #             self.connections[profile].create_db_engine(bind_tables=bind_tables, autocommit=autocommit, echo=echo, user=user, password=password)
    #         except:
    #             self.log.exception(u'Fehler bei create_db_engine')
    #             self.connections[profile].engine = None
    #         else:
    #             if profile == self.main.profile:
    #                 self.main.engine = self.connections[profile].engine
    #
    # def connect_db(self):
    #     """connect_db Methode. erzeugt für alle geladenenen Profile eine Datenbank-Connection
    #     und hinterlegt diese in BitfarmConnections[profile].db bzw. BitfarmConnections.main.db
    #     Achtung! sollte nicht in Multi-Thread-Anwendungen
    #     verwendet werden. Dort immer in den einzelnen Threads eine eigene Connection erzeugen."""
    #     for profile in self.connections:
    #         try:
    #             self.connections[profile].connect_db()
    #         except:
    #             self.log.exception(u'Fehler bei connect_db')
    #             self.connections[profile].db = None
    #         else:
    #             if profile == self.main.profile:
    #                 self.main.db = self.connections[profile].db
    #
    # def bind_tables(self):
    #     """bind_tables Methode. Bindet für alle geladenenen Profile die Datenbank-Tabellen an
    #     die entsprechenden Attribute der BitfarmCon Objekte.
    #     """
    #     for profile in self.connections:
    #         try:
    #             self.connections[profile].bind_tables()
    #         except:
    #             self.log.exception(u'Fehler bei bind_tables')
    #
    # def disconnect_db(self):
    #     """disconnect_db Methode. Schliesst für alle geladenenen Profile die Datenbank-Connection."""
    #     for profile in self.connections:
    #         try:
    #             self.connections[profile].disconnect_db()
    #         except:
    #             self.log.exception('Fehler bei disconnect_db')
    #             self.connections[profile].db = None
    #         else:
    #             if profile == self.main.profile:
    #                 self.main.db = None

    def add(self, name):
        """add - Methode: fügt eine weiteres Profil hinzu"""
        if name not in self.connections:
            self.connections[name] = BitfarmBaseCon(name)
            self.connections[name].create_db_engine()
            self.profiles.append(name)

    def read_ini(self):
        """read_ini Methode: liest aus der Scripts.ini und speichert im ConfigDictonary jedes Profils"""
        _ret = True
        for profile in self.connections.keys():
            _ret &= self.connections[profile].read_ini()
        return _ret


class BitfarmINIObject(object):
    """Hilfsklasse um eine dynamische INI-Klasse aufzubauen"""
    version = __version__

    def __init__(self):
        self = None

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            return None

    def __getitem__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            return None

    def __iter__(self):
        for key in self.__dict__.keys():
            try:
                yield key
            except Exception:
                pass