#!/usr/bin/env python
# -*- coding: latin-1 -*-

"""Dieses Modul stellt alle Bitfarm Exceptions zur Verfgung, sowie alle Errorcodes die
 alle Objekte, die vom BaseObject abstammen (BitfarmDocument, BitfarmWorkflow), erhalten knnen. Kann mit
from bitfarm_errors import *  in die jeweilige Anwendung geladen werden, falls dort bentigt.
Die bitfarm Module laden dieses Modul automatisch"""

EC_OK = 0
EC_NODATABASEENGINE = 1
EC_NORECONNECT = 2
EC_FUNCTIONNOTFOUND = 3
EC_WORKERTHREADERROR = 4
EC_SAVEWORKFLOWERROR = 5
EC_WORKFLOWERROR = 6
EC_VERSIONNOTSUPPORTED = 7
EC_NOTIMID = 8
EC_NODOCID = 9
EC_SETSTATEDONEERROR = 10
EC_RESUBISDONE = 11
EC_RESUBINCHECKBACK = 12
EC_FORWARDRESUBERROR = 13
EC_RETURNTOSENDERERROR = 14
EC_DELETERESUBERROR = 15
EC_CREATERESUBERROR = 16
EC_CHANGERESUBERROR = 17
EC_NOUSERORGROUPERROR = 18
EC_NORESUBFOUND = 19
EC_NEEDRECONNECT = 20


class BitfarmError(Exception):
    """BitfarmError Basis-Klasse """


class BitfarmConError(BitfarmError):
    """Exceptions der BitfarmCon Klasse"""


class NoneConError(BitfarmConError):
    def __str__(self):
        return "Empty or invalid BitfarmCon-Object"


class NoneConnectionsError(BitfarmConError):
    def __str__(self):
        return "Empty or invalid BitfarmConnections-Object"


class LoadConError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "Could not load BitfarmCon %s" % self._confile


class FindConError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "Could not find BitfarmCon %s" % self._confile


class ParseConError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "Could not parse BitfarmCon %s" % self._confile


class DatabaseConError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "Could not open database connection BitfarmCon %s" % self._confile


class DatabaseNotConnectedError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "The database is not connected BitfarmCon %s" % self._confile


class DatabaseEngineError(BitfarmConError):
    def __init__(self, _profile):
        self._profile = _profile

    def __str__(self):
        return "The database engine is not created for %s" % self._profile


class WrongHistoryIndexError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "History_Ext Index below 900 not allowed %s" % self._profile


class DuplicateHistoryIndexError(BitfarmConError):
    def __init__(self, _confile):
        self._confile = _confile

    def __str__(self):
        return "Duplicate Index for History_Ext %s" % self._profile


class BitfarmDocumentError(BitfarmError):
    """Exceptions der Document-Klasse"""


class LoadDocumentFromDatabaseError(BitfarmDocumentError):
    def __init__(self, gdocid):
        self.gdocid = gdocid

    def __str__(self):
        return "Cant load Document from database (%s)" % self.gdocid


class NoBitfarmDocumentError(BitfarmDocumentError):
    def __init__(self, _obj):
        self._obj = _obj

    def __str__(self):
        return "Object is not a BitfarmDocument %s" % str(self._obj.__repr__())


class DocidNotValidError(BitfarmDocumentError):
    def __init__(self, docid, tempname):
        self.docid = docid
        self.tempname = tempname

    def __str__(self):
        return "Docid %s is not valid (%s)" % (self.docid,  self.tempname)


class GdocidNotValidError(BitfarmDocumentError):
    def __init__(self, gdocid):
        self.gdocid = gdocid

    def __str__(self):
        return "Gdocid %s is not valid (%s)" % (self.gdocid,  self.tempname)


class ArcidNotFoundError(BitfarmDocumentError):
    def __init__(self, _doc, *args):
        self.profile = _doc.profile
        self.arcid = _doc.arcid
        self.archive = _doc.archive
        self.tblname = _doc.tblname
        if args:
            self.additional = str(args)

    def __str__(self):
        _ret = "profile=%s arcid=%s archive=%s tblname=%s" % (self.profile, self.arcid, self.archive, self.tblname)
        if self.additional:
            _ret += "\n%s" % self.additional
        return _ret


class CreateDocError(BitfarmDocumentError):
    def __init__(self, _doc, *args):
        self.profile = _doc.profile
        self.gdocid = _doc.gdocid
        self.arcid = _doc.arcid
        self.archive = _doc.archive
        self.tblname = _doc.tblname
        if args:
            self.additional = str(args)

    def __str__(self):
        _ret = "profile=%s gdocid=%s arcid=%s archive=%s arc_tblname=%s" % (self.profile, self.gdocid, self.arcid,  self.archive, self.tblname)
        if self.additional:
            _ret += "\n%s" % self.additional
        return _ret


class BitfarmDatabaseErrors(BitfarmError):
    """Bitfarm DataBase Error Base Klasse"""


class BitfarmWorkflowErrors(BitfarmError):
    """Bitfarm Workflow Error Base Klasse"""


class WorkflowTypeRequiredError(BitfarmWorkflowErrors):
    def __str__(self):
        return "A workflow type is required. 0 = document wf, 1 = user wf"


class WorkflowTriggerRequiredError(BitfarmWorkflowErrors):
    def __str__(self):
        return "A workflow wf_trigger is required. 0 = collector, 1 = archiving, 2 = bfaserver, 3 = viewer"


class WorkflowRequiredError(BitfarmWorkflowErrors):
    def __str__(self):
        return "A workflow or list of workflows is required."


class BitfarmServerError(BitfarmError):
    """Bitfarm Server Error Base Klasse"""


class ConfigfileNotFoundError(BitfarmServerError):
    def __init__(self, fname=''):
        self.fname = fname
    def __str__(self):
        return u'Could not find Config-File %s' % self.fname


class ConfigfileNotValidError(BitfarmServerError):
    def __init__(self, fname=''):
        self.fname = fname

    def __str__(self):
        return u'Config-File is not valid %s' % self.fname
