import argparse
import sys
import traceback

import warnings
from sqlalchemy import exc as sa_exc

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=sa_exc.SAWarning)
    warnings.simplefilter("ignore", category=sa_exc.SADeprecationWarning)
    warnings.simplefilter("ignore", category=sa_exc.SAPendingDeprecationWarning)
    warnings.simplefilter("ignore", category=sa_exc.RemovedIn20Warning)
    warnings.simplefilter("ignore", category=sa_exc.Base20DeprecationWarning)
    warnings.simplefilter("ignore", category=sa_exc.RemovedIn20Warning)
    
# disable sqlalchemy deprecation warning
from sqlalchemy.util import deprecations
deprecations.SILENCE_UBER_WARNING = True
deprecations.SQLALCHEMY_WARN_20 = False

import bitfarm_doc
import bitfarm_proc


# setup argumentparser
parser = argparse.ArgumentParser(
    prog="createnewjobfile",
    description="creates a new jobfile for a bitfarm document.",
    epilog='Either use\n    <profile> <arcid> <docid> <basepath> (--with_griddata)\nor\n    <profile> "gdocid" <gdocid> <basepath> (--with_griddata)\nto identify the bitfam document.',
    formatter_class=argparse.RawTextHelpFormatter
)

parser.add_argument("profile", help="name of the bitfarm profile")
parser.add_argument("arc_id", help='<arcid> or "gdocid"')
parser.add_argument("doc_id", help="<docid> or <gdocid>")
parser.add_argument("basepath", help='basepath for the jobfile (without ".job")')
parser.add_argument(
    "--include_griddata",
    action="store_true",
    default=False,
    required=False,
    help="use this argument to include grid-data in the jobfile",
)
parser.add_argument(
    "--print_fields",
    action="store_true",
    default=False,
    required=False,
    help="prints document field-data",
)

args = parser.parse_args()

# load profile/connect db
try:
    con = bitfarm_proc.BitfarmConnections(args.profile)
    con.create_db_engine()
    con.connect_db()
    proc = bitfarm_proc.BitfarmDocProcessing(con)
except Exception as e:
    print(
        f"error con/db/proc init with profile {args.profile}\n{e}\n{traceback.format_exc()}"
    )
    sys.exit(1)

# create doc-object with gdocid
doc = bitfarm_doc.BitfarmDocument()

if con[args.profile].version in (35, 30, 3.0, 3.5, "35", "30", "3.0", "3.5"):
    doc.version = 35
    try:
        doc.gdocid = proc.gdocid_by_docid(args.arc_id, args.doc_id)
    except Exception as e:
        print(f"error determining gdocid {e}\n{traceback.format_exc()}")
        sys.exit(1)

elif con[args.profile].version in (36, 3.6, "36", "3.6"):
    doc.version = 36
    try:
        doc.gdocid = int(args.doc_id)
    except Exception as e:
        print(f"wrong value for argument doc_id (integer required)\n{e}")
        sys.exit(1)

# load document
try:
    proc.load(doc, with_gridcolumns=True)
except Exception as e:
    print(f"error loading document:\n{e}\n{traceback.format_exc()}")
    sys.exit(1)

# print field-data if necessary
if args.print_fields is True:
    print("\nfield-values:")
    for k, v in sorted(doc.fields.items(), key=lambda x: x[1].pos):
        print(f'{k} ("{v.name}", type {v.type}): {v.value}')
    print()

# remove grid-data before writing job (if necessary)
if args.include_griddata is False:
    for field in doc.fields.values():
        if field.type == 13:
            field.value = None

# create jobfile
result = False
try:
    doc.write_empty_values = True
    result = proc.write_to_file(doc, args.basepath + ".job", con.main.usestatus35)
except Exception as e:
    print(
        f'error writing jobfile "{args.basepath}.job":\n{e}\n{traceback.format_exc()}'
    )
    sys.exit(1)

if result is True:
    print(f'jobfile "{args.basepath}.job" created')
    sys.exit(0)
else:
    print(f'\nerror writing jobfile "{args.basepath}.job"')
    sys.exit(1)
