import argparse
import base64
import codecs
import configparser
import datetime
import getpass
import imp
import importlib
import inspect
import io
import locale
import logging
import os
import re
import sys
import traceback
from decimal import Decimal  # noqa: F401
from logging.handlers import TimedRotatingFileHandler

import chardet  # noqa: F401
import lxml  # noqa: F401
import pyodbc  # noqa: F401
from lxml import etree, objectify  # noqa: F401

import bfaServerClient
import bitfarm_doc  # noqa: F401
import bitfarm_xml  # noqa: F401
import DocFromFile
import thottolib  # noqa: F401
from bitfarm_baseobj import BitfarmBaseCon  # noqa: F401
from bitfarm_con import BitfarmCon
from bitfarm_doc import BitfarmDocument  # noqa: F401
from dpapi import cryptData, decryptData  # noqa: F401

VERSION = "3.6.2.16"


LOGLEVEL = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "warning": logging.WARNING,
}


class ResponseHandler:
    def handle_respAuth(self, webclient, response):
        webclient.userid = response.userid
        webclient.loggedin = True
        return True

    def handle_default(self, client, response):
        if client.streamdata:
            data = b""
            client.streamdata.seek(0)
            while True:
                buff = client.streamdata.read()
                if buff:
                    data += buff
                else:
                    break
            client.streamdata = io.BytesIO()
            return data
        return response.data

    def handle_respDoc(self, webclient, response):
        try:
            webclient.streamdata.seek(0, os.SEEK_END)
            if webclient.streamdata.tell() > 0:
                # print "webclient streamdata.tell > 0 -> handle_default"
                return self.handle_default(webclient, response)
        except Exception:
            pass
        return response.data

    def handle_respGetBookmarks(self, client, response):
        return response.data

    def handle_respAddfields(self, client, response):
        fields = response.data
        addfields = {}
        for field in fields:
            addfields[field["id"]] = {
                "name": field["name"],
                "add": field["add"],
                "indextableid": field["indextableid"],
                "type": field["type"],
            }
            arcs = []
            for arc in field["rights"]["_rights"]:
                try:
                    arcs.append(
                        {
                            "arcid": arc["_BitfarmFieldRight__arcid"],
                            "fieldpos": arc["_BitfarmFieldRight__fieldpos"],
                        }
                    )
                except Exception:
                    pass
            addfields[field["id"]]["arcs"] = arcs
        return addfields

    def handle_respAuthError(self, webclient, response):
        return False

    def handle_respDenied(self, webclient, response):
        return False

    def handle_respOK(self, client, response):
        return True

    def handle_respArchives(self, webclient, response):
        arcs = {}
        statefields = {}
        addfields = {}
        sf_keys = [
            "addfieldid",
            "name",
            "_BitfarmStatefield__bitmask",
            "indexcolumnid",
            "pos",
            "id",
        ]
        sf_mapping = ["addfieldid", "name", "bitmask", "indexcolumnid", "pos", "id"]
        af_keys = ["indextableid", "name", "pos", "add", "mode", "type", "options"]
        af_mapping = ["indextableid", "name", "pos", "add", "mode", "type", "options"]
        for arc in response.data:
            if arc["enabled"] == "enabled" or "enabled" in arc["enabled"]:
                arcs[arc["id"]] = {
                    "name": arc["arcname"],
                    "lgrid": arc["lgrid"],
                    "type": arc["type"],
                    "statefields": [],
                    "addfields": [],
                }
                if "statefields" in arc:
                    for sf in arc["statefields"]:
                        mysf = {
                            sf_mapping[sf_keys.index(k)]: v
                            for k, v in sf.iteritems()
                            if k in sf_keys
                        }
                        arcs[arc["id"]]["statefields"].append(mysf)
                        if sf["id"] not in statefields:
                            statefields[sf["id"]] = mysf
                if "addfields" in arc:
                    for af in arc["addfields"]:
                        myaf = {
                            af_mapping[af_keys.index(k)]: v
                            for k, v in af.iteritems()
                            if k in af_keys
                        }
                        arcs[arc["id"]]["addfields"].append(myaf)
                        if af["add"] not in addfields:
                            addfields[af["add"]] = myaf
                            addfields[af["add"]]["arcids"] = [arc["id"]]
                        else:
                            addfields[af["add"]]["arcids"].append(arc["id"])
        return (arcs, statefields, addfields)

    def handle_respSendFile(self, client, response):
        return response


def get_gdocid_from_job(jobfile, log):
    job = open(jobfile, "r")
    for line in job.read().split("\n"):
        if line.lower().startswith("gdoc_id"):
            gdoc_id = line.split("=")[1]
            try:
                return int(gdoc_id)
            except Exception as e:
                log.error("error reading gdocid from job '%s': %s" % (jobfile, str(e)))
                sys.exit(1)
    return None


def check_conditions(doc, config, log):
    matchsect = []
    for section in config.sections():
        if section == "main":
            continue
        log.info("check doc-conditions section %s" % section)
        condition = True
        if "archive" in config.options(section):
            arc_cond = config.get(section, "archive").split(",")
        else:
            arc_cond = None
        sts_cond = {
            opt[4:]: config.getboolean(section, opt)
            for opt in config.options(section)
            if opt.startswith("sts_")
        }
        add_cond = {
            opt[4:]: config.get(section, opt).split("|", 1)
            for opt in config.options(section)
            if opt.startswith("add_")
        }
        docfields = {doc.fields[field].name: doc.fields[field] for field in doc.fields}
        if "filetype" in config.options(section):
            filetype = config.get(section, "filetype")
            match = re.match(filetype, doc.filetype)
            if match is None:
                log.info(
                    'condition filetype False ("%s" not matching regex "%s")'
                    % (doc.filetype, filetype)
                )
                condition = False
                continue
        # if "exclude_refreshed_docs" in config.options(section):
        try:
            exclude = config.getboolean(
                section, "exclude_refreshed_docs", fallback=False
            )
        except Exception:
            log.warning(
                f"wrong value '{config.get(section, 'exclude_refreshed_docs')}' for option exclude_refreshed_docs. valid values: True/False. assuming False..."
            )
            exclude = False
        if exclude is True:
            if 20 in [x.taskid for x in doc.history]:
                log.info("exclude_refreshed_docs=True is wrong (refreshed document)")
                condition = False
                continue
        if arc_cond is not None:
            if doc.archive not in arc_cond:
                log.info(f"condition archive False ({doc.archive} not in {arc_cond})")
                condition = False
                continue
            else:
                log.info(f"condition archive True ({doc.archive} in {arc_cond})")
        else:
            log.info("no archive condition")
        for sts, value in sts_cond.items():
            if sts not in doc.statefields:
                log.warning(
                    "statefield '%s' not found for document %d" % (sts, doc.gdocid)
                )
                condition = False
                break
            if doc.statefields[sts].value != value:
                log.info(
                    "condition statefield '%s=%s' False for document %d"
                    % (sts, value, doc.gdocid)
                )
                condition = False
                break
            else:
                log.info(
                    "condition statefield '%s=%s' True for document %d"
                    % (sts, value, doc.gdocid)
                )
        for add, (operator, cfgvalue) in add_cond.items():
            if add not in docfields:
                log.warning(
                    "addfield '%s' not found for document %d" % (add, doc.gdocid)
                )
                condition = False
                break
            cmpval = docfields[add].value

            addtype = docfields[add].type
            if addtype in (0, 2, 3):  # universal, auswahl, mehrfach
                cfg_castvalue = cfgvalue
                allowed_operators = ["<", ">", "=", "!=", "~", ".<", ".>", "re"]
                if cmpval is None:
                    cmpval = ""
            elif addtype == 1:  # datum
                # cfgvalue zu date casten
                cfg_castvalue = datetime.datetime.strptime(cfgvalue, "%d.%m.%Y").date()
                cmpval = cmpval.date()
                allowed_operators = ["<", ">", "=", "!="]
                # cfg_castvalue = datetime.date.strptime(cfgvalue, "%d.%m.%Y")
            elif addtype in (7, 8):  # Fließkomma, Währung
                if cmpval is None:
                    cmpval = 0.0
                if cfgvalue == "":
                    cfgvalue = 0
                cfg_castvalue = float(cfgvalue)
                allowed_operators = ["<", ">", "=", "!="]
            elif addtype == 5:  # interval
                cfg_castvalue = cfgvalue
                allowed_operators = ["=", "!="]
            elif addtype == 6:  # ganzzahl
                if cmpval is None:
                    cmpval = 0
                if cfgvalue == "":
                    cfgvalue = 0
                allowed_operators = ["<", ">", "=", "!="]
                if operator not in allowed_operators:
                    log.error(
                        "operator '%s' not allowed for field '%s'. allowed operators: %s"
                        % (operator, add, ", ".join(allowed_operators))
                    )
                    condition = False
                    break
                try:
                    cfg_castvalue = int(cfgvalue)
                except Exception as e:
                    log.error(
                        "field %s cannot cast config-value '%s' to integer"
                        % (add, cfgvalue)
                    )
                    log.error(e)
                    condition = False
                    break

            if operator not in allowed_operators:
                # TODO: logging!
                log.warning(
                    "error, operator '%s' not allowed for field '%s'. allowed operators: %s"
                    % (operator, add, ", ".join(allowed_operators))
                )
                condition = False
                break
            if operator == "<":
                condition = cmpval < cfg_castvalue
            elif operator == ">":
                condition = cmpval > cfg_castvalue
            elif operator == "=":
                condition = cmpval == cfg_castvalue
            elif operator == "~":
                condition = cfg_castvalue in cmpval
            elif operator == ".<":
                condition = cmpval.endswith(cfg_castvalue)
            elif operator == ".>":
                condition = cmpval.startswith(cfg_castvalue)
            elif operator == "!=":
                condition = cmpval != cfg_castvalue
            elif operator == "re":
                match = re.match(cfg_castvalue, cmpval)
                if match is None:
                    condition = False
            if condition is False:
                log.info(
                    "condition addfield '%s=%s|%s' False for document %d"
                    % (add, operator, cfgvalue, doc.gdocid)
                )
                break
            else:
                log.info(
                    "condition addfield '%s=%s|%s' True for document %d"
                    % (add, operator, cfgvalue, doc.gdocid)
                )

        if not condition:
            log.info("section %s no match" % section)
            continue
        else:
            log.info("section %s matches" % section)
            matchsect.append(section)

    log.info("matching sections '%s'" % (matchsect))
    return matchsect


def read_mylist(listfile, log):
    log.debug("loading mylist file")
    gdocids = []
    if not os.path.isfile(listfile):
        import glob

        mlp = glob.glob(r"C:\Users\bitfarm\AppData\Local\Temp\2\Viewer\mylist*")
        if len(mlp) > 0:
            listfile = mlp[0]
    if not os.path.isfile(listfile):
        log.error("mylist-file not found")
        sys.exit(1)
    else:
        f = open(listfile, "r")
        for jobfile in f.read().split("\n"):
            if os.path.splitext(jobfile)[1].lower() != ".job":
                continue
            if not os.path.isfile(jobfile):
                log.error("jobfile '%s' not found" % jobfile)
                sys.exit(1)
            else:
                gdocid = get_gdocid_from_job(jobfile, log)
                if gdocid is None:
                    log.error("no gdocid found in jobfile '%s'" % (jobfile))
                    sys.exit(1)
                gdocids.append(gdocid)
    return gdocids


def cprint(message):
    print(message.encode(locale.getpreferredencoding()))


def import_path(path):
    module_name = os.path.basename(path).replace("-", "_")
    spec = importlib.util.spec_from_loader(
        module_name, importlib.machinery.SourceFileLoader(module_name, path)
    )
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[module_name] = module
    return module


def load_python_plugins(log):
    # mydir = os.path.dirname(sys.argv[0])
    # casting = import_path(os.path.join(mydir, "bfa_sqlmapper_casting.py"))
    # castingfunctions = {k: v for k, v in inspect.getmembers(
    #     casting, inspect.isfunction)}

    python_plugins = {}
    pluginpath = os.path.join(os.path.dirname(sys.argv[0]), "plugins")
    python_plugins = {
        x: {"script": os.path.join(pluginpath, x, "plugin.py")}
        for x in os.listdir(pluginpath)
        if os.path.isfile(os.path.join(pluginpath, x, "plugin.py"))
    }
    for plugin in python_plugins:
        try:
            python_plugins[plugin]["runfunction"] = [
                x[1]
                for x in inspect.getmembers(
                    imp.load_source(plugin, python_plugins[plugin]["script"]),
                    inspect.isfunction,
                )
                if x[0] == "runplugin"
            ][0]
        except Exception as e:
            log.error("error '%s' in plugin %s" % (str(e), plugin))
            sys.exit(1)
    return python_plugins


def main():
    mydir = os.path.dirname(sys.argv[0])
    print(mydir)
    parser = argparse.ArgumentParser()

    parser.add_argument("jobfile")
    parser.add_argument("configfile")
    parser.add_argument("profile", default=None, nargs="?", type=str)
    parser.add_argument("username", default=None, nargs="?", type=str)
    parser.add_argument("password", default=None, nargs="?", type=str)
    parser.add_argument("--bookmark", default=False, action="store_true")
    parser.add_argument("--mylist", default=False, action="store_true")
    parser.add_argument("--gdocid", default=False, action="store_true")
    parser.add_argument(
        "--section",
        default=None,
        type=str,
        help="disable condition checks, run the given section",
    )

    args = parser.parse_args()
    # print(args)
    # load config
    config = configparser.ConfigParser(allow_no_value=True, interpolation=None)
    config.optionxform = str

    if not os.path.isfile(args.configfile):
        print("config-Datei %s nicht gefunden." % (args.configfile))
        sys.exit(1)
    with codecs.open(args.configfile, "r", encoding="utf-8") as cf:
        config.read_file(cf)

    # create logger
    log = logging.getLogger("plugin_runnner")
    logfilepath = os.path.expandvars(config.get("main", "logfile"))
    logdir = os.path.dirname(logfilepath)
    if not os.path.isdir(logdir):
        os.makedirs(logdir)

    logHandler = TimedRotatingFileHandler(
        logfilepath,
        when="midnight",
        backupCount=config.getint("main", "logbackup"),
        encoding="utf8",
    )
    logFormatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logHandler.setFormatter(logFormatter)
    log.addHandler(logHandler)
    log.setLevel(LOGLEVEL[config.get("main", "loglevel").lower()])
    log.root.handlers[0].setFormatter(logFormatter)
    log.info("---------------------------------------------------------")
    log.info("started by user %s" % getpass.getuser())
    log.info("version: %s" % VERSION)
    d = args.__dict__.copy()
    if "password" in d and d["password"] not in ("", None):
        d["password"] = "***************"
    log.info(f"startparams {d}")
    log.info("loading plugins")
    plugins = load_python_plugins(log)

    # sys.stderr = LogWriter(log, logging.CRITICAL)
    # sys.stdout = LogWriter(log, logging.DEBUG)
    DocFromFile.log = log
    profile = config.get("main", "profile")
    usecon = config.getboolean("main", "use_con_credentials", fallback=False)
    conobj = BitfarmCon(profile)
    # connect

    if args.username is None:
        if usecon is True:
            log.info("using con-credentials")
            bfauser = conobj.bfauser.decode("utf8")
            bfapass = conobj.bfapass.decode("utf8")
        else:
            bfauser = config.get("main", "bfauser")
            bfauser = base64.b64decode(bfauser)
            bfauser = decryptData(bfauser)
            bfapass = config.get("main", "bfapass")
            bfapass = base64.b64decode(bfapass)
            bfapass = decryptData(bfapass)
        if isinstance(bfauser, str):
            con_user = bfauser
        else:
            con_user = bfauser.decode("utf8")
        if isinstance(bfapass, str):
            con_pass = bfapass
        else:
            con_pass = bfapass.decode("utf8")

        # in standalone we get a single jobfile
        if args.bookmark is False:
            if args.gdocid is False:
                gdocids = [get_gdocid_from_job(args.jobfile, log)]
        else:
            gdocids = []
    else:
        # viewerplugin via args
        con_user = args.username
        bfauser = con_user
        con_pass = conobj.bitdecrypt(args.password)
        profile = args.profile
        gdocids = read_mylist(os.path.expandvars(args.jobfile), log)

    try:
        log.debug("loading profile")
        client = bfaServerClient.BfaServerClient(profile, auto_heartbeat=20)
        log.info("client init success")
    except Exception as e:
        log.error("error loading profile: %s" % (str(e)))
        log.error(traceback.format_exc())
        sys.exit(1)

    client.responsehandler = ResponseHandler()
    log.info(f"establish server-connection with user {con_user}")

    try:
        result = client.connect(username=con_user, userpass=con_pass)
        log.info("client connection-result with user %s: %s" % (bfauser, str(result)))
    except Exception as e:
        log.error("error connecting bfaserver: %s" % (str(e)))
        log.error(traceback.format_exc())
        sys.exit(1)

    if result is not True:
        log.error("no connection to bfaserver")
        sys.exit(1)

    if args.bookmark is True:
        lzname = args.jobfile
        log.info(f"bookmark-mode using bookmark '{lzname}'")
        lz = None
        bm_list = client.getBookmarks(asjson=True)
        if client.responseobject.reqtype == "respGetBookmarks":
            bm_dict = {x.split(";")[0]: x.split(";")[1].split(",") for x in bm_list}
            if lzname in bm_dict.keys():
                lz = bm_dict[lzname]
                log.info("bookmark found")
        if lz is None:
            log.error(f"bookmark '{lzname}' not found!")
            sys.exit(1)
        else:
            log.info("executing bookmark")
            result = client.getExecuteBookmarks(*bm_dict[lzname])
            if client.responseobject.reqtype == "respSearch":
                if result:
                    gdocids = [x[0] for x in result]
                    log.info(f"got {len(gdocids)} results for bookmark '{lzname}'")
                else:
                    log.info(f"no documents found with bookmark '{lzname}'")
                    gdocids = []
            else:
                log.error("error executing bookmark, check the bookmark and server-log")
                sys.exit(1)

    elif args.gdocid is True:
        gdocids = [int(x) for x in args.jobfile.split(",")]

    log.info("start processing documents")
    for gdocid in gdocids:
        log.info(f"gdocid {gdocid}")
        try:
            doc = client.getDoc(gdocid, include_text=True)
        except Exception as e:
            log.error(f"error loading document {gdocid}: {e}")
            continue

        if doc is False:
            log.error(f"could not load doc {gdocid}")
            log.error(f"last response from server: {client.responseobject.reqtype}")
            continue
        
        try:
            hst = client.getHistory(gdocid)
        except Exception as e:
            log.error(f"error loading history {gdocid}: {e}")
            continue
        
        doc.history = hst

        if args.section is None:
            # check conditions
            try:
                sections = check_conditions(doc, config, log)
            except Exception as e:
                log.error(f"error check_conditions document {gdocid}: {e}")
                continue
            if sections == []:
                log.info("no matching section found, quit")
                client.logout()
                sys.exit()
        else:
            if args.section not in config.sections() or args.section == "main":
                log.error(f"section {args.section} not found in config.")
                sys.exit(1)
            sections = [args.section]
        for section in sorted(sections):
            log.info("section %s" % section)

            if "plugin" in config.options(section):
                cfg_plugins = [
                    x
                    for x in config.get(section, "plugin").split(";")
                    if x.strip() != ""
                ]
                for plugin in cfg_plugins:
                    try:
                        plugin_config = None
                        if "plugin_config" in config.options(section):
                            log.debug(
                                "plugin_config: %s" % config.get(section, "plugin_config")
                            )
                            if os.path.isfile(config.get(section, "plugin_config")):
                                log.debug("plugin_config exists!")
                                plugin_config = configparser.RawConfigParser(
                                    allow_no_value=True,
                                    delimiters=("="),
                                    interpolation=None,
                                    strict=False,
                                )
                                plugin_config.optionxform = str
                                with codecs.open(
                                    config.get(section, "plugin_config"),
                                    "r",
                                    encoding="utf-8",
                                ) as cf:
                                    plugin_config.read_file(cf)
                            else:
                                log.error("plugin_config missing!")
                                log.error(
                                    os.path.join(
                                        os.getcwd(), config.get(section, "plugin_config")
                                    )
                                )
                                continue
                    except Exception as e:
                        log.error(f"error reading plugin_config: {e}")
                        continue
                    plugin_args = [client, doc, log, plugin_config]
                    if plugin in plugins:
                        log.info("starting plugin '%s'" % plugin)
                        try:
                            formatter = logging.Formatter(
                                "%(asctime)s - %(name)s - %(levelname)s - "
                                + plugin
                                + " - %(message)s"
                            )
                            for handler in log.handlers:
                                handler.setFormatter(formatter)
                            log.root.handlers[0].setFormatter(formatter)
                            old_cwd = os.getcwd()
                            os.chdir(os.path.dirname(plugins[plugin]["script"]))
                            plugins[plugin]["runfunction"](*plugin_args)
                            os.chdir(old_cwd)

                        except Exception as e:
                            log.error("exception in plugin!")
                            log.error(e)
                            log.error(traceback.format_exc())
                        finally:
                            formatter = logging.Formatter(
                                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
                            )
                            for handler in log.handlers:
                                handler.setFormatter(formatter)
                            log.root.handlers[0].setFormatter(formatter)
                        log.info("plugin finished")
                    else:
                        log.error(f"plugin {plugin} not found")

    client.logout()
    log.info("server connection closed")


if __name__ == "__main__":
    main()
