import codecs
import configparser
import copy
import imp
import inspect
import logging
import os
import pathlib
import re
import subprocess

import bs4
from lxml import etree

from bfaServerClient import BfaServerClient
from bitfarm_doc import BitfarmDocument

__version__ = "3.6.2.1"


def get_navigable_strings(element, string_elements=[]):
    if isinstance(element, bs4.element.NavigableString):
        string_elements.append(element)
    else:
        for child in element.children:
            get_navigable_strings(child, string_elements)
    return string_elements


def runplugin(
    client: BfaServerClient,
    doc: BitfarmDocument,
    log: logging.Logger,
    config: configparser.RawConfigParser = None,
):
    client.serverlogger = log
    if doc.filetype.lower() != ".xml":
        log.warning(
            "document filetype != .xml -> this plugin works only with xml-documents. You should configure plugin_runner to run this plugin only with .xml-files."
        )
        return
    mydir = os.getcwd()
    castingfunctions = {}

    if os.path.isfile(os.path.join(mydir, "casting.py")):
        casting = imp.load_source("casting", os.path.join(mydir, "casting.py"))
        castingfunctions = {
            k: v for k, v in inspect.getmembers(casting, inspect.isfunction)
        }

    section_values = {}
    for s in config.sections():
        if s in ["main"]:
            pass
        section_values[s] = []
    log.debug(f"loading gdocid {doc.gdocid}")
    ori = client.getOri(doc.gdocid)
    parser = etree.XMLParser(ns_clean=False)
    log.debug("reading xml")
    root = etree.fromstring(ori, parser)
    tree = etree.ElementTree(root)
    etree.cleanup_namespaces(tree)
    root = tree.getroot()
    values = []
    xpths = {}

    for elem in root.iter():
        xpth = tree.getpath(elem)
        xpth_clean = re.sub("\[\d+\]", "", xpth)
        val = elem.text
        if val is None:
            val = ""
        val = val.strip()
        if xpth_clean not in xpths:
            xpths[xpth_clean] = [val]
        else:
            xpths[xpth_clean].append(val)
        for s in config.sections():
            if xpth_clean in config.options(s):
                section_values[s].append(val)

    # add extra xpath-values
    log.debug("loading additional_xpaths.ini")
    cfg_file = os.path.join(mydir, "additional_xpaths.ini")
    xp_config = configparser.RawConfigParser(
        allow_no_value=True, delimiters=("="), interpolation=None, strict=False
    )
    xp_config.optionxform = str
    with open(cfg_file, "r") as f:
        xp_config.read_file(f)
    for _ in xp_config.sections():
        for sec in xp_config.options(_):
            if sec not in section_values:
                section_values[sec] = []
            pth = xp_config.get(_, sec)
            if pth.startswith("/*"):
                pass
            try:
                value = root.xpath(pth, namespaces=root.nsmap)[0].text
                section_values[sec].append(value)
            except Exception:
                continue

    html_template = pathlib.Path(
        os.path.expandvars(config.get("main", "html_template"))
    )
    log.debug(f"loading html_template {html_template}")
    with open(html_template, "rb") as fp:
        soup = bs4.BeautifulSoup(fp, "html.parser")

    # required section values
    required_section_nodes = soup.find_all(required_section=re.compile(".+"))

    log.debug("removing required_section-nodes without values")
    # remove required_section-nodes without values
    for node in required_section_nodes:
        required_sections = node.attrs["required_section"].split(";")
        for sect in required_sections:
            if sect not in config.sections():
                node.decompose()

    # multi-row handling
    multi_add_rows = soup.find_all("tr", xml_class="multi_row")
    log.debug("creating multi_rows")
    for row in multi_add_rows:
        td_nodes = get_navigable_strings(row, [])
        add_values = []
        for node in td_nodes:
            sects = [x[2:-2] for x in re.findall("%%[^%]+%%", node.text)]
            for sect in sects:
                values = section_values[sect]
                add_values.append(values)
        rowcount = max([len(x) for x in add_values])
        # new_tr_attrs = {
        #     k: v
        #     for k, v in list(row.attrs.items())
        #     if k != "xml_class" and v != "multi_row"
        # }
        for row_index in range(rowcount):
            new_tr = copy.copy(row)
            for elem in get_navigable_strings(new_tr, []):
                s = elem.string
                if s is None or s.strip() == "":
                    continue
                sect_strings = [x for x in re.findall("%%[^%]+%%", s)]
                if len(sect_strings) < 1:
                    continue
                attrs = []
                if hasattr(elem, "attrs"):
                    attrs = elem.attrs
                elif hasattr(elem, "parent"):
                    if hasattr(elem.parent, "attrs"):
                        attrs = elem.parent.attrs
                for sect in sect_strings:
                    if len(section_values[sect[2:-2]]) > row_index:
                        try:
                            value = section_values[sect[2:-2]][row_index]
                        except Exception:
                            value = ""
                    else:
                        try:
                            value = section_values[sect[2:-2]][0]
                        except Exception:
                            value = ""
                    if "castfunc" in attrs:
                        value = castingfunctions[attrs["castfunc"]](value)
                    s = s.replace(sect, value)
                try:
                    elem.replace_with(bs4.element.NavigableString(s))
                except Exception:
                    pass

            row.insert_before(new_tr)
        row.decompose()

    log.debug("creating non multi_row nodes")
    for elem in get_navigable_strings(soup.body, []):
        s = elem.string
        if s is None or s.strip() == "":
            continue
        sect_strings = [x for x in re.findall("%%[^%]+%%", s)]

        if len(sect_strings) < 1:
            continue
        attrs = []
        if hasattr(elem, "attrs"):
            attrs = elem.attrs
        elif hasattr(elem, "parent"):
            if hasattr(elem.parent, "attrs"):
                attrs = elem.parent.attrs
        for sect in sect_strings:
            try:
                value = section_values[sect[2:-2]][0]
            except Exception:
                value = ""
            if "castfunc" in attrs:
                value = castingfunctions[attrs["castfunc"]](value)
            s = s.replace(sect, value)
        try:
            elem.replace_with(bs4.element.NavigableString(s))
        except Exception:
            pass
    log.debug("creating html-file")
    tmp_html_file = f"{doc.gdocid}.html"
    with open(tmp_html_file, "wb") as f:
        f.write(soup.html.encode("utf8"))

    cfgfile = os.path.join(os.path.expandvars("%bitfarm-archiv%"), "scripts.ini")
    with codecs.open(cfgfile, "r", encoding="cp1252") as cf:
        ini_text = cf.read()

    wkhtmlexe = ''
    for line in ini_text.split("\n"):
        line = line.strip()
        if line.lower().startswith("htmlconverter="):
            wkhtmlexe = os.path.expandvars(line.split("=")[1])

    if not wkhtmlexe:
        log.error('could not find wkhtmltopdf.exe')
    else:
        tmp_pdf_file = f"{doc.gdocid}.pdf"
        tmp_tif_file = f"{doc.gdocid}.tif"
        args = [
            wkhtmlexe,
            "--enable-local-file-access",
            "--header-center",
            "[page]/[topage]",
            tmp_html_file,
            tmp_pdf_file,
        ]
        log.debug("converting html to pdf")
        log.debug(args)
        subprocess.check_call(args)

    args = [
        os.path.join(os.path.expandvars("%bitfarm-archiv%"), "bfa_pdf2tif.exe"),
        "-i",
        tmp_pdf_file,
        "-o",
        tmp_tif_file,
        "-r",
        "300",
        "-d",
        "tiff24nc",
    ]
    log.debug("converting pdf to tif")
    log.debug(args)
    subprocess.check_call(args)

    os.remove(tmp_html_file)
    os.remove(tmp_pdf_file)

    with open(tmp_tif_file, "rb") as f:
        tif_data = f.read()
    os.remove(tmp_tif_file)

    log.debug("sending new preview to bfaServer")
    [result, data] = client.getSendPreview(doc.gdocid, tif_data, ".tif")
    if result is not True:
        log.error("error sending preview to bfaServer -> please see server-log")
        log.error(data)
    else:
        log.info("preview updated")
