#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import re
import datetime
from decimal import *
import unittest

MONATE_DEUTSCH = {}
MONATE_DEUTSCH[1] = ["Januar", "Jan", "Jan."]
MONATE_DEUTSCH[2] = ["Februar", "Feb", "Feb."]
MONATE_DEUTSCH[3] = ["März", "März", "Maerz", "Mar.", "Mär", "Mär."]
MONATE_DEUTSCH[4] = ["April", "Apr", "Apr."]
MONATE_DEUTSCH[5] = ["Mai"]
MONATE_DEUTSCH[6] = ["Juni", "Jun", "Jun."]
MONATE_DEUTSCH[7] = ["Juli", "Jul", "Jul."]
MONATE_DEUTSCH[8] = ["August", "Aug", "Aug."]
MONATE_DEUTSCH[9] = ["September", "Sep", "Sep.", "Sept."]
MONATE_DEUTSCH[10] = ["Oktober", "Okt", "Okt."]
MONATE_DEUTSCH[11] = ["November", "Nov", "Nov."]
MONATE_DEUTSCH[12] = ["Dezember", "Dez", "Dez."]


def monate_deutsch_to_object(value):
    for month, monthstr in list(MONATE_DEUTSCH.items()):
        try:
            if value in monthstr:
                return month
        except:
            pass
    return None


# def monate_deutsch_to_object(text):
#     for month, monthstr in MONATE_DEUTSCH.iteritems():
#         if monthstr in text:
#             return month
#     return None


flatten = lambda l: [item for sublist in l for item in sublist]

################################################################

# RULESETS

################ DATES ################
PDATA_DATETIME = {}

PDATA_DATETIME["de-datum-kurz"] = {
    "regex": "(\s|^)((?P<days>(31|30|[012]\d|\d))\.(?P<months>(0\d|1[012]|\d))\.(?P<years>(\d\d)))(\s|$)", \
    "group": 1, \
    "txt_to_obj_func": lambda x: datetime.datetime(int(x["years"]) + 2000, int(x["months"]), int(x["days"])), \
    "obj_to_txt_func": lambda x: x.strftime("%d.%m.%y")}

# 17.06.2016 / 01.12.1998
PDATA_DATETIME["de-datum-normal"] = {
    "regex": "(\s|^)((?P<days>(31|30|[012]\d|\d))\.(?P<months>(0\d|1[012]|\d))\.(?P<years>(19[789]\d|20[0123]\d)))(\s|$)", \
    "group": 1, \
    "txt_to_obj_func": lambda x: datetime.datetime(int(x["years"]), int(x["months"]), int(x["days"])), \
    "obj_to_txt_func": lambda x: x.strftime("%d.%m.%Y")}

de_monat_values = flatten(list(MONATE_DEUTSCH.values()))
monatstr = "|".join([re.escape(x) for x in de_monat_values])
PDATA_DATETIME["de-datum-monat"] = {
    "regex": "(\s|^)((?P<days>(31|30|[012]\d|\d))(\.|\s+)\s*(?P<months>(" + monatstr + "))\s*(?P<years>(19[789]\d|20[0123]\d)))(\s|$)", \
    "group": 1, \
    "txt_to_obj_func": lambda x: datetime.datetime(int(x["years"]), monate_deutsch_to_object(x["months"]),
                                                   int(x["days"])), \
    "obj_to_txt_func": lambda x: x.strftime("%d.%m.%Y")}

# (31|30|[012]\d|\d)\.(0\d|1[012]|\d)\.(19[789]\d|20[0123]\d)                                     
# months englisch (?:J(anuary|u(ne|ly))|February|Ma(rch|y)|A(pril|ugust)|(((Sept|Nov|Dec)em)|Octo)ber)

PDATA_DECIMAL = {}
# 12311 / 12346,888 / 1,20
PDATA_DECIMAL["comma-no_sep"] = {"regex": "(\s|^)((?P<ints>\d+),\s{0,1}(?P<decimals>\d+|-))(\s|$|\$|\€)", \
                                 "group": 1, \
                                 "txt_to_obj_func": lambda x: Decimal(
                                     x["ints"].strip() + "." + x["decimals"].replace("-", "00").strip()), \
                                 "obj_to_txt_func": lambda x: str(x).replace(".", ",")}
# 123.1 / 12346.888 / 1.20
PDATA_DECIMAL["point-no_sep"] = {"regex": "(\s|^)((?P<ints>\d+)\.\s{0,1}(?P<decimals>\d+|-))(\s|$|\$|\€)", \
                                 "group": 1, \
                                 "txt_to_obj_func": lambda x: Decimal(
                                     x["ints"].strip() + "." + x["decimals"].replace("-", "00").strip()), \
                                 "obj_to_txt_func": lambda x: str(x)}

################ FLOATS / DECIMALS ################
# 1,345,112.12 / 1,123.4 / 123.145
PDATA_DECIMAL["point-comma_sep"] = {"regex": "(\s|^)((?P<ints>(\d{1,3}(,\d{3})*))\.(?P<decimals>(\d+|-)))(\s|$|\$|\€)", \
                                    "group": 1, \
                                    "txt_to_obj_func": lambda x: Decimal(
                                        x["ints"].strip().replace(",", "") + "." + x["decimals"].replace("-",
                                                                                                         "00").strip()), \
                                    "obj_to_txt_func": lambda x: str(x)}
# 1.345.112,12 / 1.123,4 / 123,145
PDATA_DECIMAL["comma-point_sep"] = {"regex": "(\s|^)((?P<ints>(\d{1,3}(\.\d{3})*)),(?P<decimals>(\d+|-)))(\s|$|\$|\€)", \
                                    "group": 1, \
                                    "txt_to_obj_func": lambda x: Decimal(
                                        x["ints"].strip().replace(".", "") + "." + x["decimals"].replace("-",
                                                                                                         "00").strip()), \
                                    "obj_to_txt_func": lambda x: str(x)}


################################################################                                     


class TestFindObjects(unittest.TestCase):

    def test_find_dates_in_text(self):
        testtext = " 12.09.2016 ajshdkalkaf 14.09.2016 soidsivnijoji fa 17.12.2016"
        expected = [(datetime.datetime(2016, 9, 12), "12.09.2016", (1, 11), 'de-datum-normal'),
                    (datetime.datetime(2016, 9, 14), "14.09.2016", (24, 34), 'de-datum-normal'),
                    (datetime.datetime(2016, 12, 17), "17.12.2016", (52, 62), 'de-datum-normal')]
        self.assertEqual(find_dates_in_text(testtext), expected)

        testtext = "14.09.2016 soidsivnijoji fa 17.12.2016 "
        expected = [(datetime.datetime(2016, 9, 14), "14.09.2016", (0, 10), 'de-datum-normal'),
                    (datetime.datetime(2016, 12, 17), "17.12.2016", (28, 38), 'de-datum-normal')]
        self.assertEqual(sorted(find_dates_in_text(testtext)), sorted(expected))

        testtext = "14. April 2016 soidsivnijoji fa 17. Mär. 2016 "
        # print sorted(find_dates_in_text(testtext))
        expected = [(datetime.datetime(2016, 4, 14), "14. April 2016", (0, 14), 'de-datum-monat'),
                    (datetime.datetime(2016, 3, 17), "17. Mär. 2016", (32, 46), 'de-datum-monat')]
        self.assertEqual(sorted(find_dates_in_text(testtext)), sorted(expected))

        testtext = "test 26 Februar 2016 "
        # print sorted(find_dates_in_text(testtext))
        expected = [(datetime.datetime(2016, 2, 26, 0, 0), '26 Februar 2016', (5, 20), 'de-datum-monat')]
        self.assertEqual(sorted(find_dates_in_text(testtext)), sorted(expected))

    def test_find_decimals_in_text(self):
        testtext = "100,23 €"
        expected = [(Decimal("100.23"), "100,23", (0, 6), "comma-point_sep"),
                    (Decimal("100.23"), "100,23", (0, 6), "comma-no_sep")]
        self.assertEqual(sorted(find_decimals_in_text(testtext)), sorted(expected))


def find_dates_in_text(text):
    result = []
    for dname, ddict in list(PDATA_DATETIME.items()):
        # print dname
        regex = ddict["regex"]
        group = ddict["group"]
        reresult = re.finditer(regex, text)
        for res in reresult:
            gdict = res.groupdict()
            value = res.groups(0)[group]
            start, end = res.span()
            vstart = start + res.group(0).find(value)
            vend = start + res.group(0).find(value) + len(value)
            try:
                dt = ddict["txt_to_obj_func"](gdict)
                result.append((dt, value, (vstart, vend), dname))
            except:
                pass

    set(result)
    return result


def find_decimals_in_text(text):
    result = []
    for dname, ddict in list(PDATA_DECIMAL.items()):
        regex = ddict["regex"]
        # print dname
        group = ddict["group"]
        reresult = re.finditer(regex, text)

        for res in reresult:
            gdict = res.groupdict()
            value = res.groups(0)[group]
            start, end = res.span()
            vstart = start + res.group(0).find(value)
            vend = start + res.group(0).find(value) + len(value)
            dt = ddict["txt_to_obj_func"](gdict)
            result.append((dt, value, (vstart, vend), dname))
    set(result)
    return result


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFindObjects)
    unittest.TextTestRunner(verbosity=2).run(suite)
