﻿#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Dieses Modul enthält alle Objektverarbeitenden Klassen und deren Methoden. Durch die Trennung von
den reinen Datenobjekten sollen die Abhängigkeiten von größeren Bibliotheken etwas aufgelöst werden.
So enthalten zB die BitfarmCon und die BitfarmConnections Klassen in diesem Modul weitere Funktionen,
um Datenbankverbindungen aufbauen zu können.

Dieses Modul enthält die bitfarm WorkflowProcessing Klasse
    Sie bietet alle Funktionen um Workflows anzulegen, zu prüfen und auszuführen.
    Es können Dokumentenworkflows und Userworkflows angelegt oder ausgeführt werden.
    Die Workflow werden in der Datenbank gespeichert, eine Con muss übergeben werden.
    In der Tabelle sys_wfd werden alle permanenten, festen Workflows definiert,
    in der Tabelle sys_workflow werden alle aktuellen Workflows gespeichert, verarbeitet
    und nach dem Beenden wieder gelöscht.
    Ein Workflow besteht aus zwei Teilen:
    Auslöser:   die Eingangsbedingungen die den Workflow auslösen (Archivierung, Import,
                Zeitpunkt, Useraktion). Die Eingangsbedingungen werden in der Datenbank-
                Tabelle hinterlegt, über Queries kann der benötigte Workflow ermittelt werden
                Es wird eine wf_trigger-ID übergeben um genau zu definieren, wer den Workflow auslöst:
                - 0 = Spooler
                - 1 = Archivierung
                - 2 = Viewer
                - usw. es können weiter wf_trigger-IDs definiert werden, wenn zB bestimmte Funktionen
                einen WF auslösen können sollen
    Tasks:  verschiedene Aktionen die durch den Workflow ausgeführt werden, zB naming, sorting
            Workflow anlegen, fortführen, beenden, erstellen einer Taskliste (für Dokumentenverarbeitung)
            Die Tasks sind BitfarmTask-Objekte die ConditionObjekte und else_tasks tragen können. Diese werden,
            falls vorhanden, nur dann ausgeführt wenn die Bedingungen zutreffen.

    Workflowdefinitionen können vererbt werden. Es ist nur eine einfache, aufeinander folgende Vererbung möglich,
    keine Mehrfachvererbung. Bei der Erstellung einer neuen Workflowdefinition wird angegeben von welchem Workflow
    (ID oder Name) der neue Workflow erbt. Wenn ein neuer Workflow von einem bestehenden erbt, werden dessen
    Auslöser und die Tasks kopiert. Die Aulöser müssen kopiert werden, da diese zum Finden des Workflows
    notwendig sind. Wenn ein Basisworkflow nachträglich geändert wird, so werden die Änderungen an die Nachfahren
    weitergegeben, in dem geprüft wird, ob sich die Auslöser zur ursprünglichen Definition verändert haben. Wenn
    sich die Auslöser der Nachfahren nicht zum Ursprung verändert haben, werden die Änderungen weitervererbt, alles was sich
    in der Zwischenzeit aber geändert hat, wird nicht überschrieben. Damit dies einfach geprüft werden kann, gibt es eine
    inherited-Liste, in der alle Auslöser stehen, die geerbt wurden. Wird ein Auslöser beim einem Nachfahren verändert,
    wird dieser Auslöser aus der inherited-Liste entfernt. Wird jetzt von einem Basisworkflow ein Auslöser geändert, so wird bei den
    Nachfahren geprüft, ob dieser Auslöser noch in der inherited Liste steht, nur dann wird der neue Wert vererbt.

    Bei Tasks/ElseTasks ist es im Prinzip ähnlich. Jeder Task erhält eine Inherited Property. Ist diese gesetzt, dann kann eine Änderung
    von einem Basistask übernommen werden, ansonsten nicht. Damit alle Tasks vererbt werden, unabhängig von ihrer Position
    in der Tasklist, erhält jeder neu erstellte Task einen microsekunden-genauen Timestamp, der beim Erben von einem Basistask
    mit übernommen wird. Ändert sich jetzt ein Task in einem Basisworkflow und trägt ein Task in einem Nachfahren denselben
    Timestamp und inherited ist auf True gesetzt dann wird die Änderung vererbt. Wenn in einem Nachfahren ein Task jedoch abgeändert wird,
    dann wird inherited auf False gesetzt und ein neuer Timestamp generiert, im Prinzip so, als wenn der Tasks neu generiert wurde.
    So kann dieser Task nicht mehr verändert/überschrieben werden.Searchbarcode

    Im Unterschied zum Doc-Objekt kann ein Workflow direkt eine condition und einen ElseTask tragen.
    (Im Doc-Objekt können nur die Tasks selbst eine condition und einen ElseTask tragen)
    Somit kann beim Laden der Workflows aus der Datenbank direkt die condition geprüft werden und nicht zutreffende workflows können wieder
    verworfen werden, bzw der ElseTask (falls vorhanden) kann nachgeladen werden. Haben wir eine Liste von Tasks und trägt der ElseTask
    eine einzelne BitfarmTask Instanz, so wird in der Taskliste nur der Task ausgetauscht, dessen condition False ergab.Trägt der ElseTask hingegen selbst
    wiederum eine Taskliste, so werden alle Elemente der ursprünglichen Liste durch die ElseTask-Liste ausgetauscht.
    Mit diesem Mechanismus kann man erreichen, dass entweder bei einer fehlgeschlagen Prüfung nur ein Aufgabenschritt sich ändert, oder aber die
    gesamte Aufgabenfolge.
"""
import copy
import os
import sys
import ctypes

# import mysql.connector
import sqlalchemy.exc

import bitfarm_errors
# import bitfarm_xml
from bitfarm_configobj import ConfigObj
import sqlalchemy as sqla
import win32api
import win32security
import win32service
import socket
import cpuinfo
# import multiprocessing
import logging
import threading
import time
import datetime
import re
import decimal
from collections import OrderedDict
import pickle
import shutil
import bitfarm_baseobj
import bitfarm_doc

from dateutil.relativedelta import *
from urllib.parse import quote as urlquote
from urllib.parse import quote_plus
from sqlalchemy.sql import default_comparator

from bitfarm_errors import *


__version__ = 'v3.6.2'
__author__ = 'F.Schneider'
__copyright = 'bitfarm Informationssystem GmbH, 57072 Siegen'
__license__ = 'GPL v2'

_ = lambda x: x


def calc_row_length(cols):
    length = 1
    for col in cols.values():
        if isinstance(col.type, sqla.types.BIGINT):
            length += 8
        elif isinstance(col.type, sqla.types.SMALLINT):
            length += 1
        elif isinstance(col.type, sqla.types.INT):
            length += 4
        elif isinstance(col.type, sqla.types.String) or isinstance(col.type, sqla.types.VARCHAR) or isinstance(col.type,
                                                                                                               sqla.types.Unicode):
            if col.type.length is None:  # das Gridfeld ist MEDIUMTEXT, gibt es aber nicht als sqla.types
                length += 12  # MEDIUMTEXT hat 16mb Kapazität, allerdings wird für die Spalte nur 12 Bytes benötigt (die Daten liegen in einem anderen Bereich)
            else:
                length += col.type.length * 4 + 2
        elif isinstance(col.type, sqla.types.Date) or isinstance(col.type, sqla.types.DateTime):
            length += 8
        elif isinstance(col.type, sqla.types.Float):
            length += 4
        else:
            length += 8
        length += 1  # wegen Null-Columns
    return length


class BitfarmSearchProcessing(object):
    u"""
    BitfarmSearchProcessing-Klasse (bfaSearchProc):
    Diese Klasse bietet die notwendigen Funktionen Suchen durchzuführen und Lesezeichen zu verwalten.
    Der Klasse muss ein bfaConnections-Objekt übergeben werden. Dadurch werden alle bekannten Profile übergeben.
    Jede Instanz von BitfarmSearchProcessing erzeugt eine eigene Datenbank-Connection auf die Main-Connection-Engine.
    Daher muss in Threads jeweils eine eigene Instanz von BitfarmSearchProcessing erzeugt werden.
    Optional kann ein Dictionary (_con_db) mit einer DB-Connection pro Profil übergeben werden, wenn von einem übergeordneten Thread
    die Db-Connections durchgereicht werden sollen.  {'Profil':sqla.engine.base.Connection}
    Diese Connections werden in self.db hinterlegt
    @version:
    """

    version = __version__
    log = logging.getLogger(u"Main.bfaSearchProc")
    log.setLevel(logging.WARNING)

    def __init__(self, _con, _con_db=None, name=u""):
        if not isinstance(_con, BitfarmConnections):
            raise NoneConnectionsError()

        self.con = _con
        self.version = _con.main.version
        self.name = name  # hier können threads ihren name bezeichner durchreichen

        if _con_db is not None and type(_con_db) is dict:
            self.db = _con_db
        else:
            self.db = {}
            for _profile in self.con.connections:
                if self.con[_profile].engine is not None:
                    self.db[_profile] = self.con[_profile].engine.connect()
                    if not isinstance(self.db[_profile], sqla.engine.base.Connection):
                        raise DatabaseConError(_con[_profile].confile)

        if self.con.main.profile in self.db:
            self.db['main'] = self.db[self.con.main.profile]

        self._addfieldtypes = {'edit': 0,
                               'date': 1,
                               'comboChecked': 3,
                               'combo': 2,
                               'interval-spin': 5,
                               'zahl': 6,
                               '0': 0,
                               '1': 1,
                               '2': 2,
                               '3': 3,
                               '4': 4,
                               '5': 5,
                               '6': 6,
                               '7': 7,
                               '8': 8,
                               '9': 9,
                               '10': 10}
        self._searchtypes = {0: 'search',
                             1: 'locale',
                             2: 'globale',
                             3: 'history',
                             4: 'resub',
                             5: 'addfields'}

    def read_xmlfile(self, path, name=None, favid=None, userid=None, groupid=None):
        u"""Funktion um eine Lesezeichendatei einzulesen. Gibt eine Liste an SearchObjekten zurück.
        Optional kann man den Namen oder die ID eines bestimmten Lesezeichen angeben, welches dann zurückgegeben wird.
        Wird eine UserId/GroupID bzw eine Liste von IDs angegeben, so werden nur die Lesezeichen zurückgegeben,
        auf die mindestens eine (oder mehrere) ID berechtigt ist. Die Eingrenzung auf UserId hat Vorrang vor GroupID"""

        if not path:
            return None

        if not os.path.exists(os.path.expandvars(path)):
            return None

        if favid is not None:
            if type(favid) is int:
                favid = str(favid)
            if type(favid) is not str:
                favid = None

        if name is not None:
            if type(name) is not str:
                name = None

        if userid is not None:
            if type(userid) is not list:
                userid = [userid]

        if groupid is not None:
            if type(groupid) is not list:
                groupid = [groupid]

        favtree = dTree.parse(path)
        root = favtree.getroot()
        return self.read_xmlroot(root, name, favid, userid, groupid, None)

    def read_xml(self, xmlstr):
        if not xmlstr:
            return None

        root = eTree.fromstring(xmlstr)
        return self.read_xmlroot(root, None, None, None, None, None)

    def read_xmlroot(self, root, name, favid, userid, groupid, ret_list):
        if len(root.findall('fav')) == 0 and len(root.findall('folder')) == 0:
            return None
        if ret_list is None:
            _ret = []
        else:
            _ret = ret_list[:]

        for folder in root.findall('folder'):
            content = self.read_xmlroot(folder, None, None, None, None, None)
            if content:
                _ret.extend(content)

        for fav in root.findall('fav'):
            if name is not None or favid is not None:
                if favid is not None:
                    if 'id' in fav.attrib:
                        if fav.attrib['id'] != favid:
                            continue
            if name is not None:
                if 'name_fav' in fav.attrib:
                    if fav.attrib['name_fav'] != name:
                        continue

                    #            so = BitfarmSearch()
            so = self.fill_search_with_node(fav)

            if so is None:
                continue

            if so.searchtype is None and 'name' in root.attrib:
                if 'lokale' in root.attrib['name'].lower():
                    so.searchtype = 1
                elif 'globale' in root.attrib['name'].lower():
                    so.searchtype = 2
                elif 'verlauf-' in root.attrib['name'].lower():
                    so.searchtype = 3
                else:
                    so.searchtype = 1

            # userids haben Vorrang vor groupids
            if userid is not None:
                if so.userid is None:  # kein SearchObject mit Userids, wir sollen aber gegen UserIds prüfen? Überspringen...
                    continue
                if not len(set(so.userid).intersection(
                        userid)) > 0:  # wenn wenigstens eine userid in so.userid enthalten ist
                    continue

            if groupid is not None:
                if so.groupid is None:  # kein SearchObject mit groupids, wir sollen aber gegen groupids prüfen? Überspringen...
                    continue
                if not len(set(so.groupid).intersection(
                        groupid)) > 0:  # wenn wenigstens eine groupid in so.groupid enthalten ist
                    continue

            _ret.append(so)

        return _ret

    def fill_search_with_node(self, fav):
        """Funktion um ein SearchObject direkt mit einem XML-Knoten (fav-Node) zu füllen,
        bei Erfolg wird ein SearchObject zurückgegeben, andernfalls None.
        Der Knoten muss ein ElementTree.Element Objekt sein"""
        if not isinstance(fav, Element):
            return None
        _ret = bitfarm_doc.BitfarmSearch()
        try:
            if 'name_fav' in fav.attrib:
                _ret.name = str(fav.attrib['name_fav'])
            if 'id' in fav.attrib:
                _ret.id = fav.attrib['id']
            if 'Suchdatum' in fav.attrib:
                _ret.date = fav.attrib['Suchdatum']
            if 'date' in fav.attrib:
                _ret.date = fav.attrib['date']
            if 'userid' in fav.attrib:
                _ret.userid = int(fav.attrib['userid'])
            if 'groupid' in fav.attrib:
                _ret.groupid = int(fav.attrib['groupid'])
            if 'userlist' in fav.attrib:
                _ret.userlist = [int(x) for x in fav.attrib['userlist'].split(',')]
            if 'grouplist' in fav.attrib:
                _ret.grouplist = [int(x) for x in fav.attrib['grouplist'].split(',')]
            if not _ret.userlist and _ret.userid:
                _ret.userlist = [_ret.userid]
            if not _ret.grouplist and _ret.groupid:
                _ret.grouplist = [_ret.groupid]

            if 'urights' in fav.attrib:
                _ret.urights = fav.attrib['urights'].split(',')
                _ret.urights = [x.strip() for x in _ret.urights]
                _ret.userid = []
                for user in _ret.urights:
                    _item = self.con.main.read_user(name=user)
                    if _item is not None:
                        _ret.userid.append(_item.id)
                if len(_ret.userid) == 0:
                    _ret.userid = None
            if 'grights' in fav.attrib:
                _ret.grights = fav.attrib['grights'].split(',')
                _ret.grights = [x.strip() for x in _ret.grights]
                _ret.groupid = []
                for group in _ret.grights:
                    _item = self.con.main.read_group(name=group)
                    if _item is not None:
                        _ret.groupid.append(_item.id)
                if len(_ret.groupid) == 0:
                    _ret.groupid = None
            for elem in fav.iter():
                if 'arcid' in fav.attrib:
                    _ret.arcid = [int(x) for x in fav.attrib['arcid'].split(',')]
                if 'gdocid' in fav.attrib:
                    _ret.gdocid = [int(x) for x in fav.attrib['gdocid'].split(',')]
                if 'timref' in fav.attrib:
                    _ret.timref = [int(x) for x in fav.attrib['timref'].split(',')]

                if elem.tag == 'arcid':
                    if ',' in elem.text:
                        _ret.arcid = [int(x) for x in elem.text.split(',')]
                    else:
                        _ret.arcid = int(elem.text)
                if elem.tag == 'gdocid':
                    if ',' in elem.text:
                        _ret.gdocid = [int(x) for x in elem.text.split(',')]
                    else:
                        _ret.gdocid = int(elem.text)
                if elem.tag == 'gdocidwithsymlink':
                    _ret.gdocidwithsymlink = elem.text == 'True'
                if elem.tag == 'timref':
                    if ',' in elem.text:
                        _ret.timref = [int(x) for x in elem.text.split(',')]
                    else:
                        _ret.timref = int(elem.text)
                if elem.tag == 'guid':
                    _ret.guid = elem.text
                if elem.tag == 'searchtype':
                    _ret.searchtype = int(elem.text)
                if elem.tag == 'Gbadvusrsrc':
                    _ret.constructive = elem.text == '0'
                if elem.tag == 'sqlwhere':
                    _ret.sqlwhere = str(elem.text) if elem.text is not None else None
                if elem.tag == 'maxvalues':
                    if elem.text.lower() != 'alle' and elem.text.lower() != 'all':
                        try:
                            _ret.maxvalues = int(elem.text)
                        except:
                            _ret.maxvalues = 50
                    else:
                        _ret.maxvalues = 0
                if elem.tag == 'iStat' or elem.tag == 'state':
                    _ret.state = int(elem.text)
                if elem.tag == 'niStat' or elem.tag == 'notstate':
                    _ret.notstate = int(elem.text)
                #                if elem.tag == 'slw' or elem.tag == 'tags':
                #                    _ret.tags = unicode(elem.text)
                #                if elem.tag == 'notslw' or elem.tag == 'tags_negated':
                #                    _ret.tags_negated = elem.text
                if elem.tag == 'slw' or elem.tag == 'tags':
                    _ret.searchstring = str(elem.text) if elem.text is not None else None
                if elem.tag == 'notslw' or elem.tag == 'tags_negated' or elem.tag == 'ftx_negated':
                    _ret.searchstring_negated = elem.text == 'True'
                if elem.tag == 'iCheckout':
                    _ret.checkout = elem.text == '-1'
                if elem.tag == 'checkout':
                    _ret.checkout = elem.text if elem.text != '-1' else None
                if elem.tag == 'inresub':
                    _ret.inresub = bool(elem.text)
                if elem.tag == 'inresub_negated':
                    _ret.inresub_negated = bool(elem.text)
                if elem.tag == 'count':
                    _ret.count = elem.text == 'True'
                #                if elem.tag == 'vtx' or elem.tag == 'ftx':
                #                    _ret.ftx = str(elem.text)
                if elem.tag == 'vtx' or elem.tag == 'ftx':
                    _ret.searchstring = str(elem.text) if elem.text is not None else None
                if elem.tag == 'ftx_matchagainst':
                    _ret.ftx_matchagainst = elem.text == 'True'
                if elem.tag == 'ftx_booleanmode':
                    _ret.ftx_booleanmode = elem.text == 'True'
                if elem.tag == 'ftx_naturalmode':
                    _ret.ftx_naturalmode = elem.text == 'True'
                if elem.tag == 'ftx_withqueryexpansion':
                    _ret.ftx_withqueryexpansion = elem.text == 'True'
                if elem.tag == 'ftx_fields':
                    _ret.ftx_fields = int(elem.text)
                if elem.tag == 'refsearch':
                    _ret.refsearch = str(elem.text)
                if elem.tag == 'refsearch_negated':
                    _ret.refsearch_negated = elem.text == 'True'
                if elem.tag == 'searchdatetype':
                    _ret.searchdatetype = int(elem.text)
                if elem.tag == 'usemodifieddate':
                    _ret.usemodifieddate = elem.text == 'True'
                if elem.tag == 'Benutzer' or elem.tag == 'searchuser':
                    _ret.searchuser = str(elem.text)
                if elem.tag == 'notBenutzer':
                    _ret.searchuser = str(elem.text)
                    _ret.searchuser_negated = True
                if elem.tag == 'searchuser_negated':
                    _ret.searchuser_negated = True
                if elem.tag == 'searchuserid':
                    _ret.searchuserid = int(elem.text)
                if elem.tag == 'searchgroupid':
                    _ret.searchgroupid = int(elem.text)
                if elem.tag == 'searchgroup':
                    _ret.searchgroup = str(elem.text)
                if elem.tag == 'searchgroup_negated':
                    _ret.searchgroup_negated = elem.text == 'True'
                #                if elem.tag == 'ftxandtags':
                #                    _ret.searchstring = elem.text == 'True'
                if elem.tag == 'constructive':
                    _ret.constructive = elem.text == 'True'
                if elem.tag == 'sqlwhere_only':
                    _ret.sqlwhere_only = elem.text == 'True'
                if elem.tag == 'sqlwhere_operator':
                    _ret.sqlwhere_operator = elem.text
                if elem.tag == 'selected_cols':
                    _ret.selected_cols = str(elem.text)
                    _ret.selected_cols = _ret.selected_cols.split(',')

                if elem.tag == 'datevom':
                    _ret.searchdate = self._convert_delphi_datefloat(elem.text)
                if elem.tag == 'datebis':
                    _ret.searchdatehigh = self._convert_delphi_datefloat(elem.text)
                if elem.tag == 'notdatevom':
                    _ret.searchdate = self._convert_delphi_datefloat(elem.text)
                    _ret.searchdate_negated = True
                if elem.tag == 'searchdate':
                    try:
                        _ret.searchdate = datetime.datetime.strptime(elem.text, '%Y-%m-%d %H:%M:%S')
                    except:
                        _ret.searchdate = None
                if elem.tag == 'searchdatehigh':
                    try:
                        _ret.searchdatehigh = datetime.datetime.strptime(elem.text, '%Y-%m-%d %H:%M:%S')
                    except:
                        _ret.searchdatehigh = None

                if elem.tag == 'resubdate':
                    try:
                        _ret.resubdate = datetime.datetime.strptime(elem.text, '%Y-%m-%d %H:%M:%S')
                    except:
                        _ret.resubdate = None
                if elem.tag == 'resubdatehigh':
                    try:
                        _ret.resubdatehigh = datetime.datetime.strptime(elem.text, '%Y-%m-%d %H:%M:%S')
                    except:
                        _ret.resubdatehigh = None
                if elem.tag == 'resubdate_negated':
                    _ret.resubdate_negated = elem.text == 'True'

                if elem.tag == 'resubalarm':
                    try:
                        _ret.resubalarm = datetime.datetime.strptime(elem.text, '%Y-%m-%d %H:%M:%S')
                    except:
                        _ret.resubalarm = None
                if elem.tag == 'resubalarm_high':
                    try:
                        _ret.resubalarmhigh = datetime.datetime.strptime(elem.text, '%Y-%m-%d %H:%M:%S')
                    except:
                        _ret.resubalarmhigh = None
                if elem.tag == 'resubalarm_negated':
                    _ret.resubalarm_negated = elem.text == 'True'
                if elem.tag == 'resubstate':
                    _ret.resubstate = int(elem.text)
                if elem.tag == 'resubnotstate':
                    _ret.resubnotstate = int(elem.text)
                if elem.tag == 'resubstate_negated':
                    _ret.resubstate_negated = elem.text == 'True'
                if elem.tag == 'resubcontrollerid':
                    _ret.resubcontrollerid = int(elem.text)
                if elem.tag == 'resubcontrollerid_negated':
                    _ret.resubcontrollerid_negated = elem.text == 'True'
                if elem.tag == 'resubuserid':
                    _ret.resubuserid = int(elem.text)
                if elem.tag == 'resubuserid_negated':
                    _ret.resubuserid_negated = elem.text == 'True'
                if elem.tag == 'resubuserid_list':
                    _ret.resubuserid_list = elem.text
                if elem.tag == 'resulttype':
                    _ret.resulttype = elem.text
                if elem.tag == 'direction':
                    _ret.direction = elem.text
                if elem.tag == 'order':
                    _ret.order = str(elem.text)
                if elem.tag == 'docname':
                    _ret.docname = str(elem.text)
                if elem.tag == 'docname_negated':
                    _ret.docname_negated = elem.text == 'True'
                if elem.tag == 'filetype':
                    _ret.filetype = str(elem.text)
                if elem.tag == 'filetype_negated':
                    _ret.filetype_negated = elem.text == 'True'
                if elem.tag == 'verify':
                    _ret.verify = elem.text == 'True'
                if elem.tag == 'debug':
                    _ret.debug = elem.text == 'True'
                if elem.tag == 'selectedarcid':
                    _ret.selectedarcid = int(elem.text)
                if elem.tag == 'dont_refresh_dataset':
                    _ret.dont_refresh_dataset = elem.text == 'True'
                if elem.tag == 'dont_inheritate':
                    _ret.dont_inheritate = elem.text == 'True'

                if elem.tag == 'zusatzname' or elem.tag == 'searchfields':
                    add = None
                    for sub in elem.iter():
                        if sub.tag == 'BitfarmAddfield' or sub.tag == 'zusatzname':
                            if add and add.id:  # ein existierendes Feld hat wenigstens eine ID
                                _ret.addfields.append(add)
                            add = bitfarm_doc.BitfarmAddfield()
                        elif sub.tag == 'BitfarmStatefield' or sub.tag == 'searchfields':
                            if add and add.id:  # ein existierendes Feld hat wenigstens eine ID
                                _ret.statefields.append(add)
                            add = bitfarm_doc.BitfarmStatefield()

                        if not add:
                            continue

                        if 'text' in sub.attrib:
                            add.add = sub.attrib['text']
                        if sub.tag == 'type':
                            add.type = self._addfieldtypes[sub.text]
                        if sub.tag == 'value' or sub.tag == '_BitfarmAddfield__value' or sub.tag == 'valueUG':
                            if add.type == 1:
                                if elem.tag == 'zusatzname':
                                    add.value = datetime.datetime.strptime(sub.text, '%d.%m.%Y')
                                else:
                                    try:
                                        add.value = datetime.datetime.strptime(sub.text, '%Y-%m-%d %H:%M:%S')
                                    except:
                                        add.value = None
                            else:
                                add.value = str(sub.text)
                        if sub.tag == '_BitfarmStatefield__value':
                            add.value = sub.text.lower() == 'true'
                        if sub.tag == '_BitfarmStatefield__bitmask':
                            add.bitmask = int(sub.text)
                        if sub.tag == 'indexcolumnid':
                            add.indexcolumnid = int(sub.text)
                        if sub.tag == 'func':
                            add.func = int(sub.text)
                        if sub.tag == 'flags':
                            add.flags = int(sub.text)
                        if sub.tag == 'addfieldid':
                            add.addfieldid = int(sub.text)
                        if sub.tag == 'displayname':
                            add.name = str(sub.text)
                        if sub.tag == 'valuehigh' or sub.tag == 'valueOG':
                            if add.type == 1:
                                if elem.tag == 'zusatzname':
                                    add.valuehigh = datetime.datetime.strptime(sub.text, '%d.%m.%Y')
                                else:
                                    try:
                                        add.valuehigh = datetime.datetime.strptime(sub.text, '%Y-%m-%d %H:%M:%S')
                                    except:
                                        add.valuehigh = None
                            else:
                                add.valuehigh = str(sub.text)
                        if sub.tag == 'NegBtn':
                            add.negated = sub.text == '0'
                        if sub.tag == 'negated':
                            add.negated = sub.text == 'True'
                        if sub.tag == 'or_junction':
                            add.or_junction = sub.text == 'True'
                        if sub.tag == 'mode':
                            add.mode = int(sub.text)
                        if sub.tag == 'id':
                            add.id = int(sub.text)
                        if sub.tag == 'add':
                            add.add = sub.text

                    if add:
                        add.modified_search = True
                        if isinstance(add, bitfarm_doc.BitfarmAddfield):
                            _ret.addfields.append(add)
                        elif isinstance(add, bitfarm_doc.BitfarmStatefield):
                            _ret.statefields.append(add)
        except:
            self.log.exception(u'Fehler bei fill_search_with_node')
            return None
        else:
            return _ret

    def write_xmlfile(self, path, items):
        """Funktion um SearchObjekte als XML-File zu schreiben. Das Format der XML entspricht der Version 3.6.
       Es kann ein Objekte oder eine Liste an Objekten übergeben werden. Funktion liefert True bei Erfolg"""

        if not path:
            return False

        if type(items) is not list:
            favs = [items]
        else:
            favs = items[:]

        for item in list(favs):
            if type(item) is not bitfarm_doc.BitfarmSearch:
                favs.remove(item)

        if len(favs) == 0:
            return False

        root = self.write_xml(favs)

        tree = eTree.ElementTree(root)
        tree.write(path, xml_declaration=True)
        return True

    def create_xml(self, items):

        if type(items) is not list:
            favs = [items]
        else:
            favs = items[:]

        for item in list(favs):
            if type(item) is not bitfarm_doc.BitfarmSearch:
                favs.remove(item)

        if len(favs) == 0:
            return False

        root = self.write_xml(favs)

        # tree = eTree.ElementTree(root)

        return '<?xml version="1.0"?>' + eTree.tostring(root).decode()

    def write_xml(self, favs):

        root = eTree.Element('bookmark')
        if favs[0].searchtype in self._searchtypes:
            root.attrib['name'] = self._searchtypes[favs[0].searchtype]

        for fav in favs:
            _elem = eTree.SubElement(root, 'fav')

            if fav.date is not None:
                _elem.attrib['date'] = fav.date
            if fav.id is not None:
                _elem.attrib['id'] = str(fav.id)
            if fav.name is not None:
                _elem.attrib['name_fav'] = fav.name
            if fav.userid is not None:
                _elem.attrib['userid'] = str(fav.userid)
            if fav.groupid is not None:
                _elem.attrib['groupid'] = str(fav.groupid)
                # deprecated
            #            if fav.urights is not None:
            #                _elem.attrib['urights'] = ",".join(fav.urights)
            #            if fav.grights is not None:
            #                _elem.attrib['grights'] = ",".join(fav.grights)
            if fav.arcid is not None:
                _sub = eTree.SubElement(_elem, 'arcid')
                if type(fav.arcid) is list:
                    _sub.text = ";".join([str(x) for x in fav.arcid])
                else:
                    _sub.text = str(fav.arcid)
            if fav.gdocid is not None:
                _sub = eTree.SubElement(_elem, 'gdocid')
                if type(fav.gdocid) is list:
                    _sub.text = ";".join(fav.gdocid)
                else:
                    _sub.text = str(fav.gdocid)
            if fav.gdocidwithsymlink:
                _sub = eTree.SubElement(_elem, 'gdocidwithsymlink')
                _sub.text = str(fav.gdocidwithsymlink)
            if fav.grouplist is not None:
                _sub = eTree.SubElement(_elem, 'grouplist')
                _sub.text = ",".join([str(x) for x in fav.grouplist])
            if fav.timref is not None:
                _sub = eTree.SubElement(_elem, 'timref')
                _sub.text = ",".join(fav.timref)
            if fav.searchdate is not None:
                _sub = eTree.SubElement(_elem, 'searchdate')
                _sub.text = fav.searchdate.strftime('%Y-%m-%d %H:%M:%S')
            if fav.searchdatehigh is not None:
                _sub = eTree.SubElement(_elem, 'searchdatehigh')
                _sub.text = fav.searchdatehigh.strftime('%Y-%m-%d %H:%M:%S')
            if fav.searchdate_negated:
                _sub = eTree.SubElement(_elem, 'searchdate_negated')
                _sub.text = str(fav.searchdate_negated)
            if fav.searchtype is not None:
                _sub = eTree.SubElement(_elem, 'searchtype')
                _sub.text = str(fav.searchtype)
            if fav.resulttype is not None:
                _sub = eTree.SubElement(_elem, 'resulttype')
                _sub.text = fav.resulttype
            if fav.sqlwhere:
                _sub = eTree.SubElement(_elem, 'sqlwhere')
                _sub.text = fav.sqlwhere
            if fav.sqlwhere_only:
                _sub = eTree.SubElement(_elem, 'sqlwhere_only')
                _sub.text = str(fav.sqlwhere_only)
            if fav.sqlwhere_operator:
                _sub = eTree.SubElement(_elem, 'sqlwhere_operator')
                _sub.text = fav.sqlwhere_operator
            if fav.maxvalues is not None:
                _sub = eTree.SubElement(_elem, 'maxvalues')
                _sub.text = str(fav.maxvalues)
            if fav.state is not None:
                _sub = eTree.SubElement(_elem, 'state')
                _sub.text = str(fav.state)
            if fav.notstate is not None:
                _sub = eTree.SubElement(_elem, 'notstate')
                _sub.text = str(fav.notstate)
            #            if fav.tags is not None:
            #                _sub = eTree.SubElement(_elem, 'tags')
            #                _sub.text = str(fav.tags)
            #            if fav.tags_negated:
            #                _sub = eTree.SubElement(_elem, 'tags_negated')
            #                _sub.text = str(fav.tags_negated)
            if fav.checkout is not None:
                _sub = eTree.SubElement(_elem, 'checkout')
                _sub.text = str(fav.checkout)
            if fav.inresub is not None:
                _sub = eTree.SubElement(_elem, 'inresub')
                _sub.text = str(fav.inresub)
            if fav.inresub_negated is not None:
                _sub = eTree.SubElement(_elem, 'inresub_negated')
                _sub.text = str(fav.inresub_negated)
            if fav.count:
                _sub = eTree.SubElement(_elem, 'count')
                _sub.text = str(fav.count)
            if fav.ftx_fields:
                _sub = eTree.SubElement(_elem, 'ftx_fields')
                _sub.text = str(fav.ftx_fields)
            if fav.searchstring is not None:
                _sub = eTree.SubElement(_elem, 'searchstring')
                _sub.text = fav.searchstring
            if fav.searchstring_negated:
                _sub = eTree.SubElement(_elem, 'searchstring_negated')
                _sub.text = str(fav.searchstring_negated)
            if fav.ftx_matchagainst:
                _sub = eTree.SubElement(_elem, 'ftx_matchagainst')
                _sub.text = str(fav.ftx_matchagainst)
            if fav.ftx_booleanmode:
                _sub = eTree.SubElement(_elem, 'ftx_booleanmode')
                _sub.text = str(fav.ftx_booleanmode)
            if fav.ftx_naturalmode:
                _sub = eTree.SubElement(_elem, 'ftx_naturalmode')
                _sub.text = str(fav.ftx_naturalmode)
            if fav.ftx_withqueryexpansion:
                _sub = eTree.SubElement(_elem, 'ftx_withqueryexpansion')
                _sub.text = str(fav.ftx_withqueryexpansion)
            if fav.refsearch is not None:
                _sub = eTree.SubElement(_elem, 'refsearch')
                _sub.text = fav.refsearch
            if fav.searchstring_negated:
                _sub = eTree.SubElement(_elem, 'refsearch_negated')
                _sub.text = str(fav.refsearch_negated)
            if fav.searchuserid is not None:
                _sub = eTree.SubElement(_elem, 'searchuserid')
                _sub.text = str(fav.searchuserid)
            if fav.searchuser is not None:
                _sub = eTree.SubElement(_elem, 'searchuser')
                _sub.text = fav.searchuser
            if fav.searchuser_negated:
                _sub = eTree.SubElement(_elem, 'searchuser_negated')
                _sub.text = str(fav.searchuser_negated)
            if fav.searchgroupid is not None:
                _sub = eTree.SubElement(_elem, 'searchgroupid')
                _sub.text = str(fav.searchgroupid)
            if fav.searchgroup is not None:
                _sub = eTree.SubElement(_elem, 'searchgroup')
                _sub.text = fav.searchgroup
            if fav.searchgroup_negated:
                _sub = eTree.SubElement(_elem, 'searchgroup_negated')
                _sub.text = str(fav.searchgroup_negated)
            #            if fav.ftxandtags:
            #                _sub = eTree.SubElement(_elem, 'ftxandtags')
            #                _sub.text = str(fav.ftxandtags)
            if fav.constructive is not None:
                _sub = eTree.SubElement(_elem, 'constructive')
                _sub.text = str(fav.constructive)
            if fav.selected_cols is not None:
                _sub = eTree.SubElement(_elem, 'selected_cols')
                if type(fav.selected_cols) is list:
                    _sub.text = ",".join(fav.selected_cols)
                else:
                    _sub.text = fav.selected_cols
            if fav.direction:
                _sub = eTree.SubElement(_elem, 'direction')
                _sub.text = fav.direction
            if fav.order:
                _sub = eTree.SubElement(_elem, 'order')
                _sub.text = fav.order
            if fav.docname:
                _sub = eTree.SubElement(_elem, 'docname')
                _sub.text = fav.docname
            if fav.docname_negated:
                _sub = eTree.SubElement(_elem, 'docname_negated')
                _sub.text = str(fav.docname_negated)
            if fav.filetype:
                _sub = eTree.SubElement(_elem, 'filetype')
                _sub.text = fav.filetype
            if fav.filetype_negated:
                _sub = eTree.SubElement(_elem, 'filetype_negated')
                _sub.text = str(fav.filetype_negated)
            if fav.resubalarm is not None:
                _sub = eTree.SubElement(_elem, 'resubalarm')
                _sub.text = fav.resubalarm.strftime('%Y-%m-%d %H:%M:%S')
            if fav.resubalarm_high is not None:
                _sub = eTree.SubElement(_elem, 'resubalarm_high')
                _sub.text = fav.resubalarm_high.strftime('%Y-%m-%d %H:%M:%S')
            if fav.resubalarm_negated:
                _sub = eTree.SubElement(_elem, 'resubalarm_negated')
                _sub.text = str(fav.resubalarm_negated)
            if fav.resubdate is not None:
                _sub = eTree.SubElement(_elem, 'resubdate')
                _sub.text = fav.resubdate.strftime('%Y-%m-%d %H:%M:%S')
            if fav.resubdate_high is not None:
                _sub = eTree.SubElement(_elem, 'resubdate_high')
                _sub.text = fav.resubdate_high.strftime('%Y-%m-%d %H:%M:%S')
            if fav.resubdate_negated:
                _sub = eTree.SubElement(_elem, 'resubdate_negated')
                _sub.text = str(fav.resubdate_negated)
            if fav.resubcontrollerid is not None:
                _sub = eTree.SubElement(_elem, 'resubcontrollerid')
                _sub.text = str(fav.resubcontrollerid)
            if fav.resubcontrollerid_negated:
                _sub = eTree.SubElement(_elem, 'resubcontrollerid_negated')
                _sub.text = str(fav.resubcontrollerid_negated)
            if fav.resubuserid is not None:
                _sub = eTree.SubElement(_elem, 'resubuserid')
                _sub.text = str(fav.resubuserid)
            if fav.resubuserid_negated:
                _sub = eTree.SubElement(_elem, 'resubuserid_negated')
                _sub.text = str(fav.resubuserid_negated)
            if fav.resubuserid_list is not None:
                _sub = eTree.SubElement(_elem, 'resubuserid_list')
                _sub.text = str(fav.resubuserid_list)
            if fav.resubstate:
                _sub = eTree.SubElement(_elem, 'resubstate')
                _sub.text = str(fav.resubstate)
            if fav.resubnotstate:
                _sub = eTree.SubElement(_elem, 'resubnotstate')
                _sub.text = str(fav.resubnotstate)
            if fav.resubstate_negated:
                _sub = eTree.SubElement(_elem, 'resubstate_negated')
                _sub.text = str(fav.resubstate_negated)
            if fav.selectedarcid is not None:
                _sub = eTree.SubElement(_elem, 'selectedarcid')
                _sub.text = str(fav.selectedarcid)
            if fav.verify:
                _sub = eTree.SubElement(_elem, 'verify')
                _sub.text = str(fav.verify)
            if fav.debug:
                _sub = eTree.SubElement(_elem, 'debug')
                _sub.text = str(fav.debug)
            if fav.guid:
                _sub = eTree.SubElement(_elem, 'guid')
                _sub.text = fav.guid

            if len(fav.addfields) > 0:
                _sub = eTree.SubElement(_elem, 'searchfields')

                for field in fav.addfields:
                    if isinstance(field, bitfarm_doc.BitfarmStatefield):
                        continue
                    _subelem = eTree.SubElement(_sub, 'BitfarmAddfield')
                    #                    _subelem.attrib['text'] = field.add
                    _add = eTree.SubElement(_subelem, 'displayname')
                    _add.text = field.name
                    _add = eTree.SubElement(_subelem, 'add')
                    _add.text = field.add
                    _add = eTree.SubElement(_subelem, 'type')
                    _add.text = str(field.type)
                    _add = eTree.SubElement(_subelem, 'indextableid')
                    _add.text = str(field.indextableid)
                    _add = eTree.SubElement(_subelem, 'value')
                    if type(field.value) is datetime.datetime:
                        _add.text = field.value.strftime('%Y-%m-%d %H:%M:%S')
                    else:
                        _add.text = field.value
                    if field.valuehigh is not None:
                        _add = eTree.SubElement(_subelem, 'valuehigh')
                        if type(field.valuehigh) is datetime.datetime:
                            _add.text = field.valuehigh.strftime('%Y-%m-%d %H:%M:%S')
                        else:
                            _add.text = field.valuehigh
                    if field.negated:
                        _add = eTree.SubElement(_subelem, 'negated')
                        _add.text = str(field.negated)
                    if field.or_junction:
                        _add = eTree.SubElement(_subelem, 'or_junction')
                        _add.text = str(field.or_junction)

            if len(fav.statefields) > 0:
                _sub = eTree.SubElement(_elem, 'searchfields')

                for field in fav.statefields:
                    if isinstance(field, bitfarm_doc.BitfarmAddfield):
                        continue
                    _subelem = eTree.SubElement(_sub, 'BitfarmStatefield')
                    _state = eTree.SubElement(_subelem, 'name')
                    _state.text = field.name
                    _state = eTree.SubElement(_subelem, 'id')
                    _state.text = str(field.id)
                    _state = eTree.SubElement(_subelem, 'indexcolumnid')
                    _state.text = str(field.indexcolumnid)
                    _state = eTree.SubElement(_subelem, 'bitmask')
                    _state.text = str(field.bitmask)
                    _state = eTree.SubElement(_subelem, 'value')
                    _state.text = str(field.value)
                    if field.or_junction:
                        _state = eTree.SubElement(_subelem, 'or_junction')
                        _state.text = str(field.or_junction)

        return root

    def _convert_delphi_datefloat(self, datestr):
        if datestr.find(',') > -1:
            days, fl = datestr.split(',')
            fl = '0.' + fl
            fl = float(fl)
        else:
            days = datestr
            fl = 0.0

        _ret = datetime.datetime(1899, 12, 30, 0, 0)
        _ret = _ret + datetime.timedelta(days=int(days))
        if fl > 0.0:
            sec = int(round(fl * 86400))
            _ret = _ret + datetime.timedelta(seconds=int(sec))

        return _ret

    def resolve_arcids(self, fav):
        _ret = []
        groupids = []

        sql = "select groupid from sys_groupuser where userid = %s" % fav.user_context
        try:
            result = self.db[fav.profile].engine.execute(sql).fetchall()
        except:
            self.log.exception(u"Fehler bei der Gruppenauflösung für die Suche")
        else:
            for row in result:
                groupids.append(row[0])

        if not groupids:
            groupids = [0]

        sql = "select * from sys_arc as a left join sys_arcxs as xs on a.arcid = xs.arcid"
        try:
            result = self.db[fav.profile].engine.execute(sql).fetchall()
        except:
            self.log.exception(u"Fehler bei der Gruppenauflösung für die Suche")
        else:
            arcid = None
            lgrid = None
            for row in result:
                if row['arcid'] == fav.arcid:
                    if row['type'] == 0:
                        _ret = self._recurse_arcids(result, fav.arcid, fav.user_context, groupids)
                    elif 'list' in row['rights']:
                        _ret = [fav.arcid]
        _ret.sort(key=lambda x: x)
        return ",".join([str(x) for x in _ret])

    def _recurse_arcids(self, result, arcid, userid, groupids):
        _ret = []
        for row in result:
            if row['arcid'] in _ret:
                continue
            if row['lgrid'] == arcid:
                if row['type'] == 0:  # lager rekursiv durchlaufen
                    rec_arcids = self._recurse_arcids(result, row['arcid'], userid, groupids)
                    for rec_arcid in rec_arcids:
                        if rec_arcid is not None and rec_arcid not in _ret:
                            _ret.append(rec_arcid)
                    if row['arcid'] is not None and row['arcid'] not in _ret and len(
                            rec_arcids) > 0:  # wenn wir unterhalb des durchsuchten lagers archive gefunden haben
                        _ret.append(row['arcid'])  # fügen wir die id des lagers auch hinzu
                else:
                    if row['rights'] is None or row['userid'] is None or row['groupid'] is None:
                        continue
                    if row['arcid'] is not None and 'list' in row['rights'] and (row['userid'] == userid or (
                            row['userid'] == 0 and row[
                        'groupid'] in groupids)):  # archiv nur hinzu wenn berechtigung besteht
                        _ret.append(row['arcid'])

        return _ret

    def search(self, fav):
        """Suchroutine. Bekommt ein SearchObjekt übergeben und liefert ein Resultset zurück, oder None bei Fehlschlag"""
        self.log.debug("Suchobjekt: %s" % str(bfaxml.object_to_xml(fav)))
        cols = None
        if type(fav) is not bitfarm_doc.BitfarmSearch:
            return None

        if self.version != 36:
            return None

        if fav.profile is None:
            fav.profile = self.con.main.profile

        if type(fav.arcid) is int:
            arcid = self.resolve_arcids(fav)
        elif type(fav.arcid) in (list, set):
            arcid = ",".join([str(x) for x in fav.arcid])
        else:
            arcid = fav.arcid

        if not fav.userlist:
            fav.userlist = [fav.user_context]

        if not fav.grouplist:
            fav.grouplist = None
        else:
            fav.grouplist = ",".join([str(x) for x in fav.grouplist])

        if fav.sqlwhere is None or fav.sqlwhere == u'None':
            fav.sqlwhere = None
            fav.sqlwhere_operator = None
        else:
            self.log.debug(fav.sqlwhere)

        for add in fav.addfields:
            if 'items' in add.__dict__:
                del add.items
            if 'rights' in add.__dict__:
                del add.rights
            if 'history' in add.__dict__:
                del add.history

        for state in fav.statefields:
            if 'rights' in state.__dict__:
                del state.rights
            if 'history' in state.__dict__:
                del state.history

        stmt = sqla.text(u'CALL sp_search(:prm_searchtype, :prm_version, :prm_arcid, :prm_order, '
                         u':prm_direction, :prm_limit, :prm_userid, :prm_searchuserid, '
                         u':prm_searchuser, :prm_searchuser_negated, :prm_sqlwhere_only, '
                         u':prm_sqlwhere, :prm_sqlwhere_operator, :prm_ftx, :prm_ftx_negated, '
                         u':prm_matchagainst, :prm_natural_mode, :prm_boolean_mode, '
                         u':prm_with_query_expansion, :prm_ftx_fields, :prm_result_type, :prm_selected_cols, '
                         u':prm_addfield_xml, :prm_stsfield_xml, :prm_xmlpath, :prm_count, '
                         u':prm_searchdate, :prm_searchdate_high, :prm_searchdate_negated, '
                         u':prm_checkout, :prm_state, :prm_notstate, :prm_searchgroupid, '
                         u':prm_searchgroup, :prm_searchgroup_negated, :prm_gdocid, :prm_gdocid_with_symlink, '
                         u':prm_refsearch, :prm_refsearch_negated, :prm_timref, :prm_resubdate, '
                         u':prm_resubdate_high, :prm_resubalarm, '
                         u':prm_resubalarm_high, :prm_resubstate, :prm_resubuserid, :prm_resubuserid_negated, '
                         u':prm_resubcontrollerid, :prm_resubcontrollerid_negated, :prm_resubdate_negated, '
                         u':prm_resubalarm_negated, :prm_grouptype, :prm_usertype, :prm_root,'
                         u':prm_mandant, :prm_store, :prm_archive, :prm_recordarchive, :prm_dmsadmin, '
                         u':prm_docname, :prm_docname_negated, :prm_grouplist, :prm_searchdatetype, '
                         u':prm_filetype, :prm_filetype_negated, :prm_resub, :prm_resubstate_negated, '
                         u':prm_resubuserid_list, :prm_debug, :prm_abo_userid, :prm_usesearch35, '
                         u':prm_resubnotstate, :prm_userlist, :prm_resub_negated, :prm_rightsmask, '
                         u':prm_groupby, :prm_selectedcols_inresub)')

        stmt = stmt.bindparams(sqla.sql.bindparam('prm_searchtype', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_version', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_arcid', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_order', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_direction', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_limit', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_userid', type_=sqla.types.BIGINT),
                               sqla.sql.bindparam('prm_searchuserid', type_=sqla.types.BIGINT),
                               sqla.sql.bindparam('prm_searchuser', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_searchuser_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_sqlwhere_only', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_sqlwhere', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_sqlwhere_operator', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_ftx', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_ftx_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_matchagainst', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_natural_mode', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_boolean_mode', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_with_query_expansion', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_ftx_fields', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_result_type', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_selected_cols', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_addfield_xml', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_stsfield_xml', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_xmlpath', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_count', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_searchdate', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_searchdate_high', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_searchdate_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_checkout', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_state', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_notstate', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_searchgroupid', type_=sqla.types.BIGINT),
                               sqla.sql.bindparam('prm_searchgroup', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_searchgroup_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_gdocid', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_gdocid_with_symlink', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_refsearch', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_refsearch_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_timref', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_resubdate', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_resubdate_high', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_resubalarm', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_resubalarm_high', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_resubstate', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_resubuserid', type_=sqla.types.BIGINT),
                               sqla.sql.bindparam('prm_resubuserid_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resubcontrollerid', type_=sqla.types.BIGINT),
                               sqla.sql.bindparam('prm_resubcontrollerid_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resubdate_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resubalarm_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_grouptype', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_usertype', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_root', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_mandant', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_store', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_archive', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_recordarchive', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_dmsadmin', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_docname', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_docname_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_grouplist', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_searchdatetype', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_filetype', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_filetype_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resub', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resubstate_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resubuserid_list', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_debug', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_abo_userid', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_usesearch35', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_resubnotstate', type_=sqla.types.INT),
                               sqla.sql.bindparam('prm_userlist', type_=sqla.types.NCHAR),
                               sqla.sql.bindparam('prm_resub_negated', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_rightsmask', type_=sqla.types.BIGINT),
                               sqla.sql.bindparam('prm_groupby', type_=sqla.types.SMALLINT),
                               sqla.sql.bindparam('prm_selectedcols_inresub', type_=sqla.types.SMALLINT)
                               )

        try:
            connection = self.db[fav.profile].engine.raw_connection()
            cursor = connection.cursor()

            addfields = bfaxml.object_to_xml(fav.addfields, encoding='utf-8', shortxml=False)
            statefields = bfaxml.object_to_xml(fav.statefields, encoding='utf-8', shortxml=False)
            usesearch35 = self.con.connections[fav.profile].usesearch35
            userlist = ",".join([str(x) for x in fav.userlist])
            xmlpath = None

            # fav.rightsmask = 1
            # fav.sqlwhere = ''
            # fav.sqlwhere_operator = 'and'
            # fav.searchuserid = -1
            # fav.searchgroupid = -1
            # addfields = ''
            # statefields = ''
            # fav.gdocid = ''
            # fav.timref = ''
            # fav.resubstate = -1
            # fav.resubnotstate = -1
            # fav.resubuserid_list = ''
            # fav.resubuserid = -1
            # fav.grouplist = ''
            # fav.searchdatetype = 0
            # fav.abo_userid = 0
            # fav.debug = 1
            # fav.searchuser_negated = 0
            # fav.sqlwhere_only = 0
            # usesearch35 = 0

            args = cursor.callproc('sp_search', [fav.searchtype, 36, arcid,
                                                 fav.order, fav.direction, fav.maxvalues, fav.user_context,
                                                 fav.searchuserid, fav.searchuser, fav.searchuser_negated,
                                                 fav.sqlwhere_only, fav.sqlwhere, fav.sqlwhere_operator,
                                                 fav.searchstring, fav.searchstring_negated, fav.ftx_matchagainst,
                                                 fav.ftx_naturalmode, fav.ftx_booleanmode, fav.ftx_withqueryexpansion,
                                                 fav.ftx_fields, fav.resulttype, fav.selected_cols,
                                                 bfaxml.object_to_xml(fav.addfields, encoding='utf-8', shortxml=False),
                                                 bfaxml.object_to_xml(fav.statefields, encoding='utf-8',
                                                                      shortxml=False),
                                                 u'', fav.count,
                                                 (fav.searchdate.strftime(
                                                     '%Y-%m-%d') if fav.searchdate is not None else None),
                                                 (fav.searchdatehigh.strftime(
                                                     '%Y-%m-%d') if fav.searchdatehigh is not None else None),
                                                 fav.searchdate_negated, fav.checkout, fav.state, fav.notstate,
                                                 fav.searchgroupid, fav.searchgroup, fav.searchgroup_negated,
                                                 fav.gdocid, fav.gdocidwithsymlink, fav.refsearch,
                                                 fav.refsearch_negated,
                                                 fav.timref,
                                                 (fav.resubdate.strftime(
                                                     '%Y-%m-%d') if fav.resubdate is not None else None),
                                                 (fav.resubdate_high.strftime(
                                                     '%Y-%m-%d') if fav.resubdate_high is not None else None),
                                                 (fav.resubalarm.strftime(
                                                     '%Y-%m-%d') if fav.resubalarm is not None else None),
                                                 (fav.resubalarm_high.strftime(
                                                     '%Y-%m-%d') if fav.resubalarm_high is not None else None),
                                                 fav.resubstate, fav.resubuserid, fav.resubuserid_negated,
                                                 fav.resubcontrollerid, fav.resubcontrollerid_negated,
                                                 fav.resubdate_negated, fav.resubalarm_negated,
                                                 0, 0, 0, 0, 0, 0, 0, 0, fav.docname, fav.docname_negated,
                                                 fav.grouplist, fav.searchdatetype, fav.filetype,
                                                 fav.filetype_negated, fav.inresub, fav.resubstate_negated,
                                                 fav.resubuserid_list,
                                                 fav.debug, fav.abo_userid,
                                                 self.con.connections[fav.profile].usesearch35,
                                                 fav.resubnotstate, ",".join([str(x) for x in fav.userlist]),
                                                 fav.inresub_negated, fav.rightsmask, fav.groupby,
                                                 fav.selectedcols_inresub])

            result = cursor.fetchall()
            cols = [x[0] for x in cursor.description]

            # for stor_res in cursor.stored_results():
            #     result = stor_res.fetchall()
            #     cols = [x for x in stor_res.column_names]

            cursor.close()
        except:
            try:
                connection.commit()
            except:
                self.log.exception('Fehler bei connection.commit (1)')
            self.log.exception('Fehler bei der Suche sp_search %s' % fav)
            return None, u''

        try:
            connection.commit()
        except:
            self.log.exception('Fehler bei connection.commit (2)')

        try:
            connection.close()
        except:
            self.log.exception('Fehler bei connection.close')

        return result, cols


class BitfarmDocProcessing(object):
    """
    BitfarmDocProcessing-Klasse (bfaProc):
    Diese Klasse bietet die notwendigen Funktionen um Dokumente laden und verarbeiten zu können.
    Der Klasse muss ein bfaConnections-Objekt übergeben werden. Dadurch werden alle bekannten Profile übergeben.
    Jede Instanz von BitfarmDocProcessing erzeugt eine eigene Datenbank-Connection auf die Main-Connection-Engine.
    Daher muss in Threads jeweils eine eigene Instanz von BitfarmDocProcessing erzeugt werden.
    Optional kann ein Dictionary (_con_db) mit einer DB-Connection pro Profil übergeben werden, wenn von einem übergeordneten Thread
    die Db-Connections durchgereicht werden sollen.  {'Profil':sqla.engine.base.Connection}
    Diese Connections werden in self.db hinterlegt
    """
    version = __version__
    log = logging.getLogger(u"Main.bfaProc")
    log.setLevel(logging.WARNING)

    def __init__(self, _con, _con_db=None, name=u"", _gridfieldlock=None, _gridfunc=None):
        if not isinstance(_con, BitfarmConnections):
            raise NoneConnectionsError()

        self.con = _con
        self.version = _con.main.version
        self.name = name  # hier können threads ihren name bezeichner durchreichen
        self.gridfieldlock = _gridfieldlock
        self.gridfunc = _gridfunc
        if _con_db is not None and type(_con_db) is dict:
            self.db = _con_db
        else:
            self.db = {}
            for _profile in self.con.connections:
                if self.con[_profile].engine is not None:
                    self.db[_profile] = self.con[_profile].engine.connect()
                    if not isinstance(self.db[_profile], sqla.engine.base.Connection):
                        raise DatabaseConError(_con[_profile].confile)

        if self.con.main.profile in self.db:
            self.db['main'] = self.db[self.con.main.profile]

    def tasks_to_queue(self):
        return None

    def load(self, _doc, _gdocid=None,
             level=['base', 'text', 'fields', 'links', 'paths', 'iev', 'history', 'commentary'],
             orderfieldsbyname=False, logid='', with_gridcolumns=False):
        """
        Lädt ein Dokument mit der übergebenen gdocid aus der Datenbank.
        Die Methode reicht den Aufruf an die privaten Methoden __load_v3 und __load_v4
        weiter, entsprechend der Version self.version.
        Wenn das übergebene Dokument keinem Profil zugeodnet ist, wird es dem Profil der Main.Connection zugeordnet.
        @param _gdocid: gdocid (int) des zu ladenen Dokuments
        """
        if not level:
            level = ['base', 'text', 'fields', 'links', 'paths', 'iev', 'history', 'commentary']
        if level == 'complete' or 'complete' in level:
            level = ['base', 'text', 'fields', 'links', 'paths', 'iev', 'history', 'commentary', 'svn', 'workflow',
                     'sap', 'ftx']
        if self.version in (3, 35, '3.0'):
            return self.__load_v3(_doc, _gdocid, level=level, logid=logid)
        elif self.version == 36:
            return self.__load_v36(_doc, _gdocid, level=level, orderfieldsbyname=orderfieldsbyname, logid=logid,
                                   with_gridcolumns=with_gridcolumns)
        elif self.version == 40:
            return self.__load_v4(_doc, _gdocid, level=level, orderfieldsbyname=orderfieldsbyname, logid=logid)
        else:
            return None

    def __load_v3(self, _doc, _gdocid, level, logid=''):
        # wir überprüfen, ob eine gdocid übergeben wurde, oder wenn nicht, ob das doc eine hat
        if _gdocid:
            if type(_gdocid) is int:
                _doc.gdocid = _gdocid
        else:
            if _doc.gdocid is None:
                # TODO: wäre es angebracht eine exception zu werfen
                return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

        if _con and _doc.gdocid:
            try:
                s2 = time.perf_counter()
                from_obj = _con.sys_gdocid.join(_con.sys_arc, _con.sys_gdocid.c.arc_id == _con.sys_arc.c.ARC_ID)
                sql = sqla.select([_con.sys_gdocid, _con.sys_arc], from_obj=[from_obj],
                                  whereclause=(_con.sys_gdocid.c.gdoc_id == _doc.gdocid))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception(u"Fehler beim Laden des Dokuments %s" % _doc.gdocid)
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)

                if result:
                    self.log.debug("%s %s" % (logid, list(result)))
                    _doc.arcid = result[1]
                    _doc._docid = result[2]
                    _doc.archive = result[5]
                    _doc.tblname = result[6]
                    _doc.encoding = _con.db_encoding

                    s = time.perf_counter()
                    if _con.arc_tbl is None:
                        _con.arc_tbl = sqla.Table("arc_" + _doc.tblname, _con.meta, autoload=True)
                    self.log.debug("%s sqla.table autoload %s" % (logid, time.perf_counter() - s))

                    # bei der Abfrage scheint ein SQL-Rawstring schneller zu sein
                    #                    s = time.perf_counter()
                    sql = 'select * from arc_%s  where doc_id = %s' % (_doc.tblname, _doc._docid)
                    try:
                        result = _db.execute(sql).fetchone()
                    except:
                        self.log.exception(u"Fehler beim Laden des Dokuments %s" % _doc.gdocid)
                        result = _con.try_again(sql=sql, db=_db, fetchone=True)

                    self.log.debug("%s sql select %s" % (logid, time.perf_counter() - s))

                    #                s = time.perf_counter()
                    #                sql = self.arc_tbl.select(self.arc_tbl.c.DOC_ID == self._docid)
                    #                result = sql.execute().fetchone()
                    #                self.log.debug("sqla.select %s" %(time.perf_counter() - s))

                    if result:
                        _doc.docloc = result['DOC_LOC'] if result['DOC_LOC'] is not None else ''
                        _doc.extloc = result['EXT_LOC'] if result['EXT_LOC'] is not None else ''
                        _doc.oriloc = result['ORI_LOC'] if result['ORI_LOC'] is not None else ''
                        _doc.user = result['USR_NAME'] if result['USR_NAME'] is not None else ''
                        _doc.title = result['DISP_NAME'] if result['DISP_NAME'] is not None else ''
                        _doc.userid = result['USR_ID']
                        _doc.date = result['TStamp']
                        _doc.sha1 = result["SHA1"]
                        _doc.ftx.text = result['DOC_TXT'] if result['DOC_TXT'] is not None else ''
                        _doc.status = result['DOC_STATE']
                    else:
                        return None

                    if _doc.user != '' and _doc.userid == 0:  # in v35 stand nur der Username, nicht die ID in der Tabelle
                        user = _con.read_user(db=_db, name=_doc.user)
                        if user is not None:
                            _doc.userid = user.id

                    _doc.in_progress = _doc.status & 1 > 0

                    if _doc.version is None:
                        _doc.version = self.version

                    _doc.fields = self.get_addfields(_con, _db, _doc, load_items=True)
                    if result:
                        for field in _doc.fields:
                            if _doc.fields[field].add in result.keys():
                                if result[_doc.fields[field].add]:
                                    _doc.fields[field].value = result[_doc.fields[field].add]

                    sql = 'select DOC_SLW from slw_%s where doc_id = %s' % (_doc.tblname, _doc._docid)
                    try:
                        result = _db.execute(sql).fetchall()
                    except:
                        self.log.exception(u"Fehler beim Laden des Dokuments %s" % _doc.gdocid)
                        result = _con.try_again(sql=sql, db=_db, fetchall=True)

                    for row in result:
                        if row[0] not in _doc.tags:
                            _doc.tags.append(row[0])

                    sql = 'select distinct(ref_name) from sys_doclnk where arc_id=%s and doc_id=%s' % (_doc.arcid, _doc._docid)
                    try:
                        result = _db.execute(sql).fetchall()
                    except:
                        self.log.exception(u"Fehler beim Laden des Dokuments %s" % _doc.gdocid)
                        result = _con.try_again(sql=sql, db=_db, fetchall=True)

                    for row in result:
                        _doc.links.append(row[0])

                    sql = 'select * from sts_%s order by sts_id' % _doc.tblname
                    try:
                        result = _db.execute(sql).fetchall()
                    except:
                        self.log.exception(u"Fehler beim Laden des Dokuments %s" % _doc.gdocid)
                        result = _con.try_again(sql=sql, db=_db, fetchall=True)

                    if len(result) > 0:
                        ind = 0
                        for y in [int(2 ** x) for x in range(1, len(result) + 1)]:
                            _doc.statusdict[result[ind][1]] = _doc.status & y > 0
                            ind += 1

                    sql = 'select * from sys_doclock where gdoc_id = %i' % _doc.gdocid
                    try:
                        result = _db.execute(sql).fetchone()
                    except:
                        self.log.exception(u'%s konnte Ausgecheckt Status nicht ermitteln [%s]' % (logid, _doc.gdocid))
                        result = _con.try_again(sql=sql, db=_db, fetchone=True)

                    if result:
                        _doc.is_checkout = True
                        _doc.checkout_userid = result[4]

                    if 'links' in level:
                        sql = sqla.select([_con.sys_doclink.c.REF_NAME],
                                          whereclause=(sqla.and_(_con.sys_doclink.c.DOC_ID == _doc._docid,
                                                                 _con.sys_doclink.c.ARC_ID == _doc.arcid)),
                                          distinct=True)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            if row[0] not in _doc.links:
                                _doc.links.append(row[0])

                    if 'sap' in level and _con.sys_sap is not None:
                        sql = sqla.select([_con.sys_sap], whereclause=(_con.sys_sap.c.gdoc_id == _doc.gdocid)).order_by(
                            _con.sys_sap.c.sap_id)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            sapfield = bitfarm_doc.BitfarmSAPField()
                            sapfield.gdocid = _doc.gdocid
                            sapfield.sap_id = row[0]
                            sapfield.sap_contrep = row[3]
                            sapfield.sap_docid = row[4]
                            sapfield.sap_compid = row[5]
                            sapfield.sap_docprot = row[6]
                            sapfield.sap_docstatus = row[7]
                            sapfield.sap_tstampc = row[8]
                            sapfield.sap_tstampm = row[9]
                            sapfield.sap_content_path = row[10]
                            sapfield.sap_content_type = row[11]
                            sapfield.sap_charset = row[12]
                            sapfield.sap_version = row[13]

                            _doc.sap.append(sapfield)

                    if 'history' in level:
                        _doc.history = self.read_history(_doc)

                    if 'svn' in level and _con.sys_svn is not None:
                        sql = sqla.select([_con.sys_svn], whereclause=(_con.sys_svn.c.GDOC_ID == _doc.gdocid)).order_by(
                            _con.sys_svn.c.SVN_VERSION)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            svnfield = bitfarm_doc.BitfarmSvnField()
                            svnfield.gdocid = _doc.gdocid
                            svnfield.svn_id = row[0]
                            svnfield.svn_link = row[4]
                            svnfield.svn_version = row[5]
                            svnfield.svn_revision = row[6]
                            svnfield.svn_date = row[7]
                            svnfield.svn_info = row[8]

                            _doc.svn_history.append(svnfield)

                    if 'commentary' in level:
                        sql = 'select * from com_%s where doc_id = %s order by tstamp desc' % (_doc.tblname, _doc._docid)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u'%s konnte Kommentare nicht lesen [%s]' % (logid, _doc._docid))

                        if result:
                            for row in result:
                                com = bitfarm_doc.BitfarmCommentary()
                                com.id = row[0]
                                com.gdocid = _doc.gdocid
                                com.userid = row[4]
                                com.tstamp = row[5]
                                com.comment = row[6]
                                _doc.commentary.append(com)
                else:
                    self.log.warning(u'%s kein Dokument mit der ID [%s] gefunden' % (logid, _doc.gdocid))
                    return None
            except:
                self.log.exception(u"Allgemeiner Fehler beim Laden des Dokuments [%s]" % _doc.gdocid)
                return None

            # self.log.debug(u"%s laden eines kompletten Dokuments %s" % (logid, time.perf_counter() - s2))

            return True
        else:
            return None

    def __load_v36(self, _doc, _gdocid, level, orderfieldsbyname, logid='', with_gridcolumns=False):
        # wir überprüfen, ob eine gdocid übergeben wurde, oder wenn nicht, ob das doc eine hat
        if _gdocid:
            if type(_gdocid) is int:
                _doc.gdocid = _gdocid
            if type(_gdocid) is str:
                _doc.gdocid = int(_gdocid)
        else:
            if _doc.gdocid is None:
                # TODO: wäre es angebracht eine exception zu werfen
                return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

        if _con and _doc.gdocid:
            try:
                # s2 = time.time()
                # from_obj = _con.sys_doc.outerjoin(_con.sys_arc, _con.sys_doc.c.arcid == _con.sys_arc.c.arcid)
                # from_obj = from_obj.outerjoin(_con.sys_user, _con.sys_user.c.userid == _con.sys_doc.c.userid)
                # sql = sqla.select([_con.sys_doc, _con.sys_arc, _con.sys_user], from_obj=[from_obj],
                #                   whereclause=(_con.sys_doc.c.gdocid == _doc.gdocid), use_labels=True)
                from_obj = _con.sys_doc.outerjoin(_con.sys_arc, _con.sys_doc.c.arcid == _con.sys_arc.c.arcid).outerjoin(
                    _con.sys_user, _con.sys_user.c.userid == _con.sys_doc.c.userid)
                sql = sqla.select([_con.sys_doc, _con.sys_arc, _con.sys_user], use_labels=True).select_from(from_obj).where(_con.sys_doc.c.gdocid == _doc.gdocid)

                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))

                if result:
                    # self.log.debug(list(result))
                    if 'base' in level:
                        _doc.encoding = _con.db_encoding
                        _doc.arcid = result['sys_arc_arcid']
                        _doc.archive = result['sys_arc_arcname']
                        _doc.tblname = result['sys_arc_tblname']
                        _doc.storageid = result['sys_arc_storageid']
                        _doc.pages = result["sys_doc_pagecount"]
                        _doc.userid = result['sys_doc_userid'] if result['sys_doc_userid'] is not None else 0
                        _doc.user = result['sys_user_username'] if _doc.userid != 0 else u"System"
                        _doc.title = result['sys_doc_docname'] if result['sys_doc_docname'] is not None else u''
                        _doc.date = result['sys_doc_docdate']
                        _doc.modified_date = result['sys_doc_modified']
                        _doc.modified_userid = result['sys_doc_modified_userid']
                        if 'expired' in _con.sys_doc.columns:
                            _doc.expired = result['sys_doc_expired']
                        if 'expired_setby' in _con.sys_doc.columns:
                            _doc.expired_setby = result['sys_doc_expired_setby']
                        _doc.in_progress = result['sys_doc_editmode']
                        _loc = result['sys_doc_oriloc'] if result['sys_doc_oriloc'] else result['sys_doc_docloc']
                        if _loc is not None:
                            _doc.is_revisionsave = _con.archive_rs.lower() in _loc.lower().split(os.path.sep)
                            # _doc.is_revisionsave = _loc.lower().find(_con.archive_rs.lower()) > -1
                        _doc.checkout_userid = result['sys_doc_checkout']
                        _doc.sha1 = result["sys_doc_sha1"]
                        _doc.is_checkout = _doc.checkout_userid > 0 if _doc.checkout_userid is not None else False
                        _doc.version = self.version

                        idx = 0
                        while 'docstate_%i' % idx in _con.sys_doc.columns:
                            _doc.states[idx] = result['sys_doc_docstate_%i' % idx] if result['sys_doc_docstate_%i' % idx] is not None else 0
                            idx += 1

                        if len(_doc.states) > 0:
                            _doc.in_resub = _doc.states[0] & 1 == 1

                        # sanity check: liegt das dokument auf wiedervorlage
                        resub = None
                        sql = 'select timid from sys_timer where status & 4 = 0 and gdocid = %d' % _doc.gdocid
                        try:
                            resub = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            resub = _con.try_again(sql=sql, db=_db, fetchall=True)

                        if resub:
                            _doc.in_resub = True
                            _doc.states[0] |= 1

                        sql = 'select arcid from sys_symlink where gdocid = %s' % _doc.gdocid
                        rows = _db.execute(sql).fetchall()
                        symlinks = []
                        if rows:
                            for row in rows:
                                symlinks.append(str(row[0]))
                            _doc.symlinks = ";".join(symlinks)

                        if _con.sys_abo is not None:
                            sql = 'select userid from sys_abo where gdocid = %s' % _doc.gdocid
                            rows = _db.execute(sql).fetchall()
                            abos = []
                            if rows:
                                for row in rows:
                                    abos.append(str(row[0]))
                                _doc.abo_userid = ";".join(abos)

                        _doc.fields = self.get_addfields(_con, _db, _doc, load_items='no_items' not in level,
                                                         orderbyname=orderfieldsbyname, with_gridcolumns=with_gridcolumns)

                    if 'paths' in level:
                        _doc.docloc = result['sys_doc_docloc'] if result['sys_doc_docloc'] is not None else ''
                        _doc.extloc = result['sys_doc_extloc'] if result['sys_doc_extloc'] is not None else ''
                        _doc.oriloc = result['sys_doc_oriloc'] if result['sys_doc_oriloc'] is not None else ''
                        _doc.srcloc = result['sys_doc_srcloc'] if result['sys_doc_srcloc'] is not None else ''
                        _doc.filename = result['sys_doc_filename'] if result['sys_doc_filename'] is not None else ''
                        _doc.filetype = result['sys_doc_filetype'] if result['sys_doc_filetype'] is not None else \
                            os.path.splitext(_doc.oriloc)[1] if _doc.oriloc else os.path.splitext(_doc.docloc)[1]

                        _doc.docloc = _doc.docloc.replace('\x00',
                                                          '')  # 0 bytes, wo immer die auch herkommen mögen, machen stress bei os.path.exists
                        _doc.extloc = _doc.extloc.replace('\x00', '')
                        _doc.oriloc = _doc.oriloc.replace('\x00', '')
                        _doc.srcloc = _doc.srcloc.replace('\x00', '')

                    # proxy.close()
                    proxy = None
                    result = None

                    if 'fields' in level:
                        if not _doc.fields and 'base' not in level:  # falls _doc.fields noch nicht geladen wurden
                            _doc.fields = self.get_addfields(_con, _db, _doc, load_items='no_items' not in level,
                                                             orderbyname=orderfieldsbyname,
                                                             with_gridcolumns=with_gridcolumns)
                        _col = ['sys_doc.gdocid']
                        _join = []
                        for field in _doc.fields.values():
                            if field.indextableid in _con.sys_docidx:
                                idx = _con.sys_docidx[field.indextableid]
                                if idx.columns.get(field.add) is not None:
                                    _col.append(field.add)
                                    if idx not in _join:
                                        _join.append(idx)
                            if field.type == 13 and field.values_query is not None:
                                if isinstance(field.values_query, bitfarm_doc.BitfarmGridConfig):
                                    grid_config = field.values_query
                                else:
                                    grid_config = bfaxml.xml_to_object(field.values_query)
                                if grid_config is not None:
                                    for column in grid_config.columns:
                                        if column.fieldname is not None:
                                            for idx in _con.sys_docidx.values():
                                                if idx.columns.get(column.fieldname) is not None:
                                                    _col.append(column.fieldname)
                                                    if idx not in _join:
                                                        _join.append(idx)

                        sql = 'select %s from sys_doc ' % ", ".join(_col)
                        for _j in _join:
                            sql += 'left join %s on sys_doc.gdocid = %s.gdocid ' % (_j, _j)
                        sql += 'where sys_doc.gdocid = %s' % _doc.gdocid

                        try:
                            result = _db.execute(sqla.text(sql)).fetchone()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchone=True)

                        if result:
                            for field in _doc.fields.values():
                                if field.add in result:
                                    if field.type == 13 and field.values_query:  # wir ergänzen bei einem Gridfeld die Addfields der einzelnen Spalten
                                        if isinstance(field.values_query, bitfarm_doc.BitfarmGridConfig):
                                            grid_config = field.values_query
                                        else:
                                            grid_config = bfaxml.xml_to_object(field.values_query)
                                        if grid_config is not None:
                                            values = {}
                                            for column in grid_config.columns:
                                                if column.fieldname is not None and column.fieldname in result:
                                                    values[column.fieldname] = result[column.fieldname].split(';') if result[column.fieldname] is not None else []
                                            keys = []
                                            maxval = 0
                                            for key, value in values.items():
                                                keys.append(key)
                                                if len(value) > maxval:
                                                    maxval = len(value)         # wieviele Reihen hat das Gridfeld maximal
                                            rows = [";".join(keys)]
                                            for i in range(maxval):
                                                row = []
                                                for key in keys:
                                                    if i < len(values[key]):
                                                        for column in grid_config.columns:
                                                            if column.fieldname is not None and column.fieldname == key and column.columntype == 1:     # sonderbehandlung datumsfelder
                                                                if values[key][i] and not re.match('^\d{4}-\d{2}-\d{2}$', values[key][i]):         # kein mysql Format? umwandeln
                                                                    try:
                                                                        date = datetime.datetime.strptime(values[key][i], '%d.%m.%Y')
                                                                        values[key][i] = datetime.datetime.strftime(date, '%Y-%m-%d')
                                                                    except:
                                                                        self.log.warning('Fehler bei der Wandlung eines Grid-Datumfeldes in MySQL-Format %s' % values[key][i])

                                                        row.append(values[key][i])
                                                    else:
                                                        row.append('')
                                                rows.append(";".join(row))

                                            field.value = "|".join(rows)
                                            field.oldvalue = field.value

                                            if self.gridfunc:  # aktualisiert die Gridwerte mit den berechneten Feldern
                                                self.log.debug("%s gridfunc before: %s" % (logid, field.value))
                                                self.gridfieldlock.acquire()
                                                try:
                                                    field.value = self.gridfunc(ctypes.c_wchar_p(_doc.profile),
                                                                                ctypes.c_int(field.id),
                                                                                ctypes.c_wchar_p(field.value))
                                                except:
                                                    self.log.exception("%s Fehler bei bfaGridField64.dll Aufruf" % logid)
                                                    field.value = field.oldvalue
                                                finally:
                                                    self.gridfieldlock.release()
                                                    self.log.debug("%s gridfunc after: %s" % (logid, field.value))

                                    elif result[field.add] is not None:
                                        field.oldvalue = result[field.add]
                                        field.value = result[field.add]

                        _doc.statefields = self.get_statefields(_con, _db, _doc)

                    if 'text' in level or 'ftx' in level:
                        sql = sqla.select(
                            [_con.sys_doctxt.c.doctxt, _con.sys_doctxt.c.doctags, _con.sys_doctxt.c.fieldtxt,
                             _con.sys_doctxt.c.annotxt],
                            whereclause=(_con.sys_doctxt.c.gdocid == _doc.gdocid))
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            if row[0] is not None and row[0] != u'':
                                # _doc.ftx.append(row[0])
                                if 'ftx' in level:
                                    _doc.ftx.text = row[0]
                                else:
                                    _doc.fulltxt = row[0]
                                    _doc.fulltxt = _doc.fulltxt.replace('\f',
                                                                        '\n')  # Delphi XML Komponente kann kein Formfeed parsen
                                    _doc.fulltxt = re.sub(r'[\x00-\x08\x0b-\x0c\x0e-\x1f\x7f-\x9f]', '',
                                                          _doc.fulltxt)  # ascii steuerzeichen parsen, außer tab und lf
                            if 'text' in level and row[1] is not None and row[1] != '':
                                _doc.doctags = row[1]
                                _doc.doctags = _doc.doctags.replace('\f', '\n')
                                _doc.doctags = re.sub(r'[\x00-\x08\x0b-\x1f\x7f-\x9f]', '',
                                                      _doc.doctags)  # ascii steuerzeichen parsen, außer tab und lf
                                for tag in _doc.doctags.split('\n'):
                                    if tag not in _doc.tags:
                                        _doc.tags.append(tag)
                            if 'text' in level and row[2] is not None and row[2] != '':
                                _doc.fieldtxt = row[2]
                                _doc.fieldtxt = _doc.fieldtxt.replace('\f', '\n')
                                _doc.fieldtxt = re.sub(r'[\x00-\x08\x0b-\x1f\x7f-\x9f]', '',
                                                       _doc.fieldtxt)  # ascii steuerzeichen parsen, außer tab und lf
                            if 'text' in level and row[3] is not None and row[3] != '':
                                _doc.annotxt = row[3]
                                _doc.annotxt = _doc.annotxt.replace('\f', '\n')
                                _doc.annotxt = re.sub(r'[\x00-\x08\x0b-\x1f\x7f-\x9f]', '',
                                                      _doc.annotxt)  # ascii steuerzeichen parsen, außer tab und lf

                    if 'links' in level:
                        sql = sqla.select([_con.sys_doclink.c.refname],
                                          whereclause=(_con.sys_doclink.c.gdocid == _doc.gdocid), distinct=True)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            if row[0] not in _doc.links:
                                _doc.links.append(row[0])

                    if 'iev' in level:
                        _doc.annotations = self.get_annotations(_doc.gdocid, _doc.profile)

                    if 'history' in level:
                        _doc.history = self.read_history(gdocid=_doc.gdocid, profile=_doc.profile)

                    if 'svn' in level and _con.sys_svn is not None:
                        sql = sqla.select([_con.sys_svn], whereclause=(_con.sys_svn.c.gdocid == _doc.gdocid)).order_by(
                            _con.sys_svn.c.svn_version)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            svnfield = bitfarm_doc.BitfarmSvnField()
                            svnfield.gdocid = _doc.gdocid
                            svnfield.svn_id = row[0]
                            svnfield.svn_link = row[2]
                            svnfield.svn_version = row[3]
                            svnfield.svn_revision = row[4]
                            svnfield.svn_info = row[6]
                            svnfield.svn_userid = row[7]
                            svnfield.svn_date = row[5]
                            _doc.svn_history.append(svnfield)

                    if 'sap' in level and _con.sys_sap is not None:
                        sql = sqla.select([_con.sys_sap], whereclause=(_con.sys_sap.c.gdocid == _doc.gdocid)).order_by(
                            _con.sys_sap.c.sapid)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            sapfield = bitfarm_doc.BitfarmSAPField()
                            sapfield.gdocid = _doc.gdocid
                            sapfield.sap_id = row[0]
                            sapfield.sap_contrep = row[2]
                            sapfield.sap_docid = row[3]
                            sapfield.sap_compid = row[4]
                            sapfield.sap_docprot = row[5]
                            sapfield.sap_docstatus = row[6]
                            sapfield.sap_tstampc = row[7]
                            sapfield.sap_tstampm = row[8]
                            sapfield.sap_content_path = row[9]
                            sapfield.sap_content_type = row[10]
                            sapfield.sap_charset = row[11]
                            sapfield.sap_version = row[12]

                            _doc.sap.append(sapfield)

                    if 'workflow' in level:
                        sql = sqla.select([_con.sys_timer],
                                          whereclause=(sqla.and_(_con.sys_timer.c.gdocid == _doc.gdocid,
                                                                 _con.sys_timer.c.status.op('&')(4) == 0)))
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        for row in result:
                            wf = bitfarm_doc.BitfarmWorkflow()
                            wf.tim_id = row[0]
                            wf.tim_ref = row[1]
                            wf.gdocid = row[2]
                            wf.groupid = row[3]
                            wf.userid_work = row[4]
                            wf.userid_shed = row[5]
                            wf.status = row[6]
                            wf.date = row[7]
                            wf.alarm_date = row[8]
                            wf.options = row[9]
                            wf.info = row[10]
                            wf.comment = row[11]
                            wf.wf_id = row[12]
                            wf.required = row[13]
                            wf.hist_id = row[14]
                            wf.modified_date = row[15]
                            _doc.workflows.append(wf)

                    if 'commentary' in level:
                        sql = 'select * from sys_com where gdocid = %s order by tstamp desc' % _doc.gdocid
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception(u'%s konnte Kommentare nicht lesen [%s]' % (logid, _doc.gdocid))
                            result = _con.try_again(sql=sql, db=_db, fetchall=True)

                        if result:
                            for row in result:
                                com = bitfarm_doc.BitfarmCommentary()
                                com.id = row[0]
                                com.gdocid = _doc.gdocid
                                com.userid = row[2]
                                com.tstamp = row[3]
                                com.comment = row[4]
                                _doc.commentary.append(com)
                else:
                    self.log.warning(u'%s kein Dokument mit der ID [%s] gefunden' % (logid, _doc.gdocid))
                    return None
            except:
                self.log.exception(u'%s Das Dokument [%s] konnte nicht geladen werden' % (logid, _doc.gdocid))
                raise LoadDocumentFromDatabaseError(_doc.gdocid)

            # self.log.debug(u"%s laden eines kompletten Dokuments %s" % (logid, time.perf_counter() - s2))
            result = None

            return True
        else:
            return None

    def __load_v4(self, _doc, _gdocid, level, orderfieldsbyname, logid=''):
        return None

    def get_ineditmode(self, _doc, _gdocid, logid=''):
        # wir überprüfen, ob eine gdocid übergeben wurde, oder wenn nicht, ob das doc eine hat
        if _gdocid:
            if type(_gdocid) is int:
                _doc.gdocid = _gdocid
            if type(_gdocid) is str:
                _doc.gdocid = int(_gdocid)
        else:
            if _doc.gdocid is None:
                # TODO: wäre es angebracht eine exception zu werfen
                return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

        sql = sqla.select([_con.sys_doc.c.gdocid],
                          whereclause=(_con.sys_doc.c.ineditmode != None), use_labels=True)
        try:
            result = _db.execute(sql).fetchall()
        except:
            self.log.exception(u"Fehler bei der Datenbankabfrage %s" % str(sql))
        ineditmode = []
        for row in result:
            ineditmode.append(str(row[0]))
        _doc.in_editmode = ",".join(ineditmode)

    def write(self, _doc, logid=''):
        """
        Schreibt das Dokument zurück in die Datenbank.
        Die Methode reicht den Aufruf an die privaten Methoden __load_v3 und __load_v4
        weiter, entsprechend der Version self.version.
        Wenn das übergebene Dokument keinem Profil zugeodnet ist, wird es dem Profil der Main.Connection zugeordnet.
        Der Instanz muss eine db-connection geöffnet haben (self.con.db)
        """
        if self.version in (3, 35, '3.0'):
            return self.__write_v3(_doc, logid=logid)
        elif self.version == 36:
            return self.__write_v36(_doc, logid=logid)
        elif self.version == 40:
            return self.__write_v4(_doc, logid=logid)
        else:
            return None

    def __write_v3(self, _doc, logid=''):
        # es kann nur ein vorhandenes Doc in die Datenbank zurückgeschrieben werden
        # ein neues Doc muss mit create erzeugt werden, dabei wird dann eine gdocid erzeugt
        if not _doc:
            return None

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if not _doc.gdocid:
            return None

        # neuen Status aus Statusdict und in_progress ermitteln
        sql = 'select * from sts_%s order by sts_id' % _doc.tblname
        try:
            result = _db.execute(sql).fetchall()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchall=True)
        _doc.status = 0
        _doc.status |= (1 if _doc.in_progress is True else 0)
        _ind = 1
        for row in result:
            if row[1] in _doc.statusdict:
                if _doc.statusdict[row[1]]:
                    _doc.status |= int(2 ** _ind)
            _ind += 1

        if _con.arc_tbl is None or _con.arc_tbl.fullname != _doc.tblname:
            s = time.perf_counter()
            _con.arc_tbl = sqla.Table("arc_" + _doc.tblname, _con.meta, autoload=True)
            self.log.debug("%s sqla.table autoload %s" % (logid, time.perf_counter() - s))

        # arcid und docid darf nicht verändert werden, soll ein doc in ein anderes archiv umziehen muss doc.move(arcid) verwendet werden
        _ftx = u''

        if type(_doc.ftx) is str:
            _ftx = _doc.ftx
        elif type(_doc.ftx) == bitfarm_doc.BitfarmFulltext:
            _ftx = _doc.ftx.text
            if (_doc.ftx.encoding != _con.db_encoding) and _ftx != '':
                _ftx = _ftx.encode(_con.db_encoding, "ignore")
                _ftx = _ftx.decode(_con.db_encoding, "ignore")
                # _ftx = _ftx.decode(_doc.ftx.encoding, "ignore")
                # _ftx = _ftx.encode(_con.db_encoding, "ignore")

        decoded_title = _doc.title
        decoded_title = decoded_title.encode(_con.db_encoding, "ignore")
        decoded_title = decoded_title.decode(_con.db_encoding, "ignore")
        # user abgleichen
        sql = ''
        if _doc.user != '' and _doc.userid is None:
            sql = "select * from sys_usr where usr_name = '" + _doc.user + "' and flags like'%%enabled%%'"
        elif _doc.user is None and _doc.userid is not None:
            sql = "select * from sys_usr where usr_id = " + str(_doc.userid) + " and flags like'%%enabled%%'"
        if sql:
            try:
                result = _db.execute(sql).fetchone()
            except:
                result = _con.try_again(sql=sql, db=_db, fetchone=True)
            if result:
                _doc.userid = result[0]
                _doc.user = result[1]

        vals = {_con.arc_tbl.c.USR_NAME: _doc.user,
                _con.arc_tbl.c.DOC_LOC: _doc.docloc,
                _con.arc_tbl.c.EXT_LOC: _doc.extloc,
                _con.arc_tbl.c.ORI_LOC: _doc.oriloc,
                _con.arc_tbl.c.DISP_NAME: decoded_title,
                _con.arc_tbl.c.USR_ID: _doc.userid,
                _con.arc_tbl.c.DOC_TXT: _ftx,
                _con.arc_tbl.c.DOC_STATE: _doc.status,
                _con.arc_tbl.c.SHA1: _doc.sha1
                }

        if _doc.arcdate is not None:
            vals[_con.arc_tbl.c.TStamp] = _doc.arcdate

        if len(_doc.fields) > 0:
            self.parse_addfields(_doc)

        for key in _doc.fields.keys():
            self.log.debug("%s %s %s %s " % (logid, key, _doc.fields[key].value, _doc.fields[key].add))
            if _doc.fields[key].add is None:
                continue
                # TODO: was machen wir mit temporäre Zusatzfeldern? in die Schlagworte?
            if _doc.fields[key].value:
                vals[_con.arc_tbl.columns.get(_doc.fields[key].add)] = (
                    ';'.join(_doc.fields[key].value) if type(_doc.fields[key].value) is list else _doc.fields[
                        key].value)

        sql = _con.arc_tbl.update(whereclause=(_con.arc_tbl.c.DOC_ID == _doc._docid), values=vals)
        #        sql = _con.arc_tbl.insert().values(dict
        self.log.debug("%s %s " % (logid, sql))
        # sql.execute()
        try:
            _db.execute(sql)
        except:
            _con.try_again(sql=sql, db=_db)
        # Verknüpfungen
        sql = 'select distinct(ref_name) from sys_doclnk where arc_id=%s and doc_id=%s' % (_doc.arcid, _doc._docid)
        vals = []
        try:
            result = _db.execute(sql).fetchall()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchall=True)
        for link in _doc.links:
            link = self.parse_sqlstring(link)
            link = link.encode(_con.db_encoding, "ignore")
            link = link.decode(_con.db_encoding, "ignore")
            if link not in [x[0] for x in result]:
                vals.append('(%s,%s,"%s")' % (_doc.arcid, _doc._docid, link))

        if len(vals) > 0:
            sql = 'Insert into sys_doclnk (arc_id, doc_id, ref_name) values %s' % (",".join(vals))
            try:
                _db.execute(sql)
            except:
                _con.try_again(sql=sql, db=_db, )
            self.log.debug("%s %s " % (logid, sql))

        # Schlagwörter
        sql = 'select DOC_SLW from slw_%s  where doc_id = %s' % (_doc.tblname, _doc._docid)
        vals = []
        try:
            result = _db.execute(sql).fetchall()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchall=True)
        for tag in _doc.tags:
            if tag not in [x[0] for x in result]:
                vals.append("(%s,'%s')" % (_doc._docid, tag))

        if len(vals) > 0:
            sql = 'Insert into slw_%s  (doc_id, DOC_SLW) values %s' % (_doc.tblname, ",".join(vals))
            try:
                _db.execute(sql)
            except:
                _con.try_again(sql=sql, db=_db)
            self.log.debug("%s %s " % (logid, sql))
        return True

    def __write_v36(self, _doc, logid=''):
        # es kann nur ein vorhandenes Doc in die Datenbank zurückgeschrieben werden
        # ein neues Doc muss mit create erzeugt werden, dabei wird dann eine gdocid erzeugt
        _ret = True
        if not _doc:
            return False

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return False

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        sql = 'select gdocid from sys_doc where gdocid = %d' % _doc.gdocid
        try:
            result = _db.execute(sql).fetchone()
        except:
            self.log.exception("%s Fehler bei der Abfrage nach schon vorhandenem Dokument [%d]" % (logid, _doc.gdocid))
            return False
        else:
            if not result:
                self.log.error("%s Das Dokument [%d] ist nicht in der Datenbank vorhanden, aktualisieren nicht möglich" % (logid, _doc.gdocid))
                return False

        # neuen Status aus Status Dictionary ermitteln
        _doc.status = {}
        for state in _doc.statefields.items():
            if state[1].value:
                if state[1].indexcolumnid in _doc.status:
                    _doc.status[state[1].indexcolumnid] |= state[1].bitmask
                else:
                    _doc.status[state[1].indexcolumnid] = state[1].bitmask
            else:
                if state[1].indexcolumnid in _doc.status:
                    _doc.status[state[1].indexcolumnid] &= ~state[1].bitmask
                else:
                    _doc.status[state[1].indexcolumnid] = 0

        decoded_title = _doc.title
        decoded_title = decoded_title.encode(_con.db_encoding, "ignore")
        decoded_title = decoded_title.decode(_con.db_encoding, "ignore")
        # user abgleichen
        s = time.perf_counter()
        sql = ''
        if _doc.user != '' and (_doc.userid is None or _doc.userid == 0):
            sql = "select userid, username, userpass, usermail, flags +0, displayname, role from sys_user where username = '" + _doc.user + "' and (flags & 1) = 1"
        elif _doc.user is None and _doc.userid is not None:
            sql = "select userid, username, userpass, usermail, flags +0, displayname, role from sys_user where userid = " + str(_doc.userid) + " and (flags & 1) = 1"
        if sql:
            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception("%s Konnte Benutzer nicht finden" % logid)

            if result:
                _doc.userid = result[0]
                _doc.user = result[1]

            # self.log.debug("%s select from sys_user %s" % (logid, time.perf_counter() - s))

        vals = {}
        # vals = {# _con.sys_doc.c.userid: _doc.userid,    wir ändern nicht mehr den initial eingetragenen user
        #         _con.sys_doc.c.arcid: _doc.arcid,
        #         _con.sys_doc.c.docloc: _doc.docloc,
        #         _con.sys_doc.c.extloc: _doc.extloc,
        #         _con.sys_doc.c.oriloc: _doc.oriloc,
        #         _con.sys_doc.c.srcloc: _doc.srcloc,
        #         _con.sys_doc.c.docname: decoded_title,
        #         _con.sys_doc.c.pagecount: _doc.pages,
        #         _con.sys_doc.c.modified_userid: _doc.userid,
        #         _con.sys_doc.c.checkout: _doc.checkout_userid,
        #         _con.sys_doc.c.filetype: _doc.filetype,
        #         _con.sys_doc.c.filename: _doc.filename,
        #         _con.sys_doc.c.sha1: _doc.sha1
        #         }
        if not _doc.expired_modified:  # wurde das expired_date geändert, müssen diese sachen nicht gespeichert werden
            vals[_con.sys_doc.c.docname] = decoded_title if decoded_title else u''
            vals[
                _con.sys_doc.c.extloc] = _doc.extloc if _doc.extloc else None  # extloc kann auch wieder entfernt werden

        if _doc.userid:
            vals[_con.sys_doc.c.userid] = _doc.userid
        if _doc.arcid:
            vals[_con.sys_doc.c.arcid] = _doc.arcid
        if _doc.docloc:
            vals[_con.sys_doc.c.docloc] = _doc.docloc
        if _doc.oriloc:
            vals[_con.sys_doc.c.oriloc] = _doc.oriloc
        if _doc.srcloc:
            vals[_con.sys_doc.c.srcloc] = _doc.srcloc
        if _doc.pages:
            vals[_con.sys_doc.c.pagecount] = _doc.pages
        if _doc.checkout_userid:
            vals[_con.sys_doc.c.checkout] = _doc.checkout_userid
        if _doc.filetype:
            vals[_con.sys_doc.c.filetype] = _doc.filetype
        if _doc.filename:
            vals[_con.sys_doc.c.filename] = _doc.filename
        if _doc.sha1:
            vals[_con.sys_doc.c.sha1] = _doc.sha1
        if _doc.expired_modified:
            vals[_con.sys_doc.c.expired] = _doc.expired
            vals[_con.sys_doc.c.expired_userid] = _doc.modified_userid
            vals[_con.sys_doc.c.expired_setby] = _doc.expired_setby

        for state in _doc.status.items():
            vals[_con.sys_doc.c['docstate_%i' % (state[0])]] = state[1]

        if len(vals) > 0:
            if not _doc.modified_userid:
                _doc.modified_userid = 1  # System
            vals[_con.sys_doc.c.modified_userid] = _doc.modified_userid
            #  vals[_con.sys_doc.c.modified] = None     # das funktioniert mit Mysql57 aber nicht mit MySql8. Wenn man es auskommentiert funktioniert es mit beiden

            sql = _con.sys_doc.update(whereclause=(_con.sys_doc.c.gdocid == _doc.gdocid), values=vals)
            # self.log.debug("%s %s " % (logid, sql))

            try:
                _db.execute(sql)
            except:
                _ret = False
                self.log.exception("%s Fehler bei Update sys_doc %s" % (logid, time.perf_counter() - s))
            # else:
            #     self.log.debug("%s Update sys_doc %s" % (logid, time.perf_counter() - s))

        #        if _doc.gdocid is None:
        #            _doc.gdocid = result.inserted_primary_key[0]
        self.log.debug('Fields count %d' % len(_doc.fields))
        if len(_doc.fields) > 0:
            self.parse_addfields(_doc)

        saved = []  # eine Referenz auf alle gespeicherten Felder um nach erfolgreicher Speicherung modified zurück zu setzen

        for field in _doc.fields.values():
            self.log.debug('Field: %s Value: %s Type: %s Modified: %s' % (field.add, field.value, str(type(field.value)), str(field.modified)))

            if field.modified:
                if self.save_addfield(field, _doc.gdocid, _doc.profile):
                    field.modified = False
                    if len(field.history) > 0:
                        self.write_history(history=field.history, profile=_doc.profile, gdocid=_doc.gdocid)
                        field.history = []
                else:
                    self.log.error("%s Das Zusatzfeld %s konnte nicht gespeichert werden" % (logid, field.name))
                    _ret = False

        # for idx in _con.sys_docidx.values():
        #     vals = {}
        #     for field in _doc.fields.values():
        #         if field.modified and field.add in idx.columns.keys():
        #             if type(field.value) is datetime.datetime:
        #                 _value = field.value.isoformat(" ").split(".")[0]
        #                 field.clear()
        #                 field.value = _value
        #             vals[idx.columns.get(field.add)] = (';'.join(field.value) if type(field.value) is list else field.value)
        #             saved.append(field)
        #
        #     if len(vals) > 0:
        #         sql = sqla.select([idx.c.gdocid], whereclause=(idx.c.gdocid == _doc.gdocid))
        #
        #         try:
        #             result = _db.execute(sql).fetchone()
        #         except:
        #             _ret = False
        #             self.log.exception(u'%s [%d] Fehler beim Suchen eines vorhandenen Eintrages' % (logid, _doc.gdocid))
        #         else:
        #             if result:
        #                 sql = idx.update(whereclause=(idx.c.gdocid == _doc.gdocid), values=vals)
        #             else:
        #                 vals[idx.c.gdocid] = _doc.gdocid
        #                 sql = idx.insert(values=vals)
        #             # self.log.debug("%s %s " % (logid, sql))
        #
        #             try:
        #                 _db.execute(sql)
        #             except:
        #                 _ret = False
        #                 self.log.exception(u'%s [%d] Fehler: Zusatzfelder konnten nicht gespeichert werden' % (logid, _doc.gdocid))
        #             else:
        #                 for save in saved:
        #                     save.modified = False
        #                     if len(save.history) > 0:
        #                         self.write_history(history=save.history, profile=_doc.profile, gdocid=_doc.gdocid)
        #                         save.history = []
        #             # TODO: Verknüpfungsfeld einbauen
        # self.log.debug("%s update sys_docidx %s" % (logid, time.perf_counter() - s))

        # es sind immer noch modified Felder übrig, die nicht gespeichert werden konnten?, dann in die Schlagworte
        for field in _doc.fields.values():
            if field.modified:
                _doc.tags.append('%s=%s' % (field.name, field.value))

                # Verknüpfungen
            #        sql = 'select distinct(refname) from sys_doclink where gdocid=%s' %(_doc.gdocid)
        sql = sqla.select([_con.sys_doclink.c.refname],
                          whereclause=(_con.sys_doclink.c.gdocid == _doc.gdocid)).distinct()
        vals = []
        try:
            result = _db.execute(sql).fetchall()
        except:
            _ret = False
            self.log.exception("%s Konnte Doc-Links nicht abfragen" % logid)
        # else:
        #     self.log.debug("%s select sys_doclink %s" % (logid, time.perf_counter() - s))

        for link in _doc.links:
            # link = self.parse_sqlstring(link)
            link = link.encode(_con.db_encoding, "ignore")
            link = link.decode(_con.db_encoding, "ignore")
            if link not in [x[0] for x in result]:
                vals.append({'gdocid': _doc.gdocid, 'refname': link})
        # self.log.debug("%s select sys_doclink %s" % (logid, time.perf_counter() - s))

        if len(vals) > 0:
            #            sql = 'Insert into sys_doclink (gdocid, refname) values %s' %(",".join(vals))
            sql = sqla.insert(_con.sys_doclink, values=vals)
            try:
                _db.execute(sql)
            except:
                _ret = False
                self.log.exception("%s Konnte Doc-Links nicht speichern" % logid)
            # else:
            #     self.log.debug("%s insert sys_doclink %s" % (logid, time.perf_counter() - s))

        # Volltext, Schlagwörter
        vals = {}
        _ftx = ''
        if type(_doc.ftx) is str:
            _ftx = _doc.ftx
        elif type(_doc.ftx) is bitfarm_doc.BitfarmFulltext:
            _ftx = _doc.ftx.text
            if (_doc.ftx.encoding != _con.db_encoding) and _ftx != '':
                _ftx = _ftx.encode(_con.db_encoding, "ignore")
                _ftx = _ftx.decode(_con.db_encoding, "ignore")

        _tags = ''
        if len(_doc.tags) > 0:
            _tags = '\n'.join(_doc.tags)
            _tags = _tags.encode(_con.db_encoding, "ignore")
            _tags = _tags.decode(_con.db_encoding, "ignore")

        if _ftx:
            vals[_con.sys_doctxt.c.doctxt] = _ftx
        if _tags:
            vals[_con.sys_doctxt.c.doctags] = _tags

        _fieldtxt = ''
        _fieldvals = []
        for field in _doc.fields.values():
            if field.value:
                if field.type == 13:
                    _rows = field.value.split('|')
                    if len(_rows) > 1:
                        for idx in range(1, len(_rows)):
                            if _rows[idx]:
                                for _val in _rows[idx].split(';'):
                                    if _val and _val not in _fieldvals:
                                        _fieldvals.append(_val)
                    continue
                if type(field.value) in [list, tuple]:
                    _fieldvals.extend(field.value)
                elif type(field.value) in [int, float, decimal.Decimal]:
                    _fieldvals.append(str(field.value))
                elif type(field.value) in [datetime.datetime]:
                    _fieldvals.append(field.value.isoformat(" ").split(".")[0])
                else:
                    _fieldvals.append(field.value)
        if len(_fieldvals) > 0:
            _fieldtxt = '\n'.join(_fieldvals)

        if _fieldtxt:
            vals[_con.sys_doctxt.c.fieldtxt] = _fieldtxt

        _annotxt = _doc.annotxt
        if _annotxt:
            vals[_con.sys_doctxt.c.annotxt] = _annotxt

        if vals:
            result = False
            if _con.sys_doctxt.fullname == 'sys_doctxt':    # Fallback für 3.6.0  (wahrscheinlich obsolet, ggf bei Migration 3.5.4 > 3.6.0 nötig)
                sql = sqla.select([_con.sys_doctxt.c.gdocid], whereclause=(_con.sys_doctxt.c.gdocid == _doc.gdocid))

                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception('%s [%d] Fehler beim Suchen eines vorhandenen Eintrages' % (logid, _doc.gdocid))
            else:
                result = True

            if result:
                sql = _con.sys_doctxt.update(whereclause=(_con.sys_doctxt.c.gdocid == _doc.gdocid), values=vals)
            else:
                vals[_con.sys_doctxt.c.gdocid] = _doc.gdocid
                sql = _con.sys_doctxt.insert(values=vals)

            try:
                _db.execute(sql)
            except:
                self.log.exception("%s Volltext konnte nicht gespeichert werden" % logid)
                _ret = False
                # else:
                #     self.log.debug(u"%s update sys_doctxt %s" % (logid, time.perf_counter() - s))

        for anno in _doc.annotations:
            if type(anno) is bitfarm_doc.BitfarmAnnotation and anno.modified:
                if not self.save_annotation(anno, _doc.gdocid, _doc.profile, logid):
                    _ret = False
                    self.log.error("%s Annotation konnte nicht gespeichert werden. Profil: %s" % (logid, _doc.profile))

        if len(_doc.history) > 0:
            self.write_history(history=_doc.history, profile=_doc.profile, gdocid=_doc.gdocid)
            _doc.history = []
        # self.log.debug(u"%s write_history %s" % (logid, time.perf_counter() - s))

        for svn in _doc.svn_history:
            if svn.svn_id:
                sql = sqla.select([_con.sys_svn], whereclause=(_con.sys_svn.c.svnid == svn.svn_id))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception('%s Fehler beim Suchen eines vorhandenen SVN-Eintrages [%s]' % (logid, svn.svn_id))
            else:
                result = None

            vals = {_con.sys_svn.c.gdocid: _doc.gdocid,
                    _con.sys_svn.c.svn_link: svn.svn_link,
                    _con.sys_svn.c.svn_version: svn.svn_version,
                    _con.sys_svn.c.svn_revision: svn.svn_revision,
                    _con.sys_svn.c.svn_info: svn.svn_info,
                    _con.sys_svn.c.svn_userid: svn.svn_userid,
                    _con.sys_svn.c.svn_date: svn.svn_date
                    }
            if result:
                sql = _con.sys_svn.update(whereclause=(_con.sys_svn.c.svnid == svn.svn_id), values=vals)
            else:
                if svn.svn_id:
                    vals[_con.sys_svn.c.svnid] = svn.svn_id
                sql = _con.sys_svn.insert(values=vals)
            try:
                res = _db.execute(sql)
            except:
                self.log.exception(u'%s Fehler beim Schreiben eines SVN-Eintrages [%s]' % (logid, svn.svn_id))
                _ret = False
            else:
                if not result and not svn.svn_id:
                    svn.svn_id = res.inserted_primary_key[0]

        for sap in _doc.sap:
            if sap.sap_id:
                sql = sqla.select([_con.sys_sap], whereclause=(_con.sys_sap.c.sapid == sap.sap_id))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception(
                        u'%s Fehler beim Suchen eines vorhandenen SAP-Eintrages [%s]' % (logid, sap.sap_id))
            else:
                result = None
            vals = {_con.sys_sap.c.gdocid: _doc.gdocid,
                    _con.sys_sap.c.contrep: sap.sap_contrep,
                    _con.sys_sap.c.docid: sap.sap_docid,
                    _con.sys_sap.c.compid: sap.sap_compid,
                    _con.sys_sap.c.docprot: sap.sap_docprot,
                    _con.sys_sap.c.docstatus: sap.sap_docstatus,
                    _con.sys_sap.c.tstampc: sap.sap_tstampc,
                    _con.sys_sap.c.tstampm: sap.sap_tstampm,
                    _con.sys_sap.c.content_path: sap.sap_content_path,
                    _con.sys_sap.c.content_type: sap.sap_content_type,
                    _con.sys_sap.c.charset: sap.sap_charset,
                    _con.sys_sap.c.version: sap.sap_version
                    }
            if result:
                sql = _con.sys_sap.update(whereclause=(_con.sys_sap.c.sapid == sap.sap_id), values=vals)
            else:
                if sap.sap_id:
                    vals[_con.sys_sap.c.sapid] = sap.sap_id
                sql = _con.sys_sap.insert(values=vals)
            try:
                res = _db.execute(sql)
            except:
                self.log.exception(u'%s Fehler beim Schreiben eines SAP-Eintrages [%s]' % (logid, sap.sap_id))
                _ret = False
            else:
                if not result and not sap.sap_id:
                    del sap.sap_id
                    sap.sap_id = res.inserted_primary_key[0]

        for com in _doc.commentary:
            if com.id:
                sql = sqla.select([_con.sys_com], whereclause=(_con.sys_com.c.commentid == com.id))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception(
                        u'%s Fehler beim Suchen eines vorhandenen Kommentar-Eintrages [%s]' % (logid, com.id))
            else:
                result = None
            vals = {_con.sys_com.c.gdocid: _doc.gdocid,
                    _con.sys_com.c.userid: com.userid,
                    _con.sys_com.c.tstamp: com.tstamp,
                    _con.sys_com.c.comment: com.comment
                    }
            if result:
                sql = _con.sys_com.update(whereclause=(_con.sys_com.c.commentid == com.id), values=vals)
            else:
                if com.id:
                    vals[_con.sys_com.c.commentid] = com.id
                sql = _con.sys_com.insert(values=vals)
            try:
                res = _db.execute(sql)
            except:
                self.log.exception(u'%s Fehler beim Schreiben eines Kommentar-Eintrages [%s]' % (logid, com.id))
                _ret = False
            else:
                if not result and not com.id:
                    com.id = res.inserted_primary_key[0]

        return _ret

    def __write_v4(self, _doc, logid=''):
        return None

    def create(self, _doc, arcid=None, logid=''):
        """
        Erstellt ein neues Dokument in der Datenbank.
        Wenn das übergebene Dokument keinem Profil zugeodnet ist, wird es dem Profil der Main.Connection zugeordnet.
        Optional kann die _arcid übergeben werden, in der das Dokument erstellt werden soll,
        ansonsten wird versucht in 'unverteilt' das Dokument zu erstellen.
        Wird nichts übergeben wird versucht self.con.undistributed auszulesen, ansonsten wird nach einem
        Archiv 'unverteilt' gesucht.
        @param _doc: Dokument als bitfarmDocument.
        @param arcid: (optional) arcid des Archivs in dem das Dokument erstellt werden soll.
        """
        if self.version in (3, 35, '3.0'):
            return self.__create_v3(_doc, arcid, logid=logid)
        elif self.version == 36:
            return self.__create_v36(_doc, arcid, logid=logid)
        elif self.version == 40:
            return self.__create_v4(_doc, arcid, logid=logid)
        else:
            return None

    def __create_v3(self, _doc, _arcid=None, logid=''):
        # erstellt einen neuen Dokumenten-Eintrag in der Datenbank
        # es kann schon eine arcid übergeben werden, ansonsten wird in unverteilt angelegt
        # TODO: das Doc-Objekt kann bereits ein Archiv-Property haben, warum nicht das verwenden?

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if _con.sys_gdocid is None:
            _con.sys_gdocid = sqla.Table('sys_gdocid', _con.meta, autoload=True)

        if not _arcid:
            if _con.undistributed_arcid:
                _arcid = int(_con.undistributed_arcid)
            else:
                if _con.undistributed:
                    try:
                        sql = _con.sys_arc.select(_con.sys_arc.c.ARC_Name == _con.undistributed)
                        try:
                            result = _db.execute(sql).fetchone()
                        except Exception as e:
                            self.log.exception(u'%s error %s in _doc.create' % (logid, str(e)))
                            result = _con.try_again(sql=sql, db=_db, fetchone=True)
                        self.log.debug("%s %s " % (logid, result))
                        if result:
                            _arcid = result[0]
                    except:
                        _arcid = None

        if type(_arcid) is int:
            try:
                vals = {}
                ins = _con.sys_gdocid.insert()
                #                if _doc.gdocid is not None:
                #                    vals[_con.sys_gdocid.c.gdoc_id] = _doc.gdocid
                try:
                    if _doc.gdocid is not None:
                        result = _db.execute(ins, arc_id=_arcid, gdoc_id=_doc.gdocid)
                    else:
                        result = _db.execute(ins, arc_id=_arcid)
                except:
                    self.log.exception(u'%s error in _doc.create' % logid)
                    return None
                else:
                    _doc.gdocid = result.inserted_primary_key[0]

                if _doc.gdocid > 0:
                    _doc.encoding = _con.db_encoding
                    self.arcid = _arcid
                    _doc.arcid = _arcid
                    sql = sqla.select([_con.sys_arc], whereclause=(_con.sys_arc.c.ARC_ID == self.arcid))

                    try:
                        result = _db.execute(sql).fetchone()
                    except:
                        self.log.exception(u'%s error in _doc.create' % logid)
                        result = _con.try_again(sql=sql, db=_db, fetchone=True)

                    self.tblname = result[3]
                    self.archive = result[2]
                    _doc.tblname = result[3]
                    _doc.archive = result[2]
                    s = time.perf_counter()
                    if _con.arc_tbl is None:
                        _con.arc_tbl = sqla.Table('arc_%s' % self.tblname, _con.meta, autoload=True)
                    elif _con.arc_tbl.name.find(_doc.tblname) == -1:
                        _con.arc_tbl = sqla.Table('arc_%s' % self.tblname, _con.meta, autoload=True)
                    self.log.debug("%s sqla.table autoload %s arc_%s" % (logid, time.perf_counter() - s, self.tblname))
                    if _doc.user == '':
                        _doc.user = u'System'
                    # wir lassen den Titel leer und setzen ihn erst beim Archivieren auf 'Dokument' falls in der Zwischenzeit nichts neues dazugekommen ist
                    # if _doc.title == '':
                    #   _doc.title = u'Dokument'
                    if _doc.date is None:
                        _doc.date = datetime.datetime.now().replace(microsecond=0)
                    vals = {_con.arc_tbl.c.USR_NAME: _doc.user,
                            _con.arc_tbl.c.TStamp: _doc.date,
                            _con.arc_tbl.c.gdoc_id: _doc.gdocid,
                            # _con.arc_tbl.c.DOC_LOC: self.docloc,
                            # _con.arc_tbl.c.EXT_LOC: self.extloc,
                            # _con.arc_tbl.c.ORI_LOC: self.oriloc,
                            _con.arc_tbl.c.DISP_NAME: _doc.title
                            # _con.arc_tbl.c.USR_ID: self.userid,
                            # _con.arc_tbl.c.DOC_TXT: self.ftx,
                            # _con.arc_tbl.c.DOC_STATE: self.status
                            }

                    sql = _con.arc_tbl.insert(values=vals)
                    #        sql = _con.arc_tbl.insert().values(dict
                    self.log.debug("%s %s" % (logid, sql))
                    try:
                        result = _db.execute(sql)
                    except:
                        self.log.exception(u'%s error in _doc.create' % logid)
                        result = _con.try_again(sql=sql, db=_db)

                    _doc._docid = result.inserted_primary_key[0]

                    if _doc._docid > 0:
                        sql = _con.sys_gdocid.update().where(_con.sys_gdocid.c.gdoc_id == _doc.gdocid).values(
                            doc_id=_doc._docid)
                        try:
                            _db.execute(sql)
                        except:
                            self.log.exception(u'%s error in _doc.create' % logid)
                            _con.try_again(sql=sql, db=_db)

                    # neues Dokument hat keine Schlagwörter oder Verknüpfungen, aber Zusatzfelder und Statusfelder !
                    if len(_doc.fields) == 0:  # wenn nicht schon vorhanden
                        _doc.fields = self.get_addfields(_con, _db, _doc)
                    if len(_doc.statusdict) == 0:
                        _doc.statusdict = self.get_statefields(_con, _db, _doc)
                    #                    _doc.version = 3    # sollte schon im Object beim __init__ gesetzt worden sein
                    if _doc.profile is None:
                        _doc.profile = _con.profile
                else:
                    _doc.gdocid = None
                    _doc.arcid = None
            except:
                self.log.exception(u'%s error in _doc.create' % logid)
                _doc.gdocid = None
                _doc.arcid = None
            else:
                self.log.debug("%s neue gdocid [%s]" % (logid, _doc.gdocid))
                return True
        else:
            _doc.gdocid = None
            _doc.arcid = None

        if not _doc.arcid:
            raise ArcidNotFoundError(_doc, 'Konnte Arcid von %s nicht ermitteln' % _con.undistributed)

        if not _doc.gdocid:
            raise CreateDocError(_doc)


    def __create_v36(self, _doc, _arcid=None, logid=''):
        # erstellt einen neuen Dokumenten-Eintrag in der Datenbank
        # es kann schon eine arcid übergeben werden, ansonsten wird in unverteilt angelegt
        # TODO: das Doc-Objekt kann bereits ein Archiv-Property haben, warum nicht das verwenden?

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
            _doc.profile = _con.profile
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if not _arcid:
            if _con.undistributed_arcid:
                _arcid = int(_con.undistributed_arcid)
            else:
                if _con.undistributed:
                    try:
                        sql = _con.sys_arc.select(_con.sys_arc.c.arcname == _con.undistributed)
                        try:
                            result = _db.execute(sql).fetchone()
                        except:
                            self.log.exception(u'%s error in _doc.create' % logid)
                            result = _con.try_again(sql=sql, db=_db, fetchone=True)

                        self.log.debug(result)
                        if result:
                            _arcid = result[0]
                            _con.undistributed_arcid = _arcid
                    except:
                        _arcid = None

        if type(_arcid) is str:
            _arcid = int(_arcid)

        if type(_arcid) is int:
            try:
                _doc.version = 36
                _doc.encoding = _con.db_encoding

                sql = sqla.select([_con.sys_arc], whereclause=(_con.sys_arc.c.arcid == _arcid))

                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception(u'%s error in _doc.create' % logid)

                if result is None:
                    # falls nicht schon in unverteilt angelegt werden sollte, versuchen wir es nochmal dort
                    if _con.undistributed_arcid is int and _arcid != _con.undistributed_arcid:
                        _arcid = _con.undistributed_arcid
                        sql = sqla.select([_con.sys_arc], whereclause=(_con.sys_arc.c.arcid == _arcid))

                        try:
                            result = _db.execute(sql).fetchone()
                        except:
                            self.log.exception(u'%s error (2) in _doc.create' % logid)

                if result:
                    _doc.arcid = _arcid
                    _doc.archive = result[2]
                    _doc.storageid = result[10]

                    if _doc.userid is None or _doc.userid == 0:
                        _doc.userid = 1  # User System

                    if _doc.date is None:
                        _doc.date = datetime.datetime.now().replace(microsecond=0)

                    # wir lassen den Titel leer und setzen ihn erst beim Archivieren auf 'Dokument' falls in der Zwischenzeit nichts neues dazugekommen ist
                    # if _doc.title == '':
                    #   _doc.title = u'Dokument'

                    if type(_doc.date) is datetime.datetime:
                        _date = _doc.date.isoformat(' ').split('.')[0]
                    else:
                        _date = _doc.date

                    vals = {_con.sys_doc.c.userid: _doc.userid,
                            _con.sys_doc.c.arcid: _arcid,
                            _con.sys_doc.c.docdate: _date,
                            _con.sys_doc.c.docname: _doc.title,
                            _con.sys_doc.c.modified_userid: _doc.userid,
                            _con.sys_doc.c.filename: _doc.filename,
                            _con.sys_doc.c.filetype: _doc.filetype,
                            }
                    if _doc.gdocid is not None:
                        vals[_con.sys_doc.c.gdocid] = _doc.gdocid

                    sql = _con.sys_doc.insert(values=vals)
                    # self.log.debug(sql)
                    try:
                        result = _db.execute(sql)
                    except sqla.exc.IntegrityError:
                        if _doc.gdocid is not None:  # IntegrityError bei übergebener gdocid? dann ist der Eintrag schon vorhanden
                            self.log.warning(u'%s Dokument %s schon vorhanden' % (logid, _doc.gdocid))
                        else:
                            self.log.exception(u'%s error (3) in _doc.create' % logid)
                            return None
                    except:
                        self.log.exception(u'%s error (4) in _doc.create' % logid)
                        return None
                    else:
                        _doc.gdocid = result.inserted_primary_key[0]

                    if _doc.gdocid > 0:
                        # neues Dokument hat keine Schlagwörter oder Verknüpfungen, aber Zusatzfelder und Statusfelder!
                        if len(_doc.fields) == 0:  # wenn nicht schon vorhanden
                            _doc.fields = self.get_addfields(_con, _db, _doc)
                        if len(_doc.statefields) == 0:
                            _doc.statefields = self.get_statefields(_con, _db, _doc)
                        if _doc.profile is None:
                            _doc.profile = _con.profile
                    else:
                        _doc.gdocid = None
                        _doc.arcid = None
                else:
                    self.log.exception(u'%s Konnte Arcid %s nicht finden' % (logid, _arcid))
                    _doc.gdocid = None
                    _doc.arcid = None
                    return None
            except:
                self.log.exception(u'%s error (7) in _doc.create' % logid)
                _doc.gdocid = None
                _doc.arcid = None
                return None
            else:
                self.log.debug("%s neue gdocid [%s]" % (logid, _doc.gdocid))
                return True
        else:
            _doc.gdocid = None
            _doc.arcid = None

        if not _doc.arcid:
            raise ArcidNotFoundError(_doc, 'Konnte Arcid von %s nicht ermitteln' % _con.undistributed)
            return None
        if not _doc.gdocid:
            raise CreateDocError(_doc)
            return None

    def __create_v4(self, _doc, _arcid=None, logid=''):
        return None

    def get_addfields(self, _con, _db, _doc, userid=None, groupid=None, load_items=False, orderbyname=False, logid='',
                      with_gridcolumns=False):
        """
        Methode um die Zusatzfelder eines Archivs aus der Datenbank zu holen.
        Zurückgegeben wird ein Dictonary {Addbezeichner des Zusatzfelds:Addfield-Instanz des Zusatzfelds}
        Wird intern auch beim Erstellen oder Laden (create/load) eines Dokuments aufgerufen
        und an self.fields zurückgegeben.
        Damit die Zusatzfelder des Archivs gelesen werden muss die doc.arcid angegeben sein.
        (bei v35 optional auch doc.tblname)
        Auf die Attribute greift man entsprechend über den Key des Dictonary zu:
        self.fields['addbezeichner '].value.
        Wird beim Aufruf load_items gesetzt, so werden auch die Vorbelegungen von Auswahlfeldern geladen.
        Siehe AddField-Klasse.
        Es kann auch eine UserID oder GroupID angegeben werden, dann werden die Addfield-Rechte mitgelesen (v36)
        UserID hat Vorrang vor GroupID, falls beides angegeben wurde
        """
        _fields = OrderedDict()

        if _doc is None:
            return _fields

        result = []

        if _con.version in [3, 35, '3.0']:
            if _doc.arcid is not None:
                sql = 'select arc_tblname from sys_arc where arc_id = %s' % _doc.arcid
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                if result:
                    _doc.tblname = result[0]

            if _doc.tblname:
                sql = 'select * from cfg_%s' % _doc.tblname
                try:
                    try:
                        result = _db.execute(sql)
                    except:
                        result = _con.try_again(sql=sql, db=_db)

                    for row in result:
                        add = bitfarm_doc.BitfarmAddfield()
                        add.name = row['Disp_Name']
                        add.add = row['Feld_Name']
                        add.type = row['Typ']
                        add.checkprogram = row['chck_prog']
                        add.mode = row['Mode']
                        add.options = row['Optionen']
                        add.rights = bitfarm_doc.BitfarmFieldRights()
                        if load_items and row['Werte']:
                            add.items.extend(row['Werte'].split(';'))
                        add.pos = row['Pos']
                        _fields[add.add] = add
                except:
                    _fields = {}

        elif _con.version == 36:
            from_obj = _con.sys_addfieldxs.join(_con.sys_addfields,
                                                _con.sys_addfieldxs.c.fieldid == _con.sys_addfields.c.addfieldid)
            sql = sqla.select(
                [_con.sys_addfieldxs.c.fieldid, _con.sys_addfieldxs.c.userid, _con.sys_addfieldxs.c.groupid,
                 _con.sys_addfieldxs.c.pos, _con.sys_addfieldxs.c.rights,
                 _con.sys_addfields.c.fieldname, _con.sys_addfields.c.displayname, _con.sys_addfields.c.type,
                 _con.sys_addfields.c['items'], _con.sys_addfields.c.mode, _con.sys_addfields.c.options,
                 _con.sys_addfields.c.checkprogram, _con.sys_addfields.c.values_query,
                 _con.sys_addfields.c.indextableid, _con.sys_addfields.c.description], from_obj=[from_obj],
                use_labels=False)
            # use_labels=False).group_by(_con.sys_addfieldxs.c.fieldid)
            arcids = [_doc.arcid]
            if _doc.fields_arcid is not None:
                arcids.append(_doc.fields_arcid)
            if _doc.symlinks:
                for arcid in _doc.symlinks.split(';'):
                    if int(arcid) not in arcids:
                        arcids.append(int(arcid))
            sql.append_whereclause(_con.sys_addfieldxs.c.arcid.in_(arcids))

            # if _doc.fields_arcid is not None:
            #     sql.append_whereclause(_con.sys_addfieldxs.c.arcid.in_([_doc.arcid, _doc.fields_arcid]))
            # else:
            #     sql.append_whereclause(_con.sys_addfieldxs.c.arcid == _doc.arcid)

            if userid is not None:
                sql.append_whereclause(_con.sys_addfieldxs.c.userid == userid)
                groupid = None  # es kann nur nach User ODER Gruppe gesucht werden, Userid hat vorrang
            if groupid is not None:
                sql.append_whereclause(_con.sys_addfieldxs.c.groupid == groupid)

            if orderbyname:
                sql = sql.order_by(_con.sys_addfields.c.displayname)
            else:
                sql = sql.order_by(_con.sys_addfieldxs.c.pos, _con.sys_addfields.c.displayname)

            try:
                result = _db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei der Datenbankabfrage %s" % (logid, str(sql)))

            if len(result) > 0:
                for row in result:
                    if row['fieldname'] not in _fields:
                        add = bitfarm_doc.BitfarmAddfield()
                        add.name = row['displayname']
                        add.add = row['fieldname']
                        add.type = row['type']
                        add.id = row['fieldid']
                        add.checkprogram = row['checkprogram']
                        add.description = row['description']
                        add.mode = row['mode']
                        add.options = row['options']
                        add.values_query = row['values_query']
                        if add.type == 13:  # Gridfeld
                            add.values_query = bfaxml.xml_to_object(add.values_query)
                            if with_gridcolumns and add.values_query:
                                for col in add.values_query.columns:
                                    if col.fieldname not in _fields:
                                        gridcol = bitfarm_doc.BitfarmAddfield()
                                        gridcol.name = col.name
                                        gridcol.add = col.fieldname
                                        gridcol.type = 14
                                        _fields[gridcol.add] = gridcol
                        add.indextableid = row['indextableid']
                        if load_items and row['items']:
                            add.items.extend(row['items'].split(';'))
                        add.pos = row['pos']
                        add.rights = bitfarm_doc.BitfarmFieldRights()

                        r = bitfarm_doc.BitfarmFieldRight()
                        r.userid = row['userid']
                        r.groupid = row['groupid']
                        r.rights = row['rights'] if row['rights'] is not None else 0
                        add.rights.add(r)
                        _fields[add.add] = add
                    else:
                        r = bitfarm_doc.BitfarmFieldRight()
                        r.userid = row['userid']
                        r.groupid = row['groupid']
                        r.rights = row['rights'] if row['rights'] is not None else 0
                        _fields[row['fieldname']].rights.add(r)

        return _fields

    def get_addfield(self, addid=None, name=None, profile=None):
        """
        Methode um ein Zusatzfeld aus der Datenbank zu holen. Wrappermethode für BitfarmCon.read_addfield.
        Damit das Zusatzfeld gelesen werden kann, muss die ID oder der Name angegeben sein.
        Wird beim Aufruf load_items gesetzt, so werden auch die Vorbelegungen von Auswahlfeldern geladen.
        Siehe AddField-Klasse.
        """

        if addid is None and name is None:
            return None

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        return _con.read_addfield(db=_db, addid=addid, name=name)

    def save_annotation(self, anno, gdocid, profile=None, logid=''):
        """Speichert eine Annotation. Es muss die Anno, die Gdocid und das Profil übergeben werden.
        Gibt den Erfolg als True oder False zurück"""

        if type(anno) is not bitfarm_doc.BitfarmAnnotation:
            return False

        if not anno.modified:
            return False

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.version == 36:
            if anno.iev is None:  # vollständig gelöschter Layer
                sql = sqla.delete(_con.sys_anno, whereclause=(sqla.and_(_con.sys_anno.c.gdocid == gdocid,
                                                                        _con.sys_anno.c.page == anno.page,
                                                                        _con.sys_anno.c.layer == anno.layer)))
            else:
                vals = {_con.sys_anno.c.iev: anno.iev.stream.getvalue()}

                sql = sqla.select([_con.sys_anno], whereclause=(sqla.and_(_con.sys_anno.c.gdocid == gdocid,
                                                                          _con.sys_anno.c.page == anno.page,
                                                                          _con.sys_anno.c.layer == anno.layer)))
                result = _db.execute(sql).fetchone()

                if result:
                    sql = _con.sys_anno.update(whereclause=(sqla.and_(_con.sys_anno.c.gdocid == gdocid,
                                                                      _con.sys_anno.c.page == anno.page,
                                                                      _con.sys_anno.c.layer == anno.layer)),
                                               values=vals)
                else:
                    vals[_con.sys_anno.c.gdocid] = gdocid
                    vals[_con.sys_anno.c.layer] = anno.layer
                    vals[_con.sys_anno.c.page] = anno.page
                    sql = _con.sys_anno.insert(values=vals)

            try:
                _db.execute(sql)
            except:
                self.log.exception(u"%s [%d] Fehler beim Speichern von Annotationen 1" % (logid, gdocid))
                return False
            # TODO: annotxt in sys_doctxt überprüfen, annotxt könnte nur von einer seite sein, in sys_doctxt wird aber pro dokument gespeichert
            if anno.text:
                vals = {_con.sys_doctxt.c.annotxt: anno.text}

                sql = sqla.select([_con.sys_doctxt], whereclause=(_con.sys_doctxt.c.gdocid == gdocid and
                                                                  _con.sys_doctxt.c.annotxt != ''))
                result = _db.execute(sql).fetchone()

                if result:
                    sql = _con.sys_doctxt.update(whereclause=(_con.sys_doctxt.c.gdocid == gdocid), values=vals)
                else:
                    vals[_con.sys_doctxt.c.gdocid] = gdocid
                    sql = _con.sys_doctxt.insert(values=vals)

                try:
                    _db.execute(sql)
                except:
                    self.log.exception(u"%s [%d] Fehler beim Speichern von Annotationen 2" % (logid, gdocid))
                    return False
            return True

    def get_annotations(self, gdocid, profile=None, userids=None, logid=''):
        """Alle Annotations eines Dokumentes werden geladen. Es muss die Gdocid und das Profil übergeben werden.
            Es kann eine Userid oder eine Liste an Userids übergeben werden, ansonsten werden nur die öffentlichen Annotations geladen.
        Gibt eine Liste von BitfarmAnnotations zurück"""
        if userids is None:
            userids = []
        result = None
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.version == 36:
            _ret = []

            if _con.sys_anno is None:
                return _ret

            if userids:
                sql = sqla.select([_con.sys_anno], whereclause=(sqla.and_(_con.sys_anno.c.gdocid == gdocid,
                                    _con.sys_anno.c.layer.in_(userids)))).order_by(_con.sys_anno.c.page)
            else:
                sql = sqla.select([_con.sys_anno], whereclause=(_con.sys_anno.c.gdocid == gdocid)).order_by(_con.sys_anno.c.page)
            try:
                result = _db.execute(sql).fetchall()
            except:
                self.log.exception(_('%s Fehler bei der Datenbankabfrage %s') % (logid, str(sql)))

            if result:
                for row in result:
                    if row[5] is None:
                        continue
                    anno = bitfarm_doc.BitfarmAnnotation()
                    anno.layer = row[2]
                    anno.page = row[3]
                    anno.modified_date = row[4]
                    anno.iev = bitfarm_doc.BitfarmStream()
                    anno.iev.stream.write(row[5])
                    _ret.append(anno)
            sql = None
            return _ret

    def set_standard_expired_date(self, arcid, gdocid, profile):
        _ret = False
        newdate = None

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.sys_doclifecycle is None:
            self.log.error(u"Fehler in set_standard_expired_date: Keine Binding zu sys_doclifecycle")
            return False, None

        doc = bitfarm_doc.BitfarmDocument()
        doc.profile = profile
        if not self.load(doc, gdocid, ['base']):
            self.log.error(u"Fehler in set_standard_expired_date: Konnte Dokument %s nicht laden" % gdocid)
            return False, None

        if not arcid:
            self.log.error(u"Fehler in set_standard_expired_date: Keine arcid übergeben, gdocid %s" % gdocid)
            return False, None

        sql = 'select * from sys_arc where arcid = %s' % arcid

        try:
            result = _db.execute(sql).fetchone()
        except:
            self.log.exception(u"Exception in set_standard_expired_date (select sys_arc)")
            return False, None
        else:
            if result:
                if result['duration'] > 0:
                    if result['interval'] == 'd':
                        newdate = doc.date + relativedelta(days=+ result['duration'])
                    elif result['interval'] == 'w':
                        newdate = doc.date + relativedelta(weeks=+ result['duration'])
                    elif result['interval'] == 'm':
                        newdate = doc.date + relativedelta(months=+ result['duration'])
                    elif result['interval'] == 'y':
                        newdate = doc.date + relativedelta(years=+ result['duration'])
                elif result['duration'] == 0 or result['interval'] == 'c':
                    newdate = None

                vals = {_con.sys_doc.c.expired: newdate,
                        _con.sys_doc.c.expired_setby: None,
                        _con.sys_doc.c.expired_userid: 1,
                        _con.sys_doc.c.expired_flags: None}  # System

                sql = sqla.update(_con.sys_doc, values=vals).where(_con.sys_doc.c.gdocid == gdocid)
                try:
                    _db.execute(sql)
                except:
                    self.log.exception(u"Exception in set_standard_expired_date (update sys_doc)")
                    return False, None
                else:
                    _ret = True

        return _ret, newdate

    def reset_expired_date(self, arcid, gdocid, profile):
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.sys_doclifecycle is None:
            return False, None

        vals = {_con.sys_doc.c.expired: None,
                _con.sys_doc.c.expired_setby: None,
                _con.sys_doc.c.expired_userid: 1,
                _con.sys_doc.c.expired_flags: None}  # System

        sql = sqla.update(_con.sys_doc, values=vals).where(_con.sys_doc.c.gdocid == gdocid)
        try:
            _db.execute(sql)
        except:
            self.log.exception("Exception in reset_expired_date (update sys_doc)")
            return False, None

        return self.recalculate_expired_date(gdocid, profile)

    def recalculate_expired_date(self, gdocid, profile=None):
        _ret = False
        newdate = None
        found = False

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.sys_doclifecycle is None:
            return False, None

        doc = bitfarm_doc.BitfarmDocument()
        doc.profile = profile
        if not self.load(doc, gdocid, ['base', 'fields']):
            return False, None

        if doc.expired_setby == 0:  # ein Admin hat das Datum gesetzt, dann dürfen wir nichts ändern
            vals = {_con.sys_doc.c.expired_flags: None}  # recalculate flag zurücksetzen

            sql = sqla.update(_con.sys_doc, values=vals).where(_con.sys_doc.c.gdocid == gdocid)
            try:
                _db.execute(sql)
            except:
                self.log.exception("Exception in recalculate_expired_date (update sys_doc) (1)")

            return False, None

        # jetzt holen wir die dlm konfiguration, arbeiten die durch, und setzten das datum beim ersten (höchsten treffer)

        sql = 'select * from sys_doclifecycle order by prio asc'
        try:
            result = _db.execute(sql).fetchall()
        except:
            self.log.exception("Exception in recalculate_expired_date (select sys_doc_lifecycle)")
            return False, None
        else:
            if result:
                for row in result:
                    found = False
                    newdate = None
                    if row['type'] == 0:  # Statusfeld
                        for state in doc.statefields.values():
                            if state.id == row['fieldid']:
                                if state.value:  # feld ist auch gesetzt?
                                    found = True
                                    break
                    elif row['type'] == 1:  # Zusatzfeld
                        for field in doc.fields.values():
                            if field.id == row['fieldid']:
                                if field.value and field.value == row['item']:  # feld ist auch gesetzt?
                                    found = True
                                    break
                    elif row['type'] == 2:  # Zusatzfeld2
                        for field in doc.fields.values():
                            if field.id == row['fieldid']:
                                if field.value:  # feld ist auch gesetzt?
                                    found = True
                                    break

                    if found:  # wir haben eine passende dlm konfiguration gefunden?
                        if row['date'] is not None:
                            newdate = row['date']
                        elif row['interval'] is None:  # Feld nicht konfiguriert? andere passende Konfigureation suchen
                            continue
                        elif row['type'] == 0:
                            if row['interval'] == 'd':
                                newdate = datetime.datetime.now() + relativedelta(days=+ row['duration'])
                            elif row['interval'] == 'w':
                                newdate = datetime.datetime.now() + relativedelta(weeks=+ row['duration'])
                            elif row['interval'] == 'm':
                                newdate = datetime.datetime.now() + relativedelta(months=+ row['duration'])
                            elif row['interval'] == 'y':
                                newdate = datetime.datetime.now() + relativedelta(years=+ row['duration'])
                            elif row['interval'] == 'c':
                                newdate = None
                        elif row['type'] == 1:
                            if row['interval'] == 'd':
                                newdate = doc.date + relativedelta(days=+ row['duration'])
                            elif row['interval'] == 'w':
                                newdate = doc.date + relativedelta(weeks=+ row['duration'])
                            elif row['interval'] == 'm':
                                newdate = doc.date + relativedelta(months=+ row['duration'])
                            elif row['interval'] == 'y':
                                newdate = doc.date + relativedelta(years=+ row['duration'])
                            elif row['interval'] == 'c':
                                newdate = None
                        elif row['type'] == 2:
                            if row['interval'] == 'd':
                                newdate = field.value + relativedelta(days=+ row['duration'])
                            elif row['interval'] == 'w':
                                newdate = field.value + relativedelta(weeks=+ row['duration'])
                            elif row['interval'] == 'm':
                                newdate = field.value + relativedelta(months=+ row['duration'])
                            elif row['interval'] == 'y':
                                newdate = field.value + relativedelta(years=+ row['duration'])
                            elif row['interval'] == 'c':
                                newdate = None

                        self.log.info(u"recalculate_expired_date neues Datum für %s %s" % (gdocid, str(newdate)))

                        vals = {_con.sys_doc.c.expired: newdate,
                                _con.sys_doc.c.expired_setby: row['prio'],
                                _con.sys_doc.c.expired_userid: 1,
                                _con.sys_doc.c.expired_flags: None}  # System

                        sql = sqla.update(_con.sys_doc, values=vals).where(_con.sys_doc.c.gdocid == gdocid)
                        try:
                            _db.execute(sql)
                        except:
                            self.log.exception("Exception in recalculate_expired_date (update sys_doc) (2)")
                            return False, None
                        else:
                            _ret = True
                            break  # wir haben den dlm eintrag mit der höchsten prio gesetzt, dann raus aus der schleife

            if not found:  # wir haben kein zutreffendes dlm feld gefunden? vielleicht gibt es eine standard lebensdauer im archiv
                return self.set_standard_expired_date(doc.arcid, gdocid, profile)

        return _ret, newdate

    def calculate_expired_date(self, field, gdocid, profile=None, userid=None):
        _ret = False
        newdate = None
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.sys_doclifecycle is None:
            return False, None

        if _con.version == 36:
            # prüfen, ob wir überhaupt ändern dürfen, alle aus der liste der gdocid die entweder NULL oder größer 0 (vom Admin gesetzt) sind
            update = None
            sql = sqla.select([_con.sys_doc], use_labels=True).where(
                sqla.and_(sqla.or_(_con.sys_doc.c.expired_setby > 0,
                                   _con.sys_doc.c.expired_setby.is_(None)),
                          _con.sys_doc.c.gdocid == gdocid))
            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception("Exception in calculate_expired_date (select sys_doc)")
                return False, None
            else:
                if result:
                    update = (result['sys_doc_gdocid'], result['sys_doc_expired_setby'], result['sys_doc_expired'],
                              result['sys_doc_docdate'])

                if update:  # zuerst die konfiguration des feldes holen
                    if field.value:  # das Feld wurde gesetzt
                        if isinstance(field, bitfarm_doc.BitfarmAddfield):
                            if field.type == 11:
                                sql = "select * from sys_doclifecycle where type = 1 and fieldid = %s and item = '%s'" % (
                                field.id, field.value)
                            else:
                                sql = "select * from sys_doclifecycle where type = 2 and fieldid = %s" % (field.id)
                        elif isinstance(field, bitfarm_doc.BitfarmStatefield):
                            sql = "select * from sys_doclifecycle where type = 0 and fieldid = %s" % field.id

                        try:
                            dlm_config = _db.execute(sql).fetchone()
                        except:
                            self.log.exception("Exception in calculate_expired_date (select sys_doclifecycle)")
                            return False, None
                        else:
                            if dlm_config:
                                docid, prio, expired, docdate = update
                                newdate = expired
                                if prio is None or prio >= dlm_config[
                                    'prio']:  # hat die Prio des Dokuments einen höheren Wert (niedrigere Prio) als die DLM Feldkonfiguration
                                    if dlm_config['date'] is not None:
                                        newdate = dlm_config['date']
                                    elif dlm_config[
                                        'interval'] is None:  # wir setzen ein dlm feld auf eine nicht konfigurierte auswahl?
                                        return self.recalculate_expired_date(gdocid,
                                                                             profile)  # das bedeutet ggf. dass eine andere auswahl zurückgesetzt wurde > neuberechnen
                                    elif dlm_config['type'] == 0:
                                        if dlm_config['interval'] == 'd':
                                            newdate = datetime.datetime.now() + relativedelta(
                                                days=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'w':
                                            newdate = datetime.datetime.now() + relativedelta(
                                                weeks=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'm':
                                            newdate = datetime.datetime.now() + relativedelta(
                                                months=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'y':
                                            newdate = datetime.datetime.now() + relativedelta(
                                                years=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'c':
                                            newdate = None
                                    elif dlm_config['type'] == 1:
                                        if dlm_config['interval'] == 'd':
                                            newdate = docdate + relativedelta(days=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'w':
                                            newdate = docdate + relativedelta(weeks=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'm':
                                            newdate = docdate + relativedelta(months=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'y':
                                            newdate = docdate + relativedelta(years=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'c':
                                            newdate = None
                                    elif dlm_config['type'] == 2:
                                        if dlm_config['interval'] == 'd':
                                            newdate = field.value + relativedelta(days=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'w':
                                            newdate = field.value + relativedelta(weeks=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'm':
                                            newdate = field.value + relativedelta(months=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'y':
                                            newdate = field.value + relativedelta(years=+ dlm_config['duration'])
                                        elif dlm_config['interval'] == 'c':
                                            newdate = None

                                    self.log.info(
                                        u"calculate_expired_date neues Datum für %s %s" % (docid, str(newdate)))

                                    vals = {_con.sys_doc.c.expired: newdate,
                                            _con.sys_doc.c.expired_setby: dlm_config['prio'],
                                            _con.sys_doc.c.expired_flags: None}
                                    if userid:
                                        vals[_con.sys_doc.c.expired_userid] = userid

                                    sql = sqla.update(_con.sys_doc, values=vals).where(_con.sys_doc.c.gdocid == docid)
                                    try:
                                        _db.execute(sql)
                                    except:
                                        self.log.exception("Exception in calculate_expired_date (update sys_doc)")
                                        return False, None
                                    else:
                                        _ret = True
                                else:
                                    _ret = False
                    else:  # das feld wurde zurückgesetzt. Jetzt müssen wir das höchste aktive element berechnen
                        return self.recalculate_expired_date(gdocid, profile)

                else:  # prüfen, ob vom Admin eine Lebesdauer gesetzt wurde
                    sql = sqla.select([_con.sys_doc], use_labels=True).where(
                        sqla.and_(_con.sys_doc.c.expired_setby == 0,
                                  _con.sys_doc.c.gdocid == gdocid))
                    try:
                        result = _db.execute(sql).fetchone()
                    except:
                        self.log.exception("Exception in calculate_expired_date (select sys_doc)")
                        return False, None
                    else:
                        if result:
                            newdate = result['sys_doc_expired']

        return _ret, newdate

    def save_addfield(self, field, gdocids, profile=None, userid=None):
        """Speichert ein Zusatzfeld. Es muss das Feld, die Gdocid als Liste und das Profil übergeben werden.
        Gibt den Erfolg als True oder False zurück"""

        if type(field) is not bitfarm_doc.BitfarmAddfield:
            return False

        if not field.modified:
            return False

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _con.version == 36:
            if type(gdocids) is int:
                gdocids = [gdocids]

            sql = 'select type, options, flags, values_query from sys_addfields where addfieldid = %s' % field.id

            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception(_('%s Fehler bei Abfrage des Zusatzfeldes.') % self.name)
                return False
            else:
                if result:
                    field.type = result['type']
                    field.options = result['options']
                    field.flags = result['flags']
                    field.values_query = result['values_query']
                    # field.flags = result['flags'].split(',') if result['flags'] is not None else []
                else:  # keine Ergebnisse?
                    return False

            tbl = _con.sys_docidx[field.indextableid]
            if type(field.value) is list:
                value = ";".join(field.value)
            elif type(field.value) is datetime.datetime:
                value = field.value.isoformat(" ").split(".")[0]
            elif type(field.value) is str:
                value = field.value.strip()
            else:
                value = field.value

            if not value and type(value) not in [int, float]:
                value = None

            oldvalues = {}
            if field.options & 2 == 2:  # oldvalues benötigen wir nur für verknüpfungsfelder
                sql = 'select gdocid, %s from sys_docidx_%i where gdocid in (%s)' % (field.add, field.indextableid, ",".join([str(x) for x in gdocids]))

                result = _db.execute(sql).fetchall()

                if result:
                    for row in result:
                        if type(row[1]) is str:
                            oldvalues[row[0]] = row[1].split(';')
                        elif type(row[1]) is datetime.datetime:
                            oldvalues[row[0]] = [row[1].isoformat(" ").split(".")[0]]
                        else:
                            oldvalues[row[0]] = [row[1]]

            if type(gdocids) is not list:
                gdocids = [int(gdocids)]
            else:
                gdocids = [int(x) for x in gdocids]

            if field.type == 13 and field.values_query is not None:  # Gridfelder benötigen eigene Behandlung
                if isinstance(field.values_query, bitfarm_doc.BitfarmGridConfig):
                    grid_config = field.values_query
                else:
                    grid_config = bfaxml.xml_to_object(field.values_query)

                if grid_config is None:
                    self.log.error('Grid Konfiguration konnte nicht gelesen werden')
                else:
                    field_values = {}
                    add_names = []
                    for column in grid_config.columns:
                        if column.fieldname is not None:
                            field_values[column.fieldname] = []

                    if field.value is not None:
                        for row in field.value.split('|'):  # in einzelne Zeilen teilem
                            if not add_names:  # erste Zeile enthält die Feldnamen
                                add_names = row.split(';')
                                continue
                            row_values = row.split(';')  # in einzelne Werte teilen

                            has_value = False
                            for val in row_values:  # prüfen, ob in einer Zeile überhaupt Werte vorhanden sind
                                if val:
                                    has_value = True

                            if not has_value:  # wenn nicht, überspringen
                                continue

                            for idx in range(len(row_values)):
                                if add_names[idx] in field_values:
                                    field_values[add_names[idx]].append(row_values[idx])

                    for key in field_values.keys():
                        if field_values[key]:
                            field_values[key] = ";".join(field_values[key])
                        else:
                            field_values[key] = None
                        tbl = None
                        for _docidx in _con.sys_docidx.values():
                            if _docidx.columns.get(key) is not None:
                                tbl = _docidx
                                break

                        if tbl is None:
                            self.log.warning('Es konnte beim Speichern des Gridzusatzfeldes %s die Spalte %s nicht gefunden werden' % (field.name, key))
                            continue

                        insert = set(gdocids)
                        update = set()

                        sql = 'select gdocid from %s where gdocid in (%s)' % (tbl, ",".join([str(x) for x in gdocids]))
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception("Exception in save_addfield (select)")
                            return False
                        else:
                            if result:
                                update = set([x[0] for x in result])

                        insert = insert - update

                        if len(insert) > 0:
                            vals = []
                            for docid in insert:
                                vals.append({'gdocid': docid, key: field_values[key]})
                            sql = sqla.insert(tbl).values(vals)

                            try:
                                _db.execute(sql)
                            except:
                                self.log.exception('Exception in save_addfield (insert)')
                                return False

                        if len(update) > 0:
                            vals = {key: field_values[key]}
                            sql = sqla.update(tbl).where(tbl.c.gdocid.in_(update)).values(vals)

                            try:
                                _db.execute(sql)
                            except:
                                self.log.exception('Exception in save_addfield (update)')
                                return False
            else:
                insert = set(gdocids)
                update = set()

                sql = 'select gdocid from %s where gdocid in (%s)' % (tbl, ",".join([str(x) for x in gdocids]))
                try:
                    result = _db.execute(sql).fetchall()
                except:
                    self.log.exception('Exception in save_addfield (select)')
                    return False
                else:
                    if result:
                        update = set([x[0] for x in result])

                insert = insert - update

                if len(insert) > 0:
                    vals = []
                    for docid in insert:
                        vals.append({'gdocid': docid, field.add: value})
                    sql = sqla.insert(tbl).values(vals)

                    try:
                        _db.execute(sql)
                    except:
                        self.log.exception('Exception in save_addfield (insert)')
                        return False

                if len(update) > 0:
                    sql = sqla.update(tbl).where(tbl.c.gdocid.in_(update)).values({field.add: value})

                    try:
                        _db.execute(sql)
                    except:
                        self.log.exception('Exception in save_addfield (update)')
                        return False

            if field.type in [2, 3] and field.options & 8 == 0:  # auswahlfeld non-strict
                sql = sqla.select([_con.sys_addfields]).where(_con.sys_addfields.c.addfieldid == field.id)
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    self.log.exception('Exception in save_addfield (items)')
                else:
                    if result:
                        if result['items'] is not None:
                            _items = result['items'].split(";")
                        else:
                            _items = []
                        len_items = len(_items)
                        if value:
                            if field.type == 2 and value not in _items:
                                _items.append(value)
                            elif field.type == 3:
                                for val in value.split(";"):
                                    if val not in _items:
                                        _items.append(val)

                            if len(_items) != len_items:
                                sql = sqla.update(_con.sys_addfields).where(_con.sys_addfields.c.addfieldid == field.id).values({'items': ";".join(_items)})
                                try:
                                    _db.execute(sql)
                                except:
                                    self.log.exception('Exception in save_addfield (items)')

            if field.type in [11, 12]:  # DLM
                for gdocid in gdocids:
                    success, expired = self.calculate_expired_date(field, gdocid, profile, userid)
                    hist = bitfarm_doc.BitfarmHistory()
                    hist.userid = userid

                    if success:
                        hist.taskid = 115
                        hist.value1 = expired.strftime('%d.%m.%Y') if expired is not None else 'leer'
                        # else:
                        #     if expired is not None:
                        #         hist.taskid = 117
                        #         hist.value1 = expired.strftime('%d.%m.%Y') if expired is not None else 'leer'
                        #     else:
                        #         hist.taskid = 116

                        self.write_history(history=hist, profile=profile, gdocid=gdocid, logid=self.name)

            if field.options & 2 == 2:  # verknüpfungsfeld
                # hatte die verknüpfung schon den alten wert, dann müssen wir die einträge aktualisieren
                new_values = []

                if value is None:
                    values = [None]
                elif type(value) is str:
                    values = value.strip().split(";")
                    values = [x.strip() for x in values]
                else:
                    values = [str(value)]
                    values = [x.strip() for x in values]

                # if oldvalue is None:
                #     oldvalues = [None]
                # elif type(oldvalue) is str:
                #     oldvalues = oldvalue.split(";")
                # else:
                #     oldvalues = [str(oldvalue)]

                for value in values:  # wir entfernen in oldvalues alle werte die sich nicht geändert haben
                    for key, oldvalue in list(oldvalues.items()):
                        if value in oldvalue:
                            oldvalues[key].remove(value)
                            if len(oldvalues[key]) == 0:
                                del oldvalues[key]

                    if value not in new_values:
                        new_values.append(value)

                for key, oldvalue in oldvalues.items():
                    for value in oldvalue:
                        if value is None:
                            continue

                        sql = "delete from sys_doclink where gdocid = %s and refname ='%s'" % (key, value)
                        try:
                            _db.execute(sql)
                        except:
                            self.log.exception(_('Exception in save_addfield (Verknüpfung 1)'))

                        hist = bitfarm_doc.BitfarmHistory()
                        hist.userid = userid
                        hist.taskid = 237
                        hist.value1 = str(value)

                        self.write_history(history=hist, profile=profile,
                                           gdocid=key, logid=self.name)

                if not new_values:  # gibt es keine neuen werte, so wird wahrscheinlich ein alter wert gelöscht
                    new_values = [None]  # hiermit wird die schleife wenigstens einmal ausgeführt

                for value in new_values:
                    ids = gdocids[:]
                    update = []

                    if value is not None:
                        sql = "select gdocid from sys_doclink where gdocid in (%s) and refname ='%s'" % (",".join([str(x) for x in ids]), value)
                        try:
                            result = _db.execute(sql).fetchall()
                        except:
                            self.log.exception("Exception in save_addfield (Verknüpfung 2)")
                        else:
                            if result:
                                for res in result:
                                    if res[0] in ids:
                                        ids.remove(res[0])
                            vals = []
                            for docid in ids:
                                vals.append(u"(%s,'%s')" % (docid, value))
                            if len(vals) > 0:
                                sql = "insert into sys_doclink (gdocid,refname) values %s" % (",".join(vals))
                                try:
                                    _db.execute(sql)
                                except:
                                    self.log.exception(_('Exception in save_addfield (Verknüpfung)'))
                                else:
                                    hist = bitfarm_doc.BitfarmHistory()
                                    hist.userid = userid
                                    hist.taskid = 236
                                    hist.value1 = str(value)

                                    self.write_history(history=hist, profile=profile,
                                                       gdocid=ids, logid=self.name)

            return True

    def save_statefield(self, field, gdocids, profile=None, userid=None, oldvalues=None):
        """Speichert ein Statusfeld. Es muss das Feld, die gdocid als Liste und das Profil übergeben werden. Gibt den Erfolg als True oder False zurück"""

        if oldvalues is None:
            oldvalues = {}
        if type(field) is not bitfarm_doc.BitfarmStatefield:
            return False

        if not field.modified:
            return False

        if field.id is None and field.bitmask is None:
            return False

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if type(gdocids) is int:
            gdocids = [gdocids]

        if _con.version == 36:
            # aktuelle konfiguration des feldes holen
            sql = 'select addfieldid, `function`, flags, bitmask, indexcolumnid from sys_statefields where stateid = %s' % field.id

            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception(_('%s Fehler bei Abfrage des Statusfeldes.') % self.name)
            else:
                if result:
                    field.addfieldid = result['addfieldid']
                    field.func = result['function']
                    if type(result['flags']) is str:
                        field.flags = result['flags'].split(',') if result['flags'] is not None else []
                    else:
                        field.flags = [x for x in result['flags']] if result['flags'] is not None else []
                    field.bitmask = result['bitmask']
                    field.indexcolumnid = result['indexcolumnid']
                else:  # keine Ergebnis und keine Bitmaske?
                    return False

            if not oldvalues:
                sql = 'select gdocid, docstate_%s from sys_doc where gdocid in (%s)' % (
                field.indexcolumnid, ",".join([str(x) for x in gdocids]))
                try:
                    result = _db.execute(sqla.text(sql)).fetchall()
                except:
                    self.log.exception(_('%s Fehler bei Abfrage des Statusfeldes.') % self.name)
                else:
                    if result:
                        for row in result:
                            oldvalues[row[0]] = row[1] & field.bitmask == field.bitmask

            tbl = _con.sys_doc
            col = _con.sys_doc.columns['docstate_%i' % field.indexcolumnid]
            if field.value:
                sql = sqla.update(tbl).where(tbl.c.gdocid.in_(gdocids)).values({col: col.op('|')(field.bitmask)})
            else:
                sql = sqla.update(tbl).where(tbl.c.gdocid.in_(gdocids)).values({col: col.op('& ~')(field.bitmask)})
            try:
                _db.execute(sql)
            except:
                return False

            for gdocid in gdocids:
                if (type(field.flags) is int and field.flags & 4 == 4) or (
                        type(field.flags) is list and 'dlm' in field.flags):
                    success, expired = self.calculate_expired_date(field, gdocid, profile, userid)
                    hist = bitfarm_doc.BitfarmHistory()
                    hist.userid = userid

                    if success:
                        hist.taskid = 115
                        hist.value1 = expired.strftime('%d.%m.%Y') if expired is not None else 'leer'
                        # else:
                        #     if expired is not None:
                        #         hist.taskid = 117
                        #         hist.value1 = expired.strftime('%d.%m.%Y') if expired is not None else 'leer'
                        #     else:
                        #         hist.taskid = 116

                        self.write_history(history=hist, profile=profile, gdocid=gdocid, logid=self.name)

            if field.func is not None and field.func != 0 and field.addfieldid is not None and field.addfieldid != 0:
                sql = 'select * from sys_addfields where addfieldid = %s' % field.addfieldid
                try:
                    result = _db.execute(sqla.text(sql)).fetchone()
                except:
                    self.log.exception(_('%s Fehler bei Abfrage der sys_addfields') % self.name)
                else:
                    indextable = int(result['indextableid'])
                    fieldname = result['fieldname']
                    displayname = result['displayname']
                    typ = result['type']
                    for gdocid in gdocids:
                        hist = bitfarm_doc.BitfarmHistory()
                        hist.userid = userid

                        oldvalue = oldvalues[gdocid] if gdocid in oldvalues else False

                        if oldvalue == field.value:  # auf den gleichen Wert gesetzt (zB wenn mehrere Dokumente verarbeitet werden) -> Funktion überspringen
                            continue

                        if field.value:
                            if field.func == 1:
                                value = datetime.datetime.now().strftime('%H:%M:%S')
                            if field.func == 2:
                                if typ in [1, 12]:
                                    value = datetime.datetime.now().strftime('%Y-%m-%d')
                                else:
                                    value = datetime.datetime.now().strftime('%d.%m.%Y')
                            if field.func == 3:
                                value = datetime.datetime.now().strftime('%d.%m.%Y %H:%M:%S')
                            if field.func == 4:
                                user = _con.read_user(db=_db, userid=userid, logid=self.name)
                                value = u'%s' % user.name if user is not None else u""
                            if field.func == 5:
                                user = _con.read_user(db=_db, userid=userid, logid=self.name)
                                value = u'%s %s' % (user.name if user is not None else u"",
                                                    datetime.datetime.now().strftime('%d.%m.%Y %H:%M:%S'))
                            sql = u"insert into sys_docidx_%s (gdocid, %s) values (%s, '%s') on duplicate key update %s = '%s'" % (
                            indextable, fieldname, gdocid, value, fieldname, value)
                            hist.taskid = 225
                            hist.value1 = displayname
                            hist.value2 = value
                        else:
                            sql = u"insert into sys_docidx_%s (gdocid, %s) values (%s, NULL) on duplicate key update %s = NULL" % (
                            indextable, fieldname, gdocid, fieldname)
                            hist.taskid = 226
                            hist.value1 = displayname

                        try:
                            _db.execute(sqla.text(sql))
                        except:
                            self.log.exception(_('%s Fehler beim Füllen des verknüpften Zusatzfeldes') % self.name)
                        else:
                            self.write_history(history=hist, profile=profile, gdocid=gdocid, logid=self.name)

            return True

    def get_statefields(self, _con, _db, _doc, userid=None, groupid=None, orderbyname=False):
        """
        Holt die Statusfelder eines Dokuments als BitfarmStatefields aus der Datenbank und übergibt diese
        in ein OrderedDict. Entsprechend des Status, der in self.status hinterlegt ist,
        werden BitfarmStatefields.value pro Feld mit True oder False hinterlegt.
        (v35 übergibt eine OrderedDict, mit Statusname als Key und True/False als Wert)
        Die Methode wird intern beim create/load eines Dokuments aufgerufen und
        in doc.statefields zurückgeschrieben.  (v35 doc.statusdict)
        Auf eine Status kann man über self.statefields['Statusname'] zugreifen, siehe BitfarmStatefields.value
        Es kann auch eine UserID oder GroupID angegeben werden, dann werden die Statefield-Rechte mitgelesen (v36)
        UserID hat Vorrang vor GroupID, falls beides angegeben wurde
        """
        _docstate = OrderedDict()
        if _con.version in [3, 35, '3.0']:
            # neuen Status aus Statusdict und in_progress ermitteln
            sql = 'select * from sts_%s order by sts_id' % _doc.tblname
            try:
                result = _db.execute(sql).fetchall()
            except:
                result = _con.try_again(sql=sql, db=_db, fetchall=True)
            if len(result) > 0:
                ind = 0
                for y in [int(2 ** x) for x in range(1, len(result) + 1)]:
                    _docstate[result[ind][1]] = _doc.status & y > 0
                    ind += 1
        elif _con.version == 36:
            from_obj = _con.sys_statefieldxs.join(_con.sys_statefields,
                                                  _con.sys_statefieldxs.c.fieldid == _con.sys_statefields.c.stateid)
            sql = sqla.select([_con.sys_statefieldxs, _con.sys_statefields], from_obj=[from_obj],
                              use_labels=True)

            # if _doc.fields_arcid is not None:
            #     sql.append_whereclause(_con.sys_statefieldxs.c.arcid.in_([_doc.arcid, _doc.fields_arcid]))
            # else:
            #     sql.append_whereclause(_con.sys_statefieldxs.c.arcid == _doc.arcid)

            arcids = [_doc.arcid]
            if _doc.fields_arcid is not None:
                arcids.append(_doc.fields_arcid)
            if _doc.symlinks:
                arcids.extend([int(x) for x in _doc.symlinks.split(';') if int(x) not in arcids])

            sql.append_whereclause(_con.sys_statefieldxs.c.arcid.in_(arcids))

            if userid is not None:
                sql.append_whereclause(_con.sys_statefieldxs.c.userid == userid)
                groupid = None  # es kann nur nach User ODER Gruppe gesucht werden, Userid hat vorrang
            if groupid is not None:
                sql.append_whereclause(_con.sys_statefieldxs.c.groupid == groupid)

            if orderbyname:
                sql = sql.order_by(_con.sys_statefields.c.displayname)
            else:
                sql = sql.order_by(_con.sys_statefieldxs.c.pos, _con.sys_statefields.c.displayname)

            try:
                result = _db.execute(sql).fetchall()
            except:
                result = _con.try_again(sql=sql, db=_db, fetchall=True)

            if len(result) > 0:
                for row in result:
                    if 'deleted' not in row['sys_statefields_flags']:
                        if row['sys_statefields_displayname'] not in _docstate:
                            state = bitfarm_doc.BitfarmStatefield()
                            state.id = row['sys_statefields_stateid']
                            if row['sys_statefields_indexcolumnid'] in _doc.states:
                                state.value = _doc.states[row['sys_statefields_indexcolumnid']] & row[
                                    'sys_statefields_bitmask'] > 0
                            state.modified = False
                            state.bitmask = row['sys_statefields_bitmask']
                            state.name = row['sys_statefields_displayname']
                            state.arcid = _doc.arcid
                            state.gdocid = _doc.gdocid
                            state.addfieldid = row['sys_statefields_addfieldid']
                            state.func = row['sys_statefields_function']
                            state.flags = 0
                            state.flags |= 1 if 'enabled' in row['sys_statefields_flags'] else state.flags
                            state.flags |= 2 if 'deleted' in row['sys_statefields_flags'] else state.flags
                            state.flags |= 4 if 'dlm' in row['sys_statefields_flags'] else state.flags
                            state.pos = row['sys_statefieldxs_pos']
                            state.rights = bitfarm_doc.BitfarmFieldRights()
                            state.indexcolumnid = row['sys_statefields_indexcolumnid']
                            r = bitfarm_doc.BitfarmFieldRight()
                            r.userid = row['sys_statefieldxs_userid']
                            r.groupid = row['sys_statefieldxs_groupid']
                            r.rights = row['sys_statefieldxs_rights'] if row['sys_statefieldxs_rights'] is not None else 0
                            state.rights.add(r)
                            _docstate[state.name] = state
                        else:
                            r = bitfarm_doc.BitfarmFieldRight()
                            r.userid = row['sys_statefieldxs_userid']
                            r.groupid = row['sys_statefieldxs_groupid']
                            r.rights = row['sys_statefieldxs_rights'] if row['sys_statefieldxs_rights'] is not None else 0
                            _docstate[row['sys_statefields_displayname']].rights.add(r)

        return _docstate

    def write_history(self, _doc=None, msg=None, userid=None, taskid=None, history=None, profile=None, gdocid=None,
                      logid=''):
        """
        Methode um für ein Dokument einen History-Eintrag in die Datenbank zu schreiben.
        Für v35 muss ein Document-Objekt und eine Meldung und optional kann eine UserId und eine TaskId übergeben werden.
        Wird keine UserId übergeben ist der eintragende User 'System'
        Für v36 wird ein einzelnes oder eine Liste von BitfarmHistory Objekte übergeben.
        Es kann mit dem History-Objekt auch eine Liste von Gdocids übergeben werden, dann wird ein Eintrag für jede ID erzeugt.
        Achtung! Die aufrufende Funktion muss sich hinterher um das entfernen der Historyeinträge aus der Liste kümmern,
        da es sonst zu Mehrfacheinträgen kommen könnte (zB bei einen weiteren BitfarmDocProcessing.write() Aufruf)
        Die Methode reicht den Aufruf an die privaten Methoden __write_history_v3 und __write_history_v4
        weiter, entsprechend der Version self.version.
        @param msg: Text des Historyeintrages
        @param userid: (optional) UserId des eintragenden Users, ansonsten 0 (System)
        @param taskid: (optional) TaskId des Historyeintrages, ansonsten 0
        """

        if self.version in (3, 35, '3.0'):
            return self.__write_history_v3(_doc, msg, userid, taskid, logid=logid)
        if self.version == 36:
            return self.__write_history_v36(history, profile, gdocid, logid=logid)
        else:
            return self.__write_history_v4(history, profile, gdocid, logid=logid)

    def __write_history_v3(self, _doc, msg, userid, taskid, logid=''):
        _ret = True

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if not msg and not _doc._docid:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if userid is None:
            userid = _doc.userid if _doc.userid else 0

        if taskid is None:
            taskid = 0
        try:
            # if _con.history is None:
            _con.history = sqla.Table('hst_%s' % _doc.tblname, _con.meta, autoload=True)

            ins = _con.history.insert()  # (task_info, usr_id, doc_id) Values (:info, :uid, :id)" %(self.tblname)
            try:
                result = _db.execute(ins, Task_Info=msg, Task=taskid, USR_ID=userid, DOC_ID=_doc._docid,
                                     TStamp=datetime.datetime.now())
            except:
                result = _con.try_again(sql=ins, db=_db, Task_Info=msg, Task=taskid, USR_ID=userid, DOC_ID=_doc._docid,
                                        TStamp=datetime.datetime.now())
            self.log.debug(u'%s History-Eintrag für [%s] geschrieben: %s' % (logid, _doc.gdocid, msg))
        except:
            self.log.exception(
                u'%s Fehler beim Schreiben eines History-Eintrags für [%s]: %s' % (logid, _doc.gdocid, msg))
            _ret = False

        return _ret

    def __write_history_v36(self, history, profile, gdocid, logid=''):
        _ret = True

        if not history or history is None:
            return _ret

        if type(history) is list:
            histories = history
        else:
            histories = [history]

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        for hist in histories:
            if type(hist) is not bitfarm_doc.BitfarmHistory:
                continue

            if hist.gdocid is None and gdocid is None:
                self.log.error(u'%s Fehler beim Schreiben eines History-Eintrags: keine gdocid' % logid)
                _ret = False
                continue
            if hist.taskid is None:
                self.log.error(
                    u'%s Fehler beim Schreiben eines History-Eintrags: keine taskid [%s]' % (logid, hist.gdocid))
                _ret = False
                continue
            if hist.userid is None:
                hist.userid = 1  # kommen wir ohne userid, dann schreiben wir als SYSTEM

            gdocids = []
            if gdocid is not None:  # wir nehmen erstmal die gdocid(s), die der funktion übergeben wurde
                if type(gdocid) is str:
                    gdocids = gdocid.split(';')
                elif type(gdocid) is list:
                    gdocids = gdocid
                elif type(gdocid) is int:
                    gdocids.append(gdocid)
                gdocids = [int(x) for x in gdocids]
            if hist.gdocid is not None and not gdocids:  # wenn nichts übergeben wurde nehmen wir die aus dem history-objekt
                gdocids = [int(hist.gdocid)]

            if type(hist.value1) is list:
                hist.value1 = u";".join(hist.value1)
            if type(hist.value2) is list:
                hist.value2 = u";".join(hist.value2)
            if type(hist.value3) is list:
                hist.value3 = u";".join(hist.value3)

            hist.value1 = str(hist.value1)
            hist.value2 = str(hist.value2)
            hist.value3 = str(hist.value3)

            # if len(hist.value1) > 255:
            #     hist.values1 = hist.value1[:250] + '(...)'
            # if len(hist.value2) > 255:
            #     hist.values2 = hist.value2[:250] + '(...)'
            # if len(hist.value3) > 255:
            #     hist.values3 = hist.value3[:250] + '(...)'

            ins = _con.sys_history.insert()
            for _gdocid in gdocids:
                try:
                    if hist.date is not None:
                        _db.execute(ins, taskid=hist.taskid, userid=hist.userid, gdocid=_gdocid,
                                    tstamp=hist.date, value1=hist.value1[:65536], value2=hist.value2[:65536],
                                    value3=hist.value3[:65536])
                    else:
                        _db.execute(ins, taskid=hist.taskid, userid=hist.userid, gdocid=_gdocid,
                                    value1=hist.value1[:65536], value2=hist.value2[:65536],
                                    value3=hist.value3[:65536])
                except:
                    self.log.exception(u'%s Fehler beim Schreiben eines History-Eintrags für [%s]' % (logid, _gdocid))
                    _ret = False
                else:
                    self.log.debug(u'%s History-Eintrag für [%s] geschrieben' % (logid, _gdocid))

            if gdocids != [0]:  # wir müssen keine Änderung am Dokument 0 registrieren (also administrative Historymeldungen ohne Dokument)
                # vals = {_con.sys_doc.c.modified_userid: hist.userid,
                #         _con.sys_doc.c.modified: None}
                vals = {_con.sys_doc.c.modified_userid: hist.userid}  # nur so funktioniert es mit MySQL57 und MySQL8
                sql = _con.sys_doc.update(whereclause=(_con.sys_doc.c.gdocid.in_(gdocids)), values=vals)

                try:
                    _db.execute(sql)
                except:
                    self.log.exception(_('%s Fehler beim Aktualisieren des Modified Date') % self.name)

        return _ret

    def __write_history_v4(self, history, profile, gdocid, logid=''):
        pass

    def read_history(self, _doc=None, gdocid=None, userid=None, taskid=None, profile=None, params=None):
        """
        Methode um für ein Dokument alle History-Einträge aus der Datenbank zu lesen.
        Gibt eine Liste mit den einzelnen Einträgen aus der Datenbank wieder.
        Optional kann eine UserId und eine TaskId übergeben werden.
        Für v35 wird ein Document-Objekt übergeben, für v36 eine gdocid und optional ein Profil
        @param userid: (optional) es werden alle Historyeinträge dieser UserId gelesen
        @param taskid: (optional) es werden alle Historyeinträge dieser TaskId gelesen
        """

        if self.version in (3, 35, '3.0'):
            return self.__read_history_v3(_doc, userid, taskid)
        if self.version == 36:
            return self.__read_history_v36(_doc, gdocid, userid, taskid, profile, params)
        else:
            return self.__read_history_v4(_doc, gdocid, userid, taskid, profile)

    def __read_history_v3(self, _doc, userid, taskid):
        _ret = []

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        try:
            _con.history = sqla.Table('hst_%s' % _doc.tblname, _con.meta, autoload=True)

            sql = sqla.select([_con.history], whereclause=(_con.history.c.DOC_ID == _doc._docid)).order_by(
                sqla.desc(_con.history.c.TStamp))
            if userid is not None:
                sql.append_whereclause(_con.history.c.USR_ID == userid)
            if taskid is not None:
                sql.append_whereclause(_con.history.c.Task == taskid)

            try:
                result = _db.execute(sql)
            except:
                result = _con.try_again(sql=sql, db=_db)
            if result:
                if _doc is not None:
                    _doc.history = []
                for row in result:
                    hist = bitfarm_doc.BitfarmHistory()
                    hist.gdocid = _doc.gdocid
                    hist.userid = row[2]
                    hist._date = row[3]
                    hist.taskid = 4
                    hist.value1 = row[5]

                    _ret.append(hist)
                    if _doc is not None:
                        _doc.history.append(hist)

            self.log.debug(u'History für [%s] gelesen' % _doc.gdocid)
        except:
            self.log.exception(u'Fehler beim Lesen der History-Einträge für [%s]' % _doc.gdocid)
            _ret = []

        return _ret

    def __read_history_v36(self, _doc, gdocid, userid, taskid, profile, params):
        _ret = []

        if _doc is not None and isinstance(_doc, bitfarm_baseobj.BitfarmBaseObject):
            gdocid = _doc.gdocid

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if params is not None and 'userid' in params:
            userid = int(params['userid'])

        if params is not None and 'gdocid' in params:
            gdocid = int(params['gdocid'])
        else:
            if gdocid is None:
                gdocid = 0

        if params is not None and 'gdocid_neg' in params:
            gdocid = int(params['gdocid_neg'])
            negate = True
        else:
            negate = False

        if params is not None and 'limit' in params:
            limit = int(params['limit'])
        else:
            limit = 1000

        if params is not None and 'offset' in params:
            offset = int(params['offset'])
        else:
            offset = 0

        if params is not None and 'datefrom' in params:
            datefrom = params['datefrom']
        else:
            datefrom = None

        if params is not None and 'dateto' in params:
            dateto = params['dateto']
        else:
            dateto = None

        try:
            from_obj = _con.sys_history.join(_con.sys_user, _con.sys_history.c.userid == _con.sys_user.c.userid)
            sql = sqla.select([_con.sys_history, _con.sys_user], from_obj=[from_obj]).order_by(
                sqla.desc(_con.sys_history.c.tstamp)).order_by(sqla.desc(_con.sys_history.c.historyid)).limit(limit).offset(offset)
            # sql = sqla.select([_con.sys_history]).order_by(
            #     sqla.desc(_con.sys_history.c.tstamp)).order_by(sqla.desc(_con.sys_history.c.historyid)).limit(limit).offset(offset)

            if gdocid is not None:
                if negate:
                    sql.append_whereclause(_con.sys_history.c.gdocid != gdocid)
                else:
                    sql.append_whereclause(_con.sys_history.c.gdocid == gdocid)
            if userid is not None:
                sql.append_whereclause(_con.sys_history.c.userid == userid)
            if taskid is not None:
                sql.append_whereclause(_con.sys_history.c.taskid == taskid)
            if datefrom is not None:
                sql.append_whereclause(_con.sys_history.c.tstamp >= datefrom)
            if dateto is not None:
                sql.append_whereclause(_con.sys_history.c.tstamp <= dateto)

            try:
                result = _db.execute(sql).fetchall()
            except:
                result = _con.try_again(sql=sql, db=_db)
            if result:
                if _doc is not None:
                    _doc.history = []
                for row in result:
                    hist = bitfarm_doc.BitfarmHistory()
                    hist.gdocid = row[1]
                    hist.userid = row[2]
                    hist._date = row[3]
                    hist.taskid = row[4]
                    hist.value1 = row[5]
                    hist.value2 = row[6]
                    hist.value3 = row[7]
                    hist.username = row[9]

                    _ret.append(hist)
                    if _doc is not None:
                        _doc.history.append(hist)

                result = None
                sql = None
                users = None
        except:
            self.log.exception(u'Fehler beim Lesen der History-Einträge für [%s]' % gdocid)
            _ret = []

        return _ret

    def __read_history_v4(self, _doc, gdocid, userid, taskid, profile):
        pass

    def checkout(self, _doc, userid=None):
        """
        Methode um ein Dokument auszuchecken. Es muss eine UserId übergeben werden
        Die Methode reicht den Aufruf an die privaten Methoden __checkout_v3 und __checkout_v4
        weiter, entsprechend der Version self.version.
        @param userid: UserId des auscheckenden Users
        """
        if _doc.gdocid is None:
            return False

        self.log.debug('checkout userid = %s' % userid)
        if self.version in (3, 35, '3.0'):
            return self.__checkout_v3(_doc, userid)
        if self.version == 36:
            return self.__checkout_v36(_doc, userid)
        else:
            return self.__checkout_v4(_doc, userid)

    def __checkout_v3(self, _doc, userid):
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        userid = userid if userid is not None else _doc.userid

        if userid is None:
            return False

        sql = "select * from sys_doclock where gdoc_id = %s" % _doc.gdocid
        try:
            result = _db.execute(sql).fetchone()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchone=True)
        if result:
            self.log.warning('Dokument [%s] ist schon ausgecheckt von %s' % (_doc.gdocid, result[4]))
            self.__is_checkout = True
            return False

        if _con.sys_doclock is None:
            _con.sys_doclock = sqla.Table('sys_doclock', _con.meta, autoload=True)

        vals = {_con.sys_doclock.c.gdoc_id: _doc.gdocid,
                _con.sys_doclock.c.arc_id: _doc.arcid,
                _con.sys_doclock.c.doc_id: _doc._docid,
                _con.sys_doclock.c.usr_id: userid
                }

        sql = _con.sys_doclock.insert(values=vals)
        self.log.debug(sql)
        try:
            result = _db.execute(sql)
        except:
            result = _con.try_again(sql=sql, db=_db)

        if result:
            lockid = result.inserted_primary_key[0]
            self.log.debug('Dokument [%s] ausgecheckt, doclock_id = %s' % (_doc.gdocid, lockid))

    def __checkout_v36(self, _doc, userid):
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        userid = userid if userid is not None else _doc.userid

        if userid is None:
            return False

        sql = "select checkout from sys_doc where gdocid = %s and checkout <> 0" % _doc.gdocid
        try:
            result = _db.execute(sql).fetchone()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchone=True)
        if result:
            self.log.warning('Dokument [%s] ist schon ausgecheckt von User (%s)' % (_doc.gdocid, result[0]))
            self.__is_checkout = True
            return False

        sql = "update sys_doc set checkout = %i where gdocid = %s" % (userid, _doc.gdocid)
        self.log.debug(sql)
        try:
            result = _db.execute(sql)
        except:
            result = _con.try_again(sql=sql, db=_db)

        if result:
            self.log.debug('Dokument [%s] ausgecheckt von User (%i)' % (_doc.gdocid, userid))
            _doc.is_checkout = True
            _doc.checkout_userid = userid
            return True
        else:
            return False

    def __checkout_v4(self, _doc, userid):
        pass

    def checkin(self, _doc, userid=None):
        """
        Methode um ein Dokument einzuchecken. Es muss eine UserId übergeben werden.
        Wenn das Dokument von diesem User ausgecheckt wurde, wird es wieder eingecheckt,
        ansonsten gibt die Funktion None zurück
        Die Methode reicht den Aufruf an die privaten Methoden __checkout_v3 und __checkout_v4
        weiter, entsprechend der Version self.version.
        @param userid: UserId des auscheckenden Users
        """
        if _doc.gdocid is None:
            return False

        self.log.debug('checkin userid = %s' % userid)
        if self.version in (3, 35, '3.0'):
            return self.__checkin_v3(_doc, userid)
        if self.version == 36:
            return self.__checkin_v36(_doc, userid)
        else:
            return self.__checkin_v4(_doc, userid)

    def __checkin_v3(self, _doc, userid):
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        userid = userid if userid is not None else _doc.userid

        if userid is None:
            return False

        sql = "select * from sys_doclock where gdoc_id = %s" % _doc.gdocid
        try:
            result = _db.execute(sql).fetchone()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchone=True)
        if result:
            if result[4] != int(userid):
                self.log.warning('Dokument [%s] von einem anderen User (%s) ausgecheckt' % (_doc.gdocid, result[4]))
                return False

            sql = "delete from sys_doclock where gdoc_id = %s and usr_id = %s" % (_doc.gdocid, userid)
            self.log.debug(sql)
            try:
                try:
                    result = _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                if result:
                    self.log.debug('Dokument [%s] wurde von %s eingecheckt' % (_doc.gdocid, userid))
                    _doc.__is_checkout = False
                    return True
            except:
                self.log.exception('Dokument [%s] konnte nicht eingeckt werden' % _doc.gdocid)
                return False

        else:
            self.log.warning('Dokument [%s] nicht ausgecheckt' % _doc.gdocid)
            return False

    def __checkin_v36(self, _doc, userid):
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        userid = userid if userid is not None else _doc.userid

        if userid is None:
            return False

        sql = "select checkout from sys_doc where gdocid = %s" % _doc.gdocid
        try:
            result = _db.execute(sql).fetchone()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchone=True)
        if result:
            if result[0] != int(userid):
                self.log.warning('Dokument [%s] von einem anderen User (%s) ausgecheckt' % (_doc.gdocid, result[0]))
                return False

            sql = "update sys_doc set checkout = NULL where gdocid = %s" % _doc.gdocid
            self.log.debug(sql)
            try:
                try:
                    result = _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                if result:
                    self.log.debug('Dokument [%s] wurde von %s eingecheckt' % (_doc.gdocid, userid))
                    _doc.is_checkout = False
                    _doc.checkout_userid = None
                    return True
            except:
                self.log.exception('Dokument [%s] konnte nicht eingeckt werden' % _doc.gdocid)
                return False

        else:
            self.log.warning('Dokument [%s] nicht ausgecheckt' % _doc.gdocid)
            return False

    def __checkin_v4(self, _doc, userid):
        pass

    def set_editmode(self, _doc, userid=None, docids=None):
        """
        Methode um ein Dokument in der Datenbank in den Editmode zu setzen. Es muss eine UserId übergeben werden
        @param userid: UserId des auscheckenden Users
        """
        if _doc.gdocid is None:
            return False

        self.log.debug('set_editmode userid = %s' % userid)

        if self.version == 36:
            return self.__set_editmode_v36(_doc, userid, docids)
        else:
            return self.__set_editmode_v4(_doc, userid, docids)

    def __set_editmode_v36(self, _doc, userid, docids):
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        userid = userid if userid is not None else _doc.userid

        if userid is None:
            return False

        if docids is None:
            gdocid = str(_doc.gdocid)
        else:
            docids.append(_doc.gdocid)
            gdocid = ','.join([str(x) for x in docids])

        sql = "select editmode from sys_doc where gdocid in (%s) and editmode <> 0" % _doc.gdocid
        try:
            result = _db.execute(sql).fetchone()
        except:
            self.log.exception(u"Fehler bei Abfrage des Editmodes")
            return False
        else:
            if result and result[0] != userid:
                self.log.warning(u'Dokument [%s] ist schon im Editmode von User (%s)' % (_doc.gdocid, result[0]))
                return False

        sql = "update sys_doc set editmode = %i where gdocid in (%s) and editmode = 0" % (
            userid, gdocid)  # wir setzen editmode nur dort, wo er noch nicht gesetzt ist
        try:
            _db.execute(sql)
        except:
            self.log.exception(u"Fehler bei Setzen des Editmodes")
            return False
        else:
            self.log.debug(u'Dokument [%s] in Editmode von User (%i)' % (gdocid, userid))
            return True

    def __set_editmode_v4(self, _doc, userid, docids):
        pass

    def unset_editmode(self, _doc, userid=None, unset_all=False, docids=None):
        """
        Methode um ein Dokument aus dem Editmode zurückzusetzen. Es muss eine UserId übergeben werden.
        Wenn das Dokument von diesem User in den Editmode gesetzt wurde, wird dieser wieder zurückgenommen,
        ansonsten gibt die Funktion None zurück. Wird unset_all übergeben, dann werden in der Datenbank alle Einträge des User zurückgesetzt.
        @param userid: UserId des auscheckenden Users
        """
        if _doc.gdocid is None and not unset_all:
            return False

        self.log.debug('unset_editmode userid = %s' % userid)
        if self.version == 36:
            return self.__unset_editmode_v36(_doc, userid, unset_all, docids)
        else:
            return self.__unset_editmode_v4(_doc, userid, unset_all, docids)

    def __unset_editmode_v36(self, _doc, userid, unset_all, docids):
        if type(_doc) not in [bitfarm_doc.BitfarmDocument, bitfarm_doc.bfaReq]:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        userid = userid if userid is not None else _doc.userid

        if userid is None:
            return False

        if docids is None:
            gdocid = str(_doc.gdocid)
        elif type(docids) is int:
            gdocid = str(docids)
        else:
            if _doc.gdocid not in docids:
                docids.append(_doc.gdocid)
            gdocid = ','.join([str(x) for x in docids])

        if unset_all:
            sql = "update sys_doc set editmode = 0 where editmode = %i" % userid
            self.log.debug(sql)
            try:
                result = _db.execute(sql)
            except:
                result = _con.try_again(sql=sql, db=_db)

            if result:
                self.log.debug('Alle Dokumente von Userid %i zurückgesetzt' % userid)
                _doc.__is_checkout = False
                return True

        else:
            sql = "select editmode from sys_doc where gdocid in (%s)" % gdocid
            try:
                result = _db.execute(sql).fetchone()
            except:
                result = _con.try_again(sql=sql, db=_db, fetchone=True)
            if result:
                if result[0] is not None:
                    if result[0] != 0 and result[0] != int(userid):
                        self.log.warning('Dokument [%s] von einem anderen User (%i) in Editmode' % (gdocid, result[0]))
                        return False

                sql = "update sys_doc set editmode = 0 where gdocid in (%s)" % gdocid
                self.log.debug(sql)
                try:
                    try:
                        result = _db.execute(sql)
                    except:
                        result = _con.try_again(sql=sql, db=_db)

                    if result:
                        self.log.debug('Editmode von Dokument [%s] wurde von %s zurückgesetzt' % (gdocid, userid))
                        return True
                except:
                    self.log.exception('Editmode von Dokument [%s] konnte nicht zurückgesetzt werden' % gdocid)
                    return False

    def __unset_editmode_v4(self, _doc, userid, unset_all, docids):
        pass

    def refresh_fieldtxt(self, _doc=None, gdocid=None, profile=None):
        """Aktualisiert die fields_txt Spalte in der sys_doctxt Tabelle"""
        self.log.debug('refresh_fieldstxt')
        if _doc is None and gdocid is None:
            return None

        if type(_doc) is not bitfarm_doc.BitfarmDocument and gdocid is None:
            return None

        if _doc is not None:
            gdocid = _doc.gdocid
            profile = _doc.profile

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        txt = []
        for idx in _con.sys_docidx.values():
            sql = sqla.select([idx]).where(idx.c.gdocid == gdocid)

            try:
                result = _db.execute(sql).fetchone()
            except:
                return False
            else:
                if result:
                    for col in result.items():
                        if col[0] not in ['docidxid', 'gdocid', 'modified'] and col[1] is not None:
                            if type(col[1]) is datetime.datetime:
                                # TODO: wir müssen noch systemabhängige Datumsformate einführen
                                # if datetime.datetime.strftime(col[1], '%d.%m.%Y') not in txt:
                                if bitfarm_doc.bfa_strftime(col[1], '%d.%m.%Y') not in txt:
                                    # txt.append(datetime.datetime.strftime(col[1], '%d.%m.%Y'))
                                    txt.append(bitfarm_doc.bfa_strftime(col[1], '%d.%m.%Y'))
                            else:
                                _vals = str(col[1]).split(
                                    ';')  # semikolon getrennte Werte (Mehrfachauswahl, Gridspalte) kommen einzelnd rein
                                for _val in _vals:
                                    if _val and _val not in txt:
                                        txt.append(_val)
        # if len(txt) > 0:
        txt = "\n".join(txt)
        # sql = u"insert into sys_doctxt (fieldtxt, gdocid) values ('%s', %i) on duplicate key update fieldtxt = '%s'" % (txt, gdocid, txt)

        sql = _con.sys_doctxt.select(_con.sys_doctxt.c.gdocid).where(_con.sys_doctxt.c.gdocid == gdocid)
        try:
            result = _db.execute(sql).fetchone()
        except:
            return False
        else:
            if result and result[0]:
                sql = _con.sys_doctxt.update().values({'fieldtxt': txt}).where(_con.sys_doctxt.c.gdocid == gdocid)
            else:
                sql = _con.sys_doctxt.insert().values({'fieldtxt': txt, 'gdocid': gdocid})

            try:
                _db.execute(sql)
            except:
                return False

        return True

    def duplicate_files(self, _con, _doc, _error=None, _right_src=None, _gdocid_src=None):
        self.log.debug('%s duplicate_files' % self.name)
        """Erstellt Kopien von docloc, extloc und oriloc und hinterlegt den neuen Pfad ins Doc Objekt"""

        result = True
        revision_safe = False
        nonrevision_safe = False
        newpath = ""
        newpath_nrs = ""
        fpath = ""

        if not _gdocid_src:
            _gdocid_src = _doc.gdocid

        if not _right_src:  # werden keine Rechte übergeben, ertsellen wir eine BitfarmRights Objekt mit write und modify
            _right_src = [bitfarm_doc.BitfarmRights()]
            _right_src[0].write = True
            _right_src[0].modify = True

        if not _error:
            _error = ''

        if _doc.docloc:
            if _doc.docloc.lower().find(_con.archive_nrs.lower()) > -1:
                nonrevision_safe = nonrevision_safe or True
            else:
                revision_safe = revision_safe or True
        if _doc.extloc:
            if _doc.extloc.lower().find(_con.archive_nrs.lower()) > -1:
                nonrevision_safe = nonrevision_safe or True
            else:
                revision_safe = revision_safe or True
        if _doc.oriloc and (_right_src[0].write or _right_src[0].modify):
            if _doc.oriloc.lower().find(_con.archive_nrs.lower()) > -1:
                nonrevision_safe = nonrevision_safe or True
            else:
                revision_safe = revision_safe or True

        self.create_basename(_doc)
        if revision_safe:
            newpath = self.create_arcpath(_doc, is_revisionsave=True)
        if nonrevision_safe:
            newpath_nrs = self.create_arcpath(_doc, is_revisionsave=False)

        if _doc.docloc:
            fext = os.path.splitext(_doc.docloc)[1]
            if _doc.docloc.lower().find(_con.archive_nrs.lower()) > -1:
                fpath = os.path.join(newpath_nrs, _doc.basename + fext)
            else:
                fpath = os.path.join(newpath, _doc.basename + fext)
            if not os.path.exists(os.path.expandvars(_doc.docloc)):
                self.log.error('Pfad nicht vorhanden %s' % _doc.docloc)
                result = False
            else:
                try:
                    shutil.copyfile(os.path.expandvars(_doc.docloc), os.path.expandvars(fpath))
                except:
                    self.log.exception("Die Datei konnte nicht kopiert werden %s nach %s" % (_doc.docloc, fpath))
                    _error += "Die Voransicht des Dokuments [%s] konnte nicht kopiert werden\n" % _gdocid_src
                    result = False
                else:
                    _doc.docloc = fpath

        if _doc.extloc:
            fext = os.path.splitext(_doc.extloc)[1]
            if _doc.extloc.lower().find(_con.archive_nrs.lower()) > -1:
                fpath = os.path.join(newpath_nrs, _doc.basename + fext)
            else:
                fpath = os.path.join(newpath, _doc.basename + fext)
            if not os.path.exists(os.path.expandvars(_doc.extloc)):
                self.log.error('Pfad nicht vorhanden %s' % _doc.extloc)
                result = False
            else:
                try:
                    shutil.copyfile(os.path.expandvars(_doc.extloc), os.path.expandvars(fpath))
                except:
                    self.log.exception("Die Datei konnte nicht kopiert werden %s nach %s" % (_doc.extloc, fpath))
                    _error += "Die Voransichtskopie des Dokuments [%s] konnte nicht kopiert werden\n" % _gdocid_src
                    result = False
                else:
                    _doc.extloc = fpath

        if _doc.oriloc and (_right_src[0].write or _right_src[0].modify):
            fext = os.path.splitext(_doc.oriloc)[1]
            if _doc.oriloc.lower().find(_con.archive_nrs.lower()) > -1:
                fpath = os.path.join(newpath_nrs, _doc.basename + fext)
            else:
                fpath = os.path.join(newpath, _doc.basename + fext)
            if not os.path.exists(os.path.expandvars(_doc.oriloc)):
                self.log.error('Pfad nicht vorhanden %s' % _doc.oriloc)
                result = False
            else:
                try:
                    shutil.copyfile(os.path.expandvars(_doc.oriloc), os.path.expandvars(fpath))
                except:
                    self.log.exception("Die Datei konnte nicht kopiert werden %s nach %s" % (_doc.oriloc, fpath))
                    _error += "Die Originaldatei des Dokuments [%s] konnte nicht kopiert werden\n" % _gdocid_src
                    result = False
                else:
                    _doc.oriloc = fpath
        else:
            _doc.oriloc = None

        return result, _error

    def read_from_file(self, _doc, jobname=None):
        """
        Liest eine Jobdatei von aus der Queue. (<= Version 3.5)
        Optional kann eine Job-Datei (oder .tpl) angegeben werden,
        ansonsten wird der aktuell in job.tempname hinterlegte Job gelesen
        @param jobname: (optional) der Pfad auf die .job-Datei. Wird nur ein Basisname übergeben,
        wird die Endung .job angehangen. Wird nichts übergeben, so wird self.tempname ausgelesen.
        """
        self.log.debug('read_from_file')
        if _doc is None:
            return None

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if jobname:
            # if type(jobname) is str:
            #     jobname = jobname.decode('utf-8')
            if not os.path.isfile(jobname):
                jobname = _doc.tempname + ".job"
        else:
            if not _doc.tempname:
                self.log.debug('es wurde kein Job gelesen oder self.tempname ist leer')
                return None
            jobname = _doc.tempname + ".job"

        if os.path.isfile(jobname) and os.path.exists(jobname):
            try:
                #                self.config = ConfigObj(jobname, list_values = False,  write_empty_values = self.write_empty_values,  encoding = _con.db_encoding, default_encoding = _con.db_encoding)
                _doc.config = ConfigObj(jobname, list_values=False, write_empty_values=_doc.write_empty_values,
                                        encoding='cp1252', default_encoding='cp1252')
            except:
                self.log.exception("Die Datei %s konnte nicht gelesen werden" % jobname)
                return False
        else:
            self.log.debug("Die Datei %s konnte nicht gefunden werden" % jobname)
            return False

        if 'Version' in _doc.config:
            if 'Version' in _doc.config['Version']:
                _doc.version = int(float(_doc.config['Version']['Version']))
                if _doc.version == 3:
                    _doc.version = 35
        if 'Archiv' in _doc.config:
            if 'Profil' in _doc.config['Archiv']:
                _doc.profile = _doc.config['Archiv']['Profil']
            if 'Name' in _doc.config['Archiv']:
                _doc.archive = _doc.config['Archiv']['Name']

                try:
                    if _con.version == 35:
                        sql = _con.sys_arc.select(_con.sys_arc.c.ARC_Name == _doc.archive)
                    elif _con.version == 36:
                        sql = _con.sys_arc.select(_con.sys_arc.c.arcname == _doc.archive)
                    try:
                        result = _db.execute(sql).fetchone()
                    except:
                        result = _con.try_again(sql=sql, db=_db, fetchone=True)
                    self.log.debug(result)
                    if result:
                        _doc.arcid = result[0]
                except:
                    _doc.arcid = None
            if 'Tabelle' in _doc.config['Archiv']:
                _doc.tabelle = _doc.config['Archiv']['Tabelle']
                _doc.tblname = _doc.tabelle

        if 'Document' in _doc.config:
            if 'Titel' in _doc.config['Document']:
                _doc.title = _doc.config['Document']['Titel']
            if 'Datum' in _doc.config['Document']:
                try:
                    _doc.date = datetime.datetime.strptime(_doc.config['Document']['Datum'], "%Y-%m-%d %H:%M:%S")
                except:
                    pass
            else:
                _doc.date = datetime.datetime.now().replace(microsecond=0)

                # wird an dieser Stelle nicht mehr benötigt, die gdocid wird vom Objekt selbst erzeugt (über self.create())
                # if 'gdoc_id' in _doc.config['Document']:
                # try:
                # self.gdocid = int(_doc.config['Document']['gdoc_id'])
                # except:
                # self.gdocid = None
            if 'DOCID' in _doc.config['Document']:
                _id = _doc.config['Document']['DOCID']
                if len(_id) > 0:
                    if '.' in _id:
                        _doc.arcid = int(_id.split('.')[0])
                        _doc._docid = int(_id.split('.')[1])
                    else:
                        _doc._docid = int(_id)
                else:
                    _id = None
                    _doc._docid = None
            if 'Orginal' in _doc.config['Document']:
                _doc.docloc = _doc.config['Document']['Orginal']
            if 'Kopie' in _doc.config['Document']:
                _doc.extloc = _doc.config['Document']['Kopie']
            if 'Quelle' in _doc.config['Document']:
                _doc.oriloc = _doc.config['Document']['Quelle']
            if 'Volltext' in _doc.config['Document']:
                _doc.txtloc = _doc.config['Document']['Volltext']
            if 'Schlagworte' in _doc.config['Document']:
                _doc.tagloc = _doc.config['Document']['Schlagworte']
            if 'Benutzer' in _doc.config['Document']:
                _doc.user = _doc.config['Document']['Benutzer']
            if 'Status' in _doc.config['Document']:
                if _doc.version == 36:
                    stat_as_string = _doc.config['Document']['Status']
                    if stat_as_string:
                        for stat in stat_as_string.split(";"):
                            try:
                                _doc.states[int(stat.split(":")[0])] = int(stat.split(":")[1])
                            except:
                                pass
                else:
                    try:
                        _doc.status = (
                            int(_doc.config['Document']['Status']) if _doc.config['Document']['Status'] else 0)
                    except:
                        _doc.status = 0
            if 'StatStr' in _doc.config['Document']:
                _doc.statstr = _doc.config['Document']['StatStr']
            if 'OCR_Typ' in _doc.config['Document']:
                _doc.ocrtype = _doc.config['Document']['OCR_Typ']
            if 'OCR_QF' in _doc.config['Document']:
                _doc.ocrqf = _doc.config['Document']['OCR_QF']
            if 'Herkunft' in _doc.config['Document']:
                _doc.origin = _doc.config['Document']['Herkunft']
            if 'Dateiname' in _doc.config['Document']:
                _doc.filename = _doc.config['Document']['Dateiname']
            if 'type' in _doc.config['Document']:
                _doc.filetype = _doc.config['Document']['type']
            if 'Filter' in _doc.config['Document']:
                _doc.filter = _doc.config['Document']['Filter'].split(";")

        if 'Task' in _doc.config:
            if 'Filter' in _doc.config['Task']:
                _doc.target_date = _doc.config['Task']['Termin']
            if 'Alarm' in _doc.config['Task']:
                _doc.alarm = _doc.config['Task']['Alarm']
            if 'Bearbeiter' in _doc.config['Task']:
                _doc.editor = _doc.config['Task']['Bearbeiter']
            if 'Notiz' in _doc.config['Task']:
                _doc.info = _doc.config['Task']['Notiz']
            if 'TimerOptionen' in _doc.config['Task']:
                _doc.timeroptions = _doc.config['Task']['TimerOptionen']

        if 'Hash' in _doc.config:
            if 'SHA1' in _doc.config['Hash']:
                _doc.sha1 = _doc.config['Hash']['SHA1']

        if 'SVN' in _doc.config:
            if 'SVN_Link' in _doc.config['SVN']:
                _doc.svn_link = _doc.config['SVN']['SVN_Link']
            if 'SVN_Version' in _doc.config['SVN']:
                _doc.svn_version = _doc.config['SVN']['SVN_Version']
            if 'SVN_Revision' in _doc.config['SVN']:
                _doc.svn_revision = _doc.config['SVN']['SVN_Revision']
            if 'SVN_Datum' in _doc.config['SVN']:
                _doc.svn_date = _doc.config['SVN']['SVN_Datum']
            if 'SVN_Info' in _doc.config['SVN']:
                _doc.svn_info = _doc.config['SVN']['SVN_Info']

        if 'Referenz' in _doc.config:
            if 'Anzahl' in _doc.config['Referenz']:
                _anz = _doc.config['Referenz']['Anzahl']
                for _ind in range(1, int(_anz)):
                    if 'Referenz%d' % _ind in _doc.config['Referenz']:
                        _doc.links.append(_doc.config['Referenz']['Referenz%d' % _ind])

        if 'Zusatzfelder' in _doc.config:
            if 'Felder' in _doc.config['Zusatzfelder']:
                _anz = _doc.config['Zusatzfelder']['Felder']
                for _ind in range(1, int(_anz) + 1):
                    _add = bitfarm_doc.BitfarmAddfield()
                    if 'ZusTitel%d' % _ind in _doc.config['Zusatzfelder']:
                        _add.name = _doc.config['Zusatzfelder']['ZusTitel%d' % _ind]
                    if 'ZusFeld%d' % _ind in _doc.config['Zusatzfelder']:
                        _add.add = _doc.config['Zusatzfelder']['ZusFeld%d' % _ind]
                    if 'ZusWert%d' % _ind in _doc.config['Zusatzfelder']:
                        if ';' in _doc.config['Zusatzfelder']['ZusWert%d' % _ind]:
                            _add.value = _doc.config['Zusatzfelder']['ZusWert%d' % _ind].split(';')
                        else:
                            _add.value = _doc.config['Zusatzfelder']['ZusWert%d' % _ind]

                    _doc.fields[_add.name] = _add

        if 'Parameter' in _doc.config:
            for item in _doc.config['Parameter'].items():
                _doc.params[item[0]] = item[1]

        _doc.__tempname = os.path.splitext(jobname)[0]
        _doc.__queue = os.path.dirname(jobname)
        _doc.path = jobname

    def write_to_file(self, _doc, jobname=None, usestatus35=False):
        """
        Schreibt das Dokument in eine Jodatei.
        Der Dateiname ist in self.tempname hinterlegt.
        Dynamisch hinzugefügte Werte werden unter der Section [Document]ergänzt
        Wenn keine entsprechenden Zusatzfelder in der Job-Datei vorhanden sind,
        so werden diese unter der Section [Document] ergänzt
        Zusatzfeldinhalte in Form einer Liste werden mit Semikolon getrennt geschrieben
        @param jobname: (optional) der Pfad auf die .job-Datei. Wird nur ein Basisname übergeben,
        wird die Endung .job angehangen. Wird nichts übergeben, so wird self.tempname ausgelesen.
        """
        if _doc is None:
            return False

        if not jobname:
            if _doc.tempname:
                jobname = _doc.tempname
            else:
                self.log.debug('kein Job.tempname vorhanden')
                return False

        if os.path.splitext(jobname)[1] == '':
            jobname += '.job'

        if not _doc.config:
            _doc.config = ConfigObj(list_values=False, write_empty_values=_doc.write_empty_values, encoding='cp1252',
                                    default_encoding='cp1252')

        _doc.config.filename = jobname

        if _doc.version in (35, 30, 3.0, 3.5, "35", "30", "3.0", "3.5"):
            if len(_doc.statusdict) > 0:
                _doc.status = 0
                _ind = 1
                for val in _doc.statusdict.values():
                    if val:
                        _doc.status |= int(2 ** _ind)
                    _ind += 1

            _doc.status |= (1 if _doc.in_progress else 0)
        else:
            # 3.6 er Status ->
            _doc.states = {}
            _doc.status = 0
            _ind = 1
            for state in _doc.statefields.items():
                if state[1].value:
                    _doc.status |= int(2 ** _ind)
                    if state[1].indexcolumnid in _doc.states:
                        _doc.states[state[1].indexcolumnid] |= state[1].bitmask
                    else:
                        _doc.states[state[1].indexcolumnid] = state[1].bitmask
                _ind += 1

            if 0 in _doc.states:
                _doc.states[0] |= (1 if _doc.in_resub else 0)
            else:
                _doc.states[0] = (1 if _doc.in_resub else 0)
            _doc.status |= (1 if _doc.in_progress else 0)

        if _doc.version or _doc.write_empty_values:
            if 'Version' not in _doc.config:
                _doc.config['Version'] = {}
            _doc.config['Version']['Version'] = (_doc.version if _doc.version else '')

        if _doc.archive or _doc.tblname or _doc.profile or _doc.arcid or _doc.write_empty_values:
            if 'Archiv' not in _doc.config:
                _doc.config['Archiv'] = {}
            _doc.config['Archiv']['Profil'] = (_doc.profile if _doc.profile else '')
            _doc.config['Archiv']['Name'] = (_doc.archive if _doc.archive else '')
            _doc.config['Archiv']['Tabelle'] = (_doc.tblname if _doc.tblname else '')
            _doc.config['Archiv']['Arcid'] = (_doc.arcid if _doc.arcid else '')

        if 'Document' not in _doc.config:
            _doc.config['Document'] = {}
        _docid = ''
        if _doc.version in (35, 30, 3.0, 3.5, "35", "30", "3.0", "3.5"):
            if _doc._docid or _doc.write_empty_values:
                if _doc.arcid and _doc._docid:
                    _docid = "%d.%d" % (_doc.arcid, _doc._docid)
                else:
                    _docid = str(_doc._docid if _doc._docid else '')
                _doc.config['Document']['DOCID'] = (_docid if _docid else '')
        elif _doc.version in (40, 4.0, 3.6, 36, "36", "40", "3.6", "4.0"):
            if _doc.arcid and _doc.gdocid:
                _docid = "%d.%d" % (_doc.arcid, _doc.gdocid)
            elif _doc.gdocid:
                _docid = "%d.%d" % (0, _doc.gdocid)
            _doc.config['Document']['DOCID'] = (_docid if _docid else '')
        if _doc.gdocid or _doc.write_empty_values:
            _doc.config['Document']['gdoc_id'] = (_doc.gdocid if _doc.gdocid else '')
        if _doc.title or _doc.write_empty_values:
            title_enc = _doc.title.encode('cp1252', 'replace')
            title_enc = title_enc.decode('cp1252')
            _doc.config['Document']['Titel'] = (title_enc if title_enc else '')
        if _doc.docloc or _doc.write_empty_values:
            _doc.config['Document']['Orginal'] = (_doc.docloc if _doc.docloc else '')
        if _doc.extloc or _doc.write_empty_values:
            _doc.config['Document']['Kopie'] = (_doc.extloc if _doc.extloc else '')
        if _doc.oriloc or _doc.write_empty_values:
            _doc.config['Document']['Quelle'] = (_doc.oriloc if _doc.oriloc else '')
        if _doc.user or _doc.write_empty_values:
            _doc.config['Document']['Benutzer'] = (_doc.user if _doc.user else '')
        if _doc.userid or _doc.write_empty_values:
            _doc.config['Document']['userid'] = (_doc.userid if _doc.userid else '')

        if _doc.ftx:
            if isinstance(_doc.ftx, bitfarm_doc.BitfarmFulltext):
                _doc.config['Document']['Volltext'] = (_doc.ftx.text if _doc.ftx.text else '')
            else:
                _doc.config['Document']['Volltext'] = (_doc.ftx if _doc.ftx else '')
        elif _doc.txtloc:
            _doc.config['Document']['Volltext'] = _doc.txtloc
        elif _doc.write_empty_values:
            _doc.config['Document']['Volltext'] = ""

        if _doc.status or _doc.states or _doc.write_empty_values:
            if _doc.version in (35, 30, 3.0, 3.5, "35", "30", "3.0", "3.5") or usestatus35:
                _doc.config['Document']['Status'] = (_doc.status if _doc.status else '')
            elif _doc.version in (40, 4.0, 3.6, 36, "36", "40", "3.6", "4.0"):
                stat_as_string = ";".join("%d:%d" % (val, key) for val, key in _doc.states.items())
                _doc.config['Document']['Status'] = stat_as_string if stat_as_string != '0:0' else ''

        if _doc.statstr or _doc.write_empty_values:
            _doc.config['Document']['StatStr'] = (_doc.statstr if _doc.statstr else '')
        if _doc.ocrqf or _doc.write_empty_values:
            _doc.config['Document']['OCR_QF'] = (_doc.ocrqf if _doc.ocrqf else -1)
        if _doc.ocrtype or _doc.write_empty_values:
            _doc.config['Document']['OCR_Typ'] = (_doc.ocrtype if _doc.ocrtype else 'OMP')
        if _doc.tags or _doc.write_empty_values:
            _doc.config['Document']['Schlagworte'] = (";".join(_doc.tags) if _doc.tags else '')
        if _doc.date or _doc.write_empty_values:
            if type(_doc.date) is str:
                _doc.config['Document']['Datum'] = (_doc.date if _doc.date else '')
            else:
                _doc.config['Document']['Datum'] = (_doc.date.strftime('%Y-%m-%d %H:%M:%S') if _doc.date else '')
        if _doc.filter or _doc.write_empty_values:
            if type(_doc.filter) is list:
                docfilter = ";".join(_doc.filter)
            else:
                docfilter = _doc.filter
            _doc.config['Document']['Filter'] = (docfilter if docfilter else '')
        #        if _doc.version  in (35,  30, 3.0,  3.5,  "35", "30",  "3.0",  "3.5"):
        if _doc.pages or _doc.write_empty_values:
            _doc.config['Document']['Pages'] = (_doc.pages if _doc.pages else '')
        #        else:
        #            if _doc.pagecount or _doc.write_empty_values:
        #                _doc.config['Document']['Pages'] = (_doc.pagecount if _doc.pagecount else '')
        if _doc.sha1 or _doc.write_empty_values:
            if 'Hash' not in _doc.config:
                _doc.config['Hash'] = {}
            _doc.config['Hash']['SHA1'] = (_doc.sha1 if _doc.sha1 else '')
        if len(_doc.links) > 0 or _doc.write_empty_values:
            if 'Referenz' not in _doc.config:
                _doc.config['Referenz'] = {}
            _doc.config['Referenz']['Anzahl'] = len(_doc.links)
            for _ind in range(len(_doc.links)):
                _doc.config['Referenz']['Referenz%d' % (_ind + 1)] = _doc.links[_ind]
        if _doc.svn_link or _doc.svn_version or _doc.svn_revision or _doc.svn_date or _doc.svn_info or _doc.write_empty_values:
            if 'SVN' not in _doc.config:
                _doc.config['SVN'] = {}
            if _doc.svn_link or _doc.write_empty_values:
                _doc.config['SVN']['SVN_Link'] = (_doc.svn_link if _doc.svn_link else '')
            if _doc.svn_version or _doc.write_empty_values:
                _doc.config['SVN']['SVN_Version'] = (_doc.svn_version if _doc.svn_version else '')
            if _doc.svn_revision or _doc.write_empty_values:
                _doc.config['SVN']['SVN_Revision'] = (_doc.svn_revision if _doc.svn_revision else '')
            if _doc.svn_date or _doc.write_empty_values:
                _doc.config['SVN']['SVN_Datum'] = (_doc.svn_date if _doc.svn_date else '')
            if _doc.svn_info or _doc.write_empty_values:
                _doc.config['SVN']['SVN_Info'] = (_doc.svn_info if _doc.svn_info else '')

        if _doc.target_date or _doc.alarm or _doc.editor or _doc.info or _doc.timeroptions or _doc.write_empty_values:
            if 'Tasks' not in _doc.config:
                _doc.config['Tasks'] = {}
            if _doc.target_date or _doc.write_empty_values:
                _doc.config['Tasks']['Termin'] = (_doc.target_date if _doc.target_date else '')
            if _doc.alarm or _doc.alarm_date or _doc.write_empty_values:
                if _doc.alarm_date is not None:
                    _doc.config['Tasks']['Alarm'] = (_doc.alarm_date if _doc.alarm_date else '')
                else:
                    alarm = None
                    if _doc.alarm is not None:
                        alarm = _doc.target_date + datetime.timedelta(_doc.alarm)
                    _doc.config['Tasks']['Alarm'] = (alarm if alarm else '')
            if _doc.editor or _doc.write_empty_values:
                _doc.config['Tasks']['Bearbeiter'] = (_doc.editor if _doc.editor else '')
            if _doc.info or _doc.write_empty_values:
                _doc.config['Tasks']['Notiz'] = (_doc.info if _doc.info else '')
            if _doc.timeroptions or _doc.write_empty_values:
                _doc.config['Tasks']['TimerOptionen'] = (_doc.timeroptions if _doc.timeroptions else '0')

        if len(_doc.fields) > 0 or _doc.write_empty_values:
            # if 'Zusatzfelder' not in _doc.config:
            _doc.config['Zusatzfelder'] = {}
            if _doc.version in (3, 35, '3.0'):
                _doc.config['Zusatzfelder']['Felder'] = len(_doc.fields)
                _ind = 1
                for field in _doc.fields.values():
                    _doc.config['Zusatzfelder']['ZusTitel%d' % _ind] = field.name
                    _doc.config['Zusatzfelder']['ZusFeld%d' % _ind] = field.add
                    if field.value:
                        val = field.value
                        if type(val) is list:
                            if field.type in (7, 8):
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = ";".join(
                                    [str(x).replace(".", ",") for x in val])
                            else:
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = ";".join(val)
                        else:
                            # Datums-Felder unterscheiden sich in 3.5 und 3.6 job-Files
                            # 3.5 Datum dd.mm.yyyy, 3.6 Datum dd.mm.yyyy HH:MM:SS
                            if field.type == 1:
                                if type(val) is str:
                                    _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = val
                                else:
                                    _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = val.strftime('%d.%m.%Y')
                            elif field.type in (7, 8):
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = str(val).replace(".", ",")
                            else:
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = val
                    else:
                        _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = ''
                    _ind += 1
            else:  # Version >= 3.6
                for field in [x for x in _doc.fields.values()]:
                    if field.type == 13 and field.values_query is not None:
                        config = bfaxml.xml_to_object(field.values_query)
                        if config:
                            for col in config.columns:
                                if col.fieldname not in _doc.fields:
                                    gridcol = bitfarm_doc.BitfarmAddfield()
                                    gridcol.name = col.name
                                    gridcol.add = col.fieldname
                                    gridcol.type = 14
                                    _doc.fields[gridcol.add] = gridcol

                        # gleichzeitig Werte im Gridfeld UND in den Spalten ist keine gute Idee, weil sich sonst die Werte beim Import mit DBImp
                        # verdoppeln. Hier gilt entweder/oder

                        # if field.value:
                        #     vals = {}
                        #     rows = field.value.split('|')
                        #     header = rows.pop(0)
                        #     header = header.split(';')
                        #     for idx in range(len(header)):
                        #         for row in rows:
                        #             if header[idx] in vals:
                        #                 vals[header[idx]].append(row.split(';')[idx])
                        #             else:
                        #                 vals[header[idx]]= [row.split(';')[idx]]
                        #         if header[idx] in _doc.fields:
                        #             _doc.fields[header[idx]].value = ';'.join(vals[header[idx]])

                _doc.config['Zusatzfelder']['Felder'] = len(_doc.fields)
                _ind = 1
                for field in _doc.fields.values():
                    _doc.config['Zusatzfelder']['ZusTitel%d' % _ind] = field.name
                    _doc.config['Zusatzfelder']['ZusFeld%d' % _ind] = field.add
                    if field.value:
                        val = field.value
                        if type(val) is list:
                            if field.type in (7, 8):
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = ";".join(
                                    [str(x).replace(".", ",") for x in val])
                            else:
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = ";".join(val)
                        else:
                            if field.type == 1:
                                if type(val) is str:
                                    _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = val
                                else:
                                    _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = val.strftime('%d.%m.%Y')
                            elif field.type in (7, 8):
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = str(val).replace(".", ",")
                            else:
                                if type(val) is str:
                                    self.log.debug(val)
                                    self.log.debug(type(val))
                                    val_enc = val.encode('cp1252', 'replace')
                                    val = val_enc.decode('cp1252')
                                _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = val
                    else:
                        _doc.config['Zusatzfelder']['ZusWert%d' % _ind] = ''
                    _ind += 1

        if len(_doc.params) > 0 or _doc.write_empty_values:
            if 'Parameter' not in _doc.config:
                _doc.config['Parameter'] = {}
            for item in _doc.params.items():
                _doc.config['Parameter'][item[0]] = item[1]

        _doc.config.write_empty_values = True
        try:
            _doc.config.write()
        except:
            self.log.exception("Fehler beim Jobfile" + _doc.config.filename)
            return False
        else:
            self.log.debug("Jobfile geschrieben: " + _doc.config.filename)
            return True

    def create_new_template(self, arcid=None, filename=None, profile=None, doc=None, fpath=None):
        """
        Methode zur Erstellung einer Template-Datei (.tpl).
        Die arcid und der Dateiname kann übergeben werden, ansonsten
        werden die in der Instanz hinterlegten Werte genommen (self.arcid und self.tempname)
        Wurde nichts übergeben oder hinterlegt wird eine Exception geworfen.
        @param arcid: arcid des Archives
        @param filename: Pfad der Template-Datei, wird nur ein Basisname übergeben wird .tpl angehangen.
        @param profile: das Profil des erzeugten Templates (optional, wenn nichts angegeben wird, wird das Profil der Main.Connectionbenutzt)
        """
        _ret = True

        deldoc = False

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if not arcid:
            if _con.undistributed_arcid:
                arcid = int(_con.undistributed_arcid)
            else:
                if _con.undistributed:
                    try:
                        if self.version == 35:
                            sql = _con.sys_arc.select(_con.sys_arc.c.ARC_Name == _con.undistributed)
                        else:
                            sql = _con.sys_arc.select(_con.sys_arc.c.arcname == _con.undistributed)
                        try:
                            result = _db.execute(sql).fetchone()
                        except:
                            self.log.exception(
                                u'Das Template konnte nicht erstellt werden, Archiv unverteilt konnte nicht ermittelt werden')
                        else:
                            if result:
                                arcid = result[0]
                    except:
                        arcid = None

        if not arcid and not _con:
            return False
        try:
            if doc is None:
                deldoc = True
                doc = bitfarm_doc.BitfarmDocument()
            doc.version = self.version
            result = None
            if self.version == 35:
                sql = 'select * from sys_arc where arc_id = %s' % arcid
            else:
                sql = 'select * from sys_arc where arcid = %s' % arcid
            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception(u"Fehler bei Abfrage der sys_arc")

            if result:
                self.log.debug(list(result))
                doc.arcid = result[0]
                doc.archive = result[2]
                doc.tblname = result[3]
            else:
                self.log.error('Arcid %s nicht gefunden' % arcid)
                return False

            if fpath is not None:
                filename = os.path.join(fpath, doc.archive)
            elif not filename:
                filename = self.__tempname

            if os.path.splitext(filename)[1] == '':
                filename += '.tpl'

            if doc.fields is None or len(doc.fields) == 0:
                doc.fields = self.get_addfields(_con, _db, doc, with_gridcolumns=True)
            else:
                new_fields = self.get_addfields(_con, _db, doc, with_gridcolumns=True)
                for new_field in new_fields.values():
                    for field in doc.fields.values():
                        if new_field.name == field.name:
                            new_field.value = field.value
                doc.fields = new_fields

            doc.write_empty_values = True
            _ret = self.write_to_file(doc, filename)

        except:
            self.log.exception('Das Template konnte nicht erstellt werden arcid=%s filename=%s' % (arcid, filename))
            _ret = False
            if deldoc:
                del doc
        return _ret

    def get_from_dir(self, _doc, _path=None):
        """
        Versucht die erste verfügbare Jobdatei aus der Queue zu ermitteln. (<= Version 3.5)
        Der Pfad wird unter _doc.tempname zurückgegeben.
        Optional kann der Pfad auf die Queue angegeben werden.
        Existiert der Queuepfad nicht wird die Standardqueue (programm\queue) genommen
        """
        if _doc is None:
            return None
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
        else:
            _con = self.con[_doc.profile]

        if _path:
            if os.path.exists(_path):
                tmp = bitfarm_doc.GetFirstFile(os.path.join(_path), "job")
                if tmp:
                    _doc.tempname = os.path.join(_path, os.path.splitext(tmp)[0])
            else:
                tmp = bitfarm_doc.GetFirstFile(os.path.join(_con.queue), "job")
                if tmp:
                    _doc.tempname = os.path.join(_con.queue, os.path.splitext(tmp)[0])
        else:
            tmp = bitfarm_doc.GetFirstFile(os.path.join(_con.queue), "job")
            if tmp:
                _doc.tempname = os.path.join(_con.queue, os.path.splitext(tmp)[0])

        if _doc.tempname:
            self.log.debug("Job gefunden: %s" % _doc.tempname)
            self.read_from_file(_doc)

    def create_new_jobfile(self, _doc, _path=None):
        """
        Erstellt eine neue Jobdatei  ( <= Version 3.5)
        Optional kann ein Pfad auf eine Job-Datei oder ein Queue-Verzeichnis angegeben werden.
        Existiert die Jobdatei schon, so wird der nächste verfügbare Name ermittelt.
        In einer Queue wird der nächste verfügbare Name ermittelt.
        Wird nichts übergeben, so wird der nächste verfügbare Name in der Standardqueue ermittelt
        """
        if _doc is None:
            return None
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
        else:
            _con = self.con[_doc.profile]

        if _path:
            _queue = os.path.split(_path)[0]
            if os.path.exists(_queue):
                if os.path.split(_path)[1] != '':

                    if os.path.exists(_path + '.job'):
                        _doc.tempname = bitfarm_doc.GetFiFo(_queue, "temp")
                    else:
                        _doc.tempname = _path
                else:
                    _doc.tempname = bitfarm_doc.GetFiFo(_queue, "temp")
            else:
                _doc.tempname = bitfarm_doc.GetFiFo(_con.queue, "temp")
        else:
            _doc.tempname = bitfarm_doc.GetFiFo(_con.queue, "temp")

        if _doc.tempname:
            self.log.debug("neuen Job erstellt: %s" % _doc.tempname)
            self.write_to_file(_doc)

    def sortto(self, _doc, _arc):
        """
        Wechselt das zugrundeliegende Archiv eines Dokuments. Dabei werden allerdings keine anderen Dokumenteneigenschaften
        gespeichert. Soll das Dokument verschoben und dabei gespeichert werden, dann verwendet man move.

        Version 35:
        In Version 35 sollte dies nur zu einer frühen Phase während des Imports mit einem leeren Dokument ausgeführt werden,
        da nur der Datenbankeintrag in der Archivtabelle umgezogen wird und die sys_gdocid aktualisiert wird (Version35).
        Weitere Metainformationen wie Schlagwörter, History, sowie Verknüpfungen, Wiedervorlagen werden nicht umgezogen, dafür bitte doc.move
        verwenden.
        sortto sollte daher nur zwischen einem doc.create() und doc.write() benutzt werden, niemals nach einem doc.write()
        """

        self.log.debug('sortto archive = %s' % _arc)
        if self.version in (3, 35, '3.0'):
            return self.__sortto_v3(_doc, _arc)
        if self.version == 36:
            return self.__sortto_v36(_doc, _arc)
        else:
            return self.__sortto_v4(_doc, _arc)

    def __sortto_v3(self, _doc, _arc):
        _arcid = None
        result = None

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None
        if _arc is None:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        src_arcid = _doc.arcid
        src_docid = _doc._docid
        src_arc_tblname = _doc.tblname

        if type(_arc) is int:
            _arcid = _arc
            try:
                sql = _con.sys_arc.select(_con.sys_arc.c.ARC_ID == _arcid)
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                self.log.debug(result)
            except:
                _arcid = None
        else:
            try:
                sql = _con.sys_arc.select(_con.sys_arc.c.ARC_Name.like(_arc))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                self.log.debug(result)
            except:
                _arcid = None

        if result:
            oldfields = _doc.fields
            oldstatusdict = _doc.statusdict
            self.arcid = result[0]
            self.archive = result[2]
            self.tblname = result[3]
            _doc.arcid = result[0]
            _doc.archive = result[2]
            _doc.tblname = result[3]
            s = time.perf_counter()
            if _con.arc_tbl is None:
                _con.arc_tbl = sqla.Table('arc_%s' % _doc.tblname, _con.meta, autoload=True)
            elif _con.arc_tbl.name.find(_doc.tblname) == -1:
                _con.arc_tbl = sqla.Table('arc_%s' % self.tblname, _con.meta, autoload=True)
            self.log.debug("sqla.table autoload %s" % (time.perf_counter() - s))
            _doc.fields = self.get_addfields(_con, _db, _doc)
            _doc.statusdict = self.get_statefields(_con, _db, _doc)

            for fieldname in oldfields:
                if fieldname in _doc.fields:
                    _doc.fields[fieldname].value = oldfields[fieldname].value  # TODO: ggf Typ prüfen und konverieren
                else:
                    _doc.params[fieldname] = oldfields[fieldname].value

            for key in oldstatusdict.keys():
                if key in _doc.statusdict:
                    _doc.statusdict[key] = oldstatusdict[key]

            # haben wir eine docid, dann gibt es da wohl einen eintrag den wir entfernen müssen
            if src_docid:
                sql = 'delete from arc_%s where doc_id = %s' % (src_arc_tblname, src_docid)
                try:
                    try:
                        _db.execute(sql)
                    except:
                        _con.try_again(sql=sql, db=_db, )
                except:
                    self.log.exception('sortto: konnte alten Eintrag nicht entfernen')

                # und einen neuen anlegen
                vals = {_con.arc_tbl.c.USR_NAME: _doc.user,
                        _con.arc_tbl.c.TStamp: _doc.date,
                        _con.arc_tbl.c.gdoc_id: _doc.gdocid,
                        _con.arc_tbl.c.DISP_NAME: _doc.title
                        }

                sql = _con.arc_tbl.insert(values=vals)
                #        sql = _con.arc_tbl.insert().values(dict
                self.log.debug(sql)
                try:
                    result = _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)
                _doc._docid = result.inserted_primary_key[0]

            # sys_gdocid aktualisieren
            if _doc.gdocid:
                try:
                    sql = _con.sys_gdocid.update().where(_con.sys_gdocid.c.gdoc_id == _doc.gdocid).values(
                        arc_id=_doc.arcid, doc_id=_doc._docid)
                    try:
                        _db.execute(sql)
                    except:
                        _con.try_again(sql=sql, db=_db)
                except:
                    self.log.exception('sortto: konnte sys_gdocid nicht aktualisieren')
                    return None
            return True
        else:
            return None

    def __sortto_v36(self, _doc, _arc):
        _arcid = None

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if type(_arc) is int:
            _arcid = _arc
            try:
                sql = _con.sys_arc.select(_con.sys_arc.c.arcid == _arcid)
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                self.log.debug(result)
            except:
                return False
        else:
            try:
                sql = _con.sys_arc.select(_con.sys_arc.c.arcname.like(_arc))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                self.log.debug(result)
            except:
                return False
        if result:
            _arcid = result[0]

        if _doc.arcid == _arcid:
            return True  # brauchen nichts machen, ist schon da, geben True zurück, damit nachfolgende Funktionen weitermachen

        _doc.arcid = _arcid

        try:
            sql = sqla.update(_con.sys_doc).where(_con.sys_doc.c.gdocid == _doc.gdocid).values({"arcid": _doc.arcid})
            try:
                result = _db.execute(sql)
            except:
                result = _con.try_again(sql=sql, db=_db, fetchone=False)
            self.log.debug(result)
            if result is not None:
                return True
        except:
            self.log.exception(u'doc.sortto: Dokument konnte nicht umsortiert werden')
            return False

    def __sortto_v4(self, _doc, _arc):
        return None

    def move(self, _doc, _arc):
        """
        Verschiebt das Dokuments in ein anderes Archiv. Dabei werden alle aktuellen Eigenschaften des Dokuments gespeichert.
        Einschränkungen gibts bei v35, siehe Docstring zu sortto.
        """
        self.log.debug('move archive = %s' % _arc)
        if self.version in (3, 35, '3.0'):
            return self.__move_v3(_doc, _arc)
        if self.version == 36:
            return self.__move_v36(_doc, _arc)
        else:
            return self.__move_v4(_doc, _arc)

    def __move_v3(self, _doc, _arc):
        _arcid = None
        _msg = u''

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        src_arcid = _doc.arcid
        src_docid = _doc._docid
        src_arc_tblname = _doc.tblname
        src_arcname = _doc.archive

        if self.__sortto_v3(_doc, _arc):  # wechsel des Archives erfolgreich, dann Metadaten umziehen
            try:
                if not self.__write_v3(_doc):
                    self.log.warning(u'doc.move: Dokument konnte nicht geschrieben werden')
                # updaten
                _msg = u"Fehler beim Aktualisieren der sys_svn"
                sql = "update sys_svn set arc_id = %s, doc_id = %s where arc_id = %s and doc_id = %s" % (
                    _doc.arcid, _doc._docid, src_arcid, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Aktualisieren der sys_doclnk"
                sql = "update sys_doclnk set arc_id = %s, doc_id = %s where arc_id = %s and  doc_id = %s" % (
                    _doc.arcid, _doc._docid, src_arcid, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Aktualisieren der sys_timer"
                sql = "update sys_timer set arc_id = %s, doc_id = %s where arc_id = %s and doc_id = %s" % (
                    _doc.arcid, _doc.gdocid, src_arcid, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Aktualisieren der sys_docxs"
                sql = "update sys_docxs set arc_id = %s, doc_id = %s where arc_id = %s and doc_id = %s" % (
                    _doc.arcid, _doc._docid, src_arcid, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                # übertragen
                _msg = u"Fehler beim Übertragen der Schlagwörter"
                sql = "insert into slw_%s (doc_id, doc_slw) select %s as docid, doc_slw from slw_%s where doc_id = %s" % (
                    _doc.tblname, _doc._docid, src_arc_tblname, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Übertragen der History"
                sql = "insert into hst_%s (doc_id, usr_id, tstamp, task, task_info) select %s as docid, usr_id, tstamp, task, task_info from hst_%s where doc_id = %s" % (
                    _doc.tblname, _doc._docid, src_arc_tblname, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Übertragen der Kommentare"
                sql = "insert into com_%s (doc_id,DISP_NAME, USR_NAME, USR_ID, TStamp, COM_TXT) select %s as docid,DISP_NAME, USR_NAME, USR_ID, TStamp, COM_TXT from com_%s where doc_id = %s" % (
                    _doc.tblname, _doc._docid, src_arc_tblname, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                # delete
                _msg = u"Fehler beim Löschen der Schlagwörter"
                sql = "delete from slw_%s where doc_id = %s" % (src_arc_tblname, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Löschen der History"
                sql = "delete from hst_%s where doc_id = %s" % (src_arc_tblname, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)

                _msg = u"Fehler beim Löschen der Kommentare"
                sql = "delete from com_%s where doc_id = %s" % (src_arc_tblname, src_docid)
                try:
                    _db.execute(sql)
                except:
                    result = _con.try_again(sql=sql, db=_db)
            except:
                self.log.exception(u'doc.move: Fehler beim Verschieben der Metadaten %s' % _msg)

            try:
                # abschliessende aktionen (disp_name, verknüpfung, history)

                _msg = u"Fehler beim automatischen Anlegen von Verknüpfungen"
                for key in _doc.fields:
                    if _doc.fields[key].options & 2 > 0:
                        if _doc.fields[key].value and type(_doc.fields[key].value) is str:
                            for link in _doc.fields[key].value.split(";"):
                                sql = "insert into sys_doclnk (arc_id, doc_id, usr_id, ref_type, ref_name) " \
                                      "values (%s, %s, %s, 1, '%s')" % (_doc.arcid, _doc._docid, 0, link)
                                try:
                                    _db.execute(sql)
                                except:
                                    result = _con.try_again(sql=sql, db=_db)

                sql = "select DefDocname from sys_arc where arc_id = %s and defdocname <> ''" % _doc.arcid
                try:
                    result = _db.execute(sql).fetchall()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchall=True)

                if result:
                    defdocname = result[0][0]
                    _msg = u"Fehler beim Setzen des Standard-Dokumentennamens"
                    sql = "select Disp_Name from arc_%s where doc_id = %s" % (_doc.tblname, _doc._docid)
                    try:
                        result = _db.execute(sql).fetchall()
                    except:
                        result = _con.try_again(sql=sql, db=_db, fetchall=True)
                    if result:
                        if result[0][0] == "Dokument" or result[0][0] == "Document" or result[0][0] == '':
                            sql = "update arc_%s set Disp_Name = '%s' where doc_id = %s" % (
                                _doc.tblname, defdocname, _doc._docid)
                            try:
                                _db.execute(sql)
                            except:
                                result = _con.try_again(sql=sql, db=_db)

                self.__write_history_v3(_doc, u'Dokument [%s] aus Archiv %s nach %s verschoben' % (
                    _doc.gdocid, src_arcname, _doc.archive), 1, 11)
            except:
                self.log.exception(u'doc.move: Fehler bei disp_name, verknüpfung, history %s' % _msg)
            else:
                return True

        else:
            self.log.warning(u'doc.move: Dokument konnte nicht umgelagert werden')
            return None

    def __move_v36(self, _doc, _arc):
        _arcid = None

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _arc is None:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        if type(_arc) is int:
            _arcid = _arc
            try:
                sql = _con.sys_arc.select(_con.sys_arc.c.arcid == _arcid)
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                self.log.debug(result)
            except:
                return False
        else:
            try:
                sql = _con.sys_arc.select(_con.sys_arc.c.arcname.like(_arc))
                try:
                    result = _db.execute(sql).fetchone()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchone=True)
                self.log.debug(result)
            except:
                return False
        if result:
            _arcid = result[0]

        if _doc.arcid == _arcid:
            return True  # brauchen nichts machen, ist schon da, geben True zurück, damit nachfolgende Funktionen weitermachen

        _doc.arcid = _arcid

        try:
            if not self.__write_v36(_doc):
                self.log.warning(u'doc.move: Dokument konnte nicht verschoben werden')
                return False
            try:
                # abschliessende aktionen (disp_name, verknüpfung, history)

                _msg = u"Fehler beim automatischen Anlegen von Verknüpfungen"
                for key in _doc.fields:
                    if _doc.fields[key].options & 2 > 0:
                        if _doc.fields[key].value and type(_doc.fields[key].value) is str:
                            for link in _doc.fields[key].value.split(";"):
                                sql = "insert into sys_doclink (gdocid, userid, reftype, refname) " \
                                      "values (%s, %s, 1, '%s')" % (_doc.gdocid, 0, link)
                                try:
                                    _db.execute(sql)
                                except:
                                    _con.try_again(sql=sql, db=_db)

                sql = "select defdocname from sys_arc where arcid = %s and defdocname <> ''" % _doc.arcid
                try:
                    result = _db.execute(sql).fetchall()
                except:
                    result = _con.try_again(sql=sql, db=_db, fetchall=True)

                if result:
                    defdocname = result[0][0]
                    _msg = u"Fehler beim Setzen des Standard-Dokumentennamens"
                    sql = "select docname from sys_doc where gdocid = %s" % _doc.gdocid
                    try:
                        result = _db.execute(sql).fetchall()
                    except:
                        result = _con.try_again(sql=sql, db=_db, fetchall=True)
                    if result:
                        if result[0][0] == "Dokument" or result[0][0] == "Document" or result[0][0] == '':
                            sql = "update sys_doc set docname = '%s' where gdocid = %s" % (defdocname, _doc._docid)
                            try:
                                _db.execute(sql)
                            except:
                                _con.try_again(sql=sql, db=_db)
                                # TODO: sinnvolle Historyeinträge erstellen
                                # self.__write_history_v36(_doc, _doc.userid, 11, 'Value1', 'Value2' )
            except:
                self.log.exception(u'doc.move: Fehler bei disp_name, verknüpfung, history %s' % _msg)
            else:
                return True

        except:
            self.log.exception(u'doc.move: Dokument konnte nicht umgelagert werden')
            return False

    def __move_v4(self, _doc, _arc):
        return None

    def parse_addfields(self, _doc):
        """
        Durchläuft alle Addfields und iteriert dort über die values
        Es werden nach Variablen gesucht die jeweils von zwei % eingefasste sind.
        Dann werden alle anderen Zusatzfelder durchlaufen und passende Werte anhand des Zusatzfeldnamens
        reingeparst.
        Standardgemäß wird die Funktion intern von doc.refresh() oder doc.write() aufgerufen
        """

        for field in _doc.fields:
            if type(_doc.fields[field].value) == list:
                for i in range(len(_doc.fields[field].value)):
                    if type(_doc.fields[field].value[i]) is str:
                        if _doc.fields[field].value[i].count('%') >= 2:
                            _doc.fields[field].value[i] = self.parse_txt(_doc, _doc.fields[field].value[i])
            else:
                if type(_doc.fields[field].value) is str:
                    if _doc.fields[field].value.count('%') >= 2:
                        _doc.fields[field].value = self.parse_txt(_doc, _doc.fields[field].value)

    def parse_txt(self, _doc, txt):
        """Parst beliebige Texte nach Bezeichner die in zwei % Zeichen gesetzt sind und versucht diese durch
                entsprechende Zusatzfeldwerte oder Dokumentwerte zu ersetzen"""
        if not isinstance(_doc, bitfarm_doc.BitfarmDocument):
            return txt

        if txt is None:
            return u''

        if txt.count('%') < 2:  # es müssen wenigstens zwei % Zeichen vorhanden sein
            return txt
        parsestring = []
        # parsestring = re.findall("%(\S*)%", txt)
        parsestring = re.findall("%(.*?)%", txt)

        for string in parsestring:
            found = False
            if string == '':
                continue

            # zuerst suchen wir in den Zusatzfeldern

            for field in _doc.fields:
                if field.lower() == string.lower():
                    txt = txt.replace('%' + string + '%', _doc.fields[field].value)
                    found = True
                    break

            if found:
                continue

            # jetzt im Dokument selbst
            if string.lower() in _doc.__dict__:
                txt = txt.replace('%' + string + '%', str(_doc.__dict__[string.lower()]))
                continue  # was gefunden, dann das nächste bitte
            if '_BitfarmDocument__%s' % string.lower() in _doc.__dict__:
                txt = txt.replace('%' + string + '%', str(_doc.__dict__['_BitfarmDocument__%s' % string.lower()]))
                continue  # was gefunden, dann das nächste bitte

        return txt

    def parse_sqlstring(self, string):
        """Hilfsfunktion um Strings nach SQL-problematischen Zeichen zu parsen"""
        if string.find("\\") >= 0:
            string = string.replace("\\", '\\\\')
        if string.find("'") >= 0:
            string = string.replace("'", "\\\\'")
        if string.find("%") >= 0:
            string = string.replace("%", '%%')
        return string

    def create_arcpath(self, _doc, is_revisionsave=None):
        """
        Erzeugt einen neuen Archivablagepfad für ein Dokument. Der Pfad wird angelegt, falls nicht vorhanden.
        Kann der Pfad nicht angelegt werden, dann wird None zurückgegeben.
        Es kann optional is_revisionsave übergeben werden, ansonsten wird die
        doc.is_revisionsave Property ausgelesen, entsprechend wird der Pfad im RS- oder NRS-Zweig
        erzeugt.
        @param is_revisionsave: (optional) wählt RS- oder NRS-Pfad
        @type is_revisionsave: bool
        """

        if is_revisionsave is None:
            is_revisionsave = _doc.is_revisionsave
            if is_revisionsave is None:
                # wenn nichts übergeben wurde und is_revisionsave für das Dokument nicht gesetzt ist
                # kann kein Pfad ermittelt werden
                return None

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
        else:
            _con = self.con[_doc.profile]

        if _doc.date is None:
            _doc.date = datetime.datetime.now()

        if _doc.version in (3, 35, '3.0'):
            if is_revisionsave:
                folder = os.path.join(_con.archive_rs, _doc.archive + _doc.tblname, _doc.date.strftime('%m%Y'),
                                      _doc.date.strftime('%d'))
            else:
                folder = os.path.join(_con.archive_nrs, _doc.archive + _doc.tblname, _doc.date.strftime('%m%Y'),
                                      _doc.date.strftime('%d'))
            if not os.path.exists(folder):
                try:
                    os.makedirs(folder)
                except:
                    self.log.warning(u'Konnte Verzeichnis %s nicht erstellen' % folder)
                    return None
            return folder
        elif _doc.version == 36:
            if _doc.storageid is None:
                _doc.storageid = 0
            if _doc.storageid >= len(_con.storages):
                _doc.storageid = 0
            if is_revisionsave:
                folder = os.path.join(_con.storages[_doc.storageid], _con.archive_rs, _doc.date.strftime('%Y'),
                                      _doc.date.strftime('%m'), _doc.date.strftime('%d'))
            else:
                folder = os.path.join(_con.storages[_doc.storageid], _con.archive_nrs, _doc.date.strftime('%Y'),
                                      _doc.date.strftime('%m'), _doc.date.strftime('%d'))
            if not os.path.exists(os.path.expandvars(folder)):
                try:
                    os.makedirs(os.path.expandvars(folder))
                except:
                    self.log.warning(u'Konnte Verzeichnis %s nicht erstellen' % folder)
                    return None
            return folder
        else:
            return "unknow_version_4_path"

    def create_tempname(self, _doc):
        """
        Erzeugt einen neuen Tempnamen für ein Dokument. Der Tempname besteht aus einem Pfad auf den
        nichtrevisionsicheren Bereich und den Basisnamen des Dokuments. Der Pfad wird angelegt, falls nicht vorhanden.
        Kann der Pfad nicht angelegt werden, dann wird tempname leer gesetzt und False zurückgegeben.
        """

        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.profile is None or _doc.profile == '':
            _con = self.con.main
        else:
            _con = self.con[_doc.profile]

        if _doc.date is None:
            _doc.date = datetime.datetime.now()

        if _doc.version in (3, 35, '3.0'):
            if _doc.basename is None or _doc.basename == '':
                self.create_basename(_doc)

            _doc.tempname = os.path.join(_con.archive_nrs, _doc.archive + _doc.tblname, _doc.date.strftime('%m%Y'),
                                         _doc.date.strftime('%d'), _doc.basename)
            if not os.path.exists(os.path.dirname(_doc.tempname)):
                try:
                    os.makedirs(os.path.dirname(_doc.tempname))
                except:
                    _doc.tempname = None
                    return False
            return True
        elif _doc.version == 36:
            if _doc.basename is None or _doc.basename == '':
                self.create_basename(_doc)
            if _doc.storageid >= len(_con.storages):
                _doc.storageid = 0
            _doc.tempname = os.path.join(_con.storages[_doc.storageid], _con.archive_nrs, _doc.date.strftime('%Y'),
                                         _doc.date.strftime('%m'), _doc.date.strftime('%d'), _doc.basename)
            if not os.path.exists(os.path.dirname(os.path.expandvars(_doc.tempname))):
                try:
                    os.makedirs(os.path.dirname(os.path.expandvars(_doc.tempname)))
                except:
                    _doc.tempname = None
                    return False
            return True
        else:
            return False

    def create_basename(self, _doc):
        """Erzeugt einen neuen Basisnamen für ein Dokument. Überschreibt einen ggf. schon vorhandenen."""
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return None

        if _doc.version in (3, 35, 36, '3.0'):
            t = datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')[:-3]
            if _doc.gdocid is not None:
                _doc.basename = "%s_%s" % (t, hex(_doc.gdocid)[2:])
            else:
                _doc.basename = "%s" % t

            return True
        else:
            return False

    def create_tmploc(self, _doc, _ext):
        """erzeugt einen tmploc-Path für die Verarbeitung und prüft ob dieser Pfad nicht belegt ist, ansonsten wird ein
        Index angehangen. Der tmploc-Path wird am Ende einer Taskverarbeitung gelöscht. Werden während einer Jobverarbeitung mehrere
        tmploc-Pfade erzeugt, so wird immer der vorherige gelöscht"""
        if _doc.tmploc is not None and _doc.tmploc != '':
            if os.path.exists(os.path.expandvars(_doc.tmploc)):
                try:
                    os.remove(os.path.expandvars(_doc.tmploc))
                except:
                    self.log.warning(u'Konnte bestehenden tmploc nicht löschen')
                else:
                    _doc.tmploc = u''
        _doc.tmploc = self._create_file(_doc, _ext)

    def create_outfile(self, _doc, _ext):
        """erzeugt einen Outfile-Path für die Verarbeitung und prüft ob dieser Pfad nicht belegt ist, ansonsten wird ein
        Index angehangen. Der Outfile-Path wird am Ende einer Taskverarbeitung in docloc gespeichert, möchte man in Extloc speichern,
        so ist dies in params.destination anzugeben.
        Danach wird doc.outfile wieder leer gemacht"""
        _doc.outfile = self._create_file(_doc, _ext)

    def _create_file(self, _doc, _ext):
        _ret = u''
        if type(_doc) is not bitfarm_doc.BitfarmDocument:
            return _ret

        if _ext == '':
            return _ret

        if _ext[0] != '.':
            _ext = '.' + _ext

        if _doc.version in (3, 35, 36, '3.0'):
            if not _doc.basename:
                self.create_basename(_doc)

            if not _doc.tempname:
                if not self.create_tempname(_doc):
                    return u''

            _ind = 0
            _filename = "%s%s" % (_doc.tempname, _ext)
            while os.path.exists(os.path.expandvars(_filename)):
                _filename = "%s_%s%s" % (_doc.tempname, _ind, _ext)
                _ind += 1
            _ret = _filename

            return _ret
        else:
            return _ret

    def gdocid_by_docid(self, arcid, docid):
        gdocid = None
        if self.version in (3, 30, 35, '3.0', '3.5'):
            result = self.db["main"].execute("SELECT gdoc_id from sys_gdocid WHERE arc_id=%s and doc_id=%s",
                                             (arcid, docid))
            result = result.fetchall()
            if len(result) == 0:
                gdocid = None
            else:
                gdocid = result[0][0]
        return gdocid


class BitfarmCon(bitfarm_baseobj.BitfarmBaseCon):
    """
    Bitfarm Configurations-Subklasse.
    Diese Klasse stellt ein einzelnes Profil zur Verfügung, beim Initialisieren kann der Name oder der komplette Pfad auf
    ein Profil übergeben werden.
    con = BitfarmCon('bitfarm')
    Eine Datenbank-Engine wird über die Methode create_db_engine hergestellt, diese wird in con.engine hinterlegt.
    Hier sind auch die Funktionen hinterlegt um User und Gruppen zu lesen, da diese profilabhängig und nicht
    workflow- oder dokumentabhängig sind. Dafür muss aber die DB-Connection übergeben werden, da diese ja erst zur Laufzeit
    erzeugt wird.
    Per Default wird das Con-Objekt als Version 35 erstellt, um abwärtskompatibiltät zur Datenbank 35 zu gewährleisten.
    Jede höhere Version (36, 40) muss in der Con-Datei in der Sektion [Server] angegeben werden.
    @version:
    """
    count = 0
    log = logging.getLogger("Main.BitfarmCon")
    log.setLevel(logging.WARNING)

    # if sys.platform == 'win32':
    #     mypath = os.path.split(os.path.abspath(sys.argv[0]))[0]
    # else:
    #     mypath = os.getcwd()
    # try:
    #     bfapath = os.environ['bitfarm-archiv']
    # except:
    #     bfapath = mypath

    # locale.setlocale(locale.LC_ALL, '')

    def __init__(self, *args, **kwargs):
        """
        Erstellt eine Profil-Instanz. Es kann entweder ein Profilname oder der Pfad auf eine .Con-Datei angegeben werden
        """
        self.log.debug("Initialisiere ein bitfarm_con Object")
        bitfarm_baseobj.BitfarmBaseCon.__init__(self, *args, **kwargs)

    def create_db_engine(self, bind_tables=True, autocommit=True, echo=False, user=None, password=None, pool_size=10):
        """erzeugt eine Datenbank-Engine und hinterlegt diese in BitfarmCon.Engine sowie eine MetaData-Bindung zu den
        Haupttabellen. con.engine ist vom Typ sqlalchemy.engine.
        Optional kann bind_tables (default=True) auf false gesetzt werden. Dann werden keine Tabellen-Bindings ans con Objekt
        vorgenommen, man muss selbst für die benötigten Bindings sorgen (bsp: self.sys_timer = sqla.Table('sys_timer', self.meta,  autoload = True))
        oder über die Funktion bind_tables()
        """

        if self.engine is None:
            connectionstr = self._makeconstr(user, password)
            con_args = self._makeconargs()

            # if self.db_ssl_ca:
            #     con_args['ssl'] = {}
            #     if not os.path.exists(self.db_ssl_ca):
            #         self.log.error('%s konnte nicht gefunden werden' % self.db_ssl_ca)
            #     else:
            #         con_args['ssl']['ca'] = self.db_ssl_ca
            # if self.db_ssl_clientcert:
            #     if 'ssl' not in con_args:
            #         con_args['ssl'] = {}
            #     if not os.path.exists(self.db_ssl_clientcert):
            #         self.log.error('%s konnte nicht gefunden werden' % self.db_ssl_clientcert)
            #     else:
            #         con_args['ssl']['client-cert'] = self.db_ssl_clientcert
            # if self.db_ssl_clientkey:
            #     if 'ssl' not in con_args:
            #         con_args['ssl'] = {}
            #     if not os.path.exists(self.db_ssl_clientkey):
            #         self.log.error('%s konnte nicht gefunden werden' % self.db_ssl_clientkey)
            #     else:
            #         con_args['ssl']['client-key'] = self.db_ssl_clientkey

            try:
                self.engine = sqla.create_engine(connectionstr, echo=echo, pool_pre_ping=True, pool_recycle=3600,
                                                 pool_size=pool_size,
                                                 max_overflow=pool_size,
                                                 # query_cache_size=0,
                                                 connect_args=con_args,
                                                 execution_options={"autocommit": autocommit})
                self.meta = sqla.MetaData()
                self.meta.bind = self.engine

                insp = sqla.inspect(self.engine)
                self.tables = insp.get_table_names()

                if bind_tables:
                    self.bind_tables()
            except sqla.exc.ProgrammingError as e:
                self.log.error('Fehler beim Erstellen der Datenbank-Engine\n%s' % (str(e)))
                raise DatabaseConError(self.confile)
            except sqla.exc.OperationalError as e:
                if hasattr(e, 'orig') and e.orig.args[0] == 1045:
                    self.log.error('Es konnte sich nicht am Datenbankserver angemeldet werden, User oder Password falsch.')
                    self.log.error(e._message())
                    raise DatabaseNotConnectedError(self.confile)
                else:
                    self.log.error('Fehler beim Erstellen der Datenbank-Engine\n%s' % (str(e)))
                    raise DatabaseConError(self.confile)
            except Exception as e:
                self.log.exception(u'Fehler beim Erstellen der Datenbank-Engine\n%s' % (str(e)))
                raise DatabaseConError(self.confile)

    def connect_db(self):
        """erzeugt eine DB-Connection und hinterlegt diese in con.db. Achtung! sollte nicht in Multi-Thread-Anwendungen
        verwendet werden. Dort immer in den einzelnen Threads eine eigene Connection erzeugen."""
        if self.engine is None:
            self.log.exception('Keine Datenbank-Engine erstellt')
            raise DatabaseEngineError(self.profile)

        if self.db is None:
            try:
                self.db = self.engine.connect()
            except:
                self.log.exception('Fehler bei Verbindung mit der Datenbank')
                raise DatabaseNotConnectedError(self.profile)
        else:
            self.log.warning('Datenbank-Connection schon vorhanden')

        if self.db and self.version == 36:
            sql = 'select name, value from sys_config'
            try:
                result = self.db.execute(sqla.text(sql)).fetchall()
            except:
                self.log.exception('Fehler bei Abfrage der sys_config')
                # raise DatabaseNotConnectedError(self.profile)
            else:
                if result:
                    for row in result:
                        self.server_config[row['name']] = row['value']

    def disconnect_db(self):
        """schliesst die DB-Connection."""
        if self.db is None:
            self.log.warning('Datenbank-Connection nicht vorhanden')
        else:
            self.db.close()
            self.db = None

    def dispose_engine(self):
        """Schliesst eine Datenbank-Engine"""
        if self.engine is not None:
            self.engine.dispose()
            self.db = None
            self.engine = None
            self.meta = None
            self.sys_gdocid = None
            self.sys_arc = None
            self.sys_queue = None
            self.sys_duplicates = None
            self.sys_garbage = None
            self.sys_timer = None

    def bind_tables(self):
        """Funktion bindet alle Tabellen and die entsprechenden Attribute des Con-Objekts"""
        try:
            if self.version == 35:
                self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)
                self.sys_gdocid = sqla.Table('sys_gdocid', self.meta, autoload=True)  # Binding an sys_gdocid nur in 35
                self.sys_user = sqla.Table("sys_usr", self.meta, autoload=True)
                self.sys_group = sqla.Table("sys_grp", self.meta, autoload=True)
                self.sys_groupuser = sqla.Table("sys_grpusr", self.meta, autoload=True)
                self.sys_timer = sqla.Table('sys_timer', self.meta, autoload=True)
                self.sys_doclink = sqla.Table('sys_doclnk', self.meta, autoload=True)
                if 'sys_holyday' in self.tables:  # in GPL v35 nicht vorhanden
                    self.sys_holiday = sqla.Table('sys_holyday', self.meta, autoload=True)
                self.sys_superior = sqla.Table('sys_superior', self.meta, autoload=True)
                if 'sys_sap' in self.tables:
                    self.sys_sap = sqla.Table('sys_sap', self.meta, autoload=True)
                if 'sys_svn' in self.tables:
                    self.sys_svn = sqla.Table('sys_svn', self.meta, autoload=True)
                if 'sys_abo' in self.tables:
                    self.sys_abo = sqla.Table('sys_abo', self.meta, autoload=True)
            if self.version >= 36:
                #                    self.meta.reflect(bind=self.engine)    #TODO: prüfen ob eine Reflection notwendig ist, oder uns self.tables reicht
                self.log.debug('sys_arc')
                self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)
                self.log.debug('sys_arcxs')
                self.sys_arcxs = sqla.Table('sys_arcxs', self.meta, autoload=True)
                self.log.debug('sys_doc')
                self.sys_doc = sqla.Table('sys_doc', self.meta, autoload=True)
                self.log.debug('sys_docxs')
                self.sys_docxs = sqla.Table('sys_docxs', self.meta, autoload=True)
                self.log.debug('sys_doctxt')
                if 'sys_doctxt' in self.tables:
                    self.sys_doctxt = sqla.Table('sys_doctxt', self.meta, autoload=True)
                else:
                    self.sys_doctxt = sqla.Table('sys_doc', self.meta, autoload=True)
                for idx in self.tables:
                    if idx[:10] == 'sys_docidx':
                        self.log.debug(idx)
                        try:
                            _ind = int(idx[11:])
                            self.sys_docidx[_ind] = sqla.Table(idx, self.meta, autoload=True)
                        except:
                            self.log.error('konnte Tabelle %s nicht binden' % idx)
                self.log.debug('sys_bookmarks')
                self.sys_bookmarks = sqla.Table('sys_bookmarks', self.meta, autoload=True)
                self.log.debug('sys_statefieldxs')
                self.sys_statefieldxs = sqla.Table('sys_statefieldxs', self.meta, autoload=True)
                self.log.debug('sys_statefields')
                self.sys_statefields = sqla.Table('sys_statefields', self.meta, autoload=True)
                self.log.debug('sys_addfieldxs')
                self.sys_addfieldxs = sqla.Table('sys_addfieldxs', self.meta, autoload=True)
                self.log.debug('sys_addfields')
                self.sys_addfields = sqla.Table('sys_addfields', self.meta, autoload=True)
                self.log.debug('sys_doclink')
                self.sys_doclink = sqla.Table('sys_doclink', self.meta, autoload=True)
                self.log.debug('sys_queue')
                self.sys_queue = sqla.Table('sys_queue', self.meta, autoload=True)
                self.log.debug('sys_garbage')
                self.sys_garbage = sqla.Table("sys_garbage", self.meta, autoload=True)
                self.log.debug('sys_duplicates')
                self.sys_duplicates = sqla.Table("sys_duplicates", self.meta, autoload=True)
                self.log.debug('sys_history')
                self.sys_history = sqla.Table("sys_history", self.meta, autoload=True)
                self.log.debug('sys_user')
                self.sys_user = sqla.Table("sys_user", self.meta, autoload=True)
                self.log.debug('sys_group')
                self.sys_group = sqla.Table("sys_group", self.meta, autoload=True)
                self.log.debug('sys_groupuser')
                self.sys_groupuser = sqla.Table("sys_groupuser", self.meta, autoload=True)
                self.log.debug('sys_timer')
                self.sys_timer = sqla.Table('sys_timer', self.meta, autoload=True)
                self.log.debug('sys_timerhist')
                self.sys_timerhist = sqla.Table('sys_timerhist', self.meta, autoload=True)
                self.log.debug('sys_holiday')
                self.sys_holiday = sqla.Table('sys_holiday', self.meta, autoload=True)
                self.log.debug('sys_annotations')
                self.sys_anno = sqla.Table('sys_annotations', self.meta, autoload=True)
                self.sys_com = sqla.Table('sys_com', self.meta, autoload=True)
                self.log.debug('sys_com')

                if 'sys_sap' in self.tables:
                    self.log.debug('sys_sap')
                    self.sys_sap = sqla.Table('sys_sap', self.meta, autoload=True)
                if 'sys_superior' in self.tables:
                    self.log.debug('sys_superior')
                    self.sys_superior = sqla.Table('sys_superior', self.meta, autoload=True)
                if 'sys_svn' in self.tables:
                    self.log.debug('sys_svn')
                    self.sys_svn = sqla.Table('sys_svn', self.meta, autoload=True)
                if 'sys_abo' in self.tables:
                    self.log.debug('sys_abo')
                    self.sys_abo = sqla.Table('sys_abo', self.meta, autoload=True)
                if 'sys_doclifecycle' in self.tables:
                    self.sys_doclifecycle = sqla.Table('sys_doclifecycle', self.meta, autoload=True)
                else:
                    if not self.gpl:
                        self.log.warning("sys_doclifecycle konnte nicht gebunden werden. Tabelle vorhanden? GGf Update durchführen.")
        except:
            self.log.exception(u'Fehler beim Binden der Datenbank-Tabellen')

    def _makeconstr(self, user=None, password=None, server=None, port=None, db_name=None, encoding=None, collation=None,
                    dbadmin=False):
        """erzeugt einen sql-connection-string passend zur gewählten Datenbank-Engine (derzeit nur mysql)
        """
        _ret = None
        if user is None:
            user = self.user
        if password is None:
            password = self.password
        if server is None:
            server = self.server
        if port is None:
            port = self.port
        if db_name is None:
            db_name = self.db_name
        if encoding is None:
            encoding = self.mysql_encoding
        if collation is None:
            collation = self.db_collation
        if dbadmin:
            host = server
            db_name = 'information_schema'
        else:
            try:
                host = socket.gethostbyname(server)
            except:
                self.log.exception(u'Fehler bei gethostbyname, verwende jetzt Servername (%s)' % self.server)
                host = server

        user = quote_plus(user)
        password = quote_plus(password)
        db_name = quote_plus(db_name)

        if self.db_engine == 'mysql':
            if self.version == 35:
                _ret = 'mysql+mysqldb://%s:%s@%s:%s/%s?charset=%s&collation=%s' % (user, password, host, port, db_name, encoding, collation)
            else:
                # hier habe ich die mysql-connector-python (von Oracle) ausgelötet, weil die leaky war
                # _ret = 'mysql+mysqlconnector://%s:%s@%s:%s/%s?charset=%s&collation=%s' % (
                # user, password, host, port, db_name, encoding, collation)

                _ret = 'mysql+mysqldb://%s:%s@%s:%s/%s?charset=%s&collation=%s' % (user, password, host, port, db_name, encoding, collation)
        elif self.db_engine == "mariadb":
            _ret = 'mysql+pymysql://%s:%s@%s:%s/%s?charset=%s&collation=%s' % (user, password, host, port, db_name, encoding, collation)

        return _ret

    def _makeconargs(self):
        _ret = {}

        if self.db_ssl_ca:
            _ret['ssl'] = {}
            if not os.path.exists(self.db_ssl_ca):
                if not os.path.exists(os.path.join(self.mypath, 'certs', self.db_ssl_ca)):
                    if not os.path.exists(os.path.join(self.mypath, 'bin', 'certs', self.db_ssl_ca)):
                        if not self.gpl:
                            self.log.warning('%s konnte nicht gefunden werden' % self.db_ssl_ca)
                    else:
                        _ret['ssl']['ca'] = os.path.join(self.mypath, 'bin', 'certs', self.db_ssl_ca)
                else:
                    _ret['ssl']['ca'] = os.path.join(self.mypath, 'certs', self.db_ssl_ca)
            else:
                _ret['ssl']['ca'] = self.db_ssl_ca
        if self.db_ssl_clientcert:
            if 'ssl' not in _ret:
                _ret['ssl'] = {}
            if not os.path.exists(self.db_ssl_clientcert):
                if not os.path.exists(os.path.join(self.mypath, 'certs', self.db_ssl_clientcert)):
                    if not os.path.exists(os.path.join(self.mypath, 'bin','certs', self.db_ssl_clientcert)):
                        if not self.gpl:
                            self.log.warning('%s konnte nicht gefunden werden' % self.db_ssl_clientcert)
                    else:
                        _ret['ssl']['client-cert'] = os.path.join(self.mypath, 'bin', 'certs', self.db_ssl_clientcert)
                else:
                    _ret['ssl']['client-cert'] = os.path.join(self.mypath, 'certs', self.db_ssl_clientcert)
            else:
                _ret['ssl']['client-cert'] = self.db_ssl_clientcert
        if self.db_ssl_clientkey:
            if 'ssl' not in _ret:
                _ret['ssl'] = {}
            if not os.path.exists(self.db_ssl_clientkey):
                if not os.path.exists(os.path.join(self.mypath, 'certs', self.db_ssl_clientkey)):
                    if not os.path.exists(os.path.join(self.mypath, 'bin', 'certs', self.db_ssl_clientkey)):
                        if not self.gpl:
                            self.log.warning('%s konnte nicht gefunden werden' % self.db_ssl_clientkey)
                    else:
                        _ret['ssl']['client-key'] = os.path.join(self.mypath, 'bin', 'certs', self.db_ssl_clientkey)
                else:
                    _ret['ssl']['client-key'] = os.path.join(self.mypath, 'certs', self.db_ssl_clientkey)
            else:
                _ret['ssl']['client-key'] = self.db_ssl_clientkey

        return _ret

    def try_again(self, sql=None, db=None, fetchall=None, fetchone=None, **kwargs):
        """führt nochmal ein connect auf der Datenbank-Engine durch und versucht den übergebenen Befehl auszuführen"""
        tests = 0
        if db is None:
            if self.db is None:
                return None
            else:
                db = self.db
        while db.invalidated:
            try:
                db = self.engine.connect()
            except:
                tests += 1
                if tests > 3:
                    raise DatabaseNotConnectedError(self.profile)
                time.sleep(1)

        if sql is not None:
            if kwargs:
                return db.execute(sql, kwargs)
            else:
                if fetchall:
                    return db.execute(sql).fetchall()
                elif fetchone:
                    return db.execute(sql).fetchone()
                else:
                    return db.execute(sql)

    def read_user(self, db=None, name=None, userid=None, logid=''):
        """erzeugt ein BitfarmUser-Objekt. Der User kann nach ID oder nach Name aus der Datenbank geholt werden.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""

        user = None
        if db is None:
            if self.db is None:
                return None
            else:
                db = self.db

        if name is None and userid is None:
            return None

        if self.version == 35:
            if userid is not None:
                sql = "select * from sys_usr where usr_id = %s and (Flags & 1) = 1" % userid  # alle die enabled sind
            else:
                sql = "select * from sys_usr where usr_name = '%s' and (Flags & 1) = 1" % name

            try:
                result = db.execute(sqla.text(sql)).fetchone()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Benutzer %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchone=True)

            if result:
                user = bitfarm_doc.BitfarmUser()
                user.id = result[0]
                user.name = result[1] if result[1] is not None else u''
                user.displayname = result[1] if result[1] is not None else u''
                user.password = result[2]
                user.email = result[3] if result[3] is not None else u''
                user.flags = result[5].split(',') if result[5] is not None else []
                user.profile = self.profile

        elif self.version == 36:
            if userid is not None:
                sql = "select userid, username, userpass, usermail, flags, displayname, role +0 from sys_user where userid = %s and (flags & 1) = 1" % userid
            else:
                sql = "select userid, username, userpass, usermail, flags, displayname, role +0 from sys_user where username = '%s' and (flags & 1) = 1" % name

            try:
                result = db.execute(sqla.text(sql)).fetchone()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Benutzer %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchone=True)

            if result:
                user = bitfarm_doc.BitfarmUser()
                user.id = result[0]
                user.name = result[1] if result[1] is not None else u''
                user.password = result[2]
                user.displayname = result[5] if result[5] is not None else u''
                user.email = result[3] if result[3] is not None else u''
                if type(result[4]) is str:
                    user.flags = result[4].split(',') if result[4] is not None else []
                else:
                    user.flags = [x for x in result[4]] if result[4] is not None else []
                user.role = result[6]
                user.profile = self.profile
                user.grouplist = self.resolve_membership(db=db, userid=user.id, logid=logid)
                if 'superior' in user.flags:
                    user.employees = self.resolve_employees(db=db, userid=user.id, logid=logid)
        # TODO: Version 40 Pfad
        return user

    def read_users(self, db=None, logid=''):
        """erzeugt eine Liste mit BitfarmUser-Objekt. Es werden alle User aus der Datenbank gelesen, die nicht gelöscht sind.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""
        _ret = []
        user = None
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if self.version == 35:
            sql = "select * from sys_usr where (Flags & 2) <> 2"  # alle die nicht deleted sind

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Benutzer %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    user = bitfarm_doc.BitfarmUser()
                    user.id = row[0]
                    user.name = row[1] if row[1] is not None else u''
                    user.displayname = row[1] if row[1] is not None else u''
                    user.password = row[2]
                    user.email = row[3] if row[3] is not None else u''
                    if type(row[5]) is str:
                        user.flags = row[5].split(',') if row[5] is not None else []
                    else:
                        user.flags = [x for x in row[5]] if row[5] is not None else []
                    user.profile = self.profile
                    _ret.append(user)

        elif self.version == 36:
            sql = "select userid, username, userpass, usermail, flags, displayname, role +0 from sys_user where (flags & 2) <> 2"

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Benutzer %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    user = bitfarm_doc.BitfarmUser()
                    user.id = row[0]
                    user.name = row[1] if row[1] is not None else u''
                    user.displayname = row[5] if row[5] is not None else u''
                    user.password = row[2]
                    user.email = row[3] if row[3] is not None else u''
                    if type(row[4]) is str:
                        user.flags = row[4].split(',') if row[4] is not None else []
                    else:
                        user.flags = [x for x in row[4]] if row[4] is not None else []
                    user.role = row[6]
                    user.profile = self.profile
                    _ret.append(user)
        # TODO: Version 40 Pfad
        return _ret

    def read_group(self, db=None, name=None, groupid=None, get_objects=False, logid=''):
        """erzeugt eine BitfarmGroup Instanz. Die Gruppe kann nach ID oder Name aus der Datenbank geholt werden.
            Unter group.members wird eine Liste mit UserIds hinterlegt, wenn get_objects übergeben wird, dann wird
            eine Liste mit BitfarmUser-Objekten hinterlegt.
            Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
            genommen, wenn mit connect_db() erzeugt"""
        group = None
        if db is None:
            if self.db is None:
                return None
            else:
                db = self.db
        if name is None and groupid is None:
            return None

        if self.version == 35:
            if groupid is not None:
                sql = "select * from sys_grp where grp_id = %s" % groupid
            else:
                sql = "select * from sys_grp where grp_name = '%s'" % name

            try:
                result = db.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler beim Lesen der Gruppen %s' % sql)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                group = bitfarm_doc.BitfarmGroup()
                groupid = result[0]
                groupname = result[1] if result[1] is not None else u''
                group.id = groupid
                group.name = groupname
                group.profile = self.profile
                sql = "select distinct(usr_id) from sys_grpusr where grp_id = %s" % groupid
                try:
                    result = db.execute(sql).fetchall()
                except:
                    self.log.exception(u'Fehler beim Lesen der Gruppenmitgliedschaft %s db = %s' % (sql, str(db)))
                    result = self.try_again(sql=sql, db=db, fetchall=True)

                if result:
                    userids = [x[0] for x in result]
                    for userid in userids:
                        if get_objects:
                            user = self.read_user(db=db, userid=userid)
                            if user is not None:
                                group.members.append(user)
                        else:
                            group.members.append(userid)

        elif self.version == 36:
            if groupid is not None:
                sql = "select groupid, groupname, role +0, flags, adpath, email from sys_group where groupid = %s" % groupid
            else:
                sql = "select groupid, groupname, role +0, flags, adpath, email from sys_group where groupname = '%s'" % name

            try:
                result = db.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler beim Lesen der Gruppen %s' % sql)
                result = self.try_again(sql=sql, db=db, fetchone=True)

            if result:
                group = bitfarm_doc.BitfarmGroup()
                groupid = result[0]
                groupname = result[1] if result[1] is not None else u''
                group.id = groupid
                group.name = groupname
                group.email = result[5] if result[5] is not None else u''
                if type(group.flags) is str:
                    group.flags = result[3].split(',') if result[3] is not None else []
                else:
                    group.flags = [x for x in result[3]] if result[3] is not None else []
                group.role = result[2]
                group.adpath = result[4]
                group.profile = self.profile

                sql = "select distinct(userid) from sys_groupuser where groupid = %s" % groupid
                try:
                    result = db.execute(sql).fetchall()
                except:
                    self.log.exception(
                        u'%s Fehler beim Lesen der Gruppenmitgliedschaft %s db = %s' % (logid, sql, str(db)))
                    result = self.try_again(sql=sql, db=db, fetchall=True)

                if result:
                    userids = [x[0] for x in result]
                    for userid in userids:
                        if get_objects:
                            user = self.read_user(db=db, userid=userid, logid=logid)
                            if user is not None:
                                group.members.append(user)
                        else:
                            group.members.append(userid)

        # TODO: Version 40 Pfad

        return group

    def read_groups2(self, db=None, logid=''):
        """erzeugt eine Liste von BitfarmGroup Instanzen aller Gruppen in der Datenbank.
            Unter group.members wird eine Liste mit BitfarmUser-Objekten hinterlegt.
            Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
            genommen, wenn mit connect_db() erzeugt"""
        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if self.version == 35:
            sql = "select * from sys_grp"

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Gruppen %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    group = bitfarm_doc.BitfarmGroup()
                    groupid = row[0]
                    groupname = row[1] if row[1] is not None else u''
                    group.id = groupid
                    group.name = groupname
                    group.profile = self.profile
                    sql = "select distinct(usr_id) from sys_grpusr where grp_id = %s" % groupid
                    try:
                        res = db.execute(sql).fetchall()
                    except:
                        self.log.exception(u'Fehler beim Lesen der Gruppenmitgliedschaft %s' % sql)
                        res = self.try_again(sql=sql, db=db, fetchall=True)

                    if res:
                        userids = [x[0] for x in res]
                        for userid in userids:
                            user = self.read_user(db, userid=userid)
                            if user is not None:
                                group.members.append(user)
                    _ret.append(group)

        elif self.version == 36:
            if self.sys_group is None:
                self.sys_group = sqla.Table('sys_group', self.meta, autoload=True)

            sql = "select groupid, groupname, role +0, flags, adpath, email from sys_group"
            try:
                result = db.execute(sqla.text(sql)).fetchall()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Gruppen %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    group = bitfarm_doc.BitfarmGroup()
                    groupid = row[0]
                    groupname = row[1] if row[1] is not None else u''
                    group.id = groupid
                    group.name = groupname
                    group.role = row[2]
                    if type(row[3]) is str:
                        group.flags = row[3].split(',') if row[3] is not None else []
                    else:
                        group.flags = [x for x in row[3]] if row[3] is not None else []
                    group.adpath = row[4]
                    group.email = row[5] if row[5] is not None else u''
                    group.profile = self.profile
                    sql = "select distinct(userid) from sys_groupuser where groupid = %s" % groupid
                    try:
                        res = db.execute(sql).fetchall()
                    except:
                        self.log.exception(u'%s Fehler beim Lesen der Gruppenmitglieder %s' % (logid, sql))
                        res = self.try_again(sql=sql, db=db, fetchall=True)

                    if res:
                        userids = [x[0] for x in res]
                        for userid in userids:
                            user = self.read_user(db, userid=userid, logid=logid)
                            if user is not None:
                                group.members.append(user)
                    _ret.append(group)
        # TODO: Version 40 Pfad

        return _ret

    def read_groups(self, db=None, logid=''):
        """erzeugt eine Liste von BitfarmGroup Instanzen aller Gruppen in der Datenbank.
            Unter group.members wird eine Liste mit BitfarmUser-Objekten hinterlegt.
            Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
            genommen, wenn mit connect_db() erzeugt"""
        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if self.version == 35:
            sql = "select * from sys_grp"

            try:
                result = db.execute(sql).fetchall()
            except:
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    group = bitfarm_doc.BitfarmGroup()
                    groupid = row[0]
                    groupname = row[1] if row[1] is not None else u''
                    group.id = groupid
                    group.name = groupname
                    group.profile = self.profile
                    sql = "select distinct(usr_id) from sys_grpusr where grp_id = %s" % groupid
                    try:
                        res = db.execute(sql).fetchall()
                    except:
                        self.log.exception(u'Fehler beim Lesen der Gruppenmitgliedschaft %s' % sql)
                    else:
                        if res:
                            userids = [x[0] for x in res]
                            for userid in userids:
                                user = self.read_user(db, userid=userid)
                                if user is not None:
                                    group.members.append(user)
                    _ret.append(group)

        elif self.version == 36:
            if self.sys_group is None:
                self.sys_group = sqla.Table('sys_group', self.meta, autoload=True)

            sql = "select groupid, groupname, role +0, flags, adpath, email from sys_group"
            try:
                result = db.execute(sqla.text(sql)).fetchall()
            except:
                self.log.exception(u'%s Fehler beim Lesen der Gruppen %s' % (logid, sql))
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    group = bitfarm_doc.BitfarmGroup()
                    groupid = row[0]
                    groupname = row[1] if row[1] is not None else u''
                    group.id = groupid
                    group.name = groupname
                    group.role = row[2]
                    if type(row[3]) is str:
                        group.flags = row[3].split(',') if row[3] is not None else []
                    else:
                        group.flags = [x for x in row[3]] if row[3] is not None else []
                    group.adpath = row[4]
                    group.email = row[5] if row[5] is not None else u''
                    group.profile = self.profile
                    sql = "select distinct(userid) from sys_groupuser where groupid = %s" % groupid
                    try:
                        res = db.execute(sql).fetchall()
                    except:
                        self.log.exception(u'%s Fehler beim Lesen der Gruppenmitglieder %s' % (logid, sql))
                    else:
                        if res:
                            # userids = [x[0] for x in res]
                            # for userid in userids:
                            #     user = self.read_user(db, userid=userid, logid=logid)
                            #     if user is not None:
                            #         group.members.append(user)
                            group.members = [x[0] for x in res]
                        _ret.append(group)
        # TODO: Version 40 Pfad

        return _ret

    def resolve_group(self, group):
        """Liefert alle UserIDs der GruppenMember auf und gibt diese als Liste zurück. Es kann eine BitfarmGroup oder eine GroupID übergeben werden"""
        _ret = []

        if type(group) is int:
            group = self.read_group(groupid=group)

        _ret = self._resolve_member(group.members, _ret)
        return _ret

    def _resolve_member(self, members, _ret):
        for member in members:
            if isinstance(member, bitfarm_doc.BitfarmUser):
                if member.id not in _ret:
                    _ret.append(member.id)
            if isinstance(member, bitfarm_doc.BitfarmGroup):
                group = self._resolve_member(member, _ret)
                for memid in group:
                    if memid not in _ret:
                        _ret.append(memid)
            if type(member) is int:
                _ret.append(member)

        return _ret

    def resolve_membership(self, db=None, userid=None, logid=''):
        """erzeugt eine Liste an GroupIDs, der Gruppen in denen der User Mitglied ist, es kann auch eine Liste an userids übergeben werden"""
        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if userid is None:
            return _ret

        if type(userid) is list:
            userid = ",".join([str(x) for x in userid])

        if self.version == 36:
            # sql = 'select distinct(groupid) from sys_groupuser where userid in (%s)' % userid
            sql = 'select distinct(groupid) from sys_groupuser where groupid in (select groupid from sys_group where flags & 2 = 0) and userid in (%s)' % userid
            try:
                result = db.execute(sqla.text(sql)).fetchall()
            except:
                self.log.exception(u'%s Fehler bei der Gruppenauflösung %s db = %s' % (logid, sql, str(db)))
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for rec in result:
                    _ret.append(rec[0])

        return _ret

    def resolve_employees(self, db=None, userid=None, logid=''):
        """erzeugt eine Liste an Userids, die Mitarbeiter eines Users sind"""
        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if userid is None:
            return _ret

        if type(userid) is str:
            try:
                userid = int(userid)
            except:
                return _ret

        if self.version == 36:
            # sql = 'select distinct(groupid) from sys_groupuser where userid in (%s)' % userid
            sql = 'select distinct(inf_userid) from sys_superior where inf_userid in (select userid from sys_user where flags & 2 = 0) and sup_userid = %s' % userid
            try:
                result = db.execute(sqla.text(sql)).fetchall()
            except:
                self.log.exception(u'%s Fehler bei der Mitarbeiterauflösung %s db = %s' % (logid, sql, str(db)))

            if result:
                for rec in result:
                    _ret.append(rec[0])

        return _ret

    def read_rights(self, db=None, arcid=None, gdocid=None, group=None, user=None, single_rights=False,
                    doc_rights=False, grouplist=None, logid=''):
        """erzeugt eine Liste mit BitfarmRights-Objekten. Es kann eine arcid oder gdocid, und optional eine groupid bzw. BitfarmGroup-Object
                oder eine userid bzw. BitfarmUser-Object übergeben werden.
                (arcid OR gdocid) {AND (groupid OR userid)}
                Bei Userrechten kann man mit single_rights alle ggf. vorhandenen Einträge einzeln holen,
                ansonsten werden die Rechte aufaddiert zurückgegeben (Effektivrechte)
                Wird eine Gdocid übergeben, so werden alle Archivrechte des entsprechenden Archives übergeben, erweitert um
                evtl. vorhandene Dokumentenrechte.
                Will man nur die Dokumentenrechte einer gdocid abfragen so übergibt man doc_rights = True beim Aufruf.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt.
                Konnten keine Rechte ermittelt werden wird eine leere Liste zurückgegeben"""

        if grouplist is None:
            grouplist = []
        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db
        if arcid is None and gdocid is None and group is None and user is None:
            return _ret

        if group is not None and type(group) is bitfarm_doc.BitfarmGroup:
            groupid = group.id
        else:
            groupid = group

        if user is not None and type(user) is bitfarm_doc.BitfarmUser:
            userid = user.id
            if len(user.grouplist) == 0:
                if len(grouplist) == 0:
                    grouplist = self.resolve_membership(db=db, userid=user.id, logid=logid)
            else:
                grouplist = user.grouplist[:]
        else:
            userid = user
            if len(grouplist) == 0:
                grouplist = self.resolve_membership(db=db, userid=userid, logid=logid)

        if len(grouplist) == 0:
            grouplist.append(0)

        if arcid is None:
            arcid = []
        elif type(arcid) is int:
            arcid = [arcid]
        elif type(arcid) is str:
            arcid = [int(arcid)]

        if self.version == 35:
            if gdocid:
                result = db.execute('select arc_id, doc_id from sys_gdocid where gdoc_id = ' + str(gdocid)).fetchone()
                if result:
                    arcid = result[0]
                    docid = result[1]
                else:
                    gdocid = None

            if arcid is None:  # keine Arcid, keine Rechte
                return _ret

            fields = 'xs.arc_id, xs.usr_id, xs.grp_id , xs.rights+0, usr.usr_id, usr.usr_name, usr.usr_mail, usr.flags, grp.grp_id, grp.grp_name'
            fields_or = 'xs.arc_id, xs.usr_id, xs.grp_id , bit_or(xs.rights), usr.usr_id, usr.usr_name, usr.usr_mail, usr.flags, grp.grp_id, grp.grp_name'
            leftjoin = 'left join sys_usr as usr on xs.usr_id = usr.usr_id left join sys_grp as grp on xs.grp_id = grp.grp_id'
            arcid_str = ",".join([str(x) for x in arcid])

            if group is None and user is None:
                if doc_rights and gdocid is not None:
                    sql = 'select ' + fields + ' from sys_docxs as xs ' + leftjoin + ' where xs.arc_id in (%s) and xs.doc_id = %i' % (
                        arcid_str, docid)
                    doc_rights = None
                    sql += ' and xs.rights is not NULL'
                else:
                    sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arc_id in (%s) and xs.grp_id = 0' % arcid_str
                    sql += ' and xs.rights is not NULL'
                    sql += ' union select ' + fields + ' from sys_arcxs as xs ' + leftjoin + \
                           ' where xs.arc_id in (%s) and xs.grp_id <> 0 and xs.rights is not NULL group by xs.grp_id' % arcid_str
            elif user is None:
                if single_rights:
                    sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arc_id in (%s) and xs.grp_id = %s and xs.usr_id <> 0' % (
                    arcid_str, groupid)
                else:
                    sql = 'select ' + fields_or + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arc_id in (%s) and xs.grp_id = %s' % (
                    arcid_str, groupid)
                sql += ' and xs.rights is not NULL'
            elif group is None:
                if single_rights:
                    sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arc_id in (%s) and xs.usr_id = %s' % (
                    arcid_str, userid)
                else:
                    sql = 'select ' + fields_or + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arc_id in (%s) and xs.usr_id = %s' % (
                    arcid_str, userid)
                sql += ' and xs.rights is not NULL'

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"Fehler bei Abfrage der Archivberechtigung")
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    if row[3] is None or int(row[3]) == 0:  # kein Recht, kein Eintrag
                        continue

                    right = bitfarm_doc.BitfarmRights()
                    right.arcid = row[0]
                    right.gdocid = gdocid
                    right.rights = int(row[3])
                    right.user = bitfarm_doc.BitfarmUser()
                    right.user.id = row[1]
                    right.user.name = row[5] if row[5] is not None else u''
                    right.user.email = row[6] if row[6] is not None else u''
                    if type(row[7]) is str:
                        right.user.flags = row[7].split(',') if row[7] is not None else []
                    else:
                        right.user.flags = [x for x in row[7]] if row[7] is not None else []
                    if row[8] is not None:
                        right.group = bitfarm_doc.BitfarmGroup()
                        right.group.id = row[8]
                        right.group.name = row[9] if row[9] is not None else u''
                    if right.user.id > 0:
                        right.userlist.append(right.user.id)
                    if userid is not None and userid not in right.userlist:
                        right.userlist.append(userid)
                    right.grouplist = grouplist[:]
                    _ret.append(right)

                if doc_rights is not None and gdocid is not None:
                    _ext = self.read_rights(db=db, gdocid=gdocid, single_rights=True, doc_rights=True,
                                            grouplist=grouplist)
                    if _ext is not None:
                        _ret.extend(_ext)

            return _ret

        elif self.version == 36:
            if gdocid:
                result = None
                try:
                    if not arcid:
                        result = db.execute(
                            'select sys_doc.arcid from sys_doc where sys_doc.gdocid = %s' % str(gdocid)).fetchone()
                except:
                    self.log.exception(u'%s Fehler bei der Abfrage der Arcid für [%s]' % (logid, gdocid))
                else:
                    if result:
                        arcid.append(result[0])

            if not arcid:  # keine Arcid, keine Rechte
                return _ret

            fields = 'xs.arcid, xs.userid, xs.groupid , xs.rights+0, usr.userid, usr.username, usr.usermail, usr.flags, grp.groupid, grp.groupname'
            # fields_or_user = 'xs.arcid, xs.userid, xs.groupid , bit_or(xs.rights) OVER (Partition by xs.userid), usr.userid, usr.username, usr.usermail, usr.flags, grp.groupid, grp.groupname'
            leftjoin = 'left join sys_user as usr on xs.userid = usr.userid left join sys_group as grp on xs.groupid = grp.groupid'
            arcid_str = ",".join([str(x) for x in arcid])

            if group is None and user is None:
                if doc_rights and gdocid is not None:
                    doc_rights = False  # nicht nochmal abfragen (s.u.)
                    sql = 'select ' + fields + ' from sys_docxs as xs ' + leftjoin + ' where xs.arcid in (%s) and xs.gdocid = %i' % (
                    arcid_str, gdocid)
                    sql += ' and xs.rights is not NULL'
                else:
                    sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arcid in (%s) and xs.groupid = 0' % arcid_str
                    sql += ' and xs.rights is not NULL'
                    sql += ' union select ' + fields + ' from sys_arcxs as xs ' + leftjoin + \
                           ' where xs.arcid in (%s) and xs.groupid <> 0 and xs.rights is not NULL group by xs.groupid, xs.userid' % arcid_str
            elif user is None:  # Gruppenrechte kommen immer einzeln, die Kombination arcid, userid, groupid ist unique, userid ist 0 bei Gruppenrechten, somit kann nichts aufsummiert werden
                sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + ' where xs.arcid in (%s) and xs.userid = 0 and xs.groupid = %i' % (
                arcid_str, groupid)
                sql += ' and xs.rights is not NULL'
            elif group is None:  # Userrechte
                sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + \
                      ' where xs.arcid in (%s) and (xs.userid = %i or (xs.userid = 0 and xs.groupid in (%s)))' % (
                      arcid_str, userid, ",".join([str(x) for x in grouplist]))
                sql += ' and xs.rights is not NULL'

                # if single_rights:
                #     sql = 'select ' + fields + ' from sys_arcxs as xs ' + leftjoin + \
                #           ' where xs.arcid in (%s) and (xs.userid = %i or (xs.userid = 0 and xs.groupid in (%s)))' % (arcid_str, userid, ",".join([str(x) for x in grouplist]))
                #     sql += ' and xs.rights is not NULL'
                # else:
                #     sql = 'select ' + fields_or_user + ' from sys_arcxs as xs ' + leftjoin + \
                #           ' where xs.arcid in (%s) and (xs.userid = %i  or (xs.userid = 0 and xs.groupid in (%s)))' % (arcid_str, userid, ",".join([str(x) for x in grouplist]))
                #     sql += ' and xs.rights is not NULL GROUP BY xs.rights, xs.userid, xs.groupid, xs.arcid'

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u'%s Fehler bei der Rechteabfrage' % logid)
                # result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    if row[3] is None or int(row[3]) == 0:  # kein Recht, kein Eintrag
                        continue
                    right = bitfarm_doc.BitfarmRights()
                    right.arcid = row[0]
                    right.gdocid = gdocid
                    right.rights = int(row[3])
                    right.user = bitfarm_doc.BitfarmUser()
                    right.user.id = row[1]
                    right.user.name = row[5] if row[5] is not None else u''
                    right.user.email = row[6] if row[6] is not None else u''
                    if type(row[7]) is str:
                        right.user.flags = row[7].split(',') if row[7] is not None else []
                    else:
                        right.user.flags = [x for x in row[7]] if row[7] is not None else []
                    if row[8] is not None:
                        right.group = self.read_group(db=db, groupid=row[8])
                    if right.user.id > 0 and right.user.id not in right.userlist:
                        right.userlist.append(right.user.id)
                    if userid is not None and userid not in right.userlist:
                        right.userlist.append(userid)
                    right.grouplist = grouplist[:]
                    _ret.append(right)

                if doc_rights and gdocid is not None:
                    _ext = self.read_rights(db=db, gdocid=gdocid, single_rights=True, doc_rights=True,
                                            grouplist=grouplist, logid=logid)
                    if _ext is not None:
                        _ret.extend(_ext)

                if not single_rights:
                    rights_sum = bitfarm_doc.BitfarmRights()
                    for right in _ret:
                        rights_sum.rights |= right.rights
                        if right.arcid:
                            rights_sum.arcid = right.arcid
                        if right.gdocid:
                            rights_sum.gdocid = right.gdocid
                        for x in right.userlist:
                            if x not in rights_sum.userlist:
                                rights_sum.userlist.append(x)
                        for x in right.grouplist:
                            if x not in rights_sum.grouplist:
                                rights_sum.grouplist.append(x)
                    _ret = [rights_sum]

            return _ret

            # TODO: Version 40 Pfad

    def read_resubright(self, db=None, gdocid=None, user=None, logid=''):
        """liest die Wiedervorlageberechtigung eines Dokuments. Es muss die gdocid und ein User (userid oder User-Objekt)
        übergeben werden. Es wird geprüft, wer der Controller der Wiedervorlage ist und dessen Rechte ausgelesen.
        Die Rechte werden als BitfarmRights-Objekt zurückgegeben
        Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
        genommen, wenn mit connect_db() erzeugt.
        Konnten keine Rechte ermittelt werden wird None zurückgegeben"""

        def bit_or(x, y):
            return x | y

        _ret = None
        if db is None:
            if self.db is None:
                return None
            else:
                db = self.db

        if gdocid is None or user is None:
            return None

        if type(user) is bitfarm_doc.BitfarmUser:
            userid = user.id
        else:
            userid = user

        if self.version == 36:
            sql = sqla.select([self.sys_timer.c.userid_shed]).where(sqla.and_(self.sys_timer.c.userid_work == userid,
                                                                              self.sys_timer.c.gdocid == gdocid,
                                                                              self.sys_timer.c.status.op('&')(4) == 0))
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"Fehler bei der Abfrage der Wiedervorlageberechtigung")
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                rights = []
                for row in result:
                    user_shed = row[0]
                    right = self.read_rights(db=db, user=user_shed, gdocid=gdocid, doc_rights=True, logid=logid)
                    if right:
                        rights.append(right[0])

                _ret = bitfarm_doc.BitfarmRights()
                if rights:
                    # _ret.rights = reduce(bit_or, [x.rights for x in rights])
                    _ret.rights = 0
                    for x in rights:
                        _ret.rights |= x.rights

        return _ret

    def read_docrights(self, db=None, gdocid=None, user=None, src_userid=None, logid=''):
        """liest die Dokumentenberechtigungen eines Dokuments aus der sys_docxs. Es muss die gdocid und ein User (userid oder User-Objekt)
        übergeben werden. Der User wird über userid gesucht und die Rechte (User oder Gruppenrechte) ausgelesen.
        Es kann auch über src_userid gesucht werden, wenn man die Rechte erhalten will, die ein bestimmter User für bestimmtes Dokument
        vergeben hat.
        Die Rechte werden als BitfarmRights-Objekte zurückgegeben (Liste), die User und Gruppen werden angehangen.
        Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
        genommen, wenn mit connect_db() erzeugt.
        Konnten keine Rechte ermittelt werden wird eine leere Liste zurückgegeben"""

        def bit_or(x, y):
            return x | y

        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if gdocid is None or user is None:
            return _ret

        if type(user) is bitfarm_doc.BitfarmUser:
            userid = user.id
        else:
            userid = user

        if self.version == 36:
            if src_userid is not None:
                sql = sqla.select([self.sys_docxs.c.gdocid, self.sys_docxs.c.src_userid, self.sys_docxs.c.userid,
                                   self.sys_docxs.c.groupid,
                                   sqla.cast(self.sys_docxs.c.rights, sqla.types.BIGINT)]).where(
                    sqla.and_(self.sys_docxs.c.src_userid == src_userid,
                              self.sys_docxs.c.gdocid == gdocid))
            else:
                sql = sqla.select([self.sys_docxs.c.gdocid, self.sys_docxs.c.src_userid, self.sys_docxs.c.userid,
                                   self.sys_docxs.c.groupid,
                                   sqla.cast(self.sys_docxs.c.rights, sqla.types.BIGINT)]).where(
                    sqla.and_(self.sys_docxs.c.userid == userid,
                              self.sys_docxs.c.gdocid == gdocid))
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_docrights" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    if row[4] is None or int(row[4]) == 0:  # kein Recht, kein Eintrag
                        continue
                    right = bitfarm_doc.BitfarmRights()
                    right.gdocid = gdocid
                    right.rights = int(row[4])
                    right.user = bitfarm_doc.BitfarmUser()
                    right.user.id = row[2]
                    right.user.src_userid = row[1]
                    if row[3] is not None and row[3] != 0:
                        right.group = self.read_group(db=db, groupid=row[3], logid=logid)
                    _ret.append(right)

        return _ret

    def get_holidaylist(self, db=None, user=None, start=None, end=None, logid=''):
        """Funktion gibt die aktuellen Urlaubsvertretungen eines User zurück.
        Wird eine Userid (int) übergeben, erhält man eine Liste von IDs, übergibt man ein User-Objekt
        erhält man eine Liste von User-Objekten
        Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
        genommen, wenn mit connect_db() erzeugt.
        Konnten keine User ermittelt werden, enthält die Liste nur den übergebenen User"""

        _ret = []
        if user is None:
            return _ret

        get_id = False
        if isinstance(user, bitfarm_doc.BitfarmUser):
            userid = user.id
        elif type(user) is int:
            userid = user
            get_id = True
        else:
            return _ret

        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if start is None:
            start = datetime.datetime.date(datetime.datetime.now())

        if end is None:
            end = datetime.datetime.date(datetime.datetime.now())

        if get_id:
            _ret = [userid]
        else:
            _ret = [user]  # wir fügen uns selbst zum Ergebnis hinzu

        self._add_holidayuser(_ret, userid, start, end, db, get_id, logid)

        return _ret

    def _add_holidayuser(self, _ret, userid, start, end, db, get_id, logid):
        """Unterfunktion für die rekursive Sammlung von Urlaubsvertretungen"""
        if self.version in (3, 35, '3.0'):
            sql = u"select * from sys_holyday where usr_id_dst = %s" % userid
        else:
            sql = u"select * from sys_holiday where userid_dst = %s" % userid
        if start is not None:
            _start = datetime.datetime.strftime(start, '%Y-%m-%d %H:%M:%S')
            _end = datetime.datetime.strftime(end, '%Y-%m-%d %H:%M:%S')
            if self.version in (3, 35, '3.0'):
                sql += u" and ((START <= '%s') and (END >= '%s')) and (END >= '%s')" % (_end, _start, _end)
            else:
                sql += u" and ((start <= '%s') and (end >= '%s')) and (end >= '%s')" % (_end, _start, _end)
        try:
            result = db.execute(sql).fetchall()
        except:
            self.log.exception(u'%s Fehler bei _add_holidayuser' % logid)
            result = self.try_again(sql=sql, db=db, fetchall=True)

        if result:
            for row in result:
                _newuser = self.read_user(db=db, userid=row[1], logid=logid)
                if _newuser:
                    if get_id:
                        _ret.append(_newuser.id)
                    else:
                        _ret.append(_newuser)
                    self._add_holidayuser(_ret, _newuser.id, start, end, db, get_id, logid)

        return _ret

    # TODO: ReadFieldRights einbauen (sobald wir eine sys_addfield-Tabelle haben)

    def read_archives(self, db=None, arcid=None, logid=''):
        """erzeugt ein Liste von Archivobjekten des gesamten oder eines bestimmten Archives. Dient zu Konfigurationszwecken.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""

        _ret = []
        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if self.sys_arc is None:
            self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)

        if self.version == 35:
            if arcid is None:
                sql = sqla.select([self.sys_arc])
            else:
                sql = sqla.select([self.sys_arc]).where(self.sys_arc.c.ARC_ID == arcid)
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_archives" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    arc = bitfarm_doc.BitfarmArchive()
                    arc.id = row[u'ARC_ID']
                    arc.lgrid = row[u'LGR_ID']
                    arc.arcname = row[u'ARC_Name']
                    arc.tblname = row[u'ARC_TBLName']
                    arc.enabled = row[u'ARC_Enabled']
                    arc.icon = row[u'Icon']
                    arc.type = row[u'Type']
                    arc.backpic = row[u'Arc_BackPic']
                    arc.options = row[u'Options']
                    arc.defdocname = row[u'DefDocname']
                    arc.rights = bitfarm_doc.BitfarmRights()
                    _ret.append(arc)

        elif self.version == 36:
            if arcid is None:
                sql = sqla.select([self.sys_arc])
            else:
                sql = sqla.select([self.sys_arc]).where(self.sys_arc.c.arcid == arcid)
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_archives" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    arc = bitfarm_doc.BitfarmArchive()
                    arc.id = row[0]
                    arc.lgrid = row[1]
                    arc.arcname = row[2]
                    arc.tblname = row[3]
                    # let's be upward compatible
                    arc.flags = row[4]
                    arc.enabled = row[4]
                    arc.backpic = row[5]
                    arc.icon = row[6]
                    arc.type = row[7]
                    arc.options = row[8]
                    arc.defdocname = row[9]
                    arc.storageid = row[10]
                    arc.rights = bitfarm_doc.BitfarmRights()
                    _ret.append(arc)
        return _ret

    def _get_archives_recurse(self, db, _ret, arcid):
        _ret.append(arcid)
        sql = 'select arcid, type from sys_arc where lgrid = %s and (flags & 2 = 0)' % arcid
        result = db.execute(sql).fetchall()
        if result:
            for row in result:
                if row[1] == 0:  # lager
                    _ret = self._get_archives_recurse(db, _ret, row[0])
                elif row[1] == 1:
                    _ret.append(row[0])

        return _ret

    def _get_parents_recurse(self, db, _ret, lgrid, logid):
        lgr = self.read_archives(db=db, arcid=lgrid, logid=logid)
        if len(lgr) > 0:
            _ret.append(lgr[0])
            if lgr[0].lgrid not in [x.id for x in _ret]:
                _ret = self._get_parents_recurse(db, _ret, lgr[0].lgrid, logid)
        return _ret

    def read_arctree(self, db=None, rootid=None, user=None, right=None, no_addfields=False, no_statefields=False,
                     logid='', no_fielditems=False):
        """erzeugt ein Liste von Archivobjekten des gesamten Archivbaums oder eines bestimmten Zweiges (rootid != None).
            Es kann ein User übergeben werden, es werden dann nur Archive zurückgeliefert, dem der User zugeordnet ist,
            oder es wird zusätzlich ein BitfarmRights Objekt übergeben, dann werden nur Archive zurückgeliefert,
            wo der User die entsprechenden Berechtigungen besitzt.
            Wird kein User übergeben, dann wird der gesamte Archivbaum zurückgegeben, ohne Prüfung auf Rechte.
            Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
            genommen, wenn mit connect_db() erzeugt"""
        _ret = []
        arclist = []
        addfields = []
        statefields = []

        if db is None:
            if self.db is None:
                return _ret
            else:
                db = self.db

        if user is None:
            right = None
            userid = None
        elif type(user) is bitfarm_doc.BitfarmUser:
            userid = user.id
        elif type(user) is int:
            userid = user
            user = self.read_user(db=db, userid=userid, logid=logid)
        else:
            return _ret

        if self.version == 36:
            if rootid:
                arclist = self._get_archives_recurse(db, [], rootid)

            if user is None:
                sql = 'select arcid, 0, 0, 0, lgrid, arcname,' \
                      'tblname, flags, backpic, icon, type, options, defdocname, storageid from sys_arc ' \
                      'where (flags & 2 = 0) ' \
                      'order by arcid'
            else:
                if len(user.grouplist) == 0:
                    user.grouplist = [0]
                sql = 'select sys_arcxs.arcid, userid, groupid, rights +0, lgrid, arcname, ' \
                      'tblname, flags, backpic, icon, type, options, defdocname, storageid from sys_arcxs ' \
                      'left join sys_arc on sys_arcxs.arcid = sys_arc.arcid where (userid = %i or (userid = 0 and groupid in (%s))) ' \
                      'and (flags & 2 = 0) ' \
                      'order by sys_arcxs.arcid' % (userid, ",".join([str(x) for x in user.grouplist]))
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_arctree" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    if right is not None:
                        if right.rights & int(row[3]) != right.rights:
                            continue
                    if rootid:
                        if row[0] not in arclist:
                            continue
                    if row[0] in [x.id for x in _ret]:  # wenn es das Archiv schon in der Liste gibt...
                        for arc in _ret:
                            if arc.id == row[0]:  # werden die Rechte aufaddiert
                                arc.rights.rights |= int(row[3])
                                break
                    else:
                        arc = bitfarm_doc.BitfarmArchive()
                        arc.id = row[0]
                        arc.lgrid = row[4]
                        arc.arcname = row[5]
                        arc.tblname = row[6]
                        # let's stay backwards compatible
                        arc.enabled = row[7]
                        arc.flags = row[7]
                        arc.backpic = row[8]
                        arc.icon = row[9]
                        arc.type = row[10]
                        arc.options = row[11]
                        arc.defdocname = row[12]
                        arc.storageid = row[13]
                        arc.rights = bitfarm_doc.BitfarmRights()
                        arc.rights.rights = int(row[3])
                        # arc.rights.user = user
                        arc.rights.userid = userid  # wir hängen nicht mehr das ganze userobjekt dran sondern nurnoch die id
                        _ret.append(arc)

                for arc in _ret:
                    if arc.lgrid != 0 and arc.lgrid not in [x.id for x in
                                                            _ret]:  # gibt es das übergeordnete Lager in unserer Liste?
                        _ret = self._get_parents_recurse(db, _ret, arc.lgrid, logid)

                # for arc in _ret:
                #     if not no_addfields:
                #         arc.addfields = self.read_addfields(db=db, arcid=arc.id, userid=userid, logid=logid, no_fielditems=no_fielditems)
                #     if not no_statefields:
                #         arc.statefields = self.read_statefields(db=db, arcid=arc.id, userid=userid, logid=logid)

                if not no_addfields:
                    addfields = self.read_addfields(db=db, arcid=None, userid=userid, logid=logid,
                                                    no_fielditems=no_fielditems)
                if not no_statefields:
                    statefields = self.read_statefields(db=db, arcid=None, userid=userid, logid=logid)
                if addfields or statefields:
                    for arc in _ret:
                        for field in [x for x in addfields]:
                            if arc.id in [x.arcid for x in field.rights]:
                                newfield = copy.deepcopy(field)
                                newfield.rights = bitfarm_doc.BitfarmFieldRights()
                                newfield.rights.add([x for x in field.rights if x.arcid == arc.id])
                                if len(newfield.rights) > 0:
                                    newfield.pos = newfield.rights[0].fieldpos
                                arc.addfields.append(newfield)
                        for field in statefields:
                            if arc.id in [x.arcid for x in field.rights]:
                                newfield = copy.deepcopy(field)
                                newfield.rights = bitfarm_doc.BitfarmFieldRights()
                                newfield.rights.add([x for x in field.rights if x.arcid == arc.id])
                                if len(newfield.rights) > 0:
                                    newfield.pos = newfield.rights[0].fieldpos
                                arc.statefields.append(newfield)

            return _ret

    def read_addfields(self, db=None, arcid=None, userid=None, groupid=None, logid='', no_fielditems=False):
        """erzeugt ein Liste von Addfieldobjekten des gesamten oder eines bestimmten Archives. Dient zu Konfigurationszwecken.
                Es kann eine userid oder groupid übergeben werden, dann werden nur die Felder zurückgegeben, auf die der User Zugriff hat.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""
        fields = []
        if db is None:
            if self.db is None:
                return fields
            else:
                db = self.db
        if self.sys_arc is None:
            self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)

        if self.version == 35:
            if arcid is None:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_TBLName != '')
            else:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_ID == arcid)

            result = db.execute(sql).fetchall()

            if result:
                for row in result:
                    tblname = 'cfg_%s' % row[0]
                    arcid = row[1]
                    if tblname in self.tables:
                        cfg = sqla.Table(tblname, self.meta, autoload=True)
                        sql = sqla.select([cfg])

                        result2 = db.execute(sql).fetchall()

                        for row2 in result2:
                            add = bitfarm_doc.BitfarmAddfield()
                            add.name = row2['Disp_Name']
                            add.add = row2['Feld_Name']
                            add.type = row2['Typ']
                            add.checkprogram = row2['chck_prog']
                            add.mode = row2['Mode']
                            add.options = row2['Optionen']
                            add.rights = bitfarm_doc.BitfarmFieldRights()
                            if row2['Werte']:
                                add.items.extend(row2['Werte'].split(';'))
                            add.pos = row2['Pos']
                            if 'sys_cfgxs' in self.tables:
                                # gibt es einen Basiseintrag, dann haben wir nur lesenden Zugriff
                                sql = 'select * from sys_cfgxs where arc_id = %i and cfg_id = 0' % arcid
                                res = db.execute(sql).fetchone()
                                if res:
                                    r = bitfarm_doc.BitfarmFieldRight()
                                    r.rights = bitfarm_doc.RIGHT_FIELD_READ
                                    r.arcid = arcid
                                    r.fieldpos = row2['Pos']
                                    add.rights.add(r)
                                else:
                                    r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                                    r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                    r.arcid = arcid
                                    r.fieldpos = row2['Pos']
                                    add.rights.add(r)
                                # jetzt noch user oder gruppen rechte prüfen
                                # erst einzelnde user holen
                                sql = 'select * from sys_cfgxs where arc_id = %i and cfg_id = %i and grp_id = 0' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        user = None
                                        if rec['usr_id'] != 0:
                                            user = self.read_user(userid=rec['usr_id'])
                                        if user is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()
                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = row2['Pos']
                                            r.user = user
                                            add.rights.add(r)
                                # dann gruppen
                                sql = 'select * from sys_cfgxs where arc_id = %i and cfg_id = %i and grp_id <> 0 group by grp_id' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        group = None
                                        if rec['grp_id'] != 0:
                                            group = self.read_group(groupid=rec['grp_id'])
                                        if group is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()

                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = row2['Pos']
                                            r.group = group
                                            add.rights.add(r)
                            else:
                                r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify für gruppe 0 / Alle
                                r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                r.arcid = arcid
                                r.group = bitfarm_doc.BitfarmGroup()
                                r.group.id = 0
                                r.fieldpos = row2['Pos']
                                add.rights.add(r)

                            fields.append(add)

        if self.version == 36:
            if self.sys_addfields is None:
                self.sys_addfields = sqla.Table('sys_addfields', self.meta, autoload=True)
            if self.sys_addfieldxs is None:
                self.sys_addfieldxs = sqla.Table('sys_addfieldxs', self.meta, autoload=True)

            # die Feldberechtigung holen, wenn arcid angegeben wurde, auf Felder dieses Archives einschränken,
            # auf user oder gruppen rechte prüfen falls angegeben, da es entweder ein gruppen eintrag ODER ein usereintrag gibt, kann man beides direkt abfragen
            # wurde userid oder groupid angegeben, das Feld nur zurückliefern, wenn Berechtigung besteht

            fieldrights = {}
            sql = 'select xsid, arcid, fieldid, userid, groupid, pos, rights+0 from sys_addfieldxs where fieldid <> 0'
            if arcid is not None:
                sql += ' and arcid = %i' % arcid
            if userid is not None:
                grouplist = self.resolve_membership(db=db, userid=userid, logid=logid)
                if not grouplist:
                    sql += ' and userid = %i and groupid = 0' % userid
                else:
                    sql += ' and (userid = %i or (userid = 0 and groupid in (%s)))' % (userid, ",".join([str(x) for x in grouplist]))

            elif groupid is not None:
                sql += ' and groupid = %i and userid = 0' % groupid
            # else:
            #     sql += ' and (userid = 0 and groupid = 0)'

            try:
                res = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_addfieldxs" % logid)
                res = self.try_again(sql=sql, db=db, fetchall=True)

            for row in res:
                if row[2] in fieldrights:
                    fieldrights[row[2]].append(row)
                else:
                    fieldrights[row[2]] = [row]

            if arcid is None:
                sql = sqla.select(
                    [self.sys_addfields.c.addfieldid, self.sys_addfields.c.fieldname, self.sys_addfields.c.displayname,
                     self.sys_addfields.c.type, self.sys_addfields.c['items'], self.sys_addfields.c.mode,
                     self.sys_addfields.c.options, self.sys_addfields.c.checkprogram, self.sys_addfields.c.values_query,
                     self.sys_addfields.c.flags, self.sys_addfields.c.indextableid, self.sys_addfields.c.description],
                    use_labels=False).order_by(self.sys_addfields.c.addfieldid)

            else:
                from_obj = self.sys_addfieldxs.join(self.sys_addfields,
                                                    self.sys_addfieldxs.c.fieldid == self.sys_addfields.c.addfieldid)
                sql = sqla.select(
                    [self.sys_addfieldxs.c.fieldid, self.sys_addfieldxs.c.pos, self.sys_addfieldxs.c.rights,
                     self.sys_addfields.c.addfieldid, self.sys_addfields.c.fieldname, self.sys_addfields.c.displayname,
                     self.sys_addfields.c.type,
                     self.sys_addfields.c['items'], self.sys_addfields.c.mode, self.sys_addfields.c.options,
                     self.sys_addfields.c.checkprogram, self.sys_addfields.c.values_query,
                     self.sys_addfields.c.flags, self.sys_addfields.c.indextableid, self.sys_addfields.c.description],
                    from_obj=[from_obj],
                    whereclause=(self.sys_addfieldxs.c.arcid == arcid), use_labels=False).group_by(
                    self.sys_addfieldxs.c.fieldid, self.sys_addfieldxs.c.pos, self.sys_addfieldxs.c.rights)
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_addfields" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    if row['fieldname'] not in [x.add for x in fields]:
                        add = bitfarm_doc.BitfarmAddfield()
                        add.name = row['displayname']
                        add.add = row['fieldname']
                        add.type = row['type']
                        add.id = row['addfieldid']
                        add.checkprogram = row['checkprogram']
                        add.description = row['description']
                        add.mode = row['mode']
                        add.options = row['options']
                        add.values_query = row['values_query']
                        if add.type == 13:  # Gridfeld
                            add.values_query = bfaxml.xml_to_object(add.values_query)
                        add.indextableid = row['indextableid']
                        if not no_fielditems and row['items']:
                            add.items.extend(row['items'].split(';'))
                        if 'pos' in row:
                            add.pos = row['pos']
                        add.flags = row['flags']
                        add.rights = bitfarm_doc.BitfarmFieldRights()

                        if add.id in fieldrights:
                            for rec in fieldrights[add.id]:
                                r = bitfarm_doc.BitfarmFieldRight()
                                r.rights = int(rec[6])
                                r.arcid = rec[1]
                                if 'pos' in rec:
                                    r.fieldpos = rec['pos']
                                r.userid = rec['userid']
                                r.groupid = rec['groupid']
                                add.rights.add(r)
                        else:
                            if userid is not None or groupid is not None:
                                continue  # es wurde gegen userid oder groupid geprüft, es gab aber keine Feldberechtigung?
                                # dann wird das Feld nicht zurückgeliefert
                        fields.append(add)

        # TODO: Version 40 Pfad
        return fields

    def read_addfields2(self, db=None, arcid=None, userid=None, groupid=None, logid=''):
        """erzeugt ein Liste von Addfieldobjekten des gesamten oder eines bestimmten Archives. Dient zu Konfigurationszwecken.
                Es kann eine userid oder groupid übergeben werden, dann werden nur die Felder zurückgegeben, auf die der User Zugriff hat.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""
        fields = []
        if db is None:
            if self.db is None:
                return fields
            else:
                db = self.db
        if self.sys_arc is None:
            self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)

        if self.version == 35:
            if arcid is None:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_TBLName != '')
            else:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_ID == arcid)

            result = db.execute(sql).fetchall()

            if result:
                for row in result:
                    tblname = 'cfg_%s' % row[0]
                    arcid = row[1]
                    if tblname in self.tables:
                        cfg = sqla.Table(tblname, self.meta, autoload=True)
                        sql = sqla.select([cfg])

                        result2 = db.execute(sql).fetchall()

                        for row2 in result2:
                            add = bitfarm_doc.BitfarmAddfield()
                            add.name = row2['Disp_Name']
                            add.add = row2['Feld_Name']
                            add.type = row2['Typ']
                            add.checkprogram = row2['chck_prog']
                            add.mode = row2['Mode']
                            add.options = row2['Optionen']
                            add.rights = bitfarm_doc.BitfarmFieldRights()
                            if row2['Werte']:
                                add.items.extend(row2['Werte'].split(';'))
                            add.pos = row2['Pos']
                            if 'sys_cfgxs' in self.tables:
                                # gibt es einen Basiseintrag, dann haben wir nur lesenden Zugriff
                                sql = 'select * from sys_cfgxs where arc_id = %i and cfg_id = 0' % arcid
                                res = db.execute(sql).fetchone()
                                if res:
                                    r = bitfarm_doc.BitfarmFieldRight()
                                    r.rights = bitfarm_doc.RIGHT_FIELD_READ
                                    r.arcid = arcid
                                    r.fieldpos = row2['Pos']
                                    add.rights.add(r)
                                else:
                                    r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                                    r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                    r.arcid = arcid
                                    r.fieldpos = row2['Pos']
                                    add.rights.add(r)
                                # jetzt noch user oder gruppen rechte prüfen
                                # erst einzelnde user holen
                                sql = 'select * from sys_cfgxs where arc_id = %i and cfg_id = %i and grp_id = 0 group by usr_id' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        user = None
                                        if rec['usr_id'] != 0:
                                            user = self.read_user(userid=rec['usr_id'])
                                        if user is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()
                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = row2['Pos']
                                            r.user = user
                                            add.rights.add(r)
                                # dann gruppen
                                sql = 'select * from sys_cfgxs where arc_id = %i and cfg_id = %i and grp_id <> 0 group by grp_id' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        group = None
                                        if rec['grp_id'] != 0:
                                            group = self.read_group(groupid=rec['grp_id'])
                                        if group is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()

                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = row2['Pos']
                                            r.group = group
                                            add.rights.add(r)
                            else:
                                r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify für gruppe 0 / Alle
                                r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                r.arcid = arcid
                                r.group = bitfarm_doc.BitfarmGroup()
                                r.group.id = 0
                                r.fieldpos = row2['Pos']
                                add.rights.add(r)

                            fields.append(add)

        if self.version == 36:
            if self.sys_addfields is None:
                self.sys_addfields = sqla.Table('sys_addfields', self.meta, autoload=True)
            if self.sys_addfieldxs is None:
                self.sys_addfieldxs = sqla.Table('sys_addfieldxs', self.meta, autoload=True)

            if arcid is None:
                sql = sqla.select(
                    [self.sys_addfields.c.addfieldid, self.sys_addfields.c.fieldname, self.sys_addfields.c.displayname,
                     self.sys_addfields.c.type, self.sys_addfields.c['items'], self.sys_addfields.c.mode,
                     self.sys_addfields.c.options, self.sys_addfields.c.checkprogram, self.sys_addfields.c.values_query,
                     self.sys_addfields.c.indextableid, self.sys_addfields.c.description],
                    use_labels=False).order_by(self.sys_addfields.c.addfieldid)

            else:
                from_obj = self.sys_addfieldxs.join(self.sys_addfields,
                                                    self.sys_addfieldxs.c.fieldid == self.sys_addfields.c.addfieldid)
                sql = sqla.select(
                    [self.sys_addfieldxs.c.fieldid, self.sys_addfieldxs.c.pos, self.sys_addfieldxs.c.rights,
                     self.sys_addfields.c.addfieldid, self.sys_addfields.c.fieldname, self.sys_addfields.c.displayname,
                     self.sys_addfields.c.type,
                     self.sys_addfields.c['items'], self.sys_addfields.c.mode, self.sys_addfields.c.options,
                     self.sys_addfields.c.checkprogram, self.sys_addfields.c.values_query,
                     self.sys_addfields.c.indextableid, self.sys_addfields.c.description], from_obj=[from_obj],
                    whereclause=(self.sys_addfieldxs.c.arcid == arcid), use_labels=False).group_by(
                    self.sys_addfieldxs.c.fieldid)
            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_addfields2" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    add = bitfarm_doc.BitfarmAddfield()
                    add.name = row['displayname']
                    add.add = row['fieldname']
                    add.type = row['type']
                    add.id = row['addfieldid']
                    add.checkprogram = row['checkprogram']
                    add.description = row['description']
                    add.mode = row['mode']
                    add.options = row['options']
                    add.values_query = row['values_query']
                    if add.type == 13:  # Gridfeld
                        add.values_query = bfaxml.xml_to_object(add.values_query)
                    add.indextableid = row['indextableid']
                    if row['items']:
                        add.items.extend(row['items'].split(';'))
                    if 'pos' in row:
                        add.pos = row['pos']
                    add.rights = bitfarm_doc.BitfarmFieldRights()

                    # die Feldberechtigung holen, wenn arcid angegeben wurde, auf Felder dieses Archives einschränken,
                    # auf user oder gruppen rechte prüfen falls angegeben, da es entweder ein gruppen eintrag ODER ein usereintrag gibt, kann man beides direkt abfragen
                    # wurde userid oder groupid angegeben, das Feld nur zurückliefern, wenn Berechtigung besteht

                    sql = 'select xsid, arcid, fieldid, userid, groupid, pos, rights+0 from sys_addfieldxs where fieldid = %i' % (
                    row[0])
                    if arcid is not None:
                        sql += ' and arcid = %i' % arcid
                    if userid is not None:
                        sql += ' and (userid = %i or userid = 0) and groupid = 0' % userid
                    elif groupid is not None:
                        sql += ' and (groupid = %i or groupid = 0) and userid = 0' % groupid
                    else:
                        sql += ' and (userid = 0 and groupid = 0)'
                    res = db.execute(sql).fetchall()
                    if res:
                        for rec in res:
                            r = bitfarm_doc.BitfarmFieldRight()
                            r.rights = int(rec[6])
                            r.arcid = rec[1]
                            if 'pos' in rec:
                                r.fieldpos = rec['pos']
                            r.userid = rec['userid']
                            r.groupid = rec['groupid']
                            add.rights.add(r)
                    else:
                        if userid is not None or groupid is not None:
                            continue  # es wurde gegen userid oder groupid geprüft, es gab aber keine Feldberechtigung?
                            # dann wird das Feld nicht zurückgeliefert
                    fields.append(add)

        # TODO: Version 40 Pfad
        return fields

    def read_addfield(self, db=None, addid=None, name=None, logid=''):
        """erzeugt ein Addfieldobjekt. Es kann der Feldname oder die ID (hat Vorrang) übergeben werden.
            Konnte das Feld nicht gefunden werden, wird None zurückgegeben.
            Das FieldRights-Objekt trägt ein Standard Modify-Eintrag, ohne Zuordnung zu User, Gruppe, Archiv.
            Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
            genommen, wenn mit connect_db() erzeugt"""

        if db is None:
            if self.db is None:
                return None
            else:
                db = self.db

        if name is None and addid is None:
            return None

        if self.sys_arc is None:
            self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)

        if self.version == 35:
            sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                self.sys_arc.c.ARC_TBLName != '')

            result = db.execute(sql).fetchall()

            if result:
                for row in result:
                    tblname = 'cfg_%s' % row[0]
                    arcid = row[1]
                    if tblname in self.tables:
                        cfg = sqla.Table(tblname, self.meta, autoload=True)
                        sql = sqla.select([cfg])

                        if addid:
                            sql.append_whereclause(cfg.c.CFG_ID == addid)
                        elif name:
                            sql.append_whereclause(cfg.c.Disp_Name == name)

                        result2 = db.execute(sql).fetchone()

                        if result2:
                            add = bitfarm_doc.BitfarmAddfield()
                            add.name = result2['Disp_Name']
                            add.add = result2['Feld_Name']
                            add.type = result2['Typ']
                            add.checkprogram = result2['chck_prog']
                            add.mode = result2['Mode']
                            add.options = result2['Optionen']
                            add.rights = bitfarm_doc.BitfarmFieldRights()
                            r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                            r.arcid = arcid
                            add.rights.add(r)

                            if result2['Werte']:
                                add.items.extend(result2['Werte'].split(';'))
                            add.pos = result2['Pos']
                            return add

            return None

        if self.version == 36:
            if self.sys_addfields is None:
                self.sys_addfields = sqla.Table('sys_addfields', self.meta, autoload=True)

            sql = sqla.select(
                [self.sys_addfields.c.addfieldid, self.sys_addfields.c.fieldname, self.sys_addfields.c.displayname,
                 self.sys_addfields.c.type, self.sys_addfields.c['items'], self.sys_addfields.c.mode,
                 self.sys_addfields.c.options, self.sys_addfields.c.checkprogram, self.sys_addfields.c.values_query,
                 self.sys_addfields.c.indextableid, self.sys_addfields.c.description],
                use_labels=False).order_by(self.sys_addfields.c.addfieldid)

            if addid:
                sql.append_whereclause(self.sys_addfields.c.addfieldid == addid)
            elif name:
                sql.append_whereclause(self.sys_addfields.c.displayname == name)

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_addfield" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    add = bitfarm_doc.BitfarmAddfield()
                    add.name = row['displayname']
                    add.add = row['fieldname']
                    add.type = row['type']
                    add.id = row['addfieldid']
                    add.checkprogram = row['checkprogram']
                    add.description = row['description']
                    add.mode = row['mode']
                    add.options = row['options']
                    add.values_query = row['values_query']
                    if add.type == 13:  # Gridfeld
                        add.values_query = bfaxml.xml_to_object(add.values_query)
                    add.indextableid = row['indextableid']
                    if row['items']:
                        add.items.extend(row['items'].split(';'))
                    if 'pos' in row:
                        add.pos = row['pos']
                    add.rights = bitfarm_doc.BitfarmFieldRights()
                    r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                    r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                    add.rights.add(r)
                    return add

            return None

            # TODO: Version 40 Pfad

    def read_statefields(self, db=None, arcid=None, userid=None, groupid=None, logid=''):
        """erzeugt ein Liste von Statefield-objekten des gesamten oder eines bestimmten Archives. Dient zu Konfigurationszwecken.
                Es kann eine userid oder groupid übergeben werden, dann werden nur die Felder zurückgegeben, auf die der User Zugriff hat.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""
        fields = []
        if db is None:
            if self.db is None:
                return fields
            else:
                db = self.db

        if self.sys_arc is None:
            self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)

        if self.version == 35:
            if arcid is None:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_TBLName != '')
            else:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_ID == arcid)
            #            result = db.execute(sql).fetchone()

            result = db.execute(sql).fetchall()

            if result:
                for row in result:
                    tblname = 'sts_%s' % row[0]
                    arcid = row[1]
                    if tblname in self.tables:
                        sts = sqla.Table(tblname, self.meta, autoload=True)
                        sql = sqla.select([sts]).order_by(sqla.asc(sts.c.sts_id))

                        result2 = db.execute(sql).fetchall()

                        pos = 0
                        for row2 in result2:
                            pos += 1
                            state = bitfarm_doc.BitfarmStatefield()
                            state.name = row2['sts_Name']
                            state.id = row2['sts_id']
                            state.addfield = row2['sts_ZusName']
                            state.addfieldid = row2['sts_ZusFeldID']
                            state.func = row2['sts_Func']
                            state.rights = bitfarm_doc.BitfarmFieldRights()
                            if 'sys_stsxs' in self.tables:
                                # gibt es einen Basiseintrag, dann haben wir nur lesenden Zugriff
                                sql = 'select * from sys_stsxs where arc_id = %i and sts_id = 0' % arcid
                                res = db.execute(sql).fetchone()
                                if res:
                                    r = bitfarm_doc.BitfarmFieldRight()
                                    r.rights = bitfarm_doc.RIGHT_FIELD_READ
                                    r.arcid = arcid
                                    r.fieldpos = pos
                                    state.rights.add(r)
                                else:
                                    r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                                    r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                    r.arcid = arcid
                                    r.fieldpos = pos
                                    state.rights.add(r)
                                # jetzt noch user oder gruppen rechte prüfen
                                # erst einzelnde user holen
                                sql = 'select * from sys_stsxs where arc_id = %i and sts_id = %i and grp_id = 0 group by usr_id' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        user = None
                                        if rec['usr_id'] != 0:
                                            user = self.read_user(userid=rec['usr_id'])
                                        if user is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()
                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = pos
                                            r.user = user
                                            state.rights.add(r)
                                # dann gruppen
                                sql = 'select * from sys_stsxs where arc_id = %i and sts_id = %i and grp_id <> 0 group by grp_id' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        group = None
                                        if rec['grp_id'] != 0:
                                            group = self.read_group(groupid=rec['grp_id'])
                                        if group is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()
                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = pos
                                            r.group = group
                                            state.rights.add(r)
                            else:
                                r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                                r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                r.arcid = arcid
                                r.fieldpos = pos
                                state.rights.add(r)

                            fields.append(state)

            return fields

        if self.version == 36:
            if self.sys_statefields is None:
                self.sys_statefields = sqla.Table('sys_statefields', self.meta, autoload=True)
            if self.sys_statefieldxs is None:
                self.sys_statefieldxs = sqla.Table('sys_statefieldxs', self.meta, autoload=True)

            fieldrights = {}
            # jetzt noch user oder gruppen rechte prüfen, da es entweder ein gruppen eintrag ODER ein usereintrag gibt, kann man beides direkt abfragen
            sql = 'select xsid, arcid, fieldid, userid, groupid, pos, rights+0 from sys_statefieldxs where fieldid <> 0'
            if arcid is not None:
                sql += ' and arcid = %i' % arcid

            if userid is not None:
                grouplist = self.resolve_membership(db=db, userid=userid, logid=logid)
                if not grouplist:
                    sql += ' and userid = %i and groupid = 0' % userid
                else:
                    sql += ' and (userid = %i or (userid = 0 and groupid in (%s)))' % (userid, ",".join([str(x) for x in grouplist]))

            elif groupid is not None:
                sql += ' and groupid = %i and userid = 0' % groupid

            # if userid is not None:
            #     sql += ' and (userid = %i or userid = 0) and groupid = 0' % userid
            # elif groupid is not None:
            #     sql += ' and (groupid = %i or groupid = 0) and userid = 0' % groupid
            # else:
            #     sql += ' and (userid = 0 and groupid = 0)'

            try:
                res = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_statefieldxs" % logid)
                res = self.try_again(sql=sql, db=db, fetchall=True)

            for row in res:
                if row[2] in fieldrights:
                    fieldrights[row[2]].append(row)
                else:
                    fieldrights[row[2]] = [row]

            if arcid is None:
                sql = sqla.select(
                    [self.sys_statefields.c.stateid, self.sys_statefields.c.displayname, self.sys_statefields.c.bitmask,
                     self.sys_statefields.c.addfieldid, self.sys_statefields.c.function,
                     self.sys_statefields.c.indexcolumnid], use_labels=False).order_by(self.sys_statefields.c.stateid)

            else:
                from_obj = self.sys_statefieldxs.join(self.sys_statefields,
                                                      self.sys_statefieldxs.c.fieldid == self.sys_statefields.c.stateid)
                sql = sqla.select(
                    [self.sys_statefields.c.stateid, self.sys_statefields.c.displayname, self.sys_statefields.c.bitmask,
                     self.sys_statefields.c.addfieldid, self.sys_statefields.c.function,
                     self.sys_statefields.c.indexcolumnid,
                     self.sys_statefieldxs.c.arcid, self.sys_statefieldxs.c.userid, self.sys_statefieldxs.c.groupid,
                     self.sys_statefieldxs.c.pos, self.sys_statefieldxs.c.rights], from_obj=[from_obj],
                    whereclause=(self.sys_statefieldxs.c.arcid == arcid), use_labels=False).group_by(
                    self.sys_statefieldxs.c.fieldid, self.sys_statefieldxs.c.userid, self.sys_statefieldxs.c.groupid)

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_statefields" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    state = bitfarm_doc.BitfarmStatefield()
                    state.name = row['displayname']
                    state.id = row['stateid']
                    state.bitmask = row['bitmask']
                    state.addfieldid = row['addfieldid']
                    state.func = row['function']
                    state.indexcolumnid = row['indexcolumnid']
                    if 'pos' in row:
                        state.pos = row['pos']
                    state.rights = bitfarm_doc.BitfarmFieldRights()

                    # # jetzt noch user oder gruppen rechte prüfen, da es entweder ein gruppen eintrag ODER ein usereintrag gibt, kann man beides direkt abfragen
                    # sql = 'select xsid, arcid, fieldid, userid, groupid, pos, rights+0 from sys_statefieldxs where fieldid = %i' % (state.id)
                    # if arcid is not None:
                    #     sql += ' and arcid = %i' % arcid
                    # if userid is not None:
                    #     sql += ' and (userid = %i or userid = 0) and groupid = 0' % userid
                    # elif groupid is not None:
                    #     sql += ' and (groupid = %i or groupid = 0) and userid = 0' % groupid
                    # else:
                    #     sql += ' and (userid = 0 and groupid = 0)'
                    #
                    # try:
                    #     res = db.execute(sql).fetchall()
                    # except:
                    #     self.log.exception(u"%s Fehler bei read_statefields" % logid)

                    if state.id in fieldrights:
                        for rec in fieldrights[state.id]:
                            r = bitfarm_doc.BitfarmFieldRight()
                            r.rights = int(rec[6])
                            r.arcid = rec[1]
                            if 'pos' in rec:
                                r.fieldpos = rec['pos']
                            r.userid = rec['userid']
                            r.groupid = rec['groupid']
                            state.rights.add(r)
                    else:
                        if userid is not None or groupid is not None:
                            continue  # es wurde gegen userid oder groupid geprüft, es gab aber keine Feldberechtigung?
                            # dann wird das Feld nicht zurückgeliefert
                    fields.append(state)

            return fields

        # TODO: Version 40 Pfad

        return fields

    def read_statefields2(self, db=None, arcid=None, userid=None, groupid=None, logid=''):
        """erzeugt ein Liste von Statefield-objekten des gesamten oder eines bestimmten Archives. Dient zu Konfigurationszwecken.
                Es kann eine userid oder groupid übergeben werden, dann werden nur die Felder zurückgegeben, auf die der User Zugriff hat.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""
        fields = []
        if db is None:
            if self.db is None:
                return fields
            else:
                db = self.db

        if self.sys_arc is None:
            self.sys_arc = sqla.Table('sys_arc', self.meta, autoload=True)

        if self.version == 35:
            if arcid is None:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_TBLName != '')
            else:
                sql = sqla.select([self.sys_arc.c.ARC_TBLName, self.sys_arc.c.ARC_ID]).where(
                    self.sys_arc.c.ARC_ID == arcid)
            #            result = db.execute(sql).fetchone()

            result = db.execute(sql).fetchall()

            if result:
                for row in result:
                    tblname = 'sts_%s' % row[0]
                    arcid = row[1]
                    if tblname in self.tables:
                        sts = sqla.Table(tblname, self.meta, autoload=True)
                        sql = sqla.select([sts]).order_by(sqla.asc(sts.c.sts_id))

                        result2 = db.execute(sql).fetchall()

                        pos = 0
                        for row2 in result2:
                            pos += 1
                            state = bitfarm_doc.BitfarmStatefield()
                            state.name = row2['sts_Name']
                            state.id = row2['sts_id']
                            state.addfield = row2['sts_ZusName']
                            state.addfieldid = row2['sts_ZusFeldID']
                            state.func = row2['sts_Func']
                            state.rights = bitfarm_doc.BitfarmFieldRights()
                            if 'sys_stsxs' in self.tables:
                                # gibt es einen Basiseintrag, dann haben wir nur lesenden Zugriff
                                sql = 'select * from sys_stsxs where arc_id = %i and sts_id = 0' % arcid
                                res = db.execute(sql).fetchone()
                                if res:
                                    r = bitfarm_doc.BitfarmFieldRight()
                                    r.rights = bitfarm_doc.RIGHT_FIELD_READ
                                    r.arcid = arcid
                                    r.fieldpos = pos
                                    state.rights.add(r)
                                else:
                                    r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                                    r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                    r.arcid = arcid
                                    r.fieldpos = pos
                                    state.rights.add(r)
                                # jetzt noch user oder gruppen rechte prüfen
                                # erst einzelnde user holen
                                sql = 'select * from sys_stsxs where arc_id = %i and sts_id = %i and grp_id = 0' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        user = None
                                        if rec['usr_id'] != 0:
                                            user = self.read_user(userid=rec['usr_id'])
                                        if user is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()
                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = pos
                                            r.user = user
                                            state.rights.add(r)
                                # dann gruppen
                                sql = 'select * from sys_stsxs where arc_id = %i and sts_id = %i and grp_id <> 0 group by grp_id' % (
                                    arcid, row2[0])
                                res = db.execute(sql).fetchall()
                                if res:
                                    for rec in res:
                                        group = None
                                        if rec['grp_id'] != 0:
                                            group = self.read_group(groupid=rec['grp_id'])
                                        if group is not None:
                                            r = bitfarm_doc.BitfarmFieldRight()
                                            r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                            r.arcid = arcid
                                            r.fieldpos = pos
                                            r.group = group
                                            state.rights.add(r)
                            else:
                                r = bitfarm_doc.BitfarmFieldRight()  # standardrecht modify
                                r.rights = bitfarm_doc.RIGHT_FIELD_MODIFY
                                r.arcid = arcid
                                r.fieldpos = pos
                                state.rights.add(r)

                            fields.append(state)

            return fields

        if self.version == 36:
            if self.sys_statefields is None:
                self.sys_statefields = sqla.Table('sys_statefields', self.meta, autoload=True)
            if self.sys_statefieldxs is None:
                self.sys_statefieldxs = sqla.Table('sys_statefieldxs', self.meta, autoload=True)

            if arcid is None:
                sql = sqla.select(
                    [self.sys_statefields.c.stateid, self.sys_statefields.c.displayname, self.sys_statefields.c.bitmask,
                     self.sys_statefields.c.addfieldid, self.sys_statefields.c.function],
                    use_labels=False).order_by(self.sys_statefields.c.stateid)

            else:
                from_obj = self.sys_statefieldxs.join(self.sys_statefields,
                                                      self.sys_statefieldxs.c.fieldid == self.sys_statefields.c.stateid)
                sql = sqla.select(
                    [self.sys_statefields.c.stateid, self.sys_statefields.c.displayname, self.sys_statefields.c.bitmask,
                     self.sys_statefields.c.addfieldid, self.sys_statefields.c.function, self.sys_statefieldxs.c.arcid,
                     self.sys_statefieldxs.c.userid, self.sys_statefieldxs.c.groupid, self.sys_statefieldxs.c.pos,
                     self.sys_statefieldxs.c.rights], from_obj=[from_obj],
                    whereclause=(self.sys_statefieldxs.c.arcid == arcid), use_labels=False).group_by(
                    self.sys_statefieldxs.c.fieldid, self.sys_statefieldxs.c.userid, self.sys_statefieldxs.c.groupid)

            try:
                result = db.execute(sql).fetchall()
            except:
                self.log.exception(u"%s Fehler bei read_statefields" % logid)
                result = self.try_again(sql=sql, db=db, fetchall=True)

            if result:
                for row in result:
                    state = bitfarm_doc.BitfarmStatefield()
                    state.name = row['displayname']
                    state.id = row['stateid']
                    state.bitmask = row['bitmask']
                    state.addfieldid = row['addfieldid']
                    state.func = row['function']
                    if 'pos' in row:
                        state.pos = row['pos']
                    state.rights = bitfarm_doc.BitfarmFieldRights()

                    # jetzt noch user oder gruppen rechte prüfen, da es entweder ein gruppen eintrag ODER ein usereintrag gibt, kann man beides direkt abfragen
                    sql = 'select xsid, arcid, fieldid, userid, groupid, pos, rights+0 from sys_statefieldxs where fieldid = %i' % state.id
                    if arcid is not None:
                        sql += ' and arcid = %i' % arcid
                    if userid is not None:
                        sql += ' and (userid = %i or userid = 0) and groupid = 0' % userid
                    elif groupid is not None:
                        sql += ' and (groupid = %i or groupid = 0) and userid = 0' % groupid
                    else:
                        sql += ' and (userid = 0 and groupid = 0)'

                    try:
                        res = db.execute(sql).fetchall()
                    except:
                        self.log.exception(u"%s Fehler bei read_statefields" % logid)
                        res = self.try_again(sql=sql, db=db, fetchall=True)

                    if res:
                        for rec in res:
                            r = bitfarm_doc.BitfarmFieldRight()
                            r.rights = int(rec[6])
                            r.arcid = rec[1]
                            if 'pos' in row:
                                r.fieldpos = row['pos']
                            r.userid = rec['userid']
                            r.groupid = rec['groupid']
                            state.rights.add(r)
                    else:
                        if userid is not None or groupid is not None:
                            continue  # es wurde gegen userid oder groupid geprüft, es gab aber keine Feldberechtigung?
                            # dann wird das Feld nicht zurückgeliefert
                    fields.append(state)

            return fields

        # TODO: Version 40 Pfad

        return fields

    def read_statefield(self, db=None, stateid=None, logid=''):
        """holt ein Statefield-objekten anhand seiner field.id. Dient zu Konfigurationszwecken.
                Es kann eine Datenbank-Connection übergeben werden (in Threads) oder es wird die Verbindung aus con.db
                genommen, wenn mit connect_db() erzeugt"""
        state = None
        if db is None:
            if self.db is None:
                return state
            else:
                db = self.db
        if stateid is None:
            return state

        if self.version == 36:
            if self.sys_statefields is None:
                self.sys_statefields = sqla.Table('sys_statefields', self.meta, autoload=True)

            sql = sqla.select(
                [self.sys_statefields.c.stateid, self.sys_statefields.c.displayname, self.sys_statefields.c.bitmask,
                 self.sys_statefields.c.addfieldid, self.sys_statefields.c.function],
                whereclause=(self.sys_statefields.c.stateid == stateid), use_labels=False)

            try:
                row = db.execute(sql).fetchone()
            except:
                self.log.exception(u"%s Fehler bei read_statefield" % logid)
                row = self.try_again(sql=sql, db=db, fetchone=True)

            if row:
                state = bitfarm_doc.BitfarmStatefield()
                state.name = row['displayname']
                state.id = row['stateid']
                state.bitmask = row['bitmask']
                state.addfieldid = row['addfieldid']
                state.func = row['function']
                state.rights = bitfarm_doc.BitfarmFieldRights()

        return state

    def get_identifier(self, _profile=None, _server=None):
        """Die Funktion ermittelt Umgebungsinformationen des Rechners"""
        if _profile is None:
            _profile = self.profile

        if _server is None:
            _server = self.server

        cpu = cpuinfo.get_cpu_info()
        brand = "unknown"
        arch = "unknown"
        if 'brand' in cpu:
            brand = cpu['brand']
        if 'brand_raw' in cpu:
            brand = cpu['brand_raw']
        if 'arch' in cpu:
            arch = cpu['arch']

        _ret = [_profile, _server, socket.getfqdn(), brand, arch]

        sid = ''
        srvuser = None
        try:
            srvctrl = win32service.OpenSCManager(None, None, win32service.SC_MANAGER_CONNECT)
        except:
            self.log.error("Keine Verbindung zum Dienst-Manager")
        else:
            try:
                hndl = win32service.OpenService(srvctrl, 'bfaServer36', win32service.SC_MANAGER_CONNECT)
            except:
                self.log.error("Dienst bfaServer36 nicht gefunden")
            else:
                srvuser = win32service.QueryServiceConfig(hndl)[7]

                if srvuser.find('@') >= 0:
                    srvuser = srvuser.split('@')[0]
                elif srvuser.find('\\') >= 0:
                    srvuser = srvuser.split('\\')[1]

        if not srvuser:
            srvuser = win32api.GetUserName()

        try:
            pysid = win32security.LookupAccountName(None, srvuser)[0]
        except:
            self.log.error('sid konnte nicht ermittelt werden')
        else:
            sid = win32security.ConvertSidToStringSid(pysid)

        _ret.append(srvuser)
        _ret.append(sid)

        return _ret


class BitfarmConnections(bitfarm_baseobj.BitfarmBaseConnections):
    """
        BitfarmConnections-Klasse. Diese Klasse stellt den Anwendungen ein oder mehrere Profile zur Verfügung.
        Das erste geladene Profil wird in BitfarmConnections.main abgelegt.
        Es können weitere Profile geladen werden in dem weitere Instanzen mit einem Profilnamen erzeugt werden.
        Auf alle Profile kann auch über den Namen zugegriffen werden  BitfarmConnections[Profil].
        Die create_db_engine-Methode erstellt eine Datenbank-Engine für alle geladenen Profile.
        Soll in Threads auf die Datenbank zugegriffen werden, so hinterlegt man innerhalb des Threads in eine lokale Variable
        ein Connection-Objekt, zB self.db = BitfarmConnections.main.engine.connect()
    @version:
    """
    version = __version__
    log = logging.getLogger("Main.BitfarmConnections")
    log.setLevel(logging.WARNING)

    def __init__(self, *name, **kwargs):
        """Erzeugt eine BitfarmConnections Instanz, es kann der Name / oder Pfad eine Profils übergeben werden, bzw. eine Liste von Profilen"""
        bitfarm_baseobj.BitfarmBaseConnections.__init__(self, *name, **kwargs)

        self.connections = {}
        """@ivar: Dictonary, welches alle Connections trägt
            @type: dictonary"""
        self.main = None
        """@ivar: hier ist das erste geladene Profil hinterlegt, die Main.Connection. Über diese Connection wird die sys_queue und die config_tabellen erreicht
            @type: BitfarmCon Instanz"""
        self.profiles = []
        """@ivar: die Namen der geladenen Profile
        @type: list"""
        self.db = {}
        """@ivar: Dictonary, welches alle Datenbank-Connections trägt, wird durch BitfarmConnections.create_db() gefüllt
            @type: dictonary"""

        if len(name) > 0:
            if type(name[0]) is list:
                name = list(name[0])
            for profile in name:
                if os.path.isfile(os.path.abspath(os.path.expandvars(profile))) and os.path.exists(
                        os.path.abspath(os.path.expandvars(profile))):
                    profilename = os.path.splitext(os.path.basename(profile))[0].lower()
                else:
                    profilename = profile.lower()

                if profilename[-4:] == '.con':  # beliebter Fehler Profilname mit Endung .con angeben! Soll man nicht, ist dann aber jetzt auch egal
                    profilename = profilename[:-4]
                self.connections[profilename] = BitfarmCon(profile, **kwargs)
                self.profiles.append(profilename)

                if self.main is None:
                    self.main = self.connections[profilename]

    # def __getitem__(self, name):
    #     if name not in self.connections:
    #         return self.main
    #     else:
    #         return self.connections[name]
    #
    # def __iter__(self):
    #     for profile in self.profiles:
    #         try:
    #             yield self.connections[profile]
    #         except Exception:
    #             pass

    def create_db_engine(self, bind_tables=True, autocommit=True, echo=False, user=None, password=None, pool_size=10):
        """create_db_engine Methode: erzeugt für alle geladenen Profile eine Datenbank-Engine.
            Optional kann bind_tables (default=True) auf False gesetzt werden. Dann werden keine Tabellen-Bindings ans con Objekt
            vorgenommen, man muss selbst für die benötigten Bindings sorgen (bsp: self.sys_timer = sqla.Table('sys_timer', self.meta,  autoload = True))
            oder über die Funktion bind_tables()
        """
        for profile in self.connections:
            try:
                self.connections[profile].create_db_engine(bind_tables=bind_tables, autocommit=autocommit, echo=echo,
                                                           user=user, password=password, pool_size=pool_size)
            except bitfarm_errors.DatabaseConError:
                self.log.error(u'Fehler bei create_db_engine')
                self.connections[profile].engine = None
            except bitfarm_errors.DatabaseNotConnectedError:
                self.log.error(u'Fehler bei create_db_engine')
                self.connections[profile].engine = None
            except:
                self.log.exception(u'Fehler bei create_db_engine')
                self.connections[profile].engine = None
            else:
                if profile == self.main.profile:
                    self.main.engine = self.connections[profile].engine

    def connect_db(self):
        """connect_db Methode. erzeugt für alle geladenenen Profile eine Datenbank-Connection
        und hinterlegt diese in BitfarmConnections[profile].db bzw. BitfarmConnections.main.db
        Achtung! sollte nicht in Multi-Thread-Anwendungen
        verwendet werden. Dort immer in den einzelnen Threads eine eigene Connection erzeugen."""
        for profile in self.connections:
            try:
                self.connections[profile].connect_db()
            except:
                self.log.exception(u'Fehler bei connect_db')
                self.connections[profile].db = None
            else:
                self.db[profile] = self.connections[profile].db
                if profile == self.main.profile:
                    self.main.db = self.connections[profile].db

    def bind_tables(self):
        """bind_tables Methode. Bindet für alle geladenenen Profile die Datenbank-Tabellen an
        die entsprechenden Attribute der BitfarmCon Objekte.
        """
        for profile in self.connections:
            try:
                self.connections[profile].bind_tables()
            except:
                self.log.exception(u'Fehler bei bind_tables')

    def disconnect_db(self):
        """disconnect_db Methode. Schliesst für alle geladenenen Profile die Datenbank-Connection."""
        for profile in self.connections:
            try:
                self.connections[profile].disconnect_db()
            except:
                self.log.exception('Fehler bei disconnect_db')
                self.connections[profile].db = None
            else:
                self.db = {}
                if profile == self.main.profile:
                    self.main.db = None

    # def add(self, name):
    #     """add - Methode: fügt eine weiteres Profil hinzu"""
    #     if name not in self.connections:
    #         self.connections[name] = BitfarmCon(name)
    #         self.connections[name].create_db_engine()
    #         self.profiles.append(name)
    #
    # def read_ini(self):
    #     """read_ini Methode: liest aus der Scripts.ini und speichert im ConfigDictonary jedes Profils"""
    #     for profile in self.connections.keys():
    #         self.connections[profile].read_ini()


class BitfarmWorkflowProcessing(object):
    """Die Workflow Processing-Klasse mit allen Methoden
    Benötigt ein BitfarmCon oder ein BitfarmConnections-Objekt. Vom Connections-Objekt wird
    die Main-Engine (BitfarmConnections.main) benutzt.
    Jede Instanz von BitfarmWorkflowProcessing erzeugt eine eigene Datenbank-Connection auf die Main-Connection-Engine.
    Daher muss in Threads jeweils eine eigene Instanz von BitfarmDocProcessing erzeugt werden.
    Optional kann ein Dictionary (_con_db) mit einer DB-Connection pro Profil übergeben werden, wenn von einem übergeordneten Thread
    die Db-Connections durchgereicht werden sollen.  {'Profil':sqla.engine.base.Connection}
    Diese Connections werden in self.db hinterlegt"""

    DOC_WF = 0
    USER_WF = 1
    TR_SPOOLER = 0
    TR_ARCHSRV = 1
    TR_VIEWER = 2
    log = logging.getLogger("Main.bfaWflowProc")
    doc_classes = {}  # Dictonary aller zur Laufzeit zur Verfügung stehenden Dokumentenklassen. {Klassenname : BitfarmDocClass}
    doc_class_lock = threading.Lock()  # Lock, wenn doc_classes eingelesen wird.
    version = __version__

    def __init__(self, _con, _con_db=None, name=u""):
        self.mypath = os.path.split(os.path.abspath(sys.argv[0]))[0]
        if _con is None or not isinstance(_con, BitfarmConnections):
            raise NoneConError()

        self.con = _con
        self.version = _con.main.version
        self.name = name  # hier können threads ihren name bezeichner durchreichen

        if _con_db is not None and type(_con_db) is dict:
            self.db = _con_db
        else:
            self.db = {}
            for _profile in self.con.connections:
                if self.con[_profile].engine is not None:
                    self.db[_profile] = self.con[_profile].engine.connect()
                    if not isinstance(self.db[_profile], sqla.engine.base.Connection):
                        raise DatabaseConError(_con[_profile].confile)

        if self.con.main.profile in self.db:
            self.db['main'] = self.db[self.con.main.profile]

    def get_workflow(self, profile=None, wf_type=None, wf_trigger=None, arcid=None, gdocid=None, userid=None,
                     groupid=None, importpath=None, date=None, use_default=False):
        """Die Methode ermittelt alle in Frage kommenden (durch die Auslöser gefundenen)
           Workflows, und gibt diese als Liste von BitfarmWorkflow-Instanzen nach Priorität zurück.
           Über die Liste kann iteriert und die Bedingungen geprüft werden.
           Die Tasks der gefundenen Workflows können dem TaskManager zur Ausführung übergeben werden.
           Sollten die übergebenen Bedingungen KEINEN Workflow finden, werden die Bedingungen Schritt für Schritt
           reduziert, solange bis eine Abfrage ein Ergebnis liefert. Die Bedingungen werden entsprechend ihrer Priorität
           abgearbeitet, je spezieller und eindeutiger eine Bedingung umso höher ihre Priorität.
           Die höchste Priorität hat dem entsprechend die gdocid, die arcid die niedrigste.
           Hier die Reihenfolge: gdocid, date, usrid, groupid, importpath, arcid
           Als Arcid kann auch die ID eines Lagers oder einer Archivgruppe angegeben werden, diese
           Id wird dann aufgelöst und alle gefundenen Arcids in einer Schleife abgearbeitet.
           Liefert wenn nichts gefunden wurde, oder im Fehlerfall eine leere Liste zurück.
           Wird use_default = True gesetzt, so werden auch ggf. angelegte Default-Workflows gefunden, (diese müssen
           'default' genannt werden). Werden default-Workflows angelegt für bestimmte Auslöser (zB arcid oder userid),
           dann werden diese gefunden unabhängig von wf_type und wf_trigger, ansonsten wird der default-workflow für
           genau den wf_type und wf_trigger gefunden.
           Entsprechend sollte man zwei Default-Workflows anlegen:
               - Default-Workflow für Dokumente, der vom Collector gefunden wird  (wf_type = 0, wf_trigger =0)
               - Default-Workflow für Dokumente, der von der Archivierung gefunden wird  (wf_type = 0, wf_trigger =1)
           wf_trigger = 0   Workflows für den Collector (in sys_wfd)
           wf_trigger = 1   Workflows für die Archivierung (in sys_wfd)
           wf_trigger = 2   Workflows für Archivierung/Server (in sys_wfd) zB. late-workflows, interval-workflows
           wf_trigger = 3   Workflows die in Verarbeitung sind (in sys_workflow)
        """
        # TODO: Archivgruppen implementieren
        _ret = []
        # eingabe prüfen

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return _ret

        if wf_type is None or wf_trigger is None:
            return _ret

        if type(wf_type) is not int:
            return _ret

        if not -1 < wf_type < 2:
            return _ret

        if type(wf_trigger) is not int:
            return _ret

        if not -1 < wf_trigger < 4:
            return _ret

        lst_args = []  # übergebene Argumente nach Priorität in die Liste aufnehmen

        if gdocid:
            lst_args.append(('gdocid', gdocid))
        if date:
            lst_args.append(('date', date))
        if userid:
            lst_args.append(('userid', userid))
        if groupid:
            lst_args.append(('groupid', groupid))
        if importpath:
            lst_args.append(('importpath', importpath))
        if arcid:
            lst_args.append(('arcid', arcid))
        lst_args.append(('base', (wf_type, wf_trigger)))

        while len(lst_args) > 0:
            # argumente für die Query sammeln
            lst_where = []
            lst_default = []
            lst_order = ['wfprio desc']

            # Argumente auf die Query verteilen
            for args in lst_args:
                if args[0] == 'gdocid':
                    lst_where.append("(gdocid = '%s')" % args[1])
                    lst_default.append("(gdocid = '%s' and wfname = 'default')" % args[1])
                    lst_order.append('gdocid desc')
                    continue
                if args[0] == 'date':
                    lst_where.append("(tstamp <= '%s')" % args[1])
                    # tstamp können wir nicht zu default hinzufügen, weil diese abfrage immer den ersten default-Workflow findet
                    # unabhängig was es sonst noch für bedingungen gab
                    # lst_default.append("(tstamp <= '%s' and wf_name = 'default')" %args[1])
                    lst_order.append('tstamp desc')
                    continue
                if args[0] == 'userid':
                    lst_where.append("(userid = '%s')" % args[1])
                    lst_default.append("(userid = '%s' and wfname = 'default')" % args[1])
                    lst_order.append('userid desc')
                    continue
                if args[0] == 'groupid':
                    lst_where.append("(groupid = '%s')" % args[1])
                    lst_default.append("(groupid = '%s' and wfname = 'default')" % args[1])
                    lst_order.append('groupid desc')
                    continue
                if args[0] == 'importpath':
                    lst_where.append("(importpath = '%s')" % args[1])
                    lst_default.append("(importpath = '%s' and wfname = 'default')" % args[1])
                    lst_order.append('importpath desc')
                    continue
                if args[0] == 'arcid':
                    lst_where.append("(arcid = %s)" % args[1])
                    lst_default.append("(arcid = %s and wfname = 'default')" % args[1])
                    lst_order.append('arcid desc')
                    continue
                if args[0] == 'base':
                    lst_where.append("(wftype = %s and wftrigger = %s)" % (args[1], wf_trigger))

            # select zusammenbauen
            if len(lst_where) > 0:
                where = " and ".join(lst_where)
            else:
                where = ""

            if use_default:
                if len(lst_default) > 0:
                    default = " or " + " or ".join(lst_default)
                else:
                    default = " and wfname = 'default'"
            else:
                default = ""

            if len(lst_order) > 0:
                order = ",".join(lst_order)
            else:
                order = ""

            if len(where) > 0:
                # condition = " and ((" + where + ")" +default +")"
                condition = "((" + where + ")" + default + ")"
            else:
                condition = default

            wf_tabelle = 'sys_wfd' if wf_trigger < 3 else 'sys_workflow'

            # sql = "select * from sys_wfd where wf_typ = %s and wf_trigger = %s %s order by %s" %(wf_type, wf_trigger, condition, order)
            sql = "select * from %s where %s order by %s" % (wf_tabelle, condition, order)

            # self.log.debug('get_workflow: %s' %sql)
            try:
                result = _db.execute(sql).fetchall()
            except:
                self.log.exception('Fehler bei der SQL-Abfrage %s' % sql)
                result = _con.try_again(sql=sql, db=_db, fetchall=True)

            if result:
                for data in result:
                    self.log.debug('Workflow gefunden wfid=%s wfname=%s' % (data[0], data[1]))
                    wf = self._create_workflow_with_data(data, _con)
                    _ret.append(wf)
                break
            else:
                lst_args.pop(0)  # wir entfernen das vordere (höchste) Argument und führen die Query nochmal durch

        return _ret

    def get_workflow_by_id(self, profile=None, wf_id=None, wf_name=None, wf_trigger=0):
        """holt einen bestimmten Workflow anhand seiner ID oder Namens aus der Datenbank.
        Als ID kann auch eine Liste an IDs (Integer) übergeben werden
        Wird als wf_trigger >= 3 übergeben, so wird der Workflow aus der sys_workflow geladen (temporärer Workflow),
        dies wird von create_workflow_step und end_workfow_step benötigt.
        Es wird eine Liste zurückgegeben, mit einem oder mehreren BitfarmWorkflow Objekten. Wurde kein Workflow gefunden,
        wird None zurückgegeben"""

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        if wf_id:
            if type(wf_id) is list:
                wf_id = [str(x) for x in wf_id if type(x) is int]
                if len(wf_id) == 0:
                    return None
            elif type(wf_id) is not int:
                return None

        if wf_name:
            if type(wf_name) is not str:
                return None

        wf_tabelle = 'sys_wfd' if wf_trigger < 3 else 'sys_workflow'

        if wf_id:
            if type(wf_id) is list:
                sql = "Select * from %s where wfid in (%s)" % (wf_tabelle, ",".join(wf_id))
            else:
                sql = "Select * from %s where wfid = %i" % (wf_tabelle, wf_id)
        elif wf_name:  # es können Namen mehrfach vergeben sein, wir nehmen den mit der höchsten Priorität
            sql = "Select * from %s where wfname = '%s' order by wfprio desc " % (wf_tabelle, wf_name)
        try:
            result = _db.execute(sql).fetchall()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchall=True)
        _ret = []
        if result:
            for row in result:
                wf = self._create_workflow_with_data(row, _con)
                wf.profile = _con.profile
                _ret.append(wf)

            return _ret
        else:
            return None

    def delete_workflow_by_id(self, profile=None, wf_id=None, wf_name=None, wf_trigger=0):
        """entfernt einen bestimmten Workflow anhand seiner ID oder Namens aus der Datenbank.
        Wird als wf_trigger >= 3 übergeben, so wird der Workflow aus der sys_workflow gelöscht (temporärer Workflow),
        dies wird von end_workfow_step benötigt"""

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        if wf_id:
            if type(wf_id) is not int:
                return None

        if wf_name:
            if type(wf_name) is not str:
                return None

        wf_tabelle = 'sys_wfd' if wf_trigger < 3 else 'sys_workflow'

        if wf_id:
            sql = "Delete from %s where wfid = %i" % (wf_tabelle, wf_id)
        elif wf_name:  # es können Namen mehrfach vergeben sein, wir nehmen den mit der höchsten Priorität
            sql = "Delete from %s where wfname = '%s' order by wf_prio desc " % (wf_tabelle, wf_name)
        try:
            _db.execute(sql)
        except:
            _con.try_again(sql=sql, db=_db)

        return True

    def _create_workflow_with_data(self, data, _con):
        wf = bitfarm_doc.BitfarmWorkflow()
        if data[0] is not None:
            wf.wf_id = data[0]
        if data[1] is not None:
            wf.wf_name = data[1].decode(_con.db_encoding)
        if data[2] is not None:
            wf.wf_type = data[2]
        if data[3] is not None:
            wf.wf_prio = data[3]
        if data[4] is not None:
            wf.wf_trigger = data[4]
        if data[5] is not None:
            wf.arcid = data[5]
        if data[6] is not None:
            wf.gdocid = data[6]
        if data[7] is not None:
            wf.usrid = data[7]
        if data[8] is not None:
            wf.grpid = data[8]
        if data[9] is not None:
            wf.importpath = data[9].decode(_con.db_encoding)
        if data[10] is not None:
            wf.date = data[10]
        if data[11] is not None:
            wf.created = data[11]
        if data[12] is not None:
            wf.modified_date = data[12]
        if data[13] is not None:
            wf.wf_pre = data[13].split(';') if len(data[13]) > 0 else []
            if len(wf.wf_pre) > 0:
                wf.wf_pre = [int(x) for x in wf.wf_pre]
        if data[14] is not None:
            wf.wf_origin = data[14]
        if data[15] is not None:
            wf.usr_counter = data[15]
        if data[16] is not None:
            wf.grp_counter = data[16]
        if data[17] is not None:
            wf.doc_counter = data[17]
        if data[18] is not None:
            wf.status = data[18]
        if data[19] is not None:
            wf.tasks = bfaxml.xml_to_object(data[19])
            if wf.tasks is None:
                wf.tasks = []
        if data[20] is not None:
            wf.condition = bfaxml.xml_to_object(data[20])
        if data[21] is not None:
            wf.else_tasks = bfaxml.xml_to_object(data[21])
        if data[22] is not None:
            wf.inherited_from = data[22]
        if data[23] is not None:
            wf.inherited_trigger = data[23].split(';') if len(data[23]) > 0 else []
        return wf

    def save_workflow(self, profile=None, wf=None, wf_type=None, wf_trigger=None, wf_prio=None, arcid=None, gdocid=None,
                      userid=None, groupid=None, importpath=None, wf_name=None, date=None, use_default=False):
        """Speichert eine Workflow-Instanz in die Datenbank.
            Der type und die Trigger_ID werden benötigt, ansonsten können alle weiteren Auslöser übergeben werden.
            Achtung, die vorhandenen Attribute der Instanz werden dann überschrieben.
            Wird ein wf_trigger >= 3 angegeben, dann wird in die sys_workflow gespeichert (für create_workflow_step, end_workflow_step)
            Wird ein wf_name angegeben wird der Workflow unter diesem Namen gespeichert, ein vorhandener Workflow wird aktualisiert.
        """
        if wf is None:
            raise WorkflowRequiredError
        if wf_type is None:
            raise WorkflowTypeRequiredError
        if wf_trigger is None:
            raise WorkflowTriggerRequiredError

        workflows = []
        if isinstance(wf, list):
            workflows = wf[:]
        else:
            workflows.append(wf)

        for wf in workflows:
            if not isinstance(wf, bitfarm_doc.BitfarmWorkflow):
                continue

            wf.wf_type = 0 if wf_type is None else wf_type

            wf.wf_trigger = 0 if wf_trigger is None else wf_trigger

            wf.wf_prio = 0 if wf_prio is None else wf_prio

            if arcid is not None:
                wf.arcid = arcid

            if gdocid is not None:
                wf.gdocid = gdocid

            if groupid is not None:
                wf.groupid = groupid

            if importpath is not None:
                wf.importpath = importpath

            if wf_name is not None:
                wf.wf_name = wf_name

            if date is not None:
                wf.date = date

            result = self.save_workflow_by_id(profile, wf=wf, wf_type=wf_type, wf_trigger=wf.wf_trigger)
            self.log.debug(str(result))

    def save_workflow_by_id(self, profile=None, wf=None, wf_id=None, wf_type=None, wf_trigger=None):
        """Speichert einen Workflow mit angegebener ID. Wird keine angegeben, so wird ein neuer Workflow angelegt,
        wird eine schon existente angegeben, so wird der Workflow aktualisiert, ansonsten unter dieser ID angelegt.
        Wird ein wf_trigger >= 3 angegeben, dann wird in die sys_workflow gespeichert (create_workflow_step, end_workflow_step).
        Der übergebene wf_type bzw. wf_trigger überschreibt die Einstellungen des übergebenen Workflow-Objekts"""
        if type(wf) is not bitfarm_doc.BitfarmWorkflow:
            return None

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

        tasks = None
        condition = None
        else_tasks = None
        now = datetime.datetime.now()
        if wf.date is None:
            wf.date = now
        if wf.created is None:
            wf.created = now
        wf.modified_date = now
        wf.wf_type = wf_type if wf_type is not None else wf.wf_type if wf.wf_type is not None else 0
        wf.wf_trigger = wf_trigger if wf_trigger is not None else wf.wf_trigger if wf.wf_trigger is not None else 0

        if wf.tasks is not None:
            tasks = bfaxml.object_to_xml(wf.tasks)

        if wf.condition is not None:
            condition = bfaxml.object_to_xml(wf.condition, shortxml=False)  # conditions niemals als shortxml
            # wegen Prüfungen gegen False oder Leer
        if wf.else_tasks is not None:
            else_tasks = bfaxml.object_to_xml(wf.else_tasks)

        #        if _con.sys_wfd is None:
        #            _con.sys_wfd =  sqla.Table('sys_wfd',  _con.meta,  autoload = True)

        if wf_trigger < 3:
            if _con.sys_wfd is None:
                _con.sys_wfd = sqla.Table('sys_wfd', _con.meta, autoload=True)
            _wf_tbl = _con.sys_wfd
            wf_tabelle = 'sys_wfd'
        else:
            if _con.sys_workflow is None:
                _con.sys_workflow = sqla.Table('sys_workflow', _con.meta, autoload=True)
            _wf_tbl = _con.sys_workflow
            wf_tabelle = 'sys_workflow'

        # knownids müssen auf wf_trigger eingegrenzt werden, sonst überschreiben sich default-Workflows
        try:
            knownids = _db.execute('select wfid from %s where wf_trigger = %s' % (wf_tabelle, wf.wf_trigger)).fetchall()
        except:
            knownids = _con.try_again(sql='select wfid from %s' % wf_tabelle, db=_db, fetchall=True)
        knownids = frozenset([x[0] for x in knownids])

        if wf_id is None and wf.wf_name is not None:
            _id = _db.execute("select wfid from %s where wfname = '%s'" % (wf_tabelle, wf.wf_name)).fetchone()
            if _id:
                wf_id = _id[0]

        vals = {_wf_tbl.c.wftype: wf.wf_type,
                _wf_tbl.c.wftrigger: wf.wf_trigger,
                _wf_tbl.c.wfname: wf.wf_name,
                _wf_tbl.c.wfprio: wf.wf_prio,
                _wf_tbl.c.arcid: wf.arcid,
                _wf_tbl.c.gdocid: wf.gdocid,
                _wf_tbl.c.userid: wf.userid,
                _wf_tbl.c.groupid: wf.groupid,
                _wf_tbl.c.importpath: wf.importpath,
                _wf_tbl.c.wfpre: ";".join([str(x) for x in wf.wf_pre]),
                _wf_tbl.c.wforigin: wf.wf_origin,
                _wf_tbl.c.usr_counter: wf.usr_counter,
                _wf_tbl.c.grp_counter: wf.grp_counter,
                _wf_tbl.c.doc_counter: wf.doc_counter,
                _wf_tbl.c.status: wf.status,
                _wf_tbl.c.tasks: tasks,
                _wf_tbl.c.condition: condition,
                _wf_tbl.c.else_tasks: else_tasks,
                _wf_tbl.c.inherited_from: wf.inherited_from,
                _wf_tbl.c.inherited_trigger: ";".join(wf.inherited_trigger)
                }

        if wf_id and wf_id in knownids:
            sql = _wf_tbl.update().where(_wf_tbl.c.wfid == wf_id).values(vals)
            # self.log.debug(sql)
            try:
                return _db.execute(sql)
            except:
                return _con.try_again(sql=sql, db=_db)

        elif wf_id and wf_id not in knownids:
            vals[_wf_tbl.c.wfid] = wf_id
            vals[_wf_tbl.c.tstamp] = wf.date
            vals[_wf_tbl.c.created] = wf.created

            sql = _wf_tbl.insert(values=vals)
            # self.log.debug(sql)
            try:
                return _db.execute(sql)
            except:
                return _con.try_again(sql=sql, db=_db)
        else:
            vals[_wf_tbl.c.tstamp] = wf.date
            vals[_wf_tbl.c.created] = wf.created
            sql = _wf_tbl.insert(values=vals)
            # self.log.debug(sql)
            try:
                return _db.execute(sql)
            except:
                return _con.try_again(sql=sql, db=_db)

    def inherit_from(self, profile=None, wid=None, name=None):
        """Methode um einen neuen Workflow zu erzeugen, der von einem bestehenden Workflow erbt. Ein Name oder eine ID
        muss übergeben werden"""

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

        if not name and not wid:
            # TODO: hier muss wohl auch eine Exception hinterlegt werden
            return None

        if name:
            sql = "Select * from sys_wfd where wfname = '%s'" % name
        elif wid:
            sql = "Select * from sys_wfd where wfid = '%s'" % wid

        try:
            result = _db.execute(sql).fetchone()
        except:
            result = _con.try_again(sql=sql, db=_db, fetchone=True)

        if not result:
            return None
        else:  # den Basisworkflow in eine neues Workflowobjekt laden
            wf = self._create_workflow_with_data(result, _con)

            # alle Tasks auf inherited setzen
            if wf.tasks is not None:
                for task in wf.tasks:
                    task.inherited = True

            # den oder alle else_task auf inherited setzen
            if wf.else_tasks is not None:
                if type(wf.else_tasks) == list:
                    for task in wf.else_tasks:
                        task.inherited = True
                else:
                    wf.else_tasks.inherited = True

            # die ID des Basistask in inherited_from übernehmen, eigene ID und Name leer machen
            wf.inherited_from = wf.wf_id
            wf.wf_id = None
            wf.name = None

            # standardgemäß erben wir ja alle verfügbaren wf_trigger (Auslöser)
            wf.inherited_trigger = ['wf_type', 'wf_prio', 'wf_trigger', '_BitfarmBaseObject__arcid',
                                    '_BitfarmBaseObject__gdocid', 'userid', 'groupid', 'importpath', 'usr_counter',
                                    'grp_counter', 'doc_counter', 'condition']

            return wf

    def inherit_to_childs(self, profile=None, wf=None):
        """Diese Methode gibt eine Workflowdefinition an alle Nachfahren weiter. Dabei werden nur die wf_trigger weitergegeben,
        die der Nachfahre geerbt hat und nicht nachträglich geändert (inherited_trigger) und alle Tasks und else_tasks, die über die gleiche inherit_id
        verfügen und inherited True ist."""

        if type(wf) is not bitfarm_doc.BitfarmWorkflow:
            return None

        # zuerst alle Nachfahren sammeln, rekursiv, zurückgegebn wird eine Liste mit wf_ids
        child_ids = self._search_childs(wf_id=wf.wf_id)
        self.log.debug(child_ids)

        # jetzt laden wir die einzelnen child-workflows
        for child_id in child_ids:
            child = self.get_workflow_by_id(profile, child_id)

            # wir vererben alles was die childs erben wollen
            if child.inherited_trigger:
                for wf_trigger in child.inherited_trigger:
                    child.__dict__[wf_trigger] = wf.__dict__[wf_trigger]

            if child.tasks:
                for task in child.tasks:
                    if task.inherited:  # wenn der task erben will suchen wir den passend ursprungstask
                        for p_task in wf.tasks:
                            if task.inherit_id == p_task.inherit_id:  # treffer, jetzt schnell vererben
                                _obj = bfaxml.copy_bitfarmobj(p_task)
                                _obj.inherited = True
                                child.tasks[child.tasks.index(task)] = _obj
                                break

            if child.else_tasks:
                for else_task in child.else_tasks:
                    if else_task.inherited:  # wenn der else_task erben will suchen wir den passend ursprungstask
                        for p_task in wf.else_tasks:
                            if else_task.inherit_id == p_task.inherit_id:  # treffer, jetzt schnell vererben
                                _obj = bfaxml.copy_bitfarmobj(p_task)
                                _obj.inherited = True
                                child.else_tasks[child.else_tasks.index(else_task)] = _obj
                                break

            self.save_workflow_by_id(profile, child, child_id)

    def _search_childs(self, profile=None, wf_id=None):
        _childs = []

        if wf_id is None:
            return _childs

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return _childs

        sql = "Select wfid from sys_wfd where inherited_from = %i" % wf_id
        try:
            result = _db.execute(sql).fetchall()
        except:
            self.log.exception(u'Fehler bei _search_childs')
            result = _con.try_again(sql=sql, db=_db, fetchall=True)
        if result:
            for cid in result:
                if cid[0] not in _childs:
                    _childs.append(cid[0])
                _childs.extend(self._search_childs(profile, cid[0]))

        return _childs

    def create_resub(self, _doc, recipients, subject, text, options=0, wf_ids=None, controller=None, date=None,
                     alarm=None, required=None, delta=7):
        """legt eine Wiedervorlage an, gibt die tim_ref der Wiedervorlage zurück. Der Wiedervorlage zugeordnete Workflows können als Liste von wf_ids übergeben werden"""
        self.log.debug(u'create_resub')
        if not isinstance(_doc, bitfarm_doc.BitfarmDocument) and not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return 0
        if type(recipients) not in [list, str] or len(recipients) == 0:
            return 0

        if not _doc.profile:
            _con = self.con.main
            _db_trans = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db_trans = self.db[_doc.profile]

        tim_ref = 0
        gdocid = _doc.gdocid

        if not wf_ids or wf_ids is None:
            wf_ids = [0]

        if controller is not None:
            if type(controller) is str:
                controller = _con.read_user(name=controller)
            if type(controller) is int:
                controller = _con.read_user(userid=controller)

        if isinstance(_doc, bitfarm_doc.BitfarmDocument):
            title = _doc.title if _doc.title is not None else u''
        elif isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            title = _doc.wf_name if _doc.wf_name is not None else u''

        if subject is None:
            subject = u''
        if text is None:
            text = u''
        if required is None:
            required = _doc.required if _doc.required is not None else 0
        if options is None:
            options = 0
        # subject = subject.replace('"', '\\"')
        # text = text.replace('"', '\\"')

        if _con.version in (3, 35, '3.0'):
            docid = _doc._docid
            arcid = _doc.arcid
            if date is None:
                date = datetime.datetime.now()
            if alarm is None:
                alarm = alarm + datetime.timedelta(delta)
            date = date.replace(microsecond=0)
            alarm = alarm.replace(microsecond=0)
            date = date.strftime('%Y-%m-%d %H:%M:%S')
            alarm = alarm.strftime('%Y-%m-%d %H:%M:%S')
            tim_ref = 0
            for item in recipients:
                vals = []
                if type(item) == bitfarm_doc.BitfarmUser:
                    if controller is None or not isinstance(controller, bitfarm_doc.BitfarmUser):
                        controllerid = item.id
                    else:
                        controllerid = controller.id
                    vals.append('(%s, %s, %s, "%s", %s, %s, %s, "%s","%s","%s","%s", %s, %s, %s)' % (
                        tim_ref, arcid, docid, title, '0', item.id, controllerid, date, alarm, subject, text, options))
                elif type(item) is str:
                    item = _con.read_user(db=_db_trans, name=item)
                    if item is None:
                        self.log.warning(u"Username %s konnte nicht aufgelöst werden" % item)
                        continue
                    if controller is None or not isinstance(controller, bitfarm_doc.BitfarmUser):
                        controllerid = item.id
                    else:
                        controllerid = controller.id
                    vals.append('(%s, %s, %s, "%s", %s, %s, %s, "%s","%s","%s","%s", %s, %s, %s)' % (
                        tim_ref, arcid, docid, title, '0', item.id, controllerid, date, alarm, subject, text, options))
                elif type(item) == bitfarm_doc.BitfarmGroup:
                    for member in item.members:
                        if controller is None or not isinstance(controller, bitfarm_doc.BitfarmUser):
                            controllerid = member.id
                        else:
                            controllerid = controller.id
                        vals.append('(%s, %s, %s, "%s", %s, %s, %s, "%s","%s","%s","%s", %s, %s, %s)' % (
                            tim_ref, arcid, docid, title, item.id, member.id, controllerid, date, alarm, subject, text,
                            options))

                if len(vals) == 0:
                    continue
                sql = 'insert into sys_timer (tim_ref, arc_id, doc_id, doc_titel, grp_id, usr_id_work, usr_id_shed, termin, erinnerung, info, wkommentar, optionen) values %s' % (
                    ",".join(vals))
                try:
                    result = _db_trans.execute(sql)
                except:
                    self.log.exception(u'Fehler beim Anlegen der Wiedervorlagen')
                    return 0
                else:
                    if tim_ref == 0:
                        tim_ref = result.lastrowid

                if tim_ref:
                    sql = 'update sys_timer set tim_ref = %s where tim_ref = 0 and arc_id = %s and doc_id = %s' % (
                    tim_ref, arcid, docid)
                    try:
                        _db_trans.execute(sql)
                    except:
                        self.log.exception(u'Fehler beim Update der tim_ref')
                        return 0

        elif _con.version in [36, 40]:
            if date is None:
                date = datetime.datetime.now()
            if alarm is None:
                alarm = date + datetime.timedelta(delta)
            date = date.replace(microsecond=0)
            alarm = alarm.replace(microsecond=0)
            date = date.strftime('%Y-%m-%d %H:%M:%S')
            alarm = alarm.strftime('%Y-%m-%d %H:%M:%S')

            tim_ref = 0
            values = []

            for item in recipients:
                if type(item) is str:
                    item = _con.read_user(db=_db_trans, name=item)
                    if item is None:
                        self.log.warning(u"Username %s konnte nicht aufgelöst werden" % item)
                        continue
                if type(item) is int:
                    item = _con.read_user(db=_db_trans, userid=item)
                    if item is None:
                        self.log.warning(u"Username %s konnte nicht aufgelöst werden" % item)
                        continue

                if type(item) == bitfarm_doc.BitfarmUser:
                    if controller is None or not isinstance(controller, bitfarm_doc.BitfarmUser):
                        controllerid = item.id
                    else:
                        controllerid = controller.id
                    # vals.append(u'(%s, %s, %s, %s, %s, "%s","%s", %s, "%s", "%s", "%s", %s)' % (
                    #     tim_ref, gdocid, '0', item.id, controllerid, date, alarm, options, subject,
                    #     text, ";".join([str(x) for x in wf_ids]), required))
                    # timref, gdocid, groupid, userid_work, userid_shed, date, alarm, options, info, comment, wfid, required''
                    vals = {}
                    vals['timref'] = tim_ref
                    vals['gdocid'] = gdocid
                    vals['groupid'] = 0
                    vals['userid_work'] = item.id
                    vals['userid_shed'] = controllerid
                    vals['date'] = date
                    vals['alarm'] = alarm
                    vals['options'] = options
                    vals['info'] = subject
                    vals['comment'] = text
                    vals['wfid'] = ";".join([str(x) for x in wf_ids])
                    vals['required'] = required
                    values.append(vals)
                elif type(item) == bitfarm_doc.BitfarmGroup:
                    for member in item.members:
                        if controller is None or not isinstance(controller, bitfarm_doc.BitfarmUser):
                            if isinstance(member, bitfarm_doc.BitfarmUser):
                                controllerid = member.id
                            else:
                                controllerid = int(member)
                        else:
                            controllerid = controller.id

                        if isinstance(member, bitfarm_doc.BitfarmUser):
                            memberid = member.id
                        else:
                            memberid = int(member)

                        # vals.append(u'(%s, %s, %s, %s, %s, "%s","%s", %s, "%s", "%s", %s, %s)' % (
                        #     tim_ref, gdocid, item.id, memberid, controllerid, date, alarm, options,
                        #     subject, text, ";".join([str(x) for x in wf_ids]), required))
                        vals = {}
                        vals['timref'] = tim_ref
                        vals['gdocid'] = gdocid
                        vals['groupid'] = item.id
                        vals['userid_work'] = memberid
                        vals['userid_shed'] = controllerid
                        vals['date'] = date
                        vals['alarm'] = alarm
                        vals['options'] = options
                        vals['info'] = subject
                        vals['comment'] = text
                        vals['wfid'] = ";".join([str(x) for x in wf_ids])
                        vals['required'] = required
                        values.append(vals)
                else:
                    self.log.warning(u"Achtung, es wurde kein User oder Gruppenobjekt übergeben (%s)" % str(item))

            if values:
                trans = _db_trans.begin()
                try:
                    if options == 256:  # Abonnement
                        sql = 'select * from sys_timer where gdocid = %s and userid_work = %s and options = 256 and status & 4 = 0' % (
                            gdocid, item.id)

                        result = _db_trans.execute(sql).fetchone()

                        if result:
                            sql = u'update sys_timer set date = "%s", alarm = "%s", info = "%s", comment = "%s" where timid = %s' % (
                                date, alarm, subject, text, result['timid'])
                            _db_trans.execute(sql)

                            vals = []
                            tim_ref = result['timref']

                    if values:
                        # sql = u'insert into sys_timer (timref, gdocid, groupid, userid_work, userid_shed, date, alarm, options, info, comment, wfid, required) values %s' % (
                        #     u",".join(vals))
                        sql = _con.sys_timer.insert().values(values)

                        result = _db_trans.execute(sql)

                        if result and tim_ref == 0:
                            tim_ref = result.lastrowid

                        if tim_ref:
                            sql = 'update sys_timer set timref = %s where timref = 0 and gdocid = %s' % (
                            tim_ref, gdocid)
                            _db_trans.execute(sql)
                    trans.commit()
                except sqla.exc.OperationalError as e:
                    self.log.exception(u"Fehler beim Ausführen der Transaction create_resub")
                    if e.orig[0] == 2006:  # mysql has gone away
                        _doc.error = EC_NEEDRECONNECT
                    else:
                        _doc.error = EC_CREATERESUBERROR
                    trans.rollback()
                    return None
                except:
                    self.log.exception(u'Fehler bei create_resub')
                    _doc.error = EC_CREATERESUBERROR
                    trans.rollback()
                    return 0

        return tim_ref

    def change_resub(self, _doc, subject=None, text=None, user=None, group=None, controller=None, date=None, alarm=None,
                     status=None):
        """Ändert eine Wiedervorlage, man kann Text und Kommentar, Datum und Alarm, sowohl den User, oder die Gruppe oder den Controller ändern.
        User und Controller können als BitframUser oder Userid übergeben werden, Gruppe als BitfarmGroup oder GroupID
        Dabei wir berücksichtigt, ob die WVL JE oder EFA ist"""
        self.log.debug(u'change_resub')
        _ret = False

        if not isinstance(_doc, bitfarm_doc.BitfarmDocument) and not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return _ret
        # TODO: bei Änderung des Controllers muss getestet werden, ob die WVL weitergeleitet wurde.
        if not _doc.profile:
            _con = self.con.main
            _db_trans = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db_trans = self.db[_doc.profile]

        change_by_timref = False  # Bei Änderungen die mehrere Einträge einer Gruppe betreffen

        if _con.version in [36, 40]:
            if _con.sys_timer is None:
                _con.sys_timer = sqla.Table('sys_timer', _con.meta, autoload=True)
            if _doc.options is None:
                _doc.options = 0

            userid = 0
            groupid = 0
            controllerid = 0

            if user is not None:
                if isinstance(user, bitfarm_doc.BitfarmUser):
                    userid = user.id
                elif type(user) is int:
                    userid = user
            if group is not None:
                if isinstance(group, bitfarm_doc.BitfarmGroup):
                    groupid = group.id
                elif type(group) is int:
                    groupid = group
            if controller is not None:
                if isinstance(controller, bitfarm_doc.BitfarmUser):
                    controller = controller.id
                elif type(controller) is int:
                    controllerid = controller

            if not _doc.tim_ref:
                sql = 'select timref from sys_timer where timid = %s' % _doc.tim_id
                try:
                    result = _db_trans.execute(sql).fetchone()
                except:
                    self.log.exception(u'Fehler beim Abfragen der tim_id = %s' % _doc.tim_id)
                else:
                    _doc.tim_ref = result[0]

            # alle einträge holen
            sys_timer = {}
            sql = sqla.select([_con.sys_timer]).where(_con.sys_timer.c.timref == _doc.tim_ref)
            try:
                result = _db_trans.execute(sql).fetchall()
            except:
                self.log.exception(u'Fehler beim Abfragen der timref = %s' % _doc.tim_ref)
            else:
                for row in result:
                    sys_timer[row[0]] = row

            if len(sys_timer) == 0:
                _doc.error = EC_NORESUBFOUND
                self.log.error(u"Einträge für tim_ref = %s konnten nicht ermittelt werden" % _doc.tim_ref)
                return False

            # ändert sich der user oder die gruppe der wvl, dann neue einträge erzeugen und alte löschen
            # ansonsten die bestehenden einträge updaten

            change_user = sys_timer[_doc.tim_id]['userid_work'] != userid or sys_timer[_doc.tim_id][
                'groupid'] != groupid

            _ret = False
            values = []

            if change_user:
                users = []
                if userid:
                    users = [userid]

                if groupid:
                    users = _con.resolve_group(groupid)

                if not users:
                    _doc.error = EC_NOUSERORGROUPERROR
                    _ret = False
                else:
                    for userid in users:  # wird der user oder die gruppe neu angelegt, müssen wir alle einträge erstellen
                        vals = {}
                        if subject is not None:
                            vals[_con.sys_timer.c.info] = subject
                        if text is not None:
                            vals[_con.sys_timer.c.comment] = text
                        if date is not None:
                            vals[_con.sys_timer.c.date] = date
                        if alarm is not None:
                            vals[_con.sys_timer.c.alarm] = alarm
                        vals[_con.sys_timer.c.timref] = _doc.tim_ref  # tim_ref bleibt bei änderung bestehen
                        vals[_con.sys_timer.c.gdocid] = sys_timer[_doc.tim_id]['gdocid']
                        vals[_con.sys_timer.c.options] = sys_timer[_doc.tim_id]['options']
                        vals[_con.sys_timer.c.wfid] = sys_timer[_doc.tim_id]['wfid']
                        vals[_con.sys_timer.c.required] = sys_timer[_doc.tim_id]['required']
                        vals[_con.sys_timer.c.histid] = sys_timer[_doc.tim_id]['histid']
                        if status is None:
                            vals[_con.sys_timer.c.status] = sys_timer[_doc.tim_id]['status']
                        else:
                            if status & 2 == 2 and _doc.options & 1 == 0:  # in Arbeit setzen?
                                if userid == sys_timer[_doc.tim_id][
                                    'userid_work']:  # eigenen eintrag auf status 2 setzen
                                    vals[_con.sys_timer.c.status] = status
                                else:
                                    vals[
                                        _con.sys_timer.c.status] = status | 4  # alle anderen auch auf fertig, damit die nicht mehr angezeigt werden
                            elif status & 2 == 0 and sys_timer[_doc.tim_id][
                                'status'] & 2 == 2 and _doc.options & 1 == 0:  # in Arbeit zurücksetzen?
                                if userid == sys_timer[_doc.tim_id]['userid_work']:  # eigenen status setzen
                                    vals[_con.sys_timer.c.status] = status
                                else:
                                    vals[_con.sys_timer.c.status] = (sys_timer[_doc.tim_id][
                                                                         'status'] & ~ 6) | status  # status 4 zurücksetzen
                            else:
                                vals[_con.sys_timer.c.status] = status
                        vals[_con.sys_timer.c.groupid] = groupid
                        vals[_con.sys_timer.c.userid_work] = userid
                        vals[_con.sys_timer.c.userid_shed] = controllerid if controllerid else userid
                        values.append(vals)
            else:
                for row in sys_timer.values():  # wird die wvl nur aktualisiert, brauchen wir nur die einträge die sich ggf ändern
                    if _doc.options & 1 == 1:  # JE? dann nur den eigenen Eintrag ändern
                        if row['timid'] != sys_timer[_doc.tim_id]['timid']:
                            continue

                    if row['histid'] == sys_timer[_doc.tim_id]['histid']:
                        vals = {}
                        vals[_con.sys_timer.c.info] = subject if subject is not None else u""
                        vals[_con.sys_timer.c.comment] = text if text is not None else u""
                        if date is not None:
                            vals[_con.sys_timer.c.date] = date
                        if alarm is not None:
                            vals[_con.sys_timer.c.alarm] = alarm
                        if status is not None:
                            if status & 2 == 2 and _doc.options & 1 == 0:  # in Arbeit setzen?
                                if userid == row['userid_work']:  # eigenen eintrag auf status 2 setzen
                                    vals[_con.sys_timer.c.status] = status
                                else:
                                    vals[
                                        _con.sys_timer.c.status] = status | 4  # alle anderen auch auf fertig, damit die nicht mehr angezeigt werden
                            elif status & 2 == 0 and row[
                                'status'] & 2 == 2 and _doc.options & 1 == 0:  # in Arbeit zurücksetzen?
                                if userid == row['userid_work']:  # eigenen status setzen
                                    vals[_con.sys_timer.c.status] = status
                                else:
                                    vals[_con.sys_timer.c.status] = (row[
                                                                         'status'] & ~ 6) | status  # status 4 zurücksetzen
                            else:
                                vals[_con.sys_timer.c.status] = status
                        vals['timid'] = row[0]
                        values.append(vals)

            if values:
                trans = _db_trans.begin()
                try:
                    if change_user:
                        sql = _con.sys_timer.insert().values(values)
                        _db_trans.execute(sql)

                        for row in sys_timer.values():
                            if row[14] == sys_timer[_doc.tim_id]['histid']:
                                sql = 'delete from sys_timer where timid = %s' % row[0]
                                _db_trans.execute(sql)
                    else:
                        for vals in values:
                            timid = vals['timid']
                            del vals['timid']
                            sql = _con.sys_timer.update().where(_con.sys_timer.c.timid == timid).values(vals)
                            _db_trans.execute(sql)
                    trans.commit()
                    _ret = True
                except sqla.exc.OperationalError as e:
                    self.log.exception(u"Fehler beim Ausführen der Transaction change_resub")
                    if e.orig[0] == 2006:  # mysql has gone away
                        _doc.error = EC_NEEDRECONNECT
                    else:
                        _doc.error = EC_SETSTATEDONEERROR
                    trans.rollback()
                    return None
                except:
                    self.log.exception(u'Fehler bei change_resub')
                    _doc.error = EC_CHANGERESUBERROR
                    trans.rollback()
                    return False

        return _ret

    def delete_resub(self, _doc):
        """Funktion um eine fertiggesetzte Wiedervorlage dauerhaft aus der sys_timer-Tabelle zu löschen."""
        if not isinstance(_doc, bitfarm_doc.BitfarmDocument) and not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return None
        if _doc.tim_id is None:
            return None

        if not _doc.profile:
            _con = self.con.main
            _db_trans = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db_trans = self.db[_doc.profile]

        _doc.result = False

        if _doc.version not in (36, 40):  # nur für v36 und v40
            self.log.warning(u'Version nicht unterstützt %s' % str(_doc.version))
            return None

        # die wvl muss fertig sein, um sie zu löschen
        sql = sqla.select([_con.sys_timer]).where(
            sqla.and_(_con.sys_timer.c.timid == _doc.tim_id, _con.sys_timer.c.status.op('&')(4) == 4))
        try:
            result = _db_trans.execute(sql).fetchone()
        except:
            self.log.exception(u'Fehler beim Abfragen der tim_id = %s' % _doc.tim_id)

        if result:
            options = int(result['options'])
            timref = int(result['timref'])
            _doc.info = result['info']
            histids = []
            if result['histid']:
                histids.append(result['histid'])

            sql = 'select * from sys_timer where timref = %d' % timref
            try:
                sys_timer = _db_trans.execute(sql).fetchall()
            except:
                self.log.exception(u'Fehler beim Löschen der timref = %s' % timref)

            so_done = True
            if sys_timer:
                if options & 1 == 1:  # bei je prüfen, ob wiedervorlage insgesamt fertig ist
                    for row in sys_timer:
                        so_done &= row[
                                       'status'] & 4 == 4  # finden wir einen nicht fertig gesetzten wird so_done False und bleibt es

                if options & 1 == 0 or so_done:  # bei efa alle histids sammeln
                    for row in sys_timer:
                        if row['histid'] > 0 and row['histid'] not in histids:
                            histids.append(row['histid'])

            trans = _db_trans.begin()
            try:
                if options & 1 == 0 or so_done:  # bei efa die gesamte timref löschen
                    sql = 'delete from sys_timer where timref = %d' % timref
                    _db_trans.execute(sql)
                else:
                    sql = 'delete from sys_timer where timid = %d' % _doc.tim_id
                    _db_trans.execute(sql)

                if histids:
                    sql = 'delete from sys_timerhist where histid in (%s)' % ",".join([str(x) for x in histids])
                    _db_trans.execute(sql)
                trans.commit()
            except sqla.exc.OperationalError as e:
                self.log.exception(u"Fehler beim Ausführen der Transaction set_statedone")
                if e.orig[0] == 2006:  # mysql has gone away
                    _doc.error = EC_NEEDRECONNECT
                else:
                    _doc.error = EC_SETSTATEDONEERROR
                trans.rollback()
                return None
            except:
                self.log.exception(u'Fehler bei delete_resub')
                _doc.error = EC_DELETERESUBERROR
                trans.rollback()
                return None

            return True
        else:
            self.log.warning(u'Die timid %d konnte nicht gefunden werden' % _doc.tim_id)
            return None

    def get_holidaydict(self, profile=None, start=None, end=None):
        """Funktion gibt alle Vertreter aller Benutzer als Dictonary zurück (key = Userid, value = Vertreter (Tupel mit Id und Name)).
        Gibt es keinen Vertreter wird der User eingetragen, so kann man das Dictonary einfach nutzen, um eine UserID
        durch einen Vertreter zu ersetzen, falls vorhanden, ansonsten wir die UserID beibehalten"""
        _ret = {}
        userdict = {}
        if not profile:
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if start is None:
            start = datetime.datetime.now()

        if end is None:
            end = datetime.datetime.now()

        if _con.version in (3, 35, '3.0'):
            sql = "select usr_id, flags +0, usr_name from sys_usr where (flags & 1) = 1"
        else:
            sql = "select userid, flags +0, username from sys_user where (flags & 1) = 1"
        try:
            result = _db.execute(sql).fetchall()
        except:
            self.log.exception(u'Fehler bei get_holidaydict')
            result = _con.try_again(sql=sql, db=_db, fetchall=True)

        if result:
            for row in result:
                userdict[row[0]] = row[2]

            for row in result:
                if row[0] is not None:
                    holidaylist = self.get_holidaylist(profile, row[0], start, end)
                    if holidaylist[
                        -1] in userdict:  # haben wir die userid eigentlich in den aktuellen benutzern? wenn nicht, dann weg
                        _ret[row[0]] = (holidaylist[-1], userdict[holidaylist[-1]])
        return _ret

    def get_holidaylist(self, profile=None, userid=None, start=None, end=None):
        """Funktion gibt die aktuellen Urlaubsvertreter für ein Userid zurück, als Liste von Userids"""
        _ret = []
        if userid is None:
            return _ret

        if not profile:
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if start is None:
            start = datetime.datetime.now()

        if end is None:
            end = datetime.datetime.now()

        _ret = [userid]  # wir fügen uns selbst zum Ergebnis hinzu

        self._add_holidayuser(_ret, userid, start, end, _con, _db)

        return _ret

    def _add_holidayuser(self, _ret, userid, start, end, _con, _db):
        """Unterfunktion für die rekursive Sammlung von Urlaubsvertretungen"""
        if userid is None:
            return _ret

        if _con.version in (3, 35, '3.0'):
            sql = u"select * from sys_holyday where usr_id_dst = %s" % userid
        else:
            sql = u"select * from sys_holiday where userid_dst = %s" % userid
        if start is not None:
            _start = datetime.datetime.strftime(start, '%Y-%m-%d %H:%M:%S')
            _end = datetime.datetime.strftime(end, '%Y-%m-%d %H:%M:%S')
            if _con.version in (3, 35, '3.0'):
                sql += u" and ((START <= '%s') and (END >= '%s')) and (END >= '%s')" % (_end, _start, _end)
            else:
                sql += u" and ((start <= '%s') and (end >= '%s')) and (end >= '%s')" % (_end, _start, _end)

        try:
            result = _db.execute(sql).fetchall()
        except:
            self.log.exception(u'Fehler bei _add_holidayuser')
            result = _con.try_again(sql=sql, db=_db, fetchall=True)

        if result:
            for row in result:
                _newuser = row[1]
                if _newuser in _ret:  # achtung rekursion
                    continue
                if _newuser:
                    _ret.append(_newuser)
                self._add_holidayuser(_ret, _newuser, start, end, _con, _db)

        return _ret

    def setstate_done(self, _doc):
        """Prüft ob eine Wiedervorlage in der sys_timer auf fertiggesetzt werden kann.
        Das übergebene Objekt muss eine tim_id und eine userid haben.
        Die Funktion gibt None bei Fehler, ansonsten True/False zurück, ob die WVL fertig gesetzt wurde.
        Gruppen mit den Optione JE oder EFA werden berücksichtigt, ebenso Urlaubsvertretung, Weiterleitung,
        Rückfrage etc. Unterstützt ab v36."""

        if not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return None
        if _doc.tim_id is None:
            _doc.error = EC_NOTIMID
            return None

        if _doc.version in (3, 35, '3.0'):  # nur für v36 und v4
            _doc.error = EC_VERSIONNOTSUPPORTED
            return None

        if not _doc.profile:
            _con = self.con.main
            _db_trans = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db_trans = self.db[_doc.profile]

        _doc.result = False
        sys_timer = {}
        rts_done = False  # return to sender done
        so_done = False  # single one done (JE)
        ofa_done = False  # one for all done (EFA)

        if not _doc.tim_ref:
            sql = u'select timref from sys_timer where timid = %s' % _doc.tim_id
            try:
                result = _db_trans.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler bei setstate_done')
                _doc.error = EC_SETSTATEDONEERROR
                return None

            if not result:
                _doc.error = EC_SETSTATEDONEERROR
                return None
            else:
                _doc.tim_ref = result[0]

        # wir holen alle einträge zu der timref ins array

        sql = u'Select Sql_No_cache ' \
              + 'userid_work, userid_shed, timid, timref, sys_timer.groupid, date, ' \
              + 'gdocid, groupname, username, ' \
              + 'alarm, info, status, options, comment, wfid, required, histid ' \
              + 'from sys_timer ' \
              + 'left join sys_user on sys_user.userid = sys_timer.userid_work ' \
              + 'left join sys_group on sys_group.groupid = sys_timer.groupid ' \
              + 'where timref = %s order by timid' % _doc.tim_ref

        try:
            result = _db_trans.execute(sql).fetchall()
        except:
            self.log.exception(u'Fehler bei setstate_done')
            _doc.error = EC_SETSTATEDONEERROR
            return None

        if not result:
            _doc.error = EC_SETSTATEDONEERROR
            return None
        else:
            for row in result:
                sys_timer[row[2]] = [x for x in row]
                # for ind in range(len(sys_timer[row[2]])):
                #     if type(sys_timer[row[2]][ind]) == str:
                #         sys_timer[row[2]][ind] = sys_timer[row[2]][ind].decode(_con.db_encoding)

        if _doc.userid is None:
            _doc.userid = sys_timer[_doc.tim_id][0]

        # holidayuser = self.get_holidaylist(profile=_doc.profile, userid=_doc.userid)

        # wir laden alle weiteren Parameter aus der sys_timer
        _doc.status = sys_timer[_doc.tim_id][11]
        _doc.groupid = sys_timer[_doc.tim_id][4]
        _doc.userid = sys_timer[_doc.tim_id][0]
        _doc.userid_work = sys_timer[_doc.tim_id][0]
        _doc.userid_shed = sys_timer[_doc.tim_id][1]
        _doc.tim_ref = sys_timer[_doc.tim_id][3]
        _doc.options = sys_timer[_doc.tim_id][12]
        _doc.wf_id = sys_timer[_doc.tim_id][14]
        _doc.required = sys_timer[_doc.tim_id][15]
        _doc.hist_id = sys_timer[_doc.tim_id][16]
        _doc.info = sys_timer[_doc.tim_id][10]
        _doc.comment = sys_timer[_doc.tim_id][13]
        _doc.groupname = sys_timer[_doc.tim_id][7]
        _doc.username = sys_timer[_doc.tim_id][8]
        _doc.date = sys_timer[_doc.tim_id][5]
        _doc.alarm = sys_timer[_doc.tim_id][9]
        _doc.gdocid = sys_timer[_doc.tim_id][6]

        next_wf = bitfarm_doc.BitfarmWorkflow()  # hier wird der nächste Workflowzustand definiert
        next_wf.tim_id = _doc.tim_id
        next_wf.status = _doc.status
        next_wf.groupid = _doc.groupid
        next_wf.userid = _doc.userid
        next_wf.userid_work = _doc.userid_work
        next_wf.userid_shed = _doc.userid_shed
        next_wf.tim_ref = _doc.tim_ref
        next_wf.options = _doc.options
        next_wf.wf_id = _doc.wf_id
        next_wf.required = _doc.required
        next_wf.hist_id = _doc.hist_id
        # next_wf.info = _doc.info.replace('"','\\"')
        next_wf.info = _doc.info
        # next_wf.comment = _doc.comment.replace('"','\\"')
        next_wf.comment = _doc.comment
        next_wf.groupname = _doc.groupname
        next_wf.username = _doc.username
        next_wf.date = _doc.date
        next_wf.alarm = _doc.alarm
        next_wf.gdocid = _doc.gdocid

        if _doc.status & 4 == 4:  # anscheinend in der Zwischenzeit schon fertig gesetzt
            self.log.warning(u'Wiedervorlage %s wurde in der Zwischenzeit schon fertig gesetzt' % _doc.tim_id)
            _doc.error = EC_RESUBISDONE
            return None

        if _doc.status & 8 == 8:  # Rückfrage, nicht fertigsetzen
            self.log.warning(u'Wiedervorlage %s befindet sich in Rückfrage, fertig setzen nicht möglich' % _doc.tim_id)
            _doc.error = EC_RESUBINCHECKBACK
            return None

        if _doc.options & 1 == 1:  # Wiedervorlage JE: ist diese Fertig?
            if _doc.required:  # alle die status 4 gesetzt haben außer meine eigene (die ja jetzt fertig gesetzt werden soll)
                cnt_done = len([x[2] for x in sys_timer.values() if
                                x[2] != _doc.tim_id and x[11] & 4 == 4 and x[16] == _doc.hist_id])
            else:  # alle die status 4 nicht gesetzt haben (offen) außer meine eigene (die ja jetzt fertig gesetzt werden soll)
                cnt_done = len([x[2] for x in sys_timer.values() if
                                x[2] != _doc.tim_id and x[11] & 4 == 0 and x[16] == _doc.hist_id])

            so_done = False  # so_done: kann die JE (single one) wvl fertig gesetzt werden

            if _doc.required:
                cnt_timref = len([x[2] for x in sys_timer.values() if x[11] & 4 == 4 and x[16] == _doc.hist_id])
                if so_done + 1 >= cnt_timref:
                    so_done = True
                else:
                    so_done = cnt_done + 1 >= _doc.required
            else:
                so_done = cnt_done == 0

            if so_done:  # sollten wir die JE wvl fertigsetzen können, müssen wir aber vorher prüfen, ob nicht jemand eine weitergeleitet hat (16 & 8 Weiterleitung oder Rückfrage)
                cnt_redirected = len([x[2] for x in sys_timer.values() if x[11] & 24 > 0 and x[16] == _doc.hist_id])
                so_done = cnt_redirected == 0

        # haben wir eine rückgabe (rts)?
        if _doc.options & 2 == 2:  # RTS
            if _doc.hist_id:  # haben wir ein hist_id, dann müssen wir den eintrag wiederherstellen
                if _doc.options & 1 == 1:  # Weiterleitung JE
                    rts_done = so_done
                    if not so_done:  # haben wir eine rückgabe und die wvl ist noch nicht fertig setzen wir die einzelne fertig, ansonsten bleibt sie offen (wird ja zurückgegeben)
                        next_wf.status = _doc.status | 4
                        next_wf.info = "Erledigt von %s: %s" % (_doc.username, _doc.info)
                    else:
                        next_wf.info = "Erledigt von %s: %s" % (
                        _doc.username if _doc.groupid == 0 else _doc.groupname, _doc.info)
                elif _doc.options & 1 == 0:  # Weiterleitung EFA
                    rts_done = True  # rts_done: die weiterleitung ist fertig, die timerhist muss restored werden
                    if _doc.options & 128 == 128:
                        next_wf.info = "Beantwortet von %s: %s" % (
                            _doc.username if _doc.groupid == 0 else _doc.groupname, _doc.info)
                    else:
                        next_wf.info = "Erledigt von %s: %s" % (
                            _doc.username if _doc.groupid == 0 else _doc.groupname, _doc.info)
            else:  # keine hist_id aber rts? an den controller zurück
                if _doc.options & 1 == 0:  # Weiterleitung EFA
                    ofa_done = True
                if _doc.userid_work == _doc.userid_shed:  # wiedervorlage liegt schon bei uns? dann beenden
                    next_wf.status = _doc.status | 4
                else:  # ansonsten zum Controller zurück
                    if _doc.options & 1 == 0 or (_doc.options & 1 == 1 and so_done):  # efa oder fertige je?
                        next_wf.userid_work = _doc.userid_shed  # zurück zum controller
                    else:  # ansonsten eintrag einfach fertig setzen
                        next_wf.status = _doc.status | 4
                    if _doc.options & 128 == 128:
                        next_wf.info = "Beantwortet von %s: %s" % (
                            _doc.username if (_doc.groupid == 0 or _doc.options & 1 == 1) else _doc.groupname,
                            _doc.info)
                    else:
                        next_wf.info = "Erledigt von %s: %s" % (
                            _doc.username if (_doc.groupid == 0 or _doc.options & 1 == 1) else _doc.groupname,
                            _doc.info)

        else:  # keine rückgabe? dann fertig setzen
            # alles vorbereiten
            next_wf.status = _doc.status | 4  # fertig setzen
            if _doc.options & 1 == 0:  # Weiterleitung EFA
                ofa_done = True
            # if _doc.hist_id:  # haben wir ein hist_id, dann müssen wir den eintrag wiederherstellen
            #     if _doc.options & 1 == 1:  # Weiterleitung JE
            #         rts_done = so_done
            #
            #     elif _doc.options & 1 == 0:  # Weiterleitung EFA
            #         rts_done = True  # rts_done: die weiterleitung ist fertig, die timerhist muss restored werden

        # fertig setzen
        timerhist = None
        timid = next_wf.tim_id
        status = next_wf.status
        if rts_done:  # wir stellen den eintrag aus timer_hist wieder her
            # entweder wir haben einen ursprünglichen eintrag mit status & 16 oder wir stellen unsere timid wieder her...
            sql = 'select * from sys_timerhist where histid = %s' % next_wf.hist_id
            try:
                timerhist = _db_trans.execute(sql).fetchone()
            except:
                self.log.exception(u"Fehler bei Abfrage der sys_timerhist")

            if timerhist:
                # timid = next_wf.tim_id
                # status = next_wf.status
                for row in sys_timer.values():
                    if _doc.options & 128 == 0:
                        if row[11] & 16 == 16 and row[16] == timerhist[
                            8]:  # wir suchen einen passenden eintrag mit status 16 und der histid der weiterleitung
                            timid = row[2]
                            status = row[11] & ~ 20
                            break
                    elif _doc.options & 128 == 128:  # in rückfrage? status 8 suchen
                        if row[11] & 8 == 8 and row[16] == timerhist[
                            8]:  # wir suchen einen passenden eintrag mit status 8 und der histid der weiterleitung
                            timid = row[2]
                            status = row[11] & ~ 12
                            break
            else:
                self.log.warning(
                    u"Es konnte kein Eintrag in der sys_timerhist mit ID %s gefunden werden" % next_wf.hist_id)
                next_wf.hist_id = 0  # wurde die histid nicht mehr in der sys_timerhist gefunden, dann setzen wir die zurück
                # sonst kann die wvl nie fertig gesetzt werden
                _doc.error = EC_RETURNTOSENDERERROR

        # eine fertigesetzte wvl kann nicht in arbeit sein, also status 2 rausnehmen
        status = status & ~ 2
        next_wf.status = next_wf.status & ~ 2

        # ab hier startet die transaction
        trans = _db_trans.begin()
        try:
            vals = {}
            vals['userid_work'] = next_wf.userid_work
            vals['userid_shed'] = next_wf.userid_shed
            vals['status'] = next_wf.status
            vals['info'] = next_wf.info
            vals['histid'] = next_wf.hist_id

            # sql = 'update sys_timer set userid_work = %s, userid_shed = %s, status = %s, info = "%s", histid = %s where timid = %s' \
            #       % (next_wf.userid_work, next_wf.userid_shed, next_wf.status, next_wf.info, next_wf.hist_id, next_wf.tim_id)
            sql = _con.sys_timer.update().values(vals).where(_con.sys_timer.c.timid == next_wf.tim_id)
            _db_trans.execute(sql)

            if (next_wf.options & 1 == 0 and (rts_done or ofa_done)) or (
                    next_wf.options & 1 == 1 and so_done):  # EFA oder JE komplett fertig: überflüssige Einträge entfernen
                sql = 'delete from sys_timer where timref = %s and histid = %s and timid <> %s' \
                      % (next_wf.tim_ref, next_wf.hist_id, next_wf.tim_id if next_wf.tim_id == timid else timid)
                _db_trans.execute(sql)

            if rts_done and timerhist:
                vals = {}
                vals['groupid'] = timerhist[2]
                vals['userid_work'] = timerhist[3]
                vals['userid_shed'] = timerhist[4]
                vals['status'] = status
                vals['info'] = next_wf.info
                vals['options'] = timerhist[5]
                vals['wfid'] = timerhist[6]
                vals['required'] = timerhist[7]
                vals['histid'] = timerhist[8]

                sql = _con.sys_timer.update().values(vals).where(_con.sys_timer.c.timid == timid)
                _db_trans.execute(sql)

                if timerhist[2] != 0 and timerhist[
                    5] & 1 == 0:  # stellen wir eine Gruppenwiedervorlage mit EFA wieder her, dann brauchen wir alle Gruppenmitglieder
                    values = []
                    sql = 'select userid from sys_groupuser where groupid = %s and userid <> %s' % (
                    timerhist[2], timerhist[3])
                    result = _db_trans.execute(sql).fetchall()

                    if result:
                        for row in result:
                            vals = {}
                            vals['timref'] = next_wf.tim_ref
                            vals['gdocid'] = next_wf.gdocid
                            vals['groupid'] = timerhist[2]
                            vals['userid_work'] = row[0]
                            vals['userid_shed'] = timerhist[4]
                            vals['status'] = status
                            vals['date'] = next_wf.date.strftime('%Y-%m-%d %H:%M:%S')
                            vals['alarm'] = next_wf.alarm.strftime('%Y-%m-%d %H:%M:%S')
                            vals['options'] = timerhist[5]
                            vals['info'] = next_wf.info
                            vals['comment'] = next_wf.comment
                            vals['wfid'] = timerhist[6]
                            vals['required'] = timerhist[7]
                            vals['histid'] = timerhist[8]
                            values.append(vals)

                    if values:
                        # sql = 'insert into sys_timer (timref, gdocid, groupid, userid_work, userid_shed, status, date, alarm, options, info, comment, wfid, required, histid) ' \
                        #       'select %s, %s, %s, g.userid, %s, %s, "%s", "%s", %s, "%s", "%s", %s, %s, %s from sys_groupuser as g ' \
                        #       'where g.groupid = %s and g.userid <> %s ' % (next_wf.tim_ref, next_wf.gdocid, timerhist[2], timerhist[4], status,
                        #                                 next_wf.date.strftime('%Y-%m-%d %H:%M:%S'), next_wf.alarm.strftime('%Y-%m-%d %H:%M:%S'),
                        #                                 timerhist[5], next_wf.info, next_wf.comment, timerhist[6], timerhist[7], timerhist[8],
                        #                                 timerhist[2], timerhist[3])
                        sql = _con.sys_timer.insert().values(values)
                        _db_trans.execute(sql)

                if (next_wf.options & 1 == 0) or (
                        next_wf.options & 1 == 1 and so_done):  # EFA oder JE komplett fertig und zurück: timerhist entfernen
                    sql = 'delete from sys_timerhist where histid = %s' % next_wf.hist_id
                    _db_trans.execute(sql)
            trans.commit()
        except sqla.exc.OperationalError as e:
            self.log.exception(u"Fehler beim Ausführen der Transaction set_statedone")
            if e.orig[0] == 2006:  # mysql has gone away
                _doc.error = EC_NEEDRECONNECT
            else:
                _doc.error = EC_SETSTATEDONEERROR
            trans.rollback()
            return None
        except:
            self.log.exception(u"Fehler beim Ausführen der Transaction set_statedone")
            _doc.error = EC_SETSTATEDONEERROR
            trans.rollback()
            return None

        # check ob auch wirklich alles fertig ist...
        sql = 'select count(*) from sys_timer where status & 4 = 0 and timref = %s' % next_wf.tim_ref  # alle offene zählen
        try:
            cnt_open = _db_trans.execute(sql).fetchone()
        except:
            self.log.exception(u"Fehler bei Abfrage der sys_timer")
            _doc.error = EC_SETSTATEDONEERROR
            return None
        else:
            _doc.result = cnt_open[0] == 0  # keine offen? wvl fertig!
            return True

    def forward_resub(self, _doc, recipients, options=None, required=None):
        """Funktion um eine Wiedervorlage an einen anderen Empfänger weiterzuleiten. Die Empfänger können als Liste von BitfarmUser
        und BitfarmGroup Objekte übergeben werden, außerdem können die Optionen der ZielWVL über options und ein eigener required-Zähler
        für die neue Wiedervorlage angegeben werden (required wird nur bei JE WVL verwendet)"""
        _ret = False
        if not isinstance(_doc, bitfarm_doc.BitfarmDocument) and not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return _ret
        if _doc.tim_id is None:
            return _ret

        if _doc.version in (3, 35, '3.0'):  # nur für v36 und v4
            return _ret

        if not _doc.profile:
            _con = self.con.main
            _db_trans = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db_trans = self.db[_doc.profile]

        _doc.result = False

        if not _doc.tim_ref:
            sql = "select timref from sys_timer where timid = %s" % _doc.tim_id
            try:
                result = _db_trans.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler bei forward_resub')
                return _ret
            if result:
                _doc.tim_ref = result[0]
            else:
                return _ret

        sql = "select * from sys_timer where timref = %s" % _doc.tim_ref
        try:
            result = _db_trans.execute(sql).fetchall()
        except:
            self.log.exception(u'Fehler bei forward_resub')
            return None

        if not result:
            return _ret

        resub = {}
        for row in result:
            row = [x for x in row]
            resub[row[0]] = row

        _doc.groupid = resub[_doc.tim_id][3]
        _doc.status = resub[_doc.tim_id][6]
        _doc.hist_id = resub[_doc.tim_id][14]
        hist_id = _doc.hist_id  # gibt es keine Weiterleitung mit RTS behalten wir den ursprünglichen Wert

        _doc.options = options if options is not None else resub[_doc.tim_id][9]
        _doc.options = _doc.options | (resub[_doc.tim_id][
                                           9] & 2)  # war ursprünglich diese wvl mit rts markiert, behalten wir diese option immer!

        if required is None:
            required = resub[_doc.tim_id][13]

        if resub[_doc.tim_id][6] & 8 == 8:  # befindet sich in Rückfrage, nicht weiterleiten
            _doc.error = EC_RESUBINCHECKBACK
            return _ret

        if resub[_doc.tim_id][9] & 128 == 128:  # die Rückfrage selbst, auch nicht weiterleiten (wird nicht unterstützt)
            _doc.error = EC_RESUBINCHECKBACK
            return _ret

        if _doc.options & 1 == 0:  # bei EFA müssen wir prüfen, ob nicht irgendjemand das auf Rückfrage hat
            sql = "select count(*) from sys_timer where timref = %s and groupid = %s and (status & 8) = 8" % (
            _doc.tim_ref, _doc.groupid)

            try:
                res = _db_trans.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler bei forward_resub')
                return _ret
            if res[0] > 0:
                return _ret

        if _doc.options & 128 == 128:  # haben wir eine weiterleitung auf rückfrage
            _doc.options = _doc.options | 2  # rückfrage ist immer rts
            _doc.options = _doc.options & ~ 1  # rückfrage ist niemals JE

        _doc.status = _doc.status & ~ 2  # weiterleitungen können nicht schon 'in Arbeit' sein

        if _con.sys_timer is None:
            _con.sys_timer = sqla.Table('sys_timer', _con.meta, autoload=True)
        if _con.sys_timerhist is None:
            _con.sys_timerhist = sqla.Table('sys_timerhist', _con.meta, autoload=True)
        if type(recipients) is not list:
            recipients = [recipients]

        # wir sortieren alle Einzeluser in eine Gruppe 0 und ergänzen mit anderen Gruppen
        groups = [bitfarm_doc.BitfarmGroup()]
        groups[0].id = 0
        for recp in recipients:
            if isinstance(recp, bitfarm_doc.BitfarmUser):
                groups[0].members.append(recp)
            if isinstance(recp, bitfarm_doc.BitfarmGroup):
                # if recp.id != _doc.groupid: #Weiterleiten an die eigene Gruppe wird nicht unterstützt
                #     groups.append(recp)
                groups.append(recp)
        _ret = False

        # wir schreiben bei weiterleitung immer in sys_timerhist um den urspünglichen controller zu behalten
        trans = _db_trans.begin()
        try:
            vals = {_con.sys_timerhist.c.timref: _doc.tim_ref,
                    _con.sys_timerhist.c.groupid: resub[_doc.tim_id][3],
                    _con.sys_timerhist.c.userid: resub[_doc.tim_id][4],
                    _con.sys_timerhist.c.shedid: resub[_doc.tim_id][5],
                    _con.sys_timerhist.c.options: resub[_doc.tim_id][9],
                    _con.sys_timerhist.c.wfid: resub[_doc.tim_id][12],
                    _con.sys_timerhist.c.required: resub[_doc.tim_id][13],
                    _con.sys_timerhist.c.prevhist: resub[_doc.tim_id][14]
                    }
            sql = _con.sys_timerhist.insert(values=vals)
            result = _db_trans.execute(sql)

            if result:
                hist_id = result.lastrowid  # neue hist_id zuweisen

            for recp in groups:
                userids = _con.resolve_group(recp)
                if len(userids) == 0:
                    continue

                vals = []

                for userid in userids:
                    vals.append({'timref': _doc.tim_ref,
                                 'gdocid': resub[_doc.tim_id][2],
                                 'groupid': recp.id,
                                 'userid_work': userid,
                                 'userid_shed': resub[_doc.tim_id][4],
                                 'status': _doc.status,
                                 'date': resub[_doc.tim_id][7],
                                 'alarm': resub[_doc.tim_id][8],
                                 'options': _doc.options,
                                 'info': u'FW: ' + resub[_doc.tim_id][10] if _doc.info is None else _doc.info,
                                 'comment': resub[_doc.tim_id][11] if _doc.comment is None else _doc.comment,
                                 'wfid': resub[_doc.tim_id][12],
                                 'required': required,
                                 'histid': hist_id})

                _db_trans.execute(_con.sys_timer.insert(), vals)

                _ret = True

            # wir entfernen unseren eigenen Eintrag, wir haben ja weitergeleitet und sind erstmal fertig damit
            if _ret:
                # neues Verhalten: wir löschen nicht unseren Eintrag sondern setzen Status 16 (Weiterleitung einer JE WVL), sowie Status 4 damit diese bei der Anzeige nicht gefunden wird

                if _doc.options & 128 == 128:
                    sql = 'update sys_timer set status = status | 12 where timid = %s ' % _doc.tim_id  # rückfrage
                else:
                    sql = 'update sys_timer set status = status | 20 where timid = %s ' % _doc.tim_id  # weiterleitung

                _db_trans.execute(sql)

                if resub[_doc.tim_id][9] & 1 == 0:
                    old_ids = [str(x) for x in resub if resub[x][14] == _doc.hist_id and resub[x][
                        0] != _doc.tim_id]  # alle einträge dieser timref mit dieser groupid, außer die eigene
                    if len(old_ids) > 0:
                        old_ids = ",".join(old_ids)
                        sql = 'delete from sys_timer where timref = %s and timid in (%s)' % (_doc.tim_ref, old_ids)
                        _db_trans.execute(sql)
            trans.commit()
        except sqla.exc.OperationalError as e:
            self.log.exception(u"Fehler beim Ausführen der Transaction forward_resub")
            if e.orig[0] == 2006:  # mysql has gone away
                _doc.error = EC_NEEDRECONNECT
            else:
                _doc.error = EC_FORWARDRESUBERROR
            trans.rollback()
            return None
        except:
            self.log.exception(u'Fehler bei forward_resub')
            _doc.error = EC_FORWARDRESUBERROR
            trans.rollback()
            return None

        return _ret

    def copy_resub(self, _doc, recipients, options=0, required=None, callback=False):
        """Funktion um einer Wiedervorlage andere Empfänger hinzuzufügen . Die Empfänger können als Liste von BitfarmUser
        und BitfarmGroup Objekte übergeben werden, alle Attribute der WVL werden übernommen (options, status, wf_id, hist_id, required.
        Wird Rückfrage gesetzt so wird der eigene Status  mit 8 verodert und die options mit 128, sowie ein Eintrag in der sys_timerhist erzeugt.
        Die hist_id wird beiden WVL zugeordnet, so kann diese auch wieder zurückgeholt werden"""
        if not isinstance(_doc, bitfarm_doc.BitfarmDocument) and not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return None
        if _doc.tim_id is None:
            return None

        if not _doc.profile:
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        _doc.result = False

        if _doc.version in (3, 35, '3.0'):  # nur für v36 und v40
            return None

        if not _doc.tim_ref:
            sql = "select tim_ref from sys_timer where timid = %s" % _doc.tim_id
            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler bei setstate_done')
                result = _con.try_again(sql=sql, db=_db, fetchone=True)

            if result:
                _doc.tim_ref = result[0]
            else:
                return None

        sql = "select * from sys_timer where timref = %s" % _doc.tim_ref
        try:
            result = _db.execute(sql).fetchall()
        except:
            self.log.exception(u'Fehler bei setstate_done')
            result = _con.try_again(sql=sql, db=_db, fetchall=True)

        if not result:
            return None

        resub = {}
        for row in result:
            row = [x for x in row]
            for ind in range(len(row)):
                if type(row[ind]) == str:
                    row[ind] = row[ind].decode(_con.db_encoding)
            resub[row[0]] = row

        _doc.groupid = resub[_doc.tim_id]['groupid']
        _doc.status = resub[_doc.tim_id]['status']
        _doc.options = resub[_doc.tim_id]['options']
        if required is None:
            required = resub[_doc.tim_id]['required']
        hist_id = resub[_doc.tim_id][
            'histid']  # gibt es keine Weiterleitung mit RTS behalten wir den ursprünglichen Wert

        if _con.sys_timer is None:
            _con.sys_timer = sqla.Table('sys_timer', _con.meta, autoload=True)
        if _con.sys_timerhist is None:
            _con.sys_timerhist = sqla.Table('sys_timerhist', _con.meta, autoload=True)
        if type(recipients) is not list:
            recipients = [recipients]

        # wir sortieren alle Einzeluser in eine Gruppe 0 und ergänzen mit anderen Gruppen
        groups = [bitfarm_doc.BitfarmGroup()]
        groups[0].id = 0
        if callback:  # bei Rückfragen nur 1 Empfänger
            for recp in recipients:
                if isinstance(recp, bitfarm_doc.BitfarmUser):
                    groups[0].members.append(recp)
                    break
                if isinstance(recp, bitfarm_doc.BitfarmGroup):
                    # if recp.id != _doc.groupid: #Weiterleiten an die eigene Gruppe wird nicht unterstützt
                    #     groups = [recp]
                    #     break
                    # else:
                    #     continue
                    groups = [recp]
                    break
        else:
            for recp in recipients:
                if isinstance(recp, bitfarm_doc.BitfarmUser):
                    groups[0].members.append(recp)
                if isinstance(recp, bitfarm_doc.BitfarmGroup):
                    # if recp.id != _doc.groupid: #Weiterleiten an die eigene Gruppe wird nicht unterstützt
                    #     groups.append(recp)
                    groups.append(recp)
        for recp in groups:
            userids = _con.resolve_group(recp)
            if len(userids) == 0:
                continue

            if callback:  # mit RTS müssen wir erst in die sys_timerhist schreiben
                callback_options = resub[_doc.tim_id][11] | 128
                vals = {_con.sys_timerhist.c.timref: _doc.tim_ref,
                        _con.sys_timerhist.c.groupid: resub[_doc.tim_id]['groupid'],
                        _con.sys_timerhist.c.userid: resub[_doc.tim_id]['userid'],
                        _con.sys_timerhist.c.shedid: resub[_doc.tim_id]['userid_shed'],
                        _con.sys_timerhist.c.options: callback_options,
                        _con.sys_timerhist.c.wfid: resub[_doc.tim_id]['wfid'],
                        _con.sys_timerhist.c.required: resub[_doc.tim_id]['required'],
                        _con.sys_timerhist.c.prevhist: resub[_doc.tim_id]['histid']
                        }
                sql = _con.sys_timerhist.insert(values=vals)
                try:
                    result = _db.execute(sql)
                except:
                    self.log.exception(u'Fehler bei copy_resub')
                if result:
                    hist_id = result.lastrowid
            elif hist_id:  # wenn die zu kopierende WVL eine hist_id hat, legen wir davon eine Kopie an
                sql = "insert into sys_timerhist (timref, groupid, userid, shedid, options, wfid, required, prevhist) select timref, groupid, userid, shedid, options, wfid, required, prevhist from sys_timerhist where histid = %s" % hist_id
                try:
                    result = _db.execute(sql)
                except:
                    self.log.exception(u'Fehler bei copy_resub')
                if result:
                    hist_id = result.lastrowid

            vals = []
            if callback:
                info = u'Rückfrage: ' + resub[_doc.tim_id]['info']
                options = options | 128
            else:
                info = resub[_doc.tim_id]['info']
            for userid in userids:
                vals.append({'timref': _doc.tim_ref,
                             'gdocid': resub[_doc.tim_id]['gdocid'],
                             'groupid': recp.id,
                             'userid_work': userid,
                             'userid_shed': resub[_doc.tim_id]['userid_shed'],
                             'status': resub[_doc.tim_id]['status'],
                             'date': resub[_doc.tim_id]['date'],
                             'alarm': resub[_doc.tim_id]['alarm'],
                             'options': options,
                             'info': info,
                             'comment': resub[_doc.tim_id]['comment'],
                             'wfid': resub[_doc.tim_id]['wfid'],
                             'required': required,
                             'histid': hist_id})
            try:
                result = _db.execute(_con.sys_timer.insert(), vals)
            except:
                self.log.exception(u'Fehler bei copy_resub')

        if callback:
            _doc.status = _doc.status | 8  # Status auf Callback setzen
            sql = "update sys_timer set status = %s, options = %s, histid = %s where timid = %s" % (
                _doc.status, _doc.options, hist_id, _doc.tim_id)
            try:
                _db.execute(sql)
            except:
                self.log.exception(u'Fehler bei copy_resub')

        return True

    def retract_resub(self, _doc):
        """Funktion um einer weitergeleitete Wiedervorlage wieder zurückzuholen/geben. Hierbei kommt es auf die Sichtweise an.
        der userid_shed holt eine Wiedervorlage zurück (die er weitergeleitet hat), ein userid_work bricht eine WVL ab (falls möglich, oder berechtigt).
        Eine Rückfrage kann von dem Fragesteller (userid_work) ebenfalls zurückgeholt werden.
        Bei GruppenWVL wird unabhängig der Option JE / EFA die gesamte Wiedervorlage zurückgeholt/gegeben"""
        if not isinstance(_doc, bitfarm_doc.BitfarmDocument) and not isinstance(_doc, bitfarm_doc.BitfarmWorkflow):
            return None
        if _doc.tim_id is None:
            return None

        if not _doc.profile:
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[_doc.profile]
            _db = self.db[_doc.profile]

        _doc.result = False

        if _doc.version in (3, 35, '3.0'):  # nur für v36 und v40
            return None

        sql = "select * from sys_timer where timid = %s" % _doc.tim_id
        try:
            result = _db.execute(sql).fetchone()
        except:
            self.log.exception(u'Fehler bei retract_resub')
            return None

        if not result:
            return None

        sys_timer = [x for x in result]
        for ind in range(len(sys_timer)):
            if type(sys_timer[ind]) == str:
                sys_timer[ind] = sys_timer[ind].decode(_con.db_encoding)

        _doc.tim_ref = sys_timer['timref']
        _doc.groupid = sys_timer['groupid']
        tim_id = _doc.tim_id
        status = sys_timer['status']
        options = sys_timer['options']
        hist_id = sys_timer['histid']
        info = u'Zurück: ' + sys_timer['info']
        result = None

        if hist_id:  # es gab eine Weiterleitung
            if status & 8 == 8:  # in Rückfrage?, dann Status zurücksetzen
                status = status & ~ 8

            sql = "select * from sys_timerhist where histid = %s" % hist_id
            try:
                result = _db.execute(sql).fetchone()
            except:
                self.log.exception(u'Fehler bei retract_resub')
                return None

        if result:  # wir haben eine Weiterleitung
            if result[5] & 128 == 128:  # wir wollen eine Rückfrage abbrechen (vom Empfänger der Rückfrage)
                options = result[5] & ~128  # Kennzeichen für Rückfrage entfernen
                # Ursprungsdokument suchen, dass ist das mit dem Status & 8
                sql = "select * from sys_timer where histid = %s and (status & 8) = 8" % hist_id
                try:
                    res = _db.execute(sql).fetchone()
                except:
                    self.log.exception(u'Fehler bei retract_resub')
                    return None
                if res:
                    status = res[8] & ~ 8  # Status zurücknehmen
                    tim_id = res[0]

            sql = u"update sys_timer set status = %s, groupid = %s, userid_work = %s, userid_shed = %s, options = %s, wfid = %s, " \
                  u"required = %s, histid = %s, info = '%s' where timid = %s" % (
                      status, result[2], result[3], result[4], options, result[6], result[7], result[8], info, tim_id)
            try:
                _db.execute(sql)
            except:
                self.log.exception(u'Fehler bei retract_resub')
                return None
            # den Eintrag noch aus der sys_timerhist entfernen
            sql = 'delete from sys_timerhist where histid = %s' % hist_id
            try:
                _db.execute(sql)
            except:
                self.log.exception(u'Fehler bei retract_resub')
                return None
            # alle Einträge mit dieser hist_id entfernen,
            sql = 'delete from sys_timer where histid = %s and timid <> %s' % (hist_id, tim_id)
            try:
                _db.execute(sql)
            except:
                self.log.exception(u'Fehler bei retract_resub')
                return None

            return True
        else:  # wenn keine Weiterleitung mit RTS, dann geben wir an den Controller zurück
            if sys_timer[6] != sys_timer[7]:  # nur wenn usr_work und usr_shed nicht derselbe User ist
                sql = u"update sys_timer set userid_work = %s, info = '%s' where timid = %s" % (
                    sys_timer[7], info, _doc.tim_id)
                try:
                    _db.execute(sql)
                except:
                    self.log.exception(u'Fehler bei retract_resub')
                    return None

                if options & 1 == 0:  # EFA? Überflüssige Gruppeneinträge entfernen
                    sql = u"delete from sys_timer where timref = %s and groupid = %s and timid <> %s" % (
                        _doc.tim_ref, _doc.groupid, _doc.tim_id)
                    try:
                        _db.execute(sql)
                    except:
                        self.log.exception(u'Fehler bei retract_resub')
                        return None
                return True

    def get_naming(self, profile=None, nameid=None, name=None):
        """Funktion um aus der sys_naming eine Naming-Regel zu holen, gibt Liste zurück"""
        _ret = []
        # eingabe prüfen
        if nameid is None and name is None:  # nichts übergeben
            return _ret

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return _ret

        sql = None
        if nameid is not None:
            sql = 'select * from sys_naming where namingid = %s' % nameid

        elif name is not None:
            sql = "select * from sys_naming where title = '%s'" % name

        if sql:
            try:
                result = _db.execute(sql).fetchall()
            except:
                result = _con.try_again(sql=sql, db=_db, fetchall=True)
            if result:
                for row in result:
                    self.log.debug(row)
                    getfield = bfaxml.xml_to_object(row[2])
                    if type(getfield) == list:
                        _ret.extend(getfield)
                    else:
                        _ret.append(getfield)
                    if row[3]:
                        for item in _ret:
                            item.ref_docs = [int(x) for x in row[3].split(';')]
        return _ret

    def save_naming(self, profile=None, nameid=None, name=None, getfield=None, ref_docs=None):
        """Funktion um in der sys_naming eine Naming-Regel zu speichern. Es kann eine einzelne GetField- Anweisung
        oder eine Liste an getField-Anweisungen übergeben werden. Wird eine ID angegeben, so wird der Eintrag aktualisiert,
        andernfalls wird ein neur angelegt und die ID zurückgegeben. Ein Name muss übergeben werden"""

        # eingabe prüfen
        if nameid is None and name is None:  # nichts übergeben
            return None

        if getfield is None:
            return None
        # TODO: auf Liste bzw Getfield-Objekt prüfen

        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

        xml = bfaxml.object_to_xml(getfield)
        sys_naming = sqla.Table('sys_naming', _con.meta, autoload=True)

        if ref_docs:
            if type(ref_docs) == list:
                ref_docs = ";".join([str(x) for x in ref_docs])
        else:
            ref_docs = []
            if type(getfield) == list:
                for gf in getfield:
                    if gf.ref_docs:
                        ref_docs.extend(gf.ref_docs)
            elif getfield.ref_docs:
                ref_docs.extend(getfield.ref_docs)
            if len(ref_docs) > 0:
                ref_docs = ";".join([str(x) for x in ref_docs])
            else:
                ref_docs = None

        if nameid:
            sql = sys_naming.update().where(sys_naming.c.namingid == nameid)
        else:
            sql = sys_naming.insert()
        try:
            result = _db.execute(sql, title=name, naming=xml, ref_docs=ref_docs)
        except:
            result = _con.try_again(sql, db=_db, title=name, naming=xml, ref_docs=ref_docs)
        if nameid:
            return nameid
        else:
            return result.lastrowid

    def create_classrules(self, profile=None, classname=u""):
        """Funktion um aus Referenzdokumenten einen Regelsatz zu erstellen, der die Dokumentenklasse definiert. Der Regelsatz wird in einer Instanz
        von BitfarmDocClasses geschrieben"""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        else:
            sql = "SELECT id FROM sys_classes WHERE classname='%s'" % classname
            result = _db.execute(sql).fetchall()
            if len(result) == 0:
                return None
            classid = result[0][0]
            sql = "SELECT doctupel from sys_classdocs WHERE classid=%d" % classid
            result = _db.execute(sql).fetchall()
            for res in result:
                # doctupeldict = bfaxml.xml_to_object(res[u'doctupel'])
                doctupeldict = pickle.loads(res[u'doctupel'])
                for k in doctupeldict.keys():
                    print(len(doctupeldict[k]))
                # key von doctupeldict ist die Anzahl der Worte eines einzelnen Tupels
            ngramslists = []
            pass

    def read_docclasses(self, profile=None):
        """Funktion um alle vorhandenen Regeln aus der Datenbank in bfaWflowProc.doc_classes einzulesen."""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        else:
            sql = "SELECT id,classname,class_rules,naming_rules FROM sys_classes"
            result = _db.execute(sql).fetchall()
            for cl in result:
                docclass = bitfarm_doc.BitfarmDocClass()
                docclass.classid = cl[0]
                docclass.classname = cl[1].decode(_con.db_encoding)
                docclass.ngrams = None
                if cl[3] is None:
                    docclass.ngrams = None
                else:
                    docclass.ngrams = pickle.loads(cl[3])
                sql = "SELECT * FROM sys_classdocs WHERE classid=%d" % docclass.classid
                docresult = _db.execute(sql).fetchall()
                for doc in docresult:
                    docclass.classdocs.append((doc[0], doc[4], doc[2].decode(_con.db_encoding)))
                self.doc_classes[docclass.classname] = docclass

    def update_docclasses(self, profile=None, classdict=None):
        """Funktion um ein oder mehrere Regeln in bfaWflowProc.doc_classes zu ändern, ergänzen oder zu löschen.
        Übergeben wird ein Dictonary {Classname: BitfarmDocClass}, vorhandene Classnames werden aktualisiert,
        nicht vorhandene hinzugefügt und wenn BitfarmDocClass = None wird die Regel entfernt"""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

    def classify(self, profile=None, _doc=None):
        """Funktion um eine Dokument zu klassifizieren. Erwartet ein Doc-Objekt."""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None

    def _add_class(self, profile=None, classname=u""):
        """Funktion um eine neu Dokumentenklasse in sys_classes anzulegen und
        bfaWflowProc.doc_classes zu aktualisieren"""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        else:
            sql = "INSERT INTO sys_classes(classname) VALUES ('%s')" % classname
            try:
                result = _db.execute(sql)
                docclass = bitfarm_doc.BitfarmDocClass()
                docclass.classid = result.lastrowid
                docclass.classname = classname
                self.doc_classes[docclass.classname] = docclass
                return result.lastrowid
            except:
                return None

    def _remove_class(self, profile=None, classname=u""):
        """Funktion um eine Dokumentenklasse aus sys_classes zu entfernen.
        Zusätzlich werden alle zugehörigen Referenzdokumente der Klasse aus sys_classdocs entfernt
        und bfaWflowProc.doc_classes wird aktualisiert"""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        else:

            try:
                sql = "DELETE FROM sys_classes where classname = '%s'" % classname
                _db.execute(sql)
                sql = "DELETE FROM sys_classdocs where classid = %d" % self.doc_classes[classname].classid
                _db.execute(sql)
                del self.doc_classes[classname]
                return True
            except:
                return None

    def _calculate(self):
        pass

    def _add_classdoc(self, profile=None, classname=u"", gdocid=None, txt=u""):
        """Funktion um einer Dokumentenklasse ein neues Referenzdokument hinzuzufügen.
        Aktualisiert sys_classdocs und bfaWflowProc.doc_classes"""
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        else:
            try:
                # print "add_classdoc, type(txt):", type(txt)
                sql = "SELECT id FROM sys_classes WHERE classname='%s'" % classname
                result = _db.execute(sql).fetchall()
                if len(result) == 0:
                    return None
                # sql = "INSERT INTO sys_classdocs(classid, gdocid, doctext) VALUES (%d, %d,'%s')" %(result[0][0], gdocid, txt)
                tbl = sqla.Table('sys_classdocs', _con.meta, autoload=True)
                try:
                    insert = sqla.insert(tbl, values={'classid': result[0][0], 'gdocid': gdocid, 'doctext': txt})
                    result = _db.execute(insert)
                except:
                    insert = sqla.insert(tbl, values={'classid': result[0][0], 'gdocid': gdocid,
                                                      'doctext': txt.encode(_con.db_encoding, "ignore")})
                    result = _db.execute(insert)
                self.doc_classes[classname].classdocs.append((result.lastrowid, gdocid, txt))
                ngrams = bitfarm_doc.get_ngrams(txt, 2, 4)
                pngrams = pickle.dumps(ngrams)
                # xmlstr = bfaxml.object_to_xml(ngrams)
                update = sqla.update(tbl).where(tbl.c.id == result.lastrowid).values(doctupel=pngrams)
                result = _db.execute(update)
                self.create_classrules(profile=None, classname=classname)
                return True
            except:
                return None

    def _remove_classdoc(self, profile=None, cid=None):
        if profile is None or profile == '':
            _con = self.con.main
            _db = self.db['main']
        else:
            _con = self.con[profile]
            _db = self.db[profile]

        if _db is None:
            # TODO: wäre es angebracht eine exception zu werfen
            return None
        else:
            try:
                sql = "SELECT id, classid, gdocid, doctext FROM sys_classdocs WHERE id=%d" % cid
                res = _db.execute(sql).fetchall()[0]
                classdoctupel = (res[0], res[2], res[3].decode(_con.db_encoding))
                classid = res[1]

                sql = "SELECT classname from sys_classes WHERE id=%d" % classid
                res = _db.execute(sql).fetchall()[0]
                classname = res[0]
                self.doc_classes[classname].classdocs.remove(classdoctupel)

                sql = "DELETE FROM sys_classdocs WHERE id = %d" % cid
                _db.execute(sql)
                return True
            except:
                return None


class BitfarmINIObject(object):
    """Hilfsklasse um eine dynamische INI-Klasse aufzubauen"""
    version = __version__

    def __init__(self):
        self = None

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            return None


from defusedxml import cElementTree as dTree
import xml.etree.cElementTree as eTree

import bitfarm_xml as bfaxml

Element = type(eTree.fromstring('<xml></xml>'))
